;+
; NAME: 
;   calibratenight
;
; PURPOSE: 
;   Automatically calibrate an entire night of data.    
;
; CALLING SEQUENCE: 
;   calibratenight, files, workingdir=workingdir, instrument=instrument, istr=istr, 
;                   /nosolve, /photometric, archivezptfile=archivezptfile, 
;                   zptradarcsec=zptradarcsec, redosolution=redosolution, 
;                   maxseeing=maxseeing, /verbose, /skiprelative
;       
; OPTIONAL INPUTS:
;   files - List of files or wildcard string containing all processed images from a
;           given night.  If unspecified, infer the probable files for specific instrument
;   workingdir - Directory to read/write the numerous files needed and produced by this
;          routine.
;   instrument - Instrument to process.  Specifying this sets a variety of settings to be
;          optimized for the instrument in question.  Supported values are:
;          P60, SEDM, LRIS, MOSFIRE, WIRC, IO:O   ("P60" means GRBCam; "SEDM"=RainbowCam)
;   istr - Short string to insert in output files to indicate the instrument.
;   /nosolve - Currently does nothing
;   /photometric - Signal that the night was probably photometric and that an airmass
;          dependent solution from calibrated fields should be applied to uncalibrated
;          fields.
;   archivezptfile - Filename of a zeropoint file (probably from another night) to use to
;          calibrate uncalibrated fields from this night.
;   zptradarcsec - Radius to define as "infinite" for zeropoint calculations; i.e. where
;          the flux encloses should be stable as a function of atmospheric conditions.
;   redosolution - Repeat the zeropoint calculation even if solution file exists.
;   maxseeing - Maximum seeing of an individual image (in arcsec) to allow zeropoint-based
;          measurements.  This should be much less than zptradarcsec
;         (default zptradarcsec/2.5)
;   /verbose - Currently does nothing
;   /skiprelative - Do not perform image-to-image relative calibration for images of the
;          same field
;       
; OUTPUTS: 
;   Numerous files are output to the working directory: a catalog file for each image
;   (calibrated to the maximum extent possible), diagnostic plots, and a zeropoint 
;   table.  Additionally, the zeropoint information is written to the headers of the 
;   original FITS files.
;
; COMMENTS:
;   This is a complex high-level code to perform photometric calibration of 
;   (ideally photometric) nights in which a large amount of imaging was taken.  Thorough
;   documentation will be provided at a later time.  In brief, the routine performs an
;   automatic photometric extraction of a star catalog for each image (.cat file) and
;   for fields where reference calibrations are available, this is then calibrated versus
;   the reference (becomes .cal file).  If the night was photometric, a photometric
;   solution is found (zeropoint as a function of airmass) and any fields without
;   reference calibration data are calibrated (becomes .caz file).
;   For the instruments this has been tested on, usually can be run fully automatically.
;   Mostly not tested over the past 5 years.
;
; Last modified 4 Sep  2020.  Documentation added 3 July 2022.
;-

; Calibrate a night (or part of night, or a block of a few nights as long as atmosphere/instrument is stable) 
;  of imaging of standardizable fields (SDSS/Landolts) and, if photometric, non-standard fields as well.
; Runs through a long list of files, determines zeropoints for all SDSS/Landolt-calibrated fields, and solves for any uncalibrated fields.

; This is always run in the directory with the (processed) files.

; To produce a calibration file, follow up with:
  ; automergecat, file_search('xxxx*.fits'), 'targetname'
  ; transformcat, 'targetname_griz.calib', 'targetname.instcal.dat'


pro writeparam
  openw, 5, 'sex.param'
  printf, 5, 'X_IMAGE'
  printf, 5, 'Y_IMAGE'
  printf, 5, 'ALPHA_J2000'
  printf, 5, 'DELTA_J2000'
  printf, 5, 'MAG_AUTO'
  printf, 5, 'MAGERR_AUTO'
  printf, 5, 'ELLIPTICITY'
  printf, 5, 'FWHM_IMAGE'
  close, 5
  return
end

pro writesex  ; this is apparently not working
  sexpath = getsexpath()
  if file_test('default.config') eq 0 then spawn, sexpath+' -d > default.config'
  openr, 5, 'default.config'
  lines = ['']
  inline = ''
  while not eof(5) do begin
    readf, 5, inline
    inline = repstr(inline, 'test.cat', 'temp.cat')
    inline = repstr(inline, 'default.conv', 'sex.conv')
    inline = repstr(inline, 'default.param', 'sex.param')
    inline = repstr(inline, 'temp.param', 'sex.param')
    inline = repstr(inline, 'DETECT_THRESH    1.5','DETECT_THRESH      3')
    inline = repstr(inline, 'ANALYSIS_THRESH  1.5','ANALYSIS_THRESH    3')
    inline = repstr(inline, 'VERBOSE_TYPE     NORMAL','VERBOSE_TYPE     QUIET')
    lines = [lines, inline]
  endwhile
  close, 5
  openw, 5, 'sex.config'
  for i = 1, n_elements(lines)-1 do begin
    printf, 5, lines[i]
  endfor
  close, 5
  return
end
pro writeconv
  openw, 5, 'sex.conv'
  printf, 5, 'CONV NORM'
  printf, 5, '# 3x3 ``all-ground'' convolution mask with FWHM = 2 pixels.'
  printf, 5, '1 2 1'
  printf, 5, '2 4 2'
  printf, 5, '1 2 1'
  close, 5
  return
end

pro calibratenight, files, workingdir=workingdir, instrument=instrument, istr=istr, nosolve=nosolve, photometric=photometric, archivezptfile=archivezptfile, zptradarcsec=zptradarcsec, redosolution=redosolution, maxseeing=maxseeing, verbose=verbose, skiprelative=skiprelative

  nir = 0

  if instrument eq 'P60' then begin  ; GRBCam only
    pixscale = 0.3777   ; P60-specific
    expkey =  'EXPTIME'
    filtkey = 'FILTER1'
    skykey =  'SKYBKG'
    instkey = 'INSTRMNT'
    objkey =  'OBJECT'
    seekey  = 'SEEING'  ; this is in arcseconds and uses the file headers.  To generalize, can't assume this is present
    timekey = 'UTSHUT'
    defaultfilestr = 'i2*p.fits'
    elev = 1680.
    if n_elements(zptradarcsec) eq 0 then zptradarcsec = 7.0  ; arcsec, ~18.6 pixels
    instfilts = ['upr','B','gpr','V','rpr','R','ipr','I','zpr']
    calfilts  = ['u',  'B','g',  'V','r',  'R','i',  'I','z']
    defaultzpt = [23, 24, 24, 24, 24, 24, 24] 
    copykey = ['FILTER1','SKYBKG','UTSHUT'] ; any keywords not automatically recognized in fieldphot
  endif
  if instrument eq 'SEDM' then begin
     pixscale = 0.3939
     expkey = 'EXPTIME'
     filtkey = 'FILTER'
     skykey = 'SKYCTS' ;?
     ; instkey = 'CAM_NAME'  doesn't get put into cat, so ignore
     objkey = 'OBJECT'
     ; seekey 
     timekey = 'UTC'  ; can it process this format?
     ; defaultfilestr = 'f*.fits' ; ?
     elev = 1680.
     if n_elements(zptradarcsec) eq 0 then zptradarcsec = 7.0  ; arcsec
     instfilts = ['u','g','r','i']
     calfilts =  ['u','g','r','i']
     defaultzpt = [21.4,24,24,23.2]
     copykey = ['CHANNEL']
  endif
  if instrument eq 'LRIS' then begin
    pixscale = 0.1351
    expkey =  'EXPTIME'
    filtkey = 'FILTER'
    filtkey2 = 'DICHNAME'
    skykey =  'COUNTS'
    instkey = 'INSTRUME'
    objkey =  'TARGNAME'
    timekey = 'UTC'
    ; copykey currently not used, but could be.
    defaultfilestr = 'azfp*r.fits'
    elev = 4200.
    if n_elements(zptradarcsec) eq 0 then zptradarcsec = 4.0  ; (30 pixels)

    instfilts = ['U','B','G','V','R','Rs','I','RG850', 'Clear']
    calfilts  = ['u','B','g','V','R','r', 'i','z',     'R']
    defaultzpt =[27, 28, 28, 27.5, 27, 27, 27, 27,   28.5]
    instfilts = [instfilts, 'NB3425','NB3640','NB4970','NB4670','NB4430','NB4535', 'NB6741', 'NB8185', 'NB8560', 'NB9135', 'NB9148']
    calfilts  = [calfilts,  'u',      'u',     'g',     'g',    'B',     'B',       'R',     'i',      'i',      'z',      'z']
    defaultzpt= [defaultzpt, 25,     25,       25,     25,      25,      25,        25,      25,       25,       25,       25]
  endif
  if instrument eq 'MOSFIRE' then begin
    instfilts = ['Y','J','H','Ks']
    calfilts = instfilts
    defaultzpt = [26.0, 25.0, 24.5, 23.7]
    nir = 1
  endif
  if instrument eq 'WIRC' then begin
    nir = 1
  endif
  if instrument eq 'IO:O' then begin
    pixscale = 0.30358
    instfilts = ['SDSS-U-02','Bessell-B-02','Bessell-V-02','SDSS-G-02','SDSS-R-02','SDSS-I-02','SDSS-Z-02'] ; ???
    calfilts = ['u','B','V','g','r','i','z']
    defaultzpt = [22.33,25,25,25,25,25,25]
    elev = 2423.
    defaultfilestr = 'h_?_*_1_1_?.fits' ; 10+ in a sequence...?
    filtkey = 'FILTERI1'
    expkey =  'EXPTIME'
    timekey = 'UTSTART'
    objkey =  'OBJECT'
    skykey = 'L1MEDIAN'
    if n_elements(zptradarcsec) eq 0 then zptradarcsec = 6.0
    copykey = [filtkey]
    sys = 'LT'
  endif


  if n_elements(istr) eq 0 then istr = legalfilestr(strlowcase(instrument))
  if n_elements(workingdir) eq 0 then workingdir = ''

  if n_elements(files) eq 0 and n_elements(defaultfilestr) gt 0 then begin
     pfiles = file_search(workingdir+defaultfilestr)
     pfiles = pfiles[sort(pfiles)]
  endif else begin
    if n_elements(files) eq 1 then begin
       if files[0] eq '' then begin
         print, 'Blank input?'
         return
       endif
       pfiles = file_search(workingdir+files)
       pfiles = pfiles[sort(pfiles)]
       realfiles = where(pfiles ne '', ct)
       if ct eq 0 then return else pfiles = pfiles[realfiles]
    endif else begin
       pfiles = files
    endelse
  endelse 

  imh = readimheaders(pfiles, expkey=expkey, filtkey=filtkey, filt2key=filt2key, skykey=skykey, instkey=instkey, objkey=objkey, seekey=seekey)


  zpt = fltarr(n_elements(pfiles))

  p = 2^((4200-elev)/4570.)*[0.12947983, -0.13458371, 0.21833111, -0.48222307, 0.41800934, -0.11447838, 0.0032198187, -0.010178335, 0.0047375054]

  if file_test('sex.config') eq 0 then writesex
  if file_test('sex.param') eq 0 then writeparam
  if file_test('sex.conv') eq 0 then writeconv


  if n_elements(maxseeing) eq 0 then $
    maxseeing = zptradarcsec/2.5



  ; Extract photometric catalog for all images

  newcalfiles = 0
  adjfiles = ['']
  adjobjects = ['']
  for f = 0, n_elements(pfiles)-1 do begin
      outfile = repstr(pfiles[f],'.fits','.cat')
      outfilec = repstr(pfiles[f],'.fits','.cal')
      outfilez = repstr(pfiles[f],'.fits','.caz')
      if file_test(outfile) or file_test(outfilec) or file_test(outfilez) then continue
      ;print, imh[f].ctrate, imh[f].exptime, imh[f].seeing, maxseeing
      if imh[f].ctrate gt 10000. then continue ; ignore worst twilight frames
      if (imh[f].exptime gt 10.5) and (imh[f].ctrate gt 5000.) and nir eq 0 then continue
      if (imh[f].exptime gt 1.5)  and (imh[f].ctrate gt 1500.) and nir eq 0 then continue
      ;if (finite(imh[f].seeing) and imh[f].seeing gt maxseeing) then continue

      filt = imh[f].filter
      fw = where(filt eq instfilts, ct)
      if ct ge 1 then begin
          calfilt = calfilts[fw[0]] 
      endif else begin
          if imh[f].instrument eq 'LRISBLUE' then calfilt = 'g'  ; LRIS-specific... should sef 'defaultfilt' at top.
          if imh[f].instrument eq 'LRIS' then calfilt = 'R'
          logprint, ' Warning - Filter not recognized!  Zeropoint calibration will attempt to match with '+calfilt+'-band.'
      endelse
      if clip(filt,1) eq 'z' or clip(filt,1) eq 'u' then errthresh=0.16 else errthresh = 0.1

      logprint, ' Extracting star catalog for '+removepath(pfiles[f])+' ('+filt+')'
      fieldcat, pfiles[f], router=zptradarcsec/imh[f].pxscale, outfile=outfile, errthresh=errthresh, copykey=copykey
         ; could replace outfile with temp.cat (and outfilec with temp.cal) to preserve resuming properly (redo target) in case of interruption
         ; during the adjustzpt call

      
      ;One way to speed things up would be to block the calls to adjustzpt by target - would have to make calfilt an array.
      ;if adjfiles[0] eq '' then adjfile = [outfile] else adjfiles = [adjfiles, outfile]
      ;if f lt n_elements(pfiles)-1 then begin
      ;   endseq = imh[f].object ne imh[f+1].object
      ;endif else begin
      ;   endseq = 1
      ;endelse
      ;if endseq then begin
      ;   adjustzpt, adjfiles, out=repstr(adjfiles,'.cat','.cal'), /rename, filter=calfiltstarg, adjusted=adjusted
      ;endif

      adjustzpt, outfile, out=outfilec, /rename, filter=calfilt, adjusted=adjusted, sys=sys
      if adjusted  then newcalfiles = 1

  endfor

  if n_elements(archivezptfile) gt 0 then begin 
       zptfile=archivezptfile
       if file_test(archivezptfile) eq 0 then print, 'WARNING: Archival zeropoint file needed but not found!'
  endif
  outzptfile = workingdir+'zeropoints.dat'
 

  ; Fit to determine zenith zeropoint
  ;   (this will not be a helpful exercise if not photometric - but, still make the plot...)

  calfileso = repstr(pfiles, '.fits', '.cal')
  calfiles = ['']
  for i = 0, n_elements(calfileso)-1 do if file_test(calfileso[i]) then calfiles = [calfiles, calfileso[i]]
  ;print, calfiles
  ctgoodcal = 0
  if n_elements(calfiles) gt 1 then begin
     calfiles = calfiles[1:*]
     cal = readcatheaders(calfiles, count=ctc,  instkey=instkey, filtkey=filtkey, filt2key=filt2key, timekey=timekey, objkey=objkey)
     wgoodsee = where(cal.seeing*pixscale lt maxseeing or finite(cal.seeing) eq 0, ctgoodcal)
     if ctgoodcal gt 0 then cal = cal[wgoodsee]
  endif

  if (  newcalfiles gt 0 or $
        (file_test(outzptfile) eq 0 and keyword_set(photometric)) or $
        keyword_set(redosolution)) $
      and ctgoodcal gt 0 then begin
     uconfig = unique(cal.config)
     nconfig = n_elements(uconfig)
     ufilt   = strarr(nconfig)
     ufilt2  = strarr(nconfig)
     uinst   = strarr(nconfig)
     for i = 0, nconfig-1 do begin
       w = where(cal.config eq uconfig[i])
       ufilt[i] = cal[w[0]].filt
       ufilt2[i] = cal[w[0]].filt2
       uinst[i] = cal[w[0]].inst
     endfor

     ; new content
     uwav = fltarr(n_elements(ufilt))
     unkfiltlist = ['']
     for i = 0, n_elements(ufilt)-1 do begin
        fw = where(ufilt[i] eq instfilts, ct)
        if ct eq 0 then begin
            print, 'No matches for filter ', ufilt[i], ' within ', instfilts
            continue
        endif
        if ct ge 1 then begin
            calfilt = calfilts[fw[0]] 
        endif else begin
            calfilt = ufilt
        endelse
 
        uwav[i] = filtwv(calfilt)

        if uwav[i] lt 3000 then begin
            unkfiltlist = [unkfiltlist, ufilt[i]]
            uwav[oth] = 6000.
        endif
     endfor
     if n_elements(unkfiltlist) gt 1 then begin
        unkfiltstr = ''
        for f = 1, ct-1 do unkfiltstr += ','+ufilt[oth[f]]
        print, 'WARNING - unrecognized filter(s): ['+ unkfiltstr + ']'
        print, '          Unable to estimate airmass correction'
     endif

     order = sort(uwav)

     ;; if file_test(archivezptfile) then archivezptdata = readzpt(zptfile, count=ctz)  ; in principle could use this to overplot archival zpt curve...

     plotopen = 0
     for ui = 0, nconfig-1 do begin
        i = order[ui]
        w = where(cal.config eq uconfig[i], ct)
        if ct eq 0 then continue

        logprint, ' '+ uconfig[i]+ '  (lambda=',clip(fix(uwav[i])),')'

        calw = cal[w]

        if plotopen eq 0 then begin
           !p.multi = [0, (nconfig > 3)]
           psopen, workingdir+istr+'cal.eps', xsize=5, ysize=4+nconfig*1, /inches, /encaps
           device, /color
           !p.font=0
           device, /helvetica, font_index=17
           colors = transpose([[0,0,0],$
                      [128,128,128],$ 
                      [128,0,0],$ 
                      [255,128,128],$ 
                      [0,128,0],$  
                      [100,255,100]])
           tvlct, colors
           ;xcbars = 0.01+findgen(ndich+1)*(0.99-0.01)/ndich
           xcbars = [0.01, 0.99]
           ycbars = 0.05+findgen(nconfig+1)*(0.95-0.01)/nconfig
           
           plotopen = 1
        endif

        ;!p.position = [xcbars[di]+0.08/ndich, ycbars[nfilt-1-fi]+0.25/(4+nfilt*1), xcbars[di+1], ycbars[nfilt-1-fi+1]]
        !p.position = [xcbars[0]+0.08, ycbars[nconfig-1-ui], xcbars[1], ycbars[nconfig-1-ui+1]]

        ;if filt[f] eq 'U' then yrange=[25,28] else yrange=[26,29]
        xrange = [1, max(cal.airmass)>1.8]
        yrange = [min(calw.zpt)-0.05, max(calw.zpt)+0.1]
        if yrange[1]-yrange[0] lt 0.5 then yrange[0]=yrange[1]-0.5
        plot, [0], [0], xrange=xrange, yrange=yrange, /ynozero, /nodata, /xsty, /ysty, charsize=1+(!p.multi[2]/2)*0.2, title='!17'


        xyouts, xrange[0]+0.05*(xrange[1]-xrange[0]), yrange[0]+0.2*(yrange[1]-yrange[0]), /data, uconfig[i]

        if nir eq 0 then defaultairterm = poly((uwav[i]-5000.)/2500.,p) else defaultairterm = 0.05
 
        wok = where(calw.type eq 'sdss' or calw.type eq 'landolt' or calw.type eq 'sdss2' or calw.type eq '2mass' and calw.zptunc lt 0.08, ctok)
        if ctok eq 0 then continue
        airterm = 0
        if uwav[i] lt 5000.  and  ctok ge 5  and  max(cal[wok].airmass)-min(cal[wok].airmass) ge 0.3 then begin
           ; actually fit (only triggers on night with large amounts of imaging, e.g. P60)
           unc = [calw[wok].zptunc > 0.03]
           extp = polyiterfit(calw[wok].airmass-1, calw[wok].zpt, unc=unc, parunc=parunc, order=1, yfit=yfit, good=wgood)
           wgood = wok[wgood] ; put in calw indexing system
           zenithzpt = extp[0]
           zptunc = parunc[0]
           airterm = -extp[1]
           airtermunc = parunc[1]
           residual = calw[wok].zpt-yfit
           residualsigma = residual/unc
        endif
        ;if uwav[i] gt 5000. or ctok lt 5 or max(cal[wok].airmass)-min(cal[wok].airmass) lt 0.3 then begin
        if airterm le 0.0 or airterm gt 1.0 then begin  ; either the two-term fit failed, or wasn't attempted
           ; use robust approach with fixed airmass term
           airvar = 0 - (calw[wok].airmass-1)*defaultairterm ; portion of variation due to airmass
           dif = [calw[wok].zpt - airvar]
           unc = [calw[wok].zptunc > 0.01]
           zenithzpt_wmean =  wmean(dif, unc, sigclip=3.5, meanunc=meanunc, good=wgood)
           wgood = wok[wgood] ; put in calw indexing system
           ;wgood = where(unc lt 0.1 > median(unc)*1.5, ct)   ; median probably isn't necessary...?
           ;if ct gt 0 then zenithzpt_med  = median(dif[wgood], /even)
           zenithzpt = zenithzpt_wmean
           residual = dif-zenithzpt
           residualsigma = residual/unc
           zptunc = meanunc
           airterm = defaultairterm
           airtermunc = -0.
        endif

        if keyword_set(photometric) then ord = sort(calw.airmass) $
                                    else ord = indgen(n_elements(calw))
        for iii = 0, n_elements(w)-1 do begin
            ii = ord[iii]
            isok   = where(ii eq wok, cto)
            isgood = where(ii eq wgood, ctg)
            isgoodstr = ctg gt 0 ? ' ' : (cto gt 0 ? '-' : 'x')
            resid = calw[ii].zpt - zenithzpt + (calw[ii].airmass-1)*airterm
            ff =  calw[ii].file    ;(strsplit(calw[ii].file,'_',/extract)) [1]
            flen = max(strlen(removepath(calw.file))) < 30
            obj = calw[ii].object
            logprint, '        '+clip(removepath(ff),flen)+'  '+clip(obj,8)+'  '+clip(calw[ii].type,9)+'  air='+fpr(calw[ii].airmass,1.3)+'  '+ $
                   'zpt='+fpr(calw[ii].zpt,2.3)+' ('+ fpr(calw[ii].zptunc,1.3)+ ')  resid='+fpr(resid,2.3)+$
                   ' '+isgoodstr+ rclip(calw[ii].nap,3)+ '/'+ clip(calw[ii].nstar,3)+ 'stars,  ap='+ $
                   fpr(calw[ii].seeing,2.2)+'px'
        endfor

        if median(residualsigma) gt 3 then logprint, ' Looks NONPHOTOMETRIC.'
        gf = fltarr(n_elements(calw))+2./8
        if ctok gt 0 then gf[wok] =   4./8
        if ctok gt 0 then gf[wgood] = 8./8
       
        wsdss = where(calw.type eq 'sdss',cts)
        wlandolt = where(calw.type eq 'landolt' or calw.type eq 'sdss2', ctl)

        ;col = lonarr(n_elements(calw)) + rgb(255*gf,255*gf,255*gf)
        ;if cts gt 0 then col[wsdss] = rgb(0,255*gf[wsdss],0)
        ;if ctl gt 0 then col[wlandolt] = rgb(255*gf[wlandolt],0,0)      
        col =  lonarr(n_elements(calw))+1
        if cts gt 0 then col[wsdss] = 4+1
        if ctl gt 0 then col[wlandolt] = 2+1  ; OK is not distinguished now
        col[wgood] -= 1
               

        megaplot, calw.airmass, calw.zpt, yerr=calw.zptunc, color=col, /noerase, size=0.7


        if keyword_set(photometric) then begin
           a = [1, 2, 3]
           oplot, a, zenithzpt-(a-1)*airterm, linestyle=2

           atu = repstr(fpr(airtermunc,1.3),'*','-') ; less glaring...
           zptfileline =  clip(uconfig[i])+' '+' : '+ $
                        fpr(zenithzpt,2.3)+' ('+fpr(zptunc,1.3)+') '+ $
                        fpr(airterm,2.3)+' ('+atu+') '+rclip(n_elements(w),4)+' ['+clipzero(zptradarcsec)+'"]'

           updatefile, outzptfile, zptfileline, sep=':'
           zptfile = outzptfile ; use the written zeropoint file
       endif else begin
           a = [1, 2, 3]
           oplot, a, zenithzpt-(a-1)*airterm, linestyle=1, color=1  ;rgb(128,128,128)

       endelse
     endfor
     !p.multi = 0
     psclose
   endif else begin
    zptfile = outzptfile
   endelse

   if n_elements(zptfile) eq 0 then zptfile = 'none'


  ; Plot transmission/zeropoint versus time
  

   if file_test(zptfile) then begin
      !p.multi = 0
      psopen, workingdir+istr+'transmission.eps', xsize=7.5, ysize=9, /inches, /encaps
      !p.multi = [0,1,2]
      !p.position = 0
      device, /color
      !p.font=0
      device, /helvetica, font_index=17
      loadfiltcolors

      cols = getfiltcolors(cal.calfilt)
      !p.position = [0.09, 0.58, 0.95, 0.9] 
      zptab = ujy2mag(mag2ujy(cal.zpt,cal.calfilt),'AB')

      yrange = [min(zptab)-0.2, max(zptab)+0.2]
      xrange = [min(cal[where(cal.time ne 0)].time), max(cal.time)]  
      if yrange[0] gt yrange[1]-1.0 then yrange[0] = yrange[1]-1.5
      megaplot, cal.time, zptab, yerr=cal.zptunc, yrange=yrange, /ysty, xrange=xrange, $ 
                xtitle='Time', ytitle='Measured zeropoint (AB)', title='!17', $
                color=cols, size=0.7

      zptdata = readzpt(zptfile, count=ct)

      predzptzenith = fltarr(n_elements(cal))
      predzpt = fltarr(n_elements(cal))
      for i = 0, n_elements(cal)-1 do begin
         ;; print, cal[i].file, cal[i].time, cal[i].airmass, '   ', clip(cal[i].filt,6), cal[i].zpt, zptab[i], cal[i].zpt-predzpt[i]
         f = where(zptdata.config eq cal[i].config, ct)
         predzptzenith[i] = zptdata[f].zpt-zptdata[f].airterm*(1)
         predzpt[i] = zptdata[f].zpt-zptdata[f].airterm*(cal[i].airmass-1)   ; some sort of unit scaling needed for below
         if i gt 0 and cal[i-1].filt eq cal[i].filt and cal[i-1].time gt cal[i].time-0.1 and cal[i-1].object eq cal[i].object then $
           oplot, cal[i-1:i].time, ujy2mag(mag2ujy(predzpt[i-1:i],cal[i].calfilt),'AB'), thick=2
         filen = filenum(cal[i].file, /str)
         ;filenum = strmid((strsplit(cal[i].file, '_', /extract)) [1],0,4)
         if fix(filen) mod 10 eq 0 then xyouts, cal[i].time, yrange[1]+0.02*(yrange[1]-yrange[0]), /data, filen, orient=90, charsize=0.5
      endfor

      !p.position = [0.09, 0.08, 0.95, 0.5]
      yrange = [min(cal.zpt-predzpt)-0.1, max(cal.zpt-predzpt)+0.1 > 0]
      megaplot, cal.time, cal.zpt-predzpt, yerr=cal.zptunc, xrange=xrange, /ysty, $ 
                xtitle='Time', ytitle='-Attenuation (mag)', title='!17', $
                color=cols, size=0.7
      oplot, xrange, 0*xrange, thick=2
      for i = 0, n_elements(cal)-1 do begin
         filen = filenum(cal[i].file, /str)
         ;filenum = strmid((strsplit(cal[i].file, '_', /extract)) [1],0,4)
         if fix(filen) mod 10 eq 0 then xyouts, cal[i].time, yrange[1]+0.02*(yrange[1]-yrange[0]), /data, filen, orient=90, charsize=0.5
      endfor

      psclose
  endif




  ; Update the zeropoints of uncalibrated fields using the airmass-dependent zeropoint solved above.
  ; Only do this if it is photometric!  (Ideally, would be useful to make sure every point within half an hour
  ; in either direction is photometric using the calculations done for the plot above.)

  ; photometric = 1

  
  if file_test(zptfile) and keyword_set(photometric) then begin

     zptdata = readzpt(zptfile, count=ct)

     if ct gt 0 then begin
        if file_test('temp.cat') then file_delete, 'temp.cat', /quiet  ; no longer relevant

        catfileso = repstr(pfiles, '.fits', '.cat')  ; we want cat files, not cal files
        if keyword_set(redosolution) then $
           cazfileso = repstr(pfiles, '.fits', '.caz')  ; in case we redo the calibration
        catfiles = ['']
        for i = 0, n_elements(catfileso)-1 do if file_test(catfileso[i]) then catfiles = [catfiles, catfileso[i]]
        for i = 0, n_elements(cazfileso)-1 do if file_test(cazfileso[i]) then catfiles = [catfiles, cazfileso[i]]
        ncatuncal = n_elements(catfiles)-1
        if n_elements(catfiles) gt 1 then catfiles = catfiles[1:*]

        for f = 0, ncatuncal-1 do begin
           outfilec = repstr(catfiles[f],'.cat','.caz')    ; .caz files are calibrated via zeropoint
           if file_test(outfilec) and (keyword_set(redosolution) eq 0) then continue
           cat = readcatheaders(catfiles[f], instkey=instkey, filtkey=filtkey, filt2key=filt2key, timekey=timekey, objkey=objkey)
           ; print, catfiles[f], ' ', cat.object, ' ', cat.filt, ' ', cat.airmass, cat.seeing
           if cat.airmass lt 0.99 or cat.airmass gt 3.3 then continue
           if finite(cat.seeing) and cat.seeing*pixscale gt maxseeing then begin  
              logprint, ' '+catfiles[f]+': seeing too poor ('+clip(cat.seeing*pixscale)+'"), not attempting to calibrate this frame.'
              continue
           endif

           w = where(zptdata.config eq cat.config, ct)
           if ct eq 0 then continue
           w = w[0] ;shouldn't ever be necessary

           predictedzpt = zptdata[w].zpt - zptdata[w].airterm*(cat.airmass-1)

           logprint, '  Applying zeropoint solution to ', removepath(catfiles[f]) , ' -> ', removepath(outfilec)

           adjustzpt, catfiles[f], predictedzpt, out=outfilec, /rename
        endfor
     endif
  endif









  ; now go through and evaluate the *relative* zeropoints to make further fine adjustments if necessary.
  ; This allows for more precise scaling of images relative to each other when stacking, in particular
  ; for any fields without direct calibrations if conditions are nonphotometric.

  ; This overwrites the P60 header seeing value.  For now just skipping it.
  if instrument eq 'P60' then return
  if keyword_set(skiprelative) then return



  ;catfiles = findfile('*.ca?')
  ;catfileso = findfile(workingdir+repstr(defaultfilestr,'.fits','.ca?'))
  catfileso = repstr(files,'.fits','.ca?')
  catfiles = ['']
  for i = 0, n_elements(catfileso)-1 do begin
       match = file_search(catfileso[i])
       if match[0] eq '' then continue
       catfiles = [catfiles, match[0]]
  endfor
  if n_elements(catfiles) gt 1 then catfiles = catfiles[1:*]

  ;transfile = 'transmission.dat'
  ;if file_test(transfile) then readcol, transfile, transfiles, format='a', skip=1

           ; need some resolution to this target vs. object business

  if n_elements(catfiles) gt 1 or catfiles[0] ne '' then begin
     catinfo = readcatheaders(catfiles, count=ct, instkey=instkey, filtkey=filtkey, filt2key=filt2key, timekey=timekey, objkey=objkey)
     objects = unique(catinfo.object)
     for t = 0, n_elements(objects)-1 do begin
       ;wl = where(objects[t] eq landoltnames, ctl) ; skip landolts
       ;if ctl gt 0 then continue
       if strmid(objects[t],0,2) eq 'SA' or strmid(objects[t],0,2) eq 'PG' or strmid(objects[t],0,2) eq 's0' then continue
       wt = where(catinfo.object eq objects[t] and catinfo.type ne 'landolt', ct)
       if ct lt 1 then continue
       configs = unique(catinfo[wt].config)
       targprinted = 0
       for c = 0, n_elements(configs)-1 do begin
           wtf = where(catinfo.object eq objects[t] and catinfo.config eq configs[c], ntf)
           catinfotf = catinfo[wtf]

           outfile = repstr(catinfotf.file,'.ca','.tca') 
           checkfiles = 0
           for i = 0, ntf-1 do checkfiles += file_test(outfile[i])
           if checkfiles eq ntf then continue   ; already did the correction for all exposures of this object

           if targprinted eq 0 then logprint, ' Field:' + objects[t]
           targprinted = 1
           logprint, '  ', configs[c]

           ; if .cat's and .cal's somehow both exist, definitely use the .cal
           wpossref = where(catinfotf.type ne 'NONE' and catinfotf.nap gt 2 and catinfotf.nstar gt 2, ct)
           if ct eq 0 then wpossref = where(catinfotf.type ne 'NONE', ct)
           if ct eq 0 then wpossref = intarr(n_elements(catinfotf))

           quality = 120. / (catinfotf[wpossref].exp > 15.) + $
                     sqrt(1.0 * catinfotf[wpossref].nstar < 64.) + $  ; best possible quality is 8 + 8 + 8 = 24
                     0.08 / (catinfotf[wpossref].zptunc > 0.01)

           ;print, catinfotf[wpossref].exp
           ;print, catinfotf[wpossref].nstar 
           ;print, catinfotf[wpossref].zptunc
           ;print, 120. / (catinfotf[wpossref].exp > 15.) 
           ;print, sqrt(1.0 * catinfotf[wpossref].nstar < 64.) 
           ;print, 0.08 / (catinfotf[wpossref].zptunc > 0.01)
           ;print, quality

           ;tshortest = min(catinfotf[wpossref].exp)     ; shortest exposure
           ;unclowest = min(catinfotf[wpossref].zptunc)  ; lowest (claimed) uncertainty on the zeropoint
           ;napmost = max(catinfotf[wpossref].nap)       ; most stars used in the aperture correction
           ;if n_elements(wpossref) gt 1 then sd = stdev(catinfotf[wpossref].zptunc) else sd = 0.
           ;if sd gt 0.02 then begin  ; if there is actual variation in the quality of the zeropoint calculation
           ;    possref = (where(catinfotf[wpossref].zptunc eq unclowest)) [0]
           ;    ;if catinfotf[ref].nap lt 3 then ref = (where(catinfotf.nap eq napmost)) [0]
           ;endif else begin
           ;    ;;;if filts[f] ne 'U' then ref = (where(catinfotf.exp eq tshortest)) [0] else $
           ;    possref = (where(catinfotf[wpossref].nap eq napmost)) [0]
           ;endelse
           ;ref = wpossref[possref]

           wbest = (where(quality eq max(quality))) [0]
           ref = wpossref[wbest]
                                                   
           dzpt = fltarr(ntf)
           dzptunc = 1 + fltarr(ntf)
           changedzpt = fltarr(ntf)
           for i = 0, ntf-1 do begin
              if i eq ref then begin
                  dzpt[i] = 0.
                  logprint, '    '+removepath(catinfotf[ref].file) + ' (reference)     '+$
                               '['+clip(catinfotf[i].type,5)+'] '+ fpr(catinfotf[i].seeing*pixscale,1.2)+$
                               '"'+ fpr(catinfotf[i].exp,4.1)+'s'
              endif else begin
                  dzpt[i] = getzptoffset(catinfotf[i].file, catinfotf[ref].file, rms=rms, unc=unc, nmatch=nmatch, getref=catname)
                  dzptunc[i] = (unc > 0.002)
                  changedzpt[i] = dzpt[i] - (catinfotf[i].zpt-catinfotf[ref].zpt)
                      ; dzpt is the difference in measured magnitudes between stars in this file and the reference.  
                      ;   This should always be very close to zero for SDSS calibrated fields
                      ; catinfotf[i]-catinfotf[ref] is the difference in header zeropoints.
                      ; changedzpt is the actual difference in true zeropoints between the images (measured now + already present in headers)
                  f1 = catinfotf[i].file
                  f2 = catinfotf[ref].file
                  logprint,  '    ' + removepath(f1) + '-' + removepath(f2) + ' : ' + fpr(dzpt[i],2.3) + $
                        ' ('+fpr(dzptunc[i],1.3)+', '+rclip(nmatch,2)+')' + $
                        '  [prev. ' + fpr(catinfotf[i].zpt,2.3) + '-' + fpr(catinfotf[ref].zpt,2.3) + ' =' + $
                        fpr(catinfotf[i].zpt-catinfotf[ref].zpt,2.3) + $
                        ' ('+fpr(sqrt(catinfotf[ref].zptunc^2+catinfotf[i].zptunc^2),1.3)+'), ' + $  
                        clip(catinfotf[i].type,5) + '] ' + fpr(catinfotf[i].seeing*pixscale,1.2)+'"' + $
                        fpr(catinfotf[i].exp,4.1)+'s' 
              endelse
           endfor

           ; If there is no evidence of any nonphotometric behavior at the 2.5% level (or within the uncertainty of the comparison) then
           ; don't do any rescaling - simply copy the .cax files to the .tcax files for this target+filter.
           ; (Could consider adjusting this behavior and letting the user/pipeline decide later to use the .cax file instead of .tcax...
           ;  a downside to current system is that the reltrans is falsely claimed to be exactly zero instead of low.)

           if (max(abs(dzpt)) gt 0.025 and max(abs(dzpt/dzptunc)) gt 2.) then begin ; or max(changedzpt gt 0.01) then begin
              logprint, '    Refining zeropoint measurements...'
              for i = 0, ntf-1 do begin
                 dum = readcat(catinfotf[i].file, href)
                 refunc = sxpar(href,'ZPTUNC')
                 ;if i eq ref then 
                 if file_test(outfile[i]) then  file_delete, outfile[i], /quiet
                 if i eq ref then spawn, 'cp -f '+catinfotf[i].file+' '+outfile[i]
                 ; unclear why these files are not always present...?
                 if i ne ref and $
                    file_test(catinfotf[i].file) and file_test(catinfotf[ref].file) $
                    then begin
                        ;print, '     adjusting ', catinfotf[i].file, ' using ', catinfotf[ref].file, ' -> ', outfile[i]
                        adjustzpt, catinfotf[i].file, catinfotf[ref].file, out=outfile[i], refunc=refunc, filter=catinfotf[ref].calfilt
                             ; generates a .tcat/.tcal/.tcaz file now, which I guess is still the best thing to do.
                             ; this repeats the derivation of the zeropoint change, but that's OK.
                             ; set the CALFILT key to the same as CALFILT in the reference file
                endif
              endfor
           endif else begin
              logprint, '    No adjustment needed  (maxdzpt = ' + fpr(max(abs(dzpt)), 1.3) + ', maxdzpt/unc =' + fpr(max(abs(dzpt/dzptunc)), 2.1)+ ')'
              for i = 0, ntf-1 do begin
                  if file_test(outfile[i]) then  file_delete, outfile[i], /quiet
                  spawn, 'cp -f '+catinfotf[i].file+' '+outfile[i]  ; the -f doesn't seem to work?
              endfor
           endelse

           ; Make some calculations and update the header information, both of the .tcax file and the .fits file.
           ct = 0
           if file_test(zptfile) then zptdata = readzpt(zptfile, count=ctzpt) else ctzpt = 0
           ;if ct gt 0 then begin
           if ctzpt gt 0 then zw = where(zptdata.config eq configs[c])
           for i = 0, ntf-1 do begin
              calibrated = catinfotf[i].type ne 'NONE'
              ocatfile = catinfotf[i].file
              tcatfile = outfile[i]
              airmass = catinfotf[i].airmass

              if ctzpt gt 0 then begin
                expzpt = zptdata[zw].zpt - zptdata[zw].airterm*(airmass-1)
                if calibrated then begin
                   extinction = zptdata[zw].zpt - catinfotf[i].zpt  ; these are absolute values before the relative calibration.
                   absorption = zptdata[zw].zpt - zptdata[zw].airterm*(airmass-1) - catinfotf[i].zpt
                   transmission = 10.0^(-extinction/2.5) 
                endif
              endif



                ; update .tcax file

                x = readcat(ocatfile, hocat)
                tcat = readcat(tcatfile, h)
                if n_elements(tcat) eq 1 then continue  ; empty file...

                diffzpt = catinfotf[ref].zpt-sxpar(h,'ZEROPT')  ; this is how much the zeropoint changed in calibration
                reltrans = 10.0^(-diffzpt/2.5)
                refcatfile = removepath(catinfotf[ref].file)


              if ctzpt gt 0 then begin
                sxaddpar, h, 'EXPZPT', zptdata[zw].zpt - zptdata[zw].airterm*(airmass-1), ' Expected airmass-adjusted zeropoint'
              endif

                if calibrated and ctzpt gt 0 then begin
                     sxaddpar, h, 'EXTINCT', extinction, ' Extinction (vs. zenith) in mag'             ; due to air+clouds
                    ;sxaddpar, h, 'EXTCLOUD', absorption, 'Extinction (vs. airmass solution) in mag'    ; due to clouds only (estimate)
                     sxaddpar, h, 'TRANS', transmission, ' Transmission (vs. zenith)'
                endif
                sxaddpar, h, 'RELEXT', diffzpt, ' Rel. extinction / zeropoint adj. (vs. reference)'
                sxaddpar, h, 'RELTRANS', reltrans, ' Rel. transmission / flux adj. (vs. reference)'
                sxaddpar, h, 'RELCFILE', refcatfile, ' Reference photometry file'
                writecat, tcatfile, tcat, h
                htcat = h


                ; update fits file
                
                fitsfile = workingdir + sxpar(h,'IMFILE')
                im = readfits(fitsfile, h, /silent)

                sxaddpar, h, 'SEEING', sxpar(htcat,'SEEING'), ' Median stellar FWHM (pixels)'
                sxaddpar, h, 'SEEARCSC', sxpar(htcat,'SEEING')*pixscale, ' Median stellar FWHM (arcsec)'
                sxaddpar, h, 'PHOTCAL', sxpar(htcat,'PHOTCAL'), ' Reference system for measured zeropoint'
                sxaddpar, h, 'CALFILT', sxpar(htcat,'CALFILT'), ' Catalog filter calibrated against'
                if ctzpt gt 0 then begin
                   sxaddpar, h, 'EXPZPT', expzpt, ' Expected airmass-adjusted zeropoint'
                endif
                if calibrated then begin
                    sxaddpar, h, 'ZEROPT', sxpar(hocat,'ZEROPT'), ' Measured zeropoint'
                    sxaddpar, h, 'RZPT', sxpar(hocat,'RZPT'), ' Aperture radius for zeropoint (pixels)'        
                    sxaddpar, h, 'RZPTARCS', sxpar(hocat,'RZPT')*pixscale, ' Aperture radius for zeropoint (arcsec)'        
                    sxaddpar, h, 'RAPER', sxpar(hocat, 'RAPER'), ' Inner aperture radius'
                    sxaddpar, h, 'APCORR', sxpar(hocat,'APCORR'), ' Inner->zeropoint aperture correction (in mag)'
                    sxaddpar, h, 'APCORUNC', sxpar(hocat,'APCORUNC'), ' Uncertainty of APCORR'
                    sxaddpar, h, 'NSTARSAP', sxpar(hocat,'NSTARSAP'), ' Number of stars used to determine APCORR'
                    sxaddpar, h, 'NSTARS', sxpar(hocat,'NSTARS'), ' Number of stars used to determine ZEROPT'
                    sxaddpar, h, 'SKYMAG', sxpar(hocat,'ZEROPT') -  2.5*alog10(sxpar(h,'CTRATE')/pixscale^2), ' Calibrated sky brightness (mag/arcsec^2/s)'
                    if ctzpt gt 0 then begin
                      sxaddpar, h, 'EXTINCT', extinction, ' Extinction (mag vs. zenith)'
                     ;sxaddpar, h, 'EXTCLOUD', absorption, 'Extinction (vs. airmass solution) in mag'    ; due to clouds only
                      sxaddpar, h, 'TRANS', transmission, ' Transmission (vs. zenith)'
                    endif
                endif
                sxaddpar, h, 'RELEXT', diffzpt, ' Rel. extinction / zeropoint adj. (vs. reference)'
                sxaddpar, h, 'RELTRANS', reltrans, ' Rel. transmission / flux adj. (vs. reference)'
                sxaddpar, h, 'RELCFILE', refcatfile, ' Reference photometry file'
                logprint, '    Updated header data in '+fitsfile
                writefits, fitsfile, im, h

             endfor ; file loop
           ;endif

         endfor ; config loop

     endfor ; object loop

   endif

end

