;+
; NAME: 
;   fieldphot
;
; PURPOSE: 
;    Extract a photometric catalog from an image and return as an IDL data structure.   
;
; CALLING SEQUENCE: 
;    fieldphot, imagename, zpt1=zpt1, airterm=airterm, 
;                rinner=rinner, router=router, skyannulus=skyannulus, seeing=seeing, 
;                /relative, getseeing=getseeing, getr1=getr1, getr2=getr2, 
;                getfwhmthresh=getfwhmthresh, errthresh=errthresh, fwhmthreshfactor=fwhmthreshfactor,
;                distthresh=distthresh, getinfo=getinfo, verbose=verbose
;       
; INPUTS:
;    imagename - Filename of the image to extract.   
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;   Calibration (if specified, output catalog will be calibrated):
;    zpt1     - Effective zeropoint for this night at airmass of 1.
;    airterm  - Airmass slope of zeropoint solution (mag per unit airmass).
;   Photometric extraction keywords:
;    rinner   - Actual photometric extraction radius (otherwise calculated from SExtractor data)
;    router   - Zeropoint calculation radius
;    skyannulus - Sky annulus inner/outer radii, as a 2-element vector
;    seeing   - Atmospheric seeing range expected/permissible (2-element array)
;    /relative - Make no attempt to tie to zeropoint; just measure flux within inner radius
;    errthresh - Maximum magnitude error allowed (on inner aperture)
;    fwhmthreshfactor - Parameter signifying maximum SEx FWHM allowed compared to median
;                object to consider an object to be a useable star for calibration
;    distthresh - Maximum distance to permit an object to be from another object to be
;                useable for calibration
;    /verbose   - Print some information to the terminal
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;    getinfo   - Output structure containing various data about the extraction parameters
;    getseeing      - Seeing estimate from SExtractor data, also in getinfo
;    getr1         - Currently does nothing (see getinfo for new method)
;    getr2         - Currently does nothing
;    getfwhmthresh - Currently does nothing
;       
; OUTPUTS: 
;
; Last modified 4 Sep 2020.  Documentation added 3 July 2022.
;-


function fieldphot, imagename, zpt1=zpt1, airtermi=airtermi, rinner=rinner, router=router, skyannulus=skyannulus, seeing=seeing, relative=relative, getseeing=getseeing, getr1=getr1, getr2=getr2, getfwhmthresh=getfwhmthresh, errthresh=errthresh, fwhmthreshfactor=fwhmthreshfactor, distthresh=distthresh, getinfo=getinfo, verbose=verbose

; r1 is the used aperture.
; r2 is a large aperture that can be treated as infinite.

;   common lrisconfig, lrisautopath, refdatapath, defaultspath
common lrisauto, lpar

if n_elements(lpar) gt 0 then begin
   if file_test('temp.param') eq 0 then spawn, 'cp '+lpar.defaultspath+'/temp.param ./'
   if file_test('sex.config') eq 0 then spawn, 'cp '+lpar.defaultspath+'/sex.config ./'
   if file_test('sex.conv') eq 0 then spawn, 'cp '+lpar.defaultspath+'/sex.conv ./'
endif
; note, temp.param seems to be needed to avoid error messages even outside LRIS processing.

if n_elements(router) eq 0 then relative=1
if n_elements(airtermi) gt 0 then begin
  if airtermi lt 0 then airterm = abs(airtermi[0]) else airterm = airtermi[0]
endif else begin
  airterm = 0
endelse

if n_elements(errthresh) eq 0 then errthresh = 0.1
if n_elements(fwhmthreshfactor) eq 0 then fwhmthreshfactor = 1.3


data = mrdfits(imagename,0,h,/silent)
filter = sxpar(h,'FILTER',count=c)
if c eq 0 then filter = sxpar(h,'FILTNAM',count=c)
if c eq 0 then filter = sxpar(h,'FILTER1',count=c)
airmass = sxpar(h,'AIRMASS',count=c)
if c eq 0 then airmass = 1
exptime = sxpar(h,'EXPTIME',count=c)
if c eq 0 then exptime = sxpar(h,'EXPOSURE',count=c)
if c eq 0 then begin
  logprint, '  Warning - No exposure time keyword found!'
  logprint, '  Setting to 1 second.'
  exptimeunk = 1
  exptime = 1.
endif else begin
  exptimeunk = 0
endelse
if exptime le 0 then logprint, '  WARNING: Exposure time is zero in header!'
saturate = sxpar(h,'SATURATE',count=c)
if c eq 0 then saturate = 55000

if n_elements(seeing) eq 0 then seeing=[1.25,100]

sexpath = getsexpath()
spawn, sexpath+' '+imagename+' -c sex.config'

check = file_test('temp.cat')
if check eq 0 then begin
   logprint, '  Sextractor failed!'
   return, -1
endif

readcol, 'temp.cat', x, y, ra, dec, sexmag, sexmagerr, ellip, fwhm, /silent
if n_Elements(x) eq 0 then begin
   logprint, '  Sextractor returned empty file'
   return, -1
endif
sex = replicate({x:0., y:0., ra:0., dec:0., mag:0., magerr:0., ellip:0., fwhm:0.}, n_elements(x))
sex.x = x
sex.y = y
sex.ra = ra
sex.dec = dec
sex.mag = sexmag
sex.magerr = sexmagerr
sex.ellip = ellip
sex.fwhm = fwhm
delvarx, x  ; can remove these later
delvarx, y
delvarx, ra
delvarx, dec
delvarx, mag
delvarx, magerr
delvarx, ellip
delvarx, fwhm

sex = sex[where(sex.fwhm gt 0.5)] ; remove some bad pixels/cosmics - arguably should keep them for mindist, though

mindist = nndist(sex.x,sex.y) ; do this before edge check

; remove everything near the edges immediately
nx = sxpar(h, 'NAXIS1')
ny = sxpar(h, 'NAXIS2')
if n_elements(router) gt 0 then if finite(router) then edgebuff = router*0.75 
if n_elements(edgebuff) eq 0 then edgebuff = 25.
notedge = where(sex.x gt edgebuff and sex.y gt edgebuff and sex.x lt nx-edgebuff and sex.y lt ny-edgebuff, ct)
if ct eq 0 then begin
   logprint, '  No objects in image except near edges!'
   return, -1
endif
sex = sex[notedge]
mindist = mindist[notedge]

;megaplot, fwhm, ellip, xrange=[0,20], size=10.0^(-(25+sexmag-16.)/10)

printsex = 0
if printsex then begin 
   for i = 0, n_elements(sex)-1 do print, sex[i].x, sex[i].y, sex[i].ra, sex[i].dec, sex[i].mag, sex[i].magerr, sex[i].ellip, sex[i].fwhm, mindist[i]
endif

;; skypix = sxpar(h,'COUNTS',count=c)  ; it may have been subtracted off, which will cause this routine to fail.
;;if c eq 0 then 
skypix = median(data)
maxpix = maxpixel(imagename, sex.x, sex.y, r=2)

; Identify the bright (probable) stars and get the mean stellar FWHM (seeing) of the image.
; This is only to determine PSF characteristics and is independent of the set we will use later.

;satval = 48000 < saturate
satval = saturate

wsat = where(maxpix gt satval, complement=wnotsat, ctsat, ncomplement=ctnotsat)
if ctnotsat eq 0 then begin
   logprint, '  No unsaturated stars?  Check sky/saturation values...'
   return, -1
endif
wnearmax = where(maxpix gt max(maxpix[wnotsat])-1000 and maxpix lt satval, ctnearmax) ; the check vs. satval isn't necessary
if (max(maxpix[wnotsat]) gt satval - 10000) and (ctnearmax gt 2) and ctnearmax gt ctsat + 2 then begin
    logprint, '  WARNING - Many stars are detected with a similar, large flux level ('+$
            clip(ctnearmax)+' between '+clip(long(max(maxpix)-1000))+' and '+clip(satval)+$
           ') but below the given saturation value ('+clip(satval)+').'
    logprint, '  Check the saturation value.'
endif

ct = 0
;abovesky = 15000
abovesky = (satval-skypix)/2.51^1.5  ; begin with stars within 1.5 mag of saturation
;skyrms = sqrt(skypix) ; crude and doesn't work on subtracted images!!!
skyrms =  sigclipstdev(data, maxiter=2)
desiredct = 7
if exptimeunk eq 0 and exptime lt 10. then desiredct -= 2
if skypix gt 20000 then desiredct -= 1
while ct lt desiredct do begin
   if abovesky le 0 then break
   pstars = where(sex.fwhm gt seeing[0] and sex.fwhm lt seeing[1]*airmass and maxpix lt satval and maxpix gt abovesky+skypix, ct)
   ;print,  '   Found '+clip(ct)+' stars above threshold '+clip(fix(abovesky+skypix))+' ('+clip(abovesky)+' above sky);  '+clip(desiredct)+' desired'
   if ct lt desiredct and abovesky gt 1600 then abovesky = abovesky - 1000
   if ct lt desiredct and abovesky le 1600 then abovesky = abovesky - 100
   if abovesky lt 8*skyrms then break
endwhile
if ct eq 0 then begin
   logprint, '  Found no stars bright enough to extract.'
   return, -1
endif
majaxis = sex.fwhm * (1+sex.ellip)
majaxispstars = majaxis[pstars]
rankedmajaxispstars = majaxispstars[sort(majaxispstars)]
baxisthresh = rankedmajaxispstars[n_elements(majaxispstars)/4]*1.1+1 < airmass*seeing[1] ; 25th percentile

;min5starfwhm = min(sex[pstars].fwhm)
;medstarfwhm = median([sex[pstars].fwhm], /even)
;bfwhmthresh = medstarfwhm*1.2 < airmass*seeing[1] ; always use factor 1.2 here
medstarfwhm = median([sex[pstars[where(majaxispstars le baxisthresh)]].fwhm], /even)

; getseeing = medstarfwhm
; could do the others too, but no need - watch out for this however

; Use 1.2*FWHM for the aperture radius if not specified externally
if n_elements(rinner) eq 0 then r1 = medstarfwhm*1.2 else r1 = rinner
if keyword_set(relative) then begin
  r2 = r1
endif else begin
  if n_elements(router) gt 0 then r2 = router else r2 = 4*medstarfwhm
endelse

if keyword_set(verbose) then print, '   Using aperture radius ', clip(r1)

if n_elements(distthresh) eq 0 then distthresh = 3.0*r1


rcheck1 = medstarfwhm*1  ; formerly 1 and 1.5    
rcheck2 = medstarfwhm*1.5

if n_elements(minval) eq 0 then begin
  skylevel = sxpar(h,'COUNTS',count=ct)
  if ct eq 0 then skylevel = median(data)
  minval = skylevel - 10000. 
  ; formerly used a flat value of -200, not appropriate for back-subtracted images.
endif


; Actually do the aperture photometry; correct for exposure time

if n_elements(skyannulus) eq 0 then skyannulus = [40, 50]
if n_elements(skyannulus) eq 1 then skyannulus = [skyannulus[0], skyannulus[0]+10]

aper, data, sex.x, sex.y, apmagarr, apmagerrarr, sky, skyerr, 1.0, [r1,r2,rcheck1,rcheck2], skyannulus, [minval,satval], /silent
; add an adjustment by 25 mag here?
apmagarr = apmagarr + 2.5*alog10(exptime) ; adjust for exposure time!!!


debugmode = 0

; Identify the stars

dmfloor = 0.01 ; some danger here if the fwhm-estimated aperture is way too large or small.  formerly 0.03

dm = (apmagarr[2,*] - apmagarr[3,*]) [*]; flux difference between two apertures
stars = where(majaxis lt baxisthresh and dm gt dmfloor, ctstars)
if debugmode then print, ctstars, ' stars '
if ctstars eq 0 then return, {x:!values.f_nan, y:!values.f_nan, ra:!values.f_nan, dec:!values.f_nan, mag:!values.f_nan, magerr:!values.f_nan}
bstars = where(majaxis lt baxisthresh and dm gt dmfloor and maxpix ge median(maxpix[stars], /even), nbstars )
if debugmode then print, nbstars, ' bstars'
dmthresh = (median(dm[bstars],/even) * [0.7, 1.1]) + [-0.2, 0.1] ; this was formerly 0.9 and -0.2, then 0.8 and -0.2, on the low end
dmthresh[0] = dmthresh[0] > 0.01 ; some danger here if the fwhm-estimated aperture is way too large or small - formerly 0.04
ellipthresh = median([sex[bstars].ellip],/even)*1.05 + 0.05 ; this is new and not carefully tested.

apmag1 = (apmagarr[0,*]) [*]
apmag2 = (apmagarr[1,*]) [*]
apmagerr1 = (apmagerrarr[0,*]) [*]
apmagerr2 = (apmagerrarr[1,*]) [*]

isgood1mag = apmag1 gt -50 and apmag1 lt 50 and dm ge 0
isgood2mag = apmag2 gt -50 and apmag2 lt 50 and dm ge 0

if debugmode then begin
  ord = sort(apmag2)
  for ii = 0, n_elements(sex)-1 do begin
     i = ord[ii]
     if apmagerr2[i] gt 0.07 then continue
     ispstar = total(i eq pstars)
     isstar = total(i eq stars)
     isbstar = total(i eq bstars)
     starstr = (ispstar ? 'p':' ') + (isstar ? 's':' ') + (isbstar ? 'b':'')
     print, fpr(sex[i].x,5.2),' ',fpr(sex[i].y,5.2),':  ',fpr(apmag1[i],3.2),' ',fpr(apmagerr1[i],2.2),'   ',fpr(apmag2[i],3.2),' ',fpr(apmagerr2[i],2.2),'   ',fpr(dm[i],2.2), ' ', fpr(mindist[i],4.1), ' ', rclip(long(maxpix[i]),6),  ' ', fpr(sex[i].fwhm,3.2),' ', fpr(sex[i].ellip,2.2), ' ', clip(isgood2mag[i]), ' ', starstr
  endfor
endif



if keyword_set(relative) eq 0 then begin
 
  ; Absolute photometry - determine the "infinity"(r2)-to-r1 aperture correction

  maxord = maxpix[sort(maxpix)]
  max3 = maxord[2 < n_elements(maxpix)-1]

  brightthresh = 0.05
  abovesky = (satval-skypix)/2.51 < max3 ; begin with stars within 1 mag of saturation OR max of 3rd brightest star
  minsep = r2+3*r1

  ct = 0
  desiredct = 7
  if sxpar(h,'EXPTIME') gt 0 and sxpar(h,'EXPTIME') lt 10. then desiredct -= 2

  if debugmode eq 1 then begin
    print, 'peakthresh0:  ', skypix+10+abovesky
    print, 'brightthresh: ', brightthresh
    print, 'minsep:       ', minsep
    print, 'baxisthresh:  ', baxisthresh
    print, 'ellipthresh:  ', ellipthresh
    print, 'dmthresh:     ', dmthresh[0], dmthresh[1]
  endif

  iter = -1
  while ct lt desiredct do begin
     iter += 1

     if abovesky gt (10*skyrms > 60) then $
        abovesky /= 1.09648 ; steps of 0.1 mag

     brightstars = where(maxpix gt skypix+10+abovesky and $
                         apmagerr2 lt brightthresh and $
                         mindist gt minsep and $
                         majaxis lt baxisthresh and $
                         sex.ellip lt ellipthresh and $
                         dm gt dmthresh[0] and dm lt dmthresh[1] and $
                         isgood2mag, ct)

     ; algorithm should ideally respond differently when the problem is low S-N (U-band, short exposure) versus crowding (dense fields).

     if debugmode then print, long(skypix+10+abovesky), ' ', fpr(brightthresh,1.2), ' ', fpr(minsep,4.1), ' ', fpr(baxisthresh,2.2), ' ', fpr(ellipthresh,2.2),' ', fpr(dmthresh[0],2.2),' ', fpr(dmthresh[1],2.2),' ', rclip(ct,3), '/', clip(desiredct)
 
     if iter eq 10 then desiredct -= 1  ;6 at 2 mag below
     if iter eq 12 then brightthresh += 0.02
     if iter eq 15 then desiredct -= 1  ;5 at 2.5 mag below
     if iter eq 17 then minsep -= r1/2.
     if iter eq 30 then desiredct -= 1  ;4 at 3.5 mag below
     if iter eq 22 then minsep -= r1/2.
     if iter eq 32 then minsep -= r1/2.
     if iter eq 45 then desiredct -= 1  ;3 at 4.5 mag below

     if ct eq n_elements(maxpix) then break   ; obviously no point in continuing if these are all the stars

     if abovesky lt (10*skyrms > 60) and iter gt 45 then break  ; only should get here for very shallow images (U-band, clouds...?)

  endwhile
 
  if ct le 3 then begin
     if abovesky lt (10*skyrms > 60) then logprint, '   Reached sky noise threshold.'
     if ct eq n_elements(maxpix) then logprint, '   Reached all bstars.'
  endif
  if ct ge 1 and ct le 3 then begin
      logprint, '   WARNING - few isolated bright stars ('+clip(ct)+') found.'
  endif
  if ct eq 0 then begin
      logprint, '   ERROR - found no isolated bright stars.'
      ;stop
      return, -1
  endif

  getseeing = median([sex[brightstars].fwhm], /even)

  apcorr = median([apmag1[brightstars]-apmag2[brightstars]], /even)
  
  for i = 0, n_elements(brightstars)-1 do begin
     bi = brightstars[i]
     
     if keyword_set(verbose) then $
     print, fpr(sex[bi].x,4.2),',',fpr(sex[bi].y,4.2),':  m1=', fpr(apmag1[bi],2.3)+'('+fpr(apmagerr1[bi],1.3)+')  m2=', fpr(apmag2[bi],2.3)+'('+fpr(apmagerr2[bi],1.3)+')   dm=',fpr(apmag1[bi]-apmag2[bi],2.3), '   resid=',fpr((apmag1[bi]-apmag2[bi])-apcorr,2.3), '  fwhm=',fpr(sex[bi].fwhm,2.2), '  ell=',fpr(sex[bi].ellip,1.2)
  endfor
  
  if ct gt 1 then begin
    apunc  = stdev(apmag1[brightstars]-apmag2[brightstars]) ; just get RMS.  assuming systematics-dominated uncertainties.
                                                  ; in general this estimate is going to be dominated if there are any outliers.
    apdev  = median(abs(apmag1[brightstars]-apmag2[brightstars] - apcorr),/even)
  endif else begin
    apunc = (apcorr*2 > 0.5) ; only a single star can happen on Landolt fields, I guess... this situation is sketchy.
    apdev = (apcorr*2 > 0.5)
  endelse
  napstars = n_elements(brightstars)

  logprint, '   Aperture correction:',fpr(apcorr,2.3), ' +/-', fpr(apunc,2.3), '(', fpr(apdev,1.3), '):  ', clip(napstars), ' stars'
endif else begin

  ; Relative photometry - no aperture correction

  apcorr = 0.
  apunc = 0.
  apdev = 0.
  napstars = 0
endelse


; Apply the aperture correction
apmag1corr = apmag1 - apcorr

if n_elements(zpt1) gt 0 then begin
  zpteff = zpt1[0] - airterm*(airmass-1) 
endif else begin
  zpteff = 0    ; make default 25 instead of 0 for new convention
endelse

; this prints out the summary, should allow a filenumber argument
;printf, 3, fpr(medstarfwhm,2.3), ' ', fpr(r1,2.3), ' ', fpr(dmthresh[0],2.3), ' ', fpr(dmthresh[1],1.3), ' | ', fpr(zpteff,2.3), ' ', fpr(apcorr,2.3), ' ', fpr(apunc,1.3), ' ', fpr(apdev,1.3), ' ', clip(napstars,3)

fwhmthresh = fwhmthreshfactor*medstarfwhm

calstars = where(apmagerr1 lt errthresh and mindist gt distthresh and sex.fwhm lt fwhmthresh and dm gt dmthresh[0] and dm lt dmthresh[1] and isgood2mag, ct)

if debugmode then begin
  print, '                ', fpr(errthresh,1.2),' ',distthresh, fwhmthresh, dmthresh[0], dmthresh[1]
  ord = sort(sex.y)
  for i = 0, n_elements(sex)-1 do begin
     c = ord[i]

     ispstar = total(c eq pstars)
     isstar = total(c eq stars)
     isbstar = total(c eq bstars)
     starstr = (ispstar ? 'p':' ') + (isstar ? 's':' ') + (isbstar ? 'b':'')

    ;if x[c] lt 455 or x[c] gt 475 or y[c] lt 750 or y[c] gt 760 then ff = '' else ff = '*'
    if apmagerr1[c] lt 0.1 then $
    print, fpr(sex[c].x,4.1),' ',fpr(sex[c].y,4.1),':  ', fpr(apmag1[c],3.2), ' ', fpr(apmag2[c],3.2), ' ', $
         fpr(apmagerr1[c],2.2),' ',fpr(mindist[c],3.1),' ',fpr(sex[c].fwhm,2.2),' ',fpr(majaxis[c],2.2),' ',fpr(dm[c],2.2), ' | ', $
                         clip(maxpix[c] gt skypix+10+abovesky)+$
                         clip(apmagerr2[c] lt brightthresh)+$
                         clip(mindist[c] gt minsep)+$
                         clip(majaxis[c] lt baxisthresh)+$
                         clip(sex[c].ellip lt ellipthresh)+$
                         clip(dm[c] gt dmthresh[0])+clip(dm[c] lt dmthresh[1])+$
                         clip(isgood2mag[c])
        ; (apmagerr1[c] lt errthresh), (mindist[c] gt distthresh), (sex[c].fwhm lt fwhmthresh), (dm[c] gt dmthresh[0]),(dm[c] lt dmthresh[1]),(isgood2mag[c]), starstr  ;, ' ', ff
  endfor
endif

phot = replicate({x:0., y:0., ra:0.D, dec:0.D, mag:0., magerr:0.}, n_elements(calstars))

phot.x = sex[calstars].x
phot.y = sex[calstars].y
phot.ra = sex[calstars].ra
phot.dec = sex[calstars].dec
phot.mag = apmag1corr[calstars] + zpteff
phot.magerr = apmagerr1[calstars]

info = {seeing:0., r1:0., r2:0., apcorr:0., apunc:0., zpteff:0., zpt1:0., airterm:0., fwhmthresh:0., fwhmthreshfactor:0.,  distthresh:0., apdev:0., minval:0., satval:0., skyr:0., skywidth:0., napstars:0, nstars:0}
  info.seeing = medstarfwhm
  info.apcorr = apcorr
  info.apunc = apunc
  info.zpteff = zpteff
  if n_elements(zpt1) gt 0 then info.zpt1 = zpt1
  info.airterm = airterm
  info.distthresh = distthresh
  info.fwhmthresh = fwhmthresh
  info.fwhmthreshfactor = fwhmthreshfactor
  info.apdev = apdev
  info.minval = minval
  info.satval = satval
  info.napstars = napstars
  info.nstars = n_elements(phot)
  info.skyr = skyannulus[0]
  info.skywidth = skyannulus[1]-skyannulus[0]
  info.r1 = r1
  info.r2 = r2
getinfo = info

return, phot

end

