;+
; NAME: 
;    findlines
;
; PURPOSE: 
;    Identify line centroid positions given a 1D spectrum.
;
; CALLING SEQUENCE: 
;    linex = findlines(f, fbaksub, linestrength=linestrength, centroidrad=centroidrad, 
;                      threshold=threshold, /debugplot, xrange=xrange)
;       
; INPUTS:
;    f - Input spectrum (usually an arc or sky spectrum)
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;    centroidrad  - Half-width (in pixels) of region used to calculate line centroids.  Default 5.
;    threshold    - Sigma level above the continuum level to detect a line.  Default 4.
;    /debugplot   - Produce some diagnostic plots
;    xrange       - Specify a limited range of x pixels over which to detect lines as [xmin,xmax]
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;    linestrength - Strengths of detected lines (paired with linex)
;       
; OUTPUTS: 
;    linex   - Centroid coordinates of detected lines.
;    fbaksub - Contains the continuum-subtracted high-pass filtered spectrum.
;
; Last modified 19 Jun 2018.  Documentation added 3 July 2022.
;-

function findlines, f, fbaksub, linestrength=linestrength, centroidrad=centroidrad, threshold=threshold, $
   debugplot=debugplot, xrange=xrange

   ; Input a 1D spectrum (usually an arc or sky); returns a list of line locations (in pixels)
   ; returns a continuum-subtracted spectrum, if desired
   ; strengths are returned in linestrength

   ; could be improved - absorption lines (solar calcium, usually) cause fake detections on the edges

      if n_elements(centroidrad) eq 0 then cr = 5 else cr = centroidrad
      if n_elements(threshold) eq 0 then threshold = 4.0 ; (it's really about 1.5x this because of convolution)

      forig = f  ; isn't altered
      farc = f   ; is altered

      nx = n_elements(farc)

      if keyword_set(debugplot) eq 0 then debugplot = 0
      ;if debugplot gt 0 then window, 0, xsize=1400, ysize=700
      x = indgen(nx) ; only for plotting
      if n_elements(xrange) ge 2 then maxy = max(f[where(x gt xrange[0] and x lt xrange[1])]) else maxy = max(f)
      if debugplot eq 1 then plot, forig, /xsty, /ysty, yrange=[-10,maxy], xrange=xrange

      ; Estimate the "continuum"
  
      ; A median filter misses the edges!!!  Need to expand, then contract, the array via reflection...
      fcont = medianfilter(farc, 301)

      ; Remove strong emission lines so they don't affect the continuum estimate
      farc_noe = farc
      emlines = where(farc gt fcont + 1000, ctem) ; pixels affected by strong emission lines ; dangerous hard coded y value here
      if ctem gt 0 then begin
        isem = intarr(n_elements(farc))
        isem[emlines] = 1
        isem = convol(isem,[1,1,1,1,1]) > 0  ; grow by +/- 2
        emlines = where(isem)
        farc_noe[emlines] = !values.f_nan
      endif
      if debugplot eq 1 and ctem gt 0 then oplot, (indgen(n_elements(farc))) [emlines], farc[emlines], col=rgb(200,255,200)
      fcont = medianfilter(farc_noe, 301) ; new median removing emission line pixels
      
      if debugplot eq 1 then oplot, fcont, col=rgb(0,255,0)

      ; Remove strong absorption lines
      abslines = where(fcont gt 150 and farc lt fcont*0.8 and farc lt fcont-150, ctabs)
      if ctabs gt 0 then begin
         isabs = intarr(n_elements(farc))
         isabs[abslines] = 1 
         isabs = convol(isabs,[1,1,1,1,1]) > 0   ; grow by +/- 2
         if debugplot eq 1 then oplot, (indgen(n_elements(farc))) [where(isabs)], farc[where(isabs)], psym=1, col=rgb(255,0,0)
         abslines = where(isabs)
         farc_noe[abslines] = !values.f_nan
      endif
      fcont = medianfilter(farc_noe, 101) ; new median filter removing both emission and absorption

      ; Do a median subtraction to remove the continuum  
      farcsub = farc - fcont

      if debugplot eq 1 then oplot, fcont, col=rgb(255,255,0)


      ; Further continuum subtraction.  Probably not necessary now?
      ;for iter = 0, 3 do begin
      ;   bakkernel =  [0.5,1,1,1,1,1,0.5]
      ;   bakkernel = bakkernel / total(bakkernel)
      ;   fbakconv = convol(farcsub, bakkernel, /edge_truncate)
      ;   for x = 0, nx-1 do begin
      ;      farcsub[x] = farcsub[x] - min(fbakconv[x-30>1:x+30<nx-1]) ; formerly 15, which seemed to create fake lines
      ;   endfor
      ;endfor

      if debugplot ge 2 then plot, farcsub > 1, /xsty, /ysty, /ylog, xrange=xrange
      if debugplot ge 2 then oplot, forig, col=rgb(50,50,255)

      arckernel = [0.2,0.4,0.9,1,0.9,0.4,0.2]
      arckernel = arckernel / total(arckernel)
      arcconv = convol(farcsub, arckernel, /edge_truncate)
      ; identify lines

      diff = abs(farcsub - shift(farcsub,-1))
      readnoise = sigclipmedian(diff)
      noise = sqrt((forig>0) + readnoise^2) ; need a gain adjustment here.

      arcthreshold = noise*threshold
      ;print, 'arc threshold ', arcthreshold

      if debugplot eq 2 then oplot, arcthreshold, col=rgb(0,128,0)
      if debugplot eq 2 then oplot, arcconv, col=rgb(255,255,0)

      farcx = findgen(n_elements(farcsub))
      linex = fltarr(1000) ; up to 1000 lines...
      linestrength = fltarr(1000)
      l = 0
      for x = cr, nx-cr-1 do begin
         if arcconv[x] lt arcthreshold[x] then continue
         if arcconv[x] gt arcconv[x-1] and arcconv[x] gt arcconv[x+1] then begin

            w = x-cr+indgen(2*cr)
            w = w[where(finite(farcsub[w]))]
            w = w[where(finite(farcx[w]))]
            farcsubwpos = farcsub[w] > 0
            if max(farcsubwpos) eq 0 then farcsubwpos = farcsub[w]-min(farcwub[w])
            linex[l] = total(farcsubwpos * farcx[w])/total(farcsubwpos)
            linestrength[l] = arcconv[x]

            ;print, x, linex[l], arcconv[x]

            improvedlinex = getlinecenter(forig, linex[l], fitrad=cr)  ; seems to produce better results,
            ; print, x, linex[l], improvedlinex                          ; but monitor this in future

            if abs(linex[l]-improvedlinex) lt 0.7 then linex[l] = improvedlinex
                ; if larger, is probably a blend or something and gaussian fit not reliable
                ; we should actually do a new iteration to look for blends and refit those as doubles

            l = l + 1

         endif
      endfor
      if l ge 1 then linex = linex[0:l-1]   else linex = -1
      if l ge 1 then linestrength = linestrength[0:l-1]   else linestrength = -1

      if debugplot eq 2 then begin
        for i = 0, n_elements(linex)-1 do oplot, [linex[i], linex[i]], [1,1e6],             col=rgb(50,50,50)
        for i = 0, n_elements(linex)-1 do oplot, [linex[i], linex[i]], [1,linestrength[i]], col=rgb(200,200,200)
        for i = 0, n_elements(linex)-1 do oplot, [linex[i]], [linestrength[i]],psym=5, col=rgb(255,0,0)

        oplot, arcconv, col=rgb(255,255,0)
      endif


      farcsub = farcsub > 1e-5 ; for plotting purposes...

  fbaksub = farcsub
  return, linex

  ; might be neat to calculate prominence versus the higher saddle elevation on either side

end




