;+
; NAME: flatproc.pro
;
; PURPOSE:
;   Divide a science image by a flat-field to produce a flat-corrected image. 
;       
; CALLING SEQUENCE: 
;   flatproc, filename, flatname, crop=crop, flatminval=flatminval, flatmaxval=flatmaxval, 
;             prefix=prefix, outname=outname, /quiet
;    
; INPUTS:
;   filename - Filename(s) to process.  [Currently does not support wildcards]
;   flatname - Filename of flat-field to use for processing.
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;   crop - Option to specify whether to crop the image after flat-fielding.  Can be an
;          array of four values (xmin ymin xmax ymax) or a string specifying the automatic
;          cropping method ('auto' for fully automatic or 'autoy' for automatic y only)
;   flatminval - Minimum flat value to treat as valid.  Set to NaN if flat counts too low.
;   flatmaxval - Maximum flat value to treat as valid.  Set to NaN if flat counts too low.
;   prefix - String to prepend to the filename(s) to produce the output file(s). 
;            Default 'f'
;   outname - Filename(s) for output.  Created from prefix+filename if unspecified.
;   /quiet -  Suppress output to console
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS:  none
;       
; OUTPUTS: 
;   Output image written to disk.
;
; REQUIREMENTS: 
;
; Last modified 10 Apr 2020.  Documentation added 2 July 2022.
;-

pro flatproc, filename, flatname, crop=crop, flatminval=flatminval, flatmaxval=flatmaxval, prefix=prefix, outname=outname, quiet=quiet

; there seems to be a bug when using a filename array.

; now corrects for different windowings (LTV); soon correct for different binnings (LTM)

if n_elements(filename) eq 1 then begin
  filearr = strsplit(filename,'.', /extract)
  fileroot = filearr[0]
  fileext = filearr[1]

  if fileext eq 'cat' or fileext eq 'lis' or fileext eq 'list' or fileext eq 'txt' then begin
    filenames = grabcolumn(filename,0,/str)
  endif else  begin
    filenames = filename
  endelse
endif else begin
  filenames = filename
endelse

if n_elements(prefix) eq 0 then prefix = 'f'

nfile = n_elements(filenames)

flat = mrdfits(flatname, 0, hflat, /silent)

flattype = sxpar(hflat,'FLATTYPE', count=cttype)
med = median(flat)
if med lt 0.5 or med gt 2.0 then begin
  logprint, ' Warning: flat is not normalized to zero'
endif

for f = 0, nfile-1 do begin
   data = mrdfits(filenames[f],0,h, /silent)
   bzero = sxpar(h,'BZERO',count=ct)
   if ct gt 0 then if float(bzero) ne 0 then begin
      data += bzero
      sxaddpar, h,'BZERO',0
   endif
   if n_elements(data) le 1 then begin
      logprint, ' '+filenames[f], ' is a blank file!'
      continue
   endif
   if n_elements(data) ne n_elements(flat) then begin
      if n_elements(quiet) eq 0 then logprint, ' WARNING - Data and flat field are different dimensions!!'
         ; runs, but doesn't work; check this later (is this still true????????)
      dataoriginx = -sxpar(h,'LTV1')
      dataoriginy = -sxpar(h,'LTV2')
      flatoriginx = -sxpar(hflat,'LTV1')
      flatoriginy = -sxpar(hflat,'LTV2')
      nxdata = (size(data)) [1]
      nydata = (size(data)) [2]
      nxflat = (size(flat)) [1]
      nyflat = (size(flat)) [2]

      ;print, 'Science image: ', clip(nxdata),'x',clip(nydata), ' [0,0]=[', clipzero(dataoriginx),',',clipzero(dataoriginy),']'
      ;print, 'Flat field:    ', clip(nxflat),'x',clip(nyflat), ' [0,0]=[', clipzero(flatoriginx),',',clipzero(flatoriginy),']'
      ;print, '               Science [0,0] = Flat [',clipzero(dataoriginx-flatoriginx),',',clipzero(dataoriginy-flatoriginy),']'

      pflat = data*0. - !values.f_nan
      xdata = lindgen(n_elements(data)) mod nxdata  ; x value of every pixel (science data)
      ydata = lindgen(n_elements(data)) / nxdata    ; y value of every pixel
      xflat = xdata + dataoriginx-flatoriginx       ; equivalent x pixels on the flat corresponding to (xdata,ydata)
      yflat = ydata + dataoriginy-flatoriginy       ; equivalent y pixels on the flat corresponding to (xdata,ydata)
      ;print, xdata[0], ydata[0], '=', xflat[0], yflat[0]

      noflatpix = where(xflat lt 0 or xflat ge nxflat or yflat lt 0 or yflat ge nyflat, ct, complement=flatpix)
      ; stop   ; cautiously  not stopping
      if n_elements(flatpix) eq 0 then begin
        logprint, ' WARNING - Flat processing failed - no overlap!'
      endif
      pflat[xdata[flatpix],ydata[flatpix]] = flat[xflat[flatpix],yflat[flatpix]]
      if ct gt 0 then begin
        logprint, ' WARNING - Flat does not fully overlap science image!  Cannot flatfield all pixels.'
        pflat[xdata[noflatpix],ydata[noflatpix]] = !values.f_nan  ; formerly 1
      endif
      if total(finite(pflat)) eq 0 then begin
        logprint, ' WARNING - Flat processing failed - no overlap!'
        return
      endif
   endif else begin
      pflat = flat
   endelse

   ; determine crop zones
   ;  this should not be in the loop?
   if n_elements(crop) eq 1 and size(crop, /tname) eq 'STRING' then begin
      if crop eq 'auto' or crop eq 'autoy' then findcroplines, pflat, xcrop, ycrop, thresh=0.2
      if      crop eq 'auto' then crop = [xcrop[0], ycrop[0], xcrop[1], ycrop[1]] $
      else if crop eq 'autoy' then crop = [0, ycrop[0], ((size(fdata)) [1])-1, ycrop[1]]
   endif

   ; set values too low/high to NaNs
   if n_elements(flatminval) gt 0 then begin
      w = where(pflat lt flatminval, ct)
      goodsignal = where(pflat ge flatminval and finite(pflat), ctg)
      if ct gt 0 then pflat[w] = -!values.f_nan    ; low-response regions are now marked by NEGATIVE nans.
      if ctg eq 0 then logprint, '  WARNING - No good pixels in flat', filename[f]
   endif else begin
      w = where(pflat lt 0.1 or finite(pflat) eq 0, complement=goodsignal) ; for sky count determination only
   endelse
   if n_elements(flatmaxval) gt 0 then begin
      w = where(pflat gt flatmaxval, ct)
      if ct gt 0 then pflat[w] = -!values.f_nan   ; Nans can complicate life, though useful as fix markers.
   endif
 
   ;plot, data[*,1157]
   ;oplot, pflat[*,1157], color=rgb(150,150,150)

   fdata = data / pflat 

   ;oplot, fdata[*,1157], color=rgb(255,0,0)


   if n_elements(crop) gt 0 then begin
      if total(crop eq 0) ne n_elements(crop) then begin
        s = size(fdata)
        if crop[2] le 0 then crop[2] = s[1] + crop[2]-1  ; can use negatives to count from right/top edges
        if crop[3] le 0 then crop[3] = s[2] + crop[3]-1
        fdata = fdata[crop[0]:crop[2],crop[1]:crop[3]]
        sxaddpar, h, 'LTV1', sxpar(h,'LTV1')-crop[0]
        sxaddpar, h, 'LTV2', sxpar(h,'LTV2')-crop[1]
        sxaddpar, h, 'CRPIX1', sxpar(h,'CRPIX1')-crop[0]
        sxaddpar, h, 'CRPIX2', sxpar(h,'CRPIX2')-crop[1]
        sxaddpar, h, 'CROPX0', crop[0]
        sxaddpar, h, 'CROPX1', crop[2]
        sxaddpar, h, 'CROPY0', crop[1]
        sxaddpar, h, 'CROPY1', crop[3]
        endif
   endif
   sxaddpar, h, 'FLATFILE', removepath(flatname), ' Flat field file used'
   if cttype gt 0 then sxaddpar, h, 'FLATTYPE', flattype, ' Type of flatfield'
   skycts = median(fdata[goodsignal])
   ;sxaddpar, h, 'COUNTS', skycts, 'Median counts after flatfielding'
   sxaddpar, h, 'SKYCTS', skycts, 'Median counts after flatfielding'
   if sxpar(h,'EXPTIME') gt 0 then $
      sxaddpar, h, 'CTRATE', skycts/sxpar(h,'EXPTIME'), after='SKYCTS', 'Counts per second'

   ; get_date, now
   hist = 'Processed by flatproc '+tstamp(/ut)+' UT'
   sxaddhist, hist, h

   
   if n_elements(outname) eq 0 then begin
      lastslashpos = strpos(filenames[f],'/',/reverse_search)
      if lastslashpos gt 0 then $
        outfilename = strmid(filenames[f],0,lastslashpos) + '/' + prefix + strmid(filenames[f],lastslashpos+1) $
      else  $
        outfilename = prefix+filenames[f]
   endif else begin
      outfilename = outname
   endelse

   if n_elements(quiet) eq 0 then logprint, filenames[f]+' / '+flatname+' -> '+outfilename

   mwrfits, fdata, outfilename, h, /create
endfor

end
