;+
; NAME: getlandolt.pro
;
; PURPOSE: 
;       Query the Landolt (2009) equatorial standard star catalog by position.
;
; CALLING SEQUENCE: 
;       c = getlandolt(ra, dec, radius, count=count)
;
; INPUTS:
;       ra  = Right ascension of the search center
;       dec = Declination of the search center
;       radius = Search radius in arcseconds
;
; OPTIONAL INPUT KEYWORD PARAMETERS: none
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS: 
;       count = The number of stars returned.  (Will be zero for most arbitrary coordinates.)
; 
; OUTPUTS: 
;      A structure containing the magnitude of all Landolt standards within the specified
;      search radius.
;      Contains the measured UBVRI magnitudes and uncertainties, as well as transformed 
;      magnitudes in the SDSS filters.
;      Note that a lazy distance calculation is used, but this shouldn't matter as there
;      are no Landolt stars at the poles or across the meridian in any case.
;
; REQUIREMENTS: 
;      The GSFC IDLAstro library is used to convert sexagesimal to decimal.
;
; Last modified 8 Feb 2013.  Documentation added 2 July 2022.
;
;-

function getlandolt, rai, deci, radius, count=count

  ; gets landolt stars around a position.
  ; (as opposed to readlandolt, which gets all of them.)

  ra = rai
  dec = deci

  if size(ra, /tn) eq 'STRING' then ra = ten(ra)*15.
  if size(dec, /tn) eq 'STRING' then dec = ten(dec)

  landolt = readlandolt()
  dist = fltarr(n_elements(landolt))

  cosdec = cos(!pi*dec/180.)
  distance = fltarr(n_elements(landolt))
  for l = 0, n_elements(landolt)-1 do begin
      dist = 3600*( abs(ra-landolt[l].ra)*cosdec + abs(dec-landolt[l].dec))  ; approximate fine for now
      if dist lt 3600. then begin
         gcirc, 2, ra, dec, landolt[l].ra, landolt[l].dec, dist
      endif
      distance[l] = dist
  endfor

  w = where(distance lt radius, count)

  if count eq 0 then return, -1

  return, landolt[w]

end

