;+
; NAME: 
;     getlinecenter
;
; PURPOSE: 
;     Fit a Gaussian profile to a line to determine its center.  
;
; CALLING SEQUENCE: 
;     c = getlinecenter(pixin, fluxin, xin, fitrad=fitrad, unc=unc, width=width, 
;                       peak=peak, bkpolyorder=bkpolyorder,
;                       pix=pix, flux=flux, ffit=ffit)
;       
; INPUTS:
;     pixin  - Array of pixel values (optional; array coordinates used if unspecified)
;     fluxin - Array of fluxes
;     xin    - Estimated central location of the line to fit
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;     fitrad - Half-width of fitting region (around x) to use to determine centroid
;     bkpolyorder - Polynomial order of the background underneath the line to use in the
;                   fitting model.  Default is zero.
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;     unc   - Estimated uncertainty on the centroid
;     width - Estimated line width (Gaussian sigma parameter)
;     peak  - Estimated line peak height
;     pix   - Contains the array of pixels used for the fit  
;     flux  - Contains the flux array used for the fit
;     ffit  - Contains the fit model
;       
; OUTPUTS: 
;     c     - Estimate of the line center pixel position.
;
; Last modified 8 May 2018.  Documentation added 3 July 2022.
;-

function getlinecenter, pixin, fluxin, xin, fitrad=fitrad, unc=unc, width=width, peak=peak, bkpolyorder=bkpolyorder, $
      pix=pix, flux=flux, ffit=ffit ; outputs for checking
      if n_elements(xin) eq 0 then begin
          flux = pixin
          x = fluxin
          pix = findgen(n_elements(flux))
      endif else begin
          pix = pixin
          flux = fluxin
          x = xin
      endelse

      if n_elements(bkpolyorder) eq 0 then bkpolyorder = 0
      if finite(x) eq 0 then begin
        print, 'x cannot be NaN'
        return, !values.f_nan
      endif   

      zone = where(pix ge x-fitrad and pix le x+fitrad, ct)
      if ct eq 0 then begin
         return, !values.f_nan
      endif
      if ct lt 3 then begin
         return, total(pix*(flux > 0))/total(flux > 0) ; simple centroid
      endif
      pix = pix[zone]
      flux = flux[zone]

      ; old mpfit method
      ; parinfo = replicate({value:0.0, fixed:0, limited:[0,0], limits:[0.,0.]}, 4)
      ; parinfo[0].value = max(flux)
      ; parinfo[1].value = x
      ; parinfo[2].value = fitrad/3.
      ; parinfo[3].value = 0            ; I turned background fitting off; why?
      ;  parinfo[3].fixed = 1
      ;
      ; sol = mpfitfun('gaussian',  pix, flux, sqrt((flux + 10.) > 10), parinfo=parinfo,$
      ;               perr=perr, bestnorm=chisq, dof=dof, yfit=yfit, /quiet)
      

      ; estimates = [max(flux), x, fitrad/3.]

       q = !quiet
       !quiet = 1   ; hide the maximum-iterations errors that occasionally appear [should look into why they appear...]
       ; sol = gaussfit(pix, flux, measure_errors=sqrt((flux + 10.) > 10), nterms=3, sigma=perr) ; old version surely did damage
       ffit = gaussfit(pix, flux, sol, measure_errors=sqrt((flux + 10.) > 10), nterms=3+bkpolyorder, sigma=perr)
       !quiet = q

       unc = perr[1]     

     peak = sol[0]
     width = sol[2]
     return, sol[1]

end

