;+
; NAME: getps1.pro
;
; PURPOSE: 
;       Query the Pan-STARRS catalog by position.
;
; CALLING SEQUENCE: 
;       c = getps1(ra, dec, radius, count=count, url=url, pauseduration=pauseduration,
;                  magtype=magtype)
;
; INPUTS:
;       ra  = Right ascension of the search center
;       dec = Declination of the search center
;       radius = Search radius in arcseconds
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       pauseduration : Currently does not do anything
;       magtype = Type of PS1 mag to look up.  Can be: 'Ap','PSF','Kron' (default Ap)
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS: 
;       count = The number of stars returned
;       url = Returns the URL used to perform the lookup query.
; 
; OUTPUTS: 
;      A structure containing the magnitudes of all PS1 objects within the specified
;      search radius.
;
; REQUIREMENTS: 
;      The GSFC IDLAstro library is used to convert sexagesimal to decimal.
;
; Last modified 3 Feb 2020.  Documentation added 2 July 2022.
;
; Todo:
;   Could use a generalization to support more of the STScI catalogs.
;-

function getps1, ra, dec, radius, url=url, count=count, pauseduration=pauseduration, magtype=magtype ;, catalog=catalog

   ; PS1 cone search.  Returns a structure of magnitudes, including magnitudes transformed to Johnson-Cousins filters.

   ; radius/boxsize is in arcsec

   catalog = 'ps1'
   if n_elements(pauseduration) eq 0 then pauseduration = 60.

   rainp = clip(string(ra))
   decinp = clip(string(dec))
   radinp = clip(radius/3600.) ; ps1 direct query only

   if float(radinp) gt 0.1 then radinp = '0.1' ; Beginning September 2018 radii > 0.1 caused the query to inexplicably fail.

   if strpos(rainp,':') gt 0 then rainp = clip(15*ten(ra))
   if strpos(decinp,':') gt 0 then decinp = clip(ten(dec))

    queryurl = '"http://gsss.stsci.edu/webservices/vo/CatalogSearch.aspx?CAT=PS1V3OBJECTS&RA='+rainp+'&DEC='+decinp+'&SR='+radinp+'&FORMAT=CSV&MAGRANGE=0,25"'

   ;# a rectangular option is also available: BBOX
   ;# can also search 2MASS, ps1, GALEX, WISE this way...!

   url = queryurl

   if file_test('tempcat.dat') then file_delete, 'tempcat.dat' ;spawn, 'rm -f tempcat.dat'
   command = 'wget ' + queryurl + ' -O tempcat.dat -q'
   spawn, command ; sometimes hangs here.

   if n_elements(magtype) eq 0 then magtype = 'Ap' ; PSF, Ap, Kron

   ns = -1
   openr, 15, 'tempcat.dat'
   inline = ''
   while not eof(15) do begin
     readf, 15, inline
     c1 = strmid(inline,0,1)
     if c1 eq '#' then continue
     if n_elements(colnames) eq 0 then begin ; first non-commented line is column header row
        colnames = strsplit(inline,',',/extract)
        wra = (where(colnames eq 'raMean' or colnames eq 'RaMean')) [0]
        wdec = (where(colnames eq 'decMean' or colnames eq 'DecMean')) [0]
        wgmag = (where(colnames eq 'gMean'+magtype+'Mag')) [0]
        wrmag = (where(colnames eq 'rMean'+magtype+'Mag')) [0]
        wimag = (where(colnames eq 'iMean'+magtype+'Mag')) [0]
        wzmag = (where(colnames eq 'zMean'+magtype+'Mag')) [0]
        wymag = (where(colnames eq 'yMean'+magtype+'Mag')) [0]
        wgmagunc = (where(colnames eq 'gMean'+magtype+'MagErr')) [0]
        wrmagunc = (where(colnames eq 'rMean'+magtype+'MagErr')) [0]
        wimagunc = (where(colnames eq 'iMean'+magtype+'MagErr')) [0]
        wzmagunc = (where(colnames eq 'zMean'+magtype+'MagErr')) [0]
        wymagunc = (where(colnames eq 'yMean'+magtype+'MagErr')) [0]
     endif     
     ns += 1
   endwhile
   close, 15

   if ns le 0 then return, -1
   if wra eq -1 or wdec eq -1 then return, -1  

   ps1 = replicate({id:0LL, sra:'',sdec:'',ra:0.D, dec:0.D, $
                     g:0.,r:0.,i:0.,z:0.,y:0.,                     Bj:0.,Vj:0.,Rc:0.,Ic:0., $
                     g_unc:0.,r_unc:0.,i_unc:0.,z_unc:0.,y_unc:0., Bj_unc:0.,Vj_unc:0.,Rc_unc:0.,Ic_unc:0., $
                     type:0},ns)

   openr, 15, 'tempcat.dat'
   i = -1
   while not eof(15) do begin
     readf, 15, inline
     inline = clip(inline)
     c1 = strmid(inline,0,1)
     if c1 eq '#' then continue
     if i ge 0 then begin
       indata = strsplit(inline,',', /extract)
       ; indata[where(indata eq '999')] = 'NaN'
       ps1[i].ra = indata[wra]
       ps1[i].dec = indata[wdec]
       ps1[i].sra = sixtystr(ps1[i].ra, /ra)
       ps1[i].sdec = sixtystr(ps1[i].dec)
       ps1[i].g = indata[wgmag] ; converts from string to float here
       ps1[i].r = indata[wrmag]
       ps1[i].i = indata[wimag]
       ps1[i].z = indata[wzmag]
       ps1[i].y = indata[wymag]
       ps1[i].g_unc = indata[wgmagunc]
       ps1[i].r_unc = indata[wrmagunc]
       ps1[i].i_unc = indata[wimagunc]
       ps1[i].z_unc = indata[wzmagunc]
       ps1[i].y_unc = indata[wymagunc]
       ; ps1.type =
     endif
     i += 1
   endwhile 
   close, 15

   ; ps1.Uj =  u - 0.0216*(u - g) - 0.799 ; combination of 2 equations, fortunately not needed

   ; alternate conversion schemes exist for B/V, not usable without u
   ps1.Bj = (ps1.g + 0.3130*(ps1.g - ps1.r) + 0.2271);    sigma = 0.0107

   ps1.Vj = (ps1.g - 0.5784*(ps1.g - ps1.r) - 0.0038);    sigma = 0.0054

   ps1.Rc = (ps1.r - 0.1837*(ps1.g - ps1.r) - 0.0971)/2+$;  sigma = 0.0106
            (ps1.r - 0.2936*(ps1.r - ps1.i) - 0.1439)/2;    sigma = 0.0072

   ps1.Ic = (ps1.r - 1.2444*(ps1.r - ps1.i) - 0.3820)*1+$;  sigma = 0.0078
            (ps1.i - 0.3780*(ps1.i - ps1.z) - 0.3974)*0;    sigma = 0.0063

   ; ps1.Uj_unc =  ps1.u_unc

   ps1.Bj_unc = sqrt((ps1.g_unc * 1.3130)^2 + (ps1.r_unc*0.2271)^2 + 0.0107^2)

   ps1.Vj_unc = sqrt((ps1.g_unc*(1-0.5784))^2 + (ps1.r_unc*0.5784)^2 + 0.0054^2)

   ps1.Rc_unc = sqrt(ps1.r_unc^2 + 0.0106^2)

   ps1.Ic_unc = sqrt((ps1.r_unc*0.2444)^2 + (ps1.i_unc*1.2444)^2 + 0.0078^2)




   return, ps1

end
