;+
; NAME: getsdss2.pro
;
; PURPOSE: 
;       Query the SDSS calibration telescope bright star catalog by position.
;       (This is different from the all-sky SDSS catalog: it is a vastly smaller number of
;        stars, measured at higher precision.)
;
; CALLING SEQUENCE: 
;       c = getsdss2(ra, dec, radius, count=count, liverpool=liverpool)
;
; INPUTS:
;       ra  = Right ascension of the search center
;       dec = Declination of the search center
;       radius = Search radius in arcseconds
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       liverpool : Set to tranform the magnitudes to the Liverpool Telescope system.
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS: 
;       count = The number of stars returned.  (Will be zero for most arbitrary coordinates.)
; 
; OUTPUTS: 
;      A structure containing the magnitudes of SDSS secondary standards within the search
;      radius.
;
; REQUIREMENTS: 
;      The GSFC IDLAstro library is used to convert sexagesimal to decimal.
;
; Last modified 17 Aug 2019.  Documentation added 2 July 2022.
;-


function getsdss2, rai, deci, radius, count=count, liverpool=liverpool

  ; gets secondary SDSS standards

  ra = rai
  dec = deci

  if size(ra, /tn) eq 'STRING' then ra = ten(ra)*15.
  if size(dec, /tn) eq 'STRING' then dec = ten(dec)

  sdss2 = readsdss2(liverpool=liverpool) ; reads everything
  dist = fltarr(n_elements(sdss2))

  cosdec = cos(!pi*dec/180.)
  distance = fltarr(n_elements(sdss2))
  for l = 0, n_elements(sdss2)-1 do begin
      dist = 3600*( abs(ra-sdss2[l].ra)*cosdec + abs(dec-sdss2[l].dec))  ; approximate fine for now
      if dist lt 3600. then begin
         gcirc, 2, ra, dec, sdss2[l].ra, sdss2[l].dec, dist
      endif
      distance[l] = dist
  endfor

  w = where(distance lt radius, count)
  if count eq 0 then return, -1

  return, sdss2[w]

end

