;+
; NAME: getskymapper.pro
;
; PURPOSE: 
;       Query the Skymapper catalog by position.
;
; CALLING SEQUENCE: 
;       c = getskymapper(ra, dec, radius, count=count, url=url, pauseduration=pauseduration,
;                   magtype=magtype)
;
; INPUTS:
;       ra  = Right ascension of the search center
;       dec = Declination of the search center
;       radius = Search radius in arcseconds
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       pauseduration : Currently does not do anything
;       magtype = Type of mag to look up.  Can be: 'petro', 'Ap','PSF','Kron' (default petro)
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS: 
;       count = The number of stars returned.
;       url = Returns the URL used to perform the lookup query.
; 
; OUTPUTS: 
;      A structure containing the magnitudes of all Skymapper objects within the specified
;      search radius.  Transformation equations (from SDSS) are applied to also estimate
;      magnitudes in the BVRI system.
;
; REQUIREMENTS: 
;      The GSFC IDLAstro library is used to convert sexagesimal to decimal.
;
; Last modified 24 Sep 2018.  Documentation added 2 July 2022.
;-

function getskymapper, ra, dec, radius, url=url, count=count, pauseduration=pauseduration, magtype=magtype ;, catalog=catalog

   ; skymapper cone search.  Returns a structure of magnitudes, including magnitudes transformed to Johnson-Cousins filters.

   ; radius/boxsize is in arcsec

   catalog = 'skymapper'
   if n_elements(pauseduration) eq 0 then pauseduration = 60.

   rainp = clip(string(ra))
   decinp = clip(string(dec))
   radinp = clip(radius/3600.) ; convert from arcsec to degrees

   if float(radinp) gt 1 then radinp = '1' 

   if strpos(rainp,':') gt 0 then rainp = clip(15*ten(ra))
   if strpos(decinp,':') gt 0 then decinp = clip(ten(dec))

   queryurl = '"http://skymapper.anu.edu.au/sm-cone/query?RA='+rainp+'&DEC='+decinp+'&SR='+radinp+'&RESPONSEFORMAT=CSV"'

   url = queryurl

   if file_test('tempcat.dat') then file_delete, 'tempcat.dat'
   command = 'wget ' + queryurl + ' -O tempcat.dat -q'
   spawn, command

   if n_elements(magtype) eq 0 then magtype = 'petro' ; PSF, Ap, Kron

   ns = -1
   openr, 15, 'tempcat.dat'
   inline = ''
   while not eof(15) do begin
     readf, 15, inline
     c1 = strmid(inline,0,1)
     if c1 eq '#' then continue
     if n_elements(colnames) eq 0 then begin ; first non-commented line is column header row
        colnames = strsplit(inline,',',/extract)
        wra = (where(colnames eq 'raj2000')) [0]
        wdec = (where(colnames eq 'dej2000')) [0]
        wstar = (where(colnames eq 'class_star')) [0]
        wumag = (where(colnames eq 'u_'+magtype)) [0]
        wvmag = (where(colnames eq 'v_'+magtype)) [0]
        wgmag = (where(colnames eq 'g_'+magtype)) [0]
        wrmag = (where(colnames eq 'r_'+magtype)) [0]
        wimag = (where(colnames eq 'i_'+magtype)) [0]
        wzmag = (where(colnames eq 'z_'+magtype)) [0]
        wumagunc = (where(colnames eq 'e_u_'+magtype)) [0]
        wvmagunc = (where(colnames eq 'e_v_'+magtype)) [0]
        wgmagunc = (where(colnames eq 'e_g_'+magtype)) [0]
        wrmagunc = (where(colnames eq 'e_r_'+magtype)) [0]
        wimagunc = (where(colnames eq 'e_i_'+magtype)) [0]
        wzmagunc = (where(colnames eq 'e_z_'+magtype)) [0]
     endif     
     ns += 1
   endwhile
   close, 15

   if ns le 0 then return, -1
   if wra eq -1 or wdec eq -1 then return, -1  

   sm = replicate({id:0LL, sra:'',sdec:'',ra:0.D, dec:0.D, $
                     u:0.,g:0.,r:0.,i:0.,z:0.,                     Bj:0.,Vj:0.,Rc:0.,Ic:0., $
                     u_unc:0.,g_unc:0.,r_unc:0.,i_unc:0.,z_unc:0., Bj_unc:0.,Vj_unc:0.,Rc_unc:0.,Ic_unc:0., $
                     star:0.},ns)

   openr, 15, 'tempcat.dat'
   i = -1
   while not eof(15) do begin
     readf, 15, inline
     inline = clip(inline)
     c1 = strmid(inline,0,1)
     if c1 eq '#' then continue
     if i ge 0 then begin
       indata = strsplit(inline,',', /extract,/preserve_null)
       ; indata[where(indata eq '999')] = 'NaN'
       sm[i].ra = indata[wra]
       sm[i].dec = indata[wdec]
       sm[i].sra = sixtystr(sm[i].ra, /ra)
       sm[i].sdec = sixtystr(sm[i].dec)
       sm[i].u = indata[wumag]
       sm[i].g = indata[wgmag] ; converts from string to float here
       sm[i].r = indata[wrmag]
       sm[i].i = indata[wimag]
       sm[i].z = indata[wzmag]
       sm[i].u_unc = indata[wumagunc]
       sm[i].g_unc = indata[wgmagunc]
       sm[i].r_unc = indata[wrmagunc]
       sm[i].i_unc = indata[wimagunc]
       sm[i].z_unc = indata[wzmagunc]
       sm[i].star = indata[wstar]
     endif
     i += 1
   endwhile 
   close, 15

   ; sm.Uj =  u - 0.0216*(u - g) - 0.799 ; combination of 2 equations, fortunately not needed

   wgr = where(sm.g gt 0 and sm.g lt 30 and sm.r gt 0 and sm.r lt 30, ctgr)
   wgri= where(sm.g gt 0 and sm.g lt 30 and sm.r gt 0 and sm.r lt 30 and sm.i gt 0 and sm.i lt 30, ctgri)
   wri = where(sm.r gt 0 and sm.r lt 30 and sm.i gt 0 and sm.i lt 30, ctri)

   ; alternate conversion schemes exist for B/V!
   if ctgr gt 0 then $
   sm[wgr].Bj = (sm[wgr].g + 0.3130*(sm[wgr].g - sm[wgr].r) + 0.2271);    sigma = 0.0107

   if ctgr gt 0 then $
   sm[wgr].Vj = (sm[wgr].g - 0.5784*(sm[wgr].g - sm[wgr].r) - 0.0038);    sigma = 0.0054

   if ctgr gt 0 then $
   sm[wgr].Rc = (sm[wgr].r - 0.1837*(sm[wgr].g - sm[wgr].r) - 0.0971);  sigma = 0.0106
   if ctri gt 0 then $
   sm[wri].Rc = (sm[wri].r - 0.2936*(sm[wri].r - sm[wri].i) - 0.1439);    sigma = 0.0072
   if ctgri gt 0 then $
   sm[wgri].Rc = (sm[wgri].r - 0.1837*(sm[wgri].g - sm[wgri].r) - 0.0971)/2+$;  sigma = 0.0106
                 (sm[wgri].r - 0.2936*(sm[wgri].r - sm[wgri].i) - 0.1439)/2;    sigma = 0.0072

   if ctri gt 0 then $
   sm[wri].Ic = (sm[wri].r - 1.2444*(sm[wri].r - sm[wri].i) - 0.3820)*1+$;  sigma = 0.0078
                (sm[wri].i - 0.3780*(sm[wri].i - sm[wri].z) - 0.3974)*0;    sigma = 0.0063

   ; sm.Uj_unc =  sm.u_unc

   sm[wgr].Bj_unc = sqrt((sm[wgr].g_unc * 1.3130)^2 + (sm[wgr].r_unc*0.2271)^2 + 0.0107^2)

   sm[wgr].Vj_unc = sqrt((sm[wgr].g_unc*(1-0.5784))^2 + (sm[wgr].r_unc*0.5784)^2 + 0.0054^2)

   sm.Rc_unc = sqrt(sm.r_unc^2 + 0.0106^2)

   sm[wri].Ic_unc = sqrt((sm[wri].r_unc*0.2444)^2 + (sm[wri].i_unc*1.2444)^2 + 0.0078^2)




   return, sm

end
