;+
; NAME: 
;      lookupstandardlines
;
; PURPOSE: 
;      A routine to obtain the wavelengths of various contaminating lines in commonly used
;      spectroscopic standard stars to aid with excising these wavelength regions from
;      response function modeling. 
;
; CALLING SEQUENCE: 
;      lookupstandardlines, stdname, lines, linev, linesnt, linevnt
;       
; INPUTS:
;      stdname - The name of the standard.  Must be the exact name used in the database.
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;      lines   - List of lines, including tellurics.
;      linev   - List of line relative widths ([half-width]/[central wavelength])
;      linesnt - List of lines without tellurics.
;      linevnt - List of line relative widths without tellurics.
;       
; OUTPUTS:  none (see above)
;
; Last modified 18 Sep 2019.  Documentation added 3 July 2022.
;-

pro lookupstandardlines, stdname, lines, linev, linesnt, linevnt

         ; get the list of lines and widths for a standard star.

       ;  hydrogen = [6562.80, 4861.34, 4340.48, 4101.75, 3970.09, 3889.07, $
       ;              3835.40, 3797.92, 3770.65, 3750.17, 3734.39, 3721.96, $
       ;              3711.99, 3703.88, 3697.17, 3691.58, 3686.85, 3682.83]
       ;  hv = 0.003 + (hydrogen gt 3900)*0.007
       ;  calcium = [3968.5, 3933.7]
       ;  cav = 0.003 + fltarr(n_elements(calcium))
       ;  telluric = [6890., 7280., 7670., 8150., 9060., 9480.]
       ;  tellv = [0.008, 0.018, 0.011, 0.0045, 0.014, 0.022]
       ;  helium = [4685., 5411.]
       ;  hev =   [0.004, 0.004]

         hydrogen = [6562.80, 4861.34, 4340.48, 4101.75, 3970.09, 3889.07, $
                     3835.40, 3797.92, 3770.65, 3750.17, 3734.39, 3721.96, $
                     3711.99, 3703.88, 3697.17, 3691.58, 3686.85, 3682.83] ; Balmer
         hv =       [0.006,   0.006,   0.0065,   0.007,   0.0065,   0.0048, $
                     0.004,   0.003,   replicate(0.003,10)]

         calcium = [3968.5, 3933.7]
         cav = 0.003 + fltarr(n_elements(calcium))

         helium = [4685., 5411.]
         hev =   [0.004, 0.004]

         paschen = [1.87561,1.28216,1.09411,1.00521,0.95486,0.92315,0.90174,0.88652,$
                    0.87529,0.86674,0.86008,0.85477,0.85048,0.84696,0.84403]*10000*0.9995
         paschenv = replicate(0.0025, n_elements(paschen))
               ;6863-6940
               ;7167-7361 (weak)
               ;7588-7749
               ;8131-8180 (weak)
               ;8950-9177
               ;9276-9684
         telluric = [6900., 7280., 7656., 8150., 9060., 9480.]
         tellv = [0.0065, 0.018, 0.01, 0.0045, 0.014, 0.022]

         ; Default set of lines for stars without a specific linelist below
         linesnt = [hydrogen, calcium, helium]
         linevnt = [hv, cav, hev]

            z1 = 1.

         ; Or, if it's a standard we know and love, hard-code some custom lines instead
         if stdname eq 'G191-B2B' then begin
            linesnt = [hydrogen[0:5], 6279.5, 8228.]   ; No helium or calcium, higher-order hydrogen weak
            linevnt = [hv[0:5]+0.002, 0.0015, 0.0005]
         endif
         if stdname eq 'BD+28_4211' then begin
            linesnt = [[hydrogen[0:6]], helium, calcium[0], 6279.5, 4542.]
            linevnt = [[hv[0:6]],       hev,    cav[0],     0.0015, 0.0015]
         endif
         if stdname eq 'Feige34' then begin
            linesnt = [[hydrogen[0:8]], helium, calcium[0], 6279.5, 4542.]
            linevnt = [[hv[0:8]],       hev,    cav[0],     0.0012, 0.0015]
         endif
         if stdname eq 'Feige67' then begin
            linesnt = [hydrogen[0:7], helium, 6279.5,  4542.]
            linevnt = [hv[0:7],       hev,    0.0012, 0.0015]
         endif
         if stdname eq 'Feige110' then begin
            linesnt = [hydrogen[0:7], helium, 5876, 6279.5,  4472.5,   4542., 8228.]
            linevnt = [hv[0:7]      , hev,    0.0009, 0.0012, 0.0013, 0.004, 0.0005]
         endif
         if stdname eq 'HZ44' then begin
            linesnt = [hydrogen[0:5], calcium, helium, 3819.4, 4026.0, 4388.8, 4471.6, 4922.0, 5875., 6677., 7065.]
            linevnt = [hv[0:5], cav, hev,               0.002, 0.002,  0.002,  0.002,  0.002, 0.0025, 0.003, 0.002]
         endif
         if stdname eq 'BD+33_2642' then begin ; a really terrible standard
            linesnt = [hydrogen[0:11], 5876, 6279.5, 6677., 7064., 7772., 8230., paschen, $
                      4923, 4473, 4390, 4145, 4123, 4010, 4028, 3820]
            linevnt = [hv[0:11]-0.0025 > 0.0025,    0.0009,  0.0012, 0.001, 0.001, 0.001, 0.001, paschenv, $ 
                      0.0015, 0.0015, 0.0015, 0.0015, 0.0015, 0.0015, 0.0015, 0.0015]
         endif 
         if stdname eq 'BD+17_4708' then begin
            ; RV=291.31, i.e. 0.999 wav shift
            linesnt = [[hydrogen[0:15]]*0.9985, calcium*0.9985, 5169, $ 
                       5890, 6279.5, 8229,  8490, 8535, 8654, 8742, 8853, 9219,10039] ; last few are paschen
            linevnt = [[hv[0:15]],             cav,             0.0025, $
                       0.0017, 0.0012, 0.001, 0.0008, 0.0012, 0.0015, 0.0012, 0.0012, 0.001, 0.001]
         endif 
         if stdname eq 'BD+26_2606' then begin
            linesnt = [hydrogen[0:11]*1.0001, calcium*1.0001, 5893, 6279.5,   8230., paschen[1:12]*1.0001, $
                      5328.5,5270, 5228, 5184, 5170, 4958, 4923, 4390, 3820]
            linevnt = [hv[0:11],  cav-0.001,  0.0013,  0.0012, 0.001, paschenv[1:12]-0.001, $ 
                      0.0005,0.0005,0.0005,0.0005, 0.0013, 0.0008, 0.0015, 0.0015, 0.0015]
                      ;-0.001 > 0.0025
         endif 
         if stdname eq 'Feige66' then begin ; seems to be a pretty good standard
            linesnt = [[hydrogen[0:6]], 5876, 6279.5, 7064, $
                      4542.] ; no blue data to use with checking
            linevnt = [[hv[0:6]],       0.0016, 0.0015, 0.0008, $
                      0.0015]
         endif 
         if stdname eq 'Hiltner102' then begin
            z1 = 1.0002
            linesnt = [hydrogen[0:11]*z1, helium*z1, calcium*z1, paschen*z1,  $
                      5876, 6279.5, 6613.5, 6678., 6993, 7065.,  8230., $
                      5016, 4923, 4713, 4648, 4473, 4390, 4145, 4123, 4028, 3820, 4428]
            linevnt = [hv[0:11], hev*0+0.0016, cav*0+0.0015, paschenv,$
                      0.0009,  0.0012, 0.001, 0.001, 0.0006, 0.001, 0.001,  $ 
                      0.001, 0.0015, 0.001, 0.0022, 0.0015, 0.0015, 0.0015, 0.0015, 0.0015, 0.0015, 0.0035]
                      ;-0.0025 > 0.0025
            ; only explored out to 8500 Ang
         endif 
         if stdname eq 'BD+75_325' then begin
            z1=1-0.00006
            linesnt = [hydrogen[0:5]*z1,helium*z1, $
                      3819.4, 4026.0, 4200,  4471.6, 4540, $
                      5875., 6279., 6684., 7065., 8240]
            linevnt = [hv[0:5]*z1,      hev*z1,    $
                      0.002, 0.002, 0.0065,  0.002, 0.0065, $
                      0.002, 0.001, 0.0033, 0.0013, 0.004]
         endif 
         if stdname eq 'Hiltner600' then begin
            linesnt = [hydrogen[0:13]*z1, calcium*z1, paschen[0:12]*z1,  $
                      5876, 6279.5, 6613.5, 6678., 6993, 7065.,  8230., 8781, $
                      5047, 5016, 5003, 4923, 4713, 4648, 4471, 4388, 4143, 4123, 4026, 4008, 3820, 4428]
            linevnt = [hv[0:13],  cav*0+0.001, paschenv[0:12]+0.001,$
                      0.0009,  0.0012, 0.001, 0.001, 0.0006, 0.001, 0.001, 0.0012, $ 
                      0.001, 0.001, 0.0009, 0.0015, 0.001, 0.0022, 0.0015, 0.0015, 0.0015, 0.0015, 0.0018, 0.0012,0.0015, 0.0035]
                      ;-0.0025 > 0.0025
         endif 
         if stdname eq 'BD+25_4655' then begin ; not to be confused with BD+25_3941
            z1 = 1. ; this is a really terrible standard
            linesnt = [hydrogen[0:3]*z1, helium*z1, $
                       3590,3615,3636,3708,3820,3870,3890,3967, 4713, 4027, 4145, 4390, 4474, 4923.,5004,5016, 5047, $
                      5875,  6279., 6678, 7065, 8230.]
            linevnt = [[hv[0:3]],        hev,      $
                       0.0025,0.0015,0.005,0.005,0.007,0.002,0.002,0.003, 0.0025,0.003,0.0035, 0.003, 0.005, 0.0035,0.001,0.0018,0.0018, $
                      0.004, 0.0013, 0.003,0.0016, 0.0017]  
         endif 
         if stdname eq 'LTT3218' then begin
            linesnt = [hydrogen[0:10]*z1, $
                      8229]
            linevnt = [hv[0:10]+[0,0.01,0.006,0.0045,replicate(0.,7)], $
                      0.001]
         endif 
         if stdname eq 'Wolf1346' then begin
            linesnt = [hydrogen[0:6]*z1, $
                      6281,8229]
            linevnt = [hv[0:6]+[0.01,0.01,0.01,0.005,0,0,0,0], $
                      0.001,0.001]
         endif 
         if stdname eq 'HD19445' then begin
            z1 = 0.9995 ; SIMBAD says it's blueshifted
            linesnt = [[hydrogen[0:15]]*z1, calcium*z1, $
                       4222.5, 4303, 4917, 4954, 5171, 5223.5, 5266.5, 5323, $ 
                       5890, 6279.5, 8229,  8493, 8537, 8656, 8800, 8857, 9221,10043] ; last few are paschen
            linevnt = [[hv[0:15]],             cav,     $
                       0.0015, 0.0025, 0.0014, 0.001, 0.0025, 0.001, 0.001, 0.001, $
                       0.0017, 0.0012, 0.001, 0.0008, 0.0012, 0.0015, 0.0012, 0.0009, 0.001, 0.001]
            ; -0.002 > 0.003
         endif 
         if stdname eq 'HD84937' then begin
            z1 = 0.9997
            linesnt = [[hydrogen[0:15]]*z1, calcium*z1, $
                       4222.5,  4921, 4957, 5175, 5226, 5268, 5327, $ 
                       5892, 6281, 8229,  8493, 8540, 8660, 8748, 8860, 9225,10046] ; last few are paschen
            linevnt = [[hv[0:15]],             cav,     $
                       0.0015,  0.0014, 0.001, 0.0025, 0.001, 0.001, 0.001, $
                       0.0017, 0.0012, 0.001, 0.0008, 0.0012, 0.0015,0.0015,  0.0013, 0.001, 0.001]
             ; -0.002 > 0.003
         endif 

;BD+28') ge 0) or (strpos(names,'Feige34') ge 0) or $
;                   (strpos(names,'Feige67') ge 0) or (strpos(names,'Feige110

         ; Always add telluric lines (which sadly can't be avoided)
         lines = [linesnt, telluric]
         linev = [linevnt, tellv]

end

; Popularity of standards:
;       2  GD153   
;       2  HZ15    
;       2  Feige98 
;       2  G158-100
;       2  G93-48  
;       2  GD50    
;       4  LTT6248 
;       4  Feige56 
;       5  LDS749B 
;       6  GD71    
;       6  Kopff27 
;       7  HD93521 
;       7  PG0823+546
;      12  Feige66 
;      12  Hiltner102
;      16  BD+75   
;      18  Hiltner600
;      20  BD+25   
;      24  LTT3218 
;      30  Wolf1346
;      32  HD19445 
;      38  HD84937 
;      52  BD+26   
;      62  Feige110
;      70  Feige67 
;      71  BD+17   
;      79  BD+33   
;      97  HZ44    
;     223  Feige34 
;     225  G191-B2B
;     295  BD+28   

