;+
; NAME: 
;    plotarccheck
;
; PURPOSE: 
;    Creates a check plot to aid in examination of an arc solution.
;    The plot contains several pages showing the solution fit on the first page, the input
;    spectrum with line positions marked on the second page, and zoom-ins on bright lines
;    on the third page.  
;
; CALLING SEQUENCE: 
;    plotarccheck, pix, wav, arcsol, farc, linex, brightlinex, 
;                  reflinewav, brightreflinewav, faintlinewav, 
;                  nx=nx, refid=refid, brightrefid=brightrefid, 
;                  /separate, plotfile=plotfile
;       
; INPUTS:
;    pix -  pixel values of arc spectrum
;    wav -  solved wavelength values of arc spectrum
;    arcsol - polynomial terms of wavelength solution
;    farc   - flux values of arc spectrum
;    linex  - pixel locations of lines
;    brightlinex - pixel locations of bright lines
;    reflinewav  - wavelengths of reference lines
;    brightreflinewav - wavelengths of bright reference lines
;    faintlinewav - wavelengths of faint reference lines
;    nx - Length of input spectrum.  (Needed to obtain central pixel.)
;    plotfile - Name of file to write
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;    refid - Species IDs of reference lines.
;    brightrefid - Species IDs of bright reference lines.
;    separate - Produce several separate plot files instead of a multi-page plot.
;       
; OUTPUTS:  none
;
; Last modified 18 Sep 2019.  Documentation added 3 July 2022.
;-

function wavtox, wav, arcsol, nx=nx
   if n_elements(wav) gt 1 then begin
      allx = fltarr(n_elements(wav))
      for k = 0, n_elements(wav)-1 do allx[k] = wavtox(wav[k], arcsol, nx=nx)
      return, allx
   endif
   if n_elements(wav) eq 1 then begin
       poly2 = arcsol                ; wav = C0 + C1 x + C2 x^2 +... 
       poly2[0] -= wav     ; 0 = (C0-wav)  +  C1 x + ...
       poly2 = poly2[where(poly2 ne 0)] ; risky
       roots = fz_roots(poly2)
       imag = imaginary(roots)
       real = real_part(roots)
       good = where(imag eq 0 and real gt -nx/2 and real lt nx/2, ct)
       if ct gt 1 then good = max(good)
       if ct eq 0 then out = !values.f_nan else out = real[good]
       return, out+nx/2
   endif
   return, -1
end


pro plotarccheck, pix, wav, arcsol, farc, linex, brightlinex, reflinewav, brightreflinewav, faintlinewav, nx=nx, $
                  refid=refid, brightrefid=brightrefid, separate=separate, plotfile=plotfile

      ; pix not actually used (formerly, used for one plot now)

      xmin = 0
      xmax = n_elements(wav)-1

      wavrange = poly([xmin,xmax]-nx/2.,arcsol)
      meddisp = median(wav-shift(wav,1))
      fullrangedisp = clip((wavrange[1]-wavrange[0])/(xmax-xmin))
      logprint,   '     Wavelength range: ', clip(wavrange[0]), '-', clip(wavrange[1]), $
                  ' (median dispersion ', clip(meddisp), ' A/pix)' 
      if n_elements(brightlinex) gt 0 then brightlinesolwav = poly(brightlinex-nx/2., arcsol)
      if n_elements(linex) gt 0 then       linesolwav = poly(linex-nx/2., arcsol)
      if n_elements(brightlinex) gt 0 then brightliney = farc[round(brightlinex)]
      if n_elements(linex) gt 0 then       liney       = farc[round(linex)]

      if keyword_set(separate) then begin
         if !d.name eq 'PS' then psclose
         psopen, repstr(plotfile,'.eps','.1.eps'), xsize=7, ysize=5, /inch, /encaps
         device, /color
         colors = transpose([[0,0,0],$
                       [170,170,200],$  ; 1 = grey          (reference line)
                       [210,210,220],$  ; 2 = light grey    (faint reference line)
                       [64, 158, 64],$  ; 3 = green         (faint detected line)
                       [112,112, 64],$ 
                       [0,    0,192],$  ; 5 = blue          (strong detected line)
                       [192, 64, 64],$  ; 6 = reddish       (strong reference line)
                       [128, 64, 128]$
                    ])
         tvlct, colors
      endif
      !p.position = 0
      !p.multi = [0,1,3]
     
      nplot = 3
      wavlines = min(wav) + findgen(nplot+1)*(max(wav)-min(wav))/nplot ; confusing name
      ybot = reverse(0.1 + (0.98-0.1)*findgen(nplot)/nplot)
      ytop = ybot + abs(ybot[1]-ybot[0])*0.8

      ; main plot
      yr = [median(farc)/3. > 10., 1.02*max(farc)]
;      for pi = 0, nplot-1 do begin
;         xr = wavlines[[pi,pi+1]]
;         !p.position = [0.05, ybot[pi], 0.98, ytop[pi]]
;         plot, [0], [0], xrange=xr, yrange=yr, /xstyle, /ystyle, /ylog
;         for l = 0, n_elements(faintlinewav)-1 do oplot, faintlinewav[l]*[1,1], yr, color=2
;         for l = 0, n_elements(reflinewav)-1 do oplot, reflinewav[l]*[1,1], yr, color=1
;         for l = 0, n_elements(brightreflinewav)-1 do oplot, brightreflinewav[l]*[1,1], yr, color=6, thick=2
;         ;for l = 0, n_elements(brightlinesolwav)-1 do oplot, brightlinesolwav[l]*[1,1], yr*0.01, color=5, thick=3
;         for l = 0, n_elements(linex)-1 do oplot, [linesolwav[l]], [liney[l]], psym=5, color=3, symsize=0.6
;         for l = 0, n_elements(brightlinex)-1 do oplot, [brightlinesolwav[l]], [brightliney[l]], psym=5, color=5, symsize=0.8, thick=2
;         oplot, wav, farc
;      endfor


     reflinex = wavtox(reflinewav, arcsol, nx=nx)
     brightreflinex = wavtox(brightreflinewav, arcsol, nx=nx)
     ymajr = [yr[1]*(yr[0]/yr[1])^0.05,yr[1]]
     yminr = [yr[1]*(yr[0]/yr[1])^0.025,yr[1]]
     for pi = 0, nplot-1 do begin
         !p.position = [0.05, ybot[pi], 0.98, ytop[pi]]
         xr = ([pi,pi+1]*nx)/nplot
         wavr = wav[xr]
         plot, [0], [0], xrange=xr, yrange=yr, xstyle=9, ystyle=1, /ylog
         ;for l = 0, n_elements(faintlinewav)-1 do oplot, faintlinex[l]*[1,1], yr, color=2
         for l = 0, n_elements(reflinewav)-1 do oplot, reflinex[l]*[1,1], yr, color=1
         for l = 0, n_elements(brightreflinewav)-1 do oplot, brightreflinex[l]*[1,1], yr, color=6, thick=2
         for l = 0, n_elements(linex)-1 do oplot, [linex[l]], [liney[l]], psym=5, color=3, symsize=0.6
         for l = 0, n_elements(brightlinex)-1 do oplot, [brightlinex[l]], [brightliney[l]], psym=5, color=5, symsize=0.8, thick=2


         if n_elements(refid) ge n_elements(reflinewav) then $
         for l = 0, n_elements(reflinewav)-1 do $
           if reflinex[l] ge xr[0] and reflinex[l] le xr[1] then $
             xyouts, reflinex[l]*[1,1]-(xr[1]-xr[0])*0.003, yr[0]+yr[0]*(yr[1]/yr[0])^0.15, $
               refid[l]+' '+clip(reflinewav[l]), color=1, charsize=0.4, orient=90

         if n_elements(brightrefid) ge n_elements(brightreflinewav) then $
         for l = 0, n_elements(brightreflinewav)-1 do $
           if brightreflinex[l] ge xr[0] and brightreflinex[l] le xr[1] then $
             xyouts, brightreflinex[l]*[1,1]-(xr[1]-xr[0])*0.003, yr[0]+yr[0]*(yr[1]/yr[0])^0.15, $
               brightrefid[l]+' '+clip(brightreflinewav[l]), color=6, charsize=0.4, orient=90

         oplot, pix, farc
         axis, xaxis=1, xminor=0, xtickformat='(A1)', xticklen=0.0001
         for w = 2000, 11000, 100 do begin 
            x = wavtox(w, arcsol, nx=nx)
            oplot, [x,x], ymajr
            if w gt wavr[0] and w lt wavr[1] and w mod 200 eq 0 $
                then xyouts, x-(0.005*nx), yr[1]*1.14, /data, clip(w), charsize=0.5
            for wmin = w+10, w+90, 10 do begin
               x = wavtox(wmin, arcsol, nx=nx)
               oplot, [x,x], yminr
            endfor
            if w gt wavr[1] then break
         endfor
      endfor

      if keyword_set(separate) then begin
         if !d.name eq 'PS' then psclose
         !p.position = 0
         !p.multi = [0, 2, 1]
         psopen, repstr(plotfile,'.eps','.2.eps'), xsize=7, ysize=5, /inch, /encaps
         device, /color
         colors = transpose([[0,0,0],$
                       [170,170,200],$  ; 1 = grey          (reference line)
                       [210,210,220],$  ; 2 = light grey    (faint reference line)
                       [64, 158, 64],$  ; 3 = green         (faint detected line)
                       [112,112, 64],$ 
                       [0,    0,192],$  ; 5 = blue          (strong detected line)
                       [192, 64, 64],$  ; 6 = reddish       (strong reference line)
                       [128, 64, 128]$
                    ])
         tvlct, colors
      endif


      ; line zoom-ins

      !p.position = 0
      if n_elements(brightreflinewav) le 20 then begin
         nplotx = ceil(n_elements(brightreflinewav)/4.)
         nploty = 4
      endif else begin
         nplotx = ceil(n_elements(brightreflinewav)/5.)
         nploty = 5
      endelse
      xbars = 0.02 + 0.96*findgen(nplotx+1)/(nplotx)
      ybars = 0.07 + 0.9*findgen(nploty+1)/(nploty)
      px = 0
      py = nploty-1
      !p.multi=[0,3,4]
      for l = 0, n_elements(brightreflinewav)-1 do begin
         !p.position = [xbars[px], ybars[py], xbars[px+1], ybars[py+1]] + [0.03, 0.03, 0, 0]
         plot, wav, farc, yrange=yr, /xstyle, /ystyle, /ylog, xrange=brightreflinewav[l]+[-10,+10], psym=10, xtickformat='(I)'
         for ll = 0, n_elements(faintlinewav)-1 do oplot, [faintlinewav[ll], faintlinewav[ll]], yr, color=2
         for ll = 0, n_elements(reflinewav)-1 do oplot, [reflinewav[ll], reflinewav[ll]], yr, color=1
         oplot,  brightreflinewav[l]*[1,1], yr, color=6, thick=3
         for ll = 0, n_elements(brightlinesolwav)-1 do oplot, brightlinesolwav[ll]*[1,1], yr*0.01, color=5, thick=4
         outstr = clip(brightreflinewav[l])
         ytext = 1e2
         if n_elements(brightrefid) gt l then outstr = brightrefid[l] + '  ' + outstr
         if n_elements(brightrefid) gt l then ytext = 40.
         xyouts, brightreflinewav[l]-1, ytext, outstr, color=6, orientation=90, charsize=0.5
         px += 1
         if px eq nplotx then begin
            px = 0
            py -= 1
         endif
         !p.multi[0] = 1
      endfor

      !p.position = 0
end


