;+
; NAME: printcatalog.pro
;
; PURPOSE: 
;       Print out a basic text table containing photometry of objects in a field from
;       a specified reference survey, within the footprint of a given image or set of
;       images.
;
; CALLING SEQUENCE: 
;       printcatalog, filesstr, catalog, galaxiesok=galaxiesok, all=all, 
;                     unc=unc, outfile=outfile, quiet=quiet, count=count, 
;                     magthresh=magthresh, edgebuffer=edgebuffer, sys=sys
;
; INPUTS:
;       filesstr = File name string or array of strings.  Wildcard specifier OK.
;                  If multiple files are given, the catalogs will be combined.
;       catalog  = Reference catalog.  One of:
;                  'sdss'  (SDSS catalog)
;                  'ps1'   (Pan-STARRS 1 catalog)
;                  'skymapper' (Skymapper catalog)
;                  '2mass' (2MASS catalog)
;                  'usno' (USNO B1.0 catalog)
;                  'sdss2' (SDSS primary standards catalog - for brighter objects)
;                  'landolt' (Landolt catalog)
;
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       galaxiesok - Set this not to exclude galaxies if SDSS (by default, stars are excluded)
;       all - Set to require that the object be in every one of the multiple images.
;             (Otherwise, the default is that the object need be present in any image.)
;       unc - Set to also include the uncertainties and not just the magnitude measurements.
;       outfile - Write the catalog to disk.  Otherwise, it is printed out to the screen.
;       quiet - Set to hide any warning messages printed to the screen.
;       magthresh - R/r-band magnitude threshold; fainter stars are excluded. (Optical only.)
;       edgebuffer - Exclude sources within this many arcsec of the field edge.  
;              Defaults to 7 if unspecified.
;       sys - Transform to a nonstandard filter system.  Currently only works for transforming 
;              SDSS primary standards to the Liverpool Telescope system (set sys='LT')
;
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;       count - The number of objects in the final catalog
; 
; OUTPUTS: The filter wavelength, or width, in Angstroms.
;
; REQUIREMENTS: Uses the GSFC IDLAstro library for reading FITS files.
;          Also uses the catalog lookup codes getsdss, getps1, get2mass, etc.
;
; Last modified 17 Aug 2019.  Documentation added 2 July 2022.
;-


; Prints out a table of [cat/USNO/2MASS/custom] objects matching an image or series of images.

; ToDo:  enable Landolt.


function prm, mag
   if mag ge 0.0 and mag lt 90. then return, fpr(mag,2.2) else return, ' -   '
end
function pru, mag
   if mag ge 0.0 and mag lt 90. then return, fpr(mag,1.2) else return, ' -  '
end

pro printcatalog, filesstr, catalog, galaxiesok=galaxiesok, all=all, minsep=minsep, unc=unc, outfile=outfile, quiet=quiet, count=count, magthresh=magthresh, edgebuffer=edgebuffer, sys=sys

; minsep doesn't do anything

count = 0
if n_elements(edgebuffer) eq 0 then edgebuffer = 7
if catalog eq 'tmc' then catalog = '2mass'

imagenames = ['']
for i = 0, n_elements(filesstr)-1 do begin
  imagenames = [imagenames, file_search(filesstr[i])]
endfor
real = where(imagenames ne '', ct)
if ct eq 0 then begin
   print, 'Image(s) not found:', filesstr
   return
endif
imagenames = imagenames[real]
nimages = n_elements(imagenames)

ras = [0]
decs = [0]
es = [0]
nxs = [0]
nys = [0]
for i = 0, nimages-1 do begin

  fits_info, imagenames[i], /silent, N_ext=next
  h = headfits(imagenames[i])

  extast, h, astr, noparams, alt=alt
  e = 0
  ; next = sxpar(h, 'NEXTEND')
  while noparams lt 0 do begin
    e += 1
    if e gt next then break
    h = headfits(imagenames[i], ext=e)
    extast, h, astr, noparams, alt=alt
  endwhile 
  if noparams lt 0 then begin
    print, 'No header astrometry for '+imagenames[i]+' - cannot print matched catalog.'  ; In the future, could add provisional astrometry right here.
    return
  endif

  ;rastr = sxpar(h, 'RA', count=ctra)
  ;decstr = sxpar(h, 'DEC', count=crdec)
  ;if strpos(rastr,  ':') ge 0 then ra  = 15*ten(rastr)  else ra  = float(rastr)
  ;if strpos(decstr, ':') ge 0 then dec = ten(decstr) else dec = float(decstr)
  nx = sxpar(h, 'NAXIS1')
  ny = sxpar(h, 'NAXIS2')
  if sxpar(h,'PC001001') eq -1 then sxaddpar,h, 'CDELT1',-sxpar(h,'CDELT1')  ; some weird Pan-STARRS thing I don't fully understand
  if sxpar(h,'PC002002') eq -1 then sxaddpar,h, 'CDELT2',-sxpar(h,'CDELT2') ; some weird Pan-STARRS thing I don't fully understand
  xyad, h, nx/2, ny/2, ra, dec
  nxs = [nxs, nx]
  nys = [nys, ny]
  ras = [ras, ra]
  decs = [decs, dec]
  es = [es, e]
endfor
ras = ras[1:*]
decs = decs[1:*]
es = es[1:*]
nxs = nxs[1:*]
nys = nys[1:*]
avgra = median(ras, /even)
avgdec = median(decs, /even)
pixscale = getpixelscale(h)
imsize = max(sqrt(nxs^2. + nys^2.)) * pixscale
maxoff = max( sqrt ( (abs(ras-avgra)*cos(avgdec*!pi/180.)*3600.)^2 + $
                     (abs(decs-avgdec)*3600.)^2 ))
queryrad = maxoff + imsize/2. - (edgebuffer*pixscale < 0)
if queryrad gt 1800 then begin
   print, "    Images of this field are too far apart ("+clip(queryrad/2/60.)+"') to download a reference catalog."
   return
endif
if queryrad gt 900 then begin
   print, "    Images cover too much area (>30'), catalog query may not be reliable"
endif

liverpool = 0
if n_elements(sys) gt 0 then if sys eq 'LT' then liverpool = 1

if catalog eq 'sdss' then cat = getsdss(avgra, avgdec, queryrad)
if catalog eq 'ps1' then cat = getps1(avgra, avgdec, queryrad)
if catalog eq 'skymapper' then cat = getskymapper(avgra, avgdec, queryrad)
if catalog eq '2mass' then cat = get2mass(avgra, avgdec, queryrad)
if catalog eq 'usno' then cat = getusno(avgra, avgdec, queryrad)
if catalog eq 'sdss2' then cat = getsdss2(avgra, avgdec, queryrad, liverpool=liverpool)
if catalog eq 'landolt' then cat = getlandolt(avgra, avgdec, queryrad)
if n_elements(cat) eq 0 then begin
   print, 'Could not recognize catalog ', catalog
   return ; no catalog recognized
endif

; other

starcount = intarr(n_elements(cat))
ncat = n_elements(cat)
if (ncat eq 0) or ncat eq 1 and size(ncat,/tn) eq 'LONG' then begin ; or (ncat eq 1 and catalog ne 'landolt' and catalog ne 'sdss2') then begin
  if keyword_set(quiet) eq 0 then print, '    No objects in catalog for this field.'
  return
endif

; Star/galaxy/marginal-source filtering for SDSS
if catalog eq 'sdss' then begin
   if keyword_set(galaxiesok) eq 0 then cat = cat[where(cat.type eq 6)] ; stars only for SDSS
endif
;if catalog eq 'skymapper' then begin
;   if keyword_set(galaxiesok) eq 0 then cat = cat[where(cat.star ge 0)] ; stars only for skymapper.
;   Currently it's not clear what parameter threshold to use. 
;   "Maximum stellarity index from photometry table (between 0=no star and 1=star"
;endif
ngooddet = intarr(n_elements(cat))
if catalog eq 'sdss' then begin
   ngooddet += cat.g_unc lt 0.33
   ngooddet += cat.r_unc lt 0.25
   ngooddet += cat.i_unc lt 0.25
   cat = cat[where(ngooddet ge 3)]
   cat = cat[sort(cat.r)]
endif

; (No star-galaxy/marginal-source filtering for USNO or 2MASS)



for i = 0, nimages-1 do begin
  h = headfits(imagenames[i], ext=es[i])
  nx = sxpar(h, 'NAXIS1')
  ny = sxpar(h, 'NAXIS2')
  if sxpar(h,'PC001001') eq -1 and sxpar(h,'PC002002') eq -1 then begin
    sxaddpar,h, 'CDELT1',-sxpar(h,'CDELT1')  ; some weird Pan-STARRS thing I don't fully understand
    sxaddpar,h, 'CDELT2',-sxpar(h,'CDELT2')  ; some weird Pan-STARRS thing I don't fully understand
  endif
  

  for s = 0, n_elements(cat)-1 do begin
    adxy, h, cat[s].ra, cat[s].dec, x, y ; seems to generate wrong values sometimes
    ; if s le 10 or s mod 100 eq 0 then print, s, cat[s].ra, cat[s].dec, x, y

    if finite(edgebuffer) eq 0 or $
       (x ge edgebuffer and x lt nx-edgebuffer and $
        y ge edgebuffer and y lt ny-edgebuffer) then $
       starcount[s] += 1
  endfor
endfor

ncat = total(starcount)

if total(starcount) eq 0 then begin
  if keyword_set(quiet) eq 0 then print, '    At the edge of a catalog field, stars are nearby but do not cover this image.'
    ;stop
  return
endif


if keyword_set(all) then cat = cat[where(starcount eq nimage)] $
                    else cat = cat[where(starcount gt 0)]

ncat = n_elements(cat)

if n_elements(outfile) gt 0 then  openw, 1, outfile

leg = clip('#ra',11)+' '+clip('dec',10)
if catalog eq 'sdss' or catalog eq 'skymapper' or catalog eq 'landolt' or catalog eq 'sdss2' then begin
  if catalog eq 'landolt' or catalog eq 'sdss2' then leg += ' '+clip('U',5)
  leg +=' '+clip('u',5)+' '+clip('B',5)+' '+clip('g',5)+' '+clip('V',5)+' '+clip('r',5)+' '+clip('R',5)+' '+clip('i',5)+' '+clip('I',5)+' '+clip('z',5)
  if keyword_set(unc) then begin
    if catalog eq 'landolt' or catalog eq 'sdss2' then leg += ' '+clip('Uerr',5)
    leg += ' '+clip('uerr',4)+' '+clip('Berr',4)+' '+clip('gerr',4)+' '+clip('Verr',4)+' '+clip('rerr',4)+' '+clip('Rerr',4)+' '+clip('ierr',4)+' '+clip('Ierr',4)+' '+clip('zerr',4)
  endif
endif
if catalog eq 'ps1' then begin
  leg += ' '+clip('B',5)+' '+clip('g',5)+' '+clip('V',5)+' '+clip('r',5)+' '+clip('R',5)+' '+clip('i',5)+' '+clip('I',5)+' '+clip('z',5)+' '+clip('y',5)
  if keyword_set(unc) then $
  leg += ' '+clip('Berr',5)+' '+clip('gerr',5)+' '+clip('Verr',5)+' '+clip('rerr',5)+' '+clip('Rerr',5)+' '+clip('ierr',5)+' '+clip('Ierr',5)+' '+clip('zerr',5)+' '+clip('yerr',5)
endif
if catalog eq '2mass' then begin
  leg +=' '+clip('Y',5)+' '+clip('J',5)+' '+clip('H',5)+' '+clip('Ks',5)+' '+clip('Kp',5)+' '+clip('K',5)
  if keyword_set(unc) then $
    leg += ' '+clip('e_Y',4)+' '+clip('e_J',4)+' '+clip('e_H',4)+' '+clip('e_Ks',4)+' '+clip('e_Kp',4)+' '+clip('e_K',4)
endif
if catalog eq 'usno' then begin
  leg +=' '+clip('B',5)+' '+clip('R',5)+' '+clip('I',5)
endif

if n_elements(outfile) eq 0 then print, leg $
                            else printf, 1, leg

for s = 0, n_elements(cat)-1 do begin
   ss = cat[s]
   out = fpr(ss.ra,3.7)+' '+fpr(ss.dec,3.6)+' '

   if catalog eq 'sdss' or catalog eq 'skymapper' then begin
      if n_elements(magthresh) gt 0 then if ss.r gt magthresh then continue
      out += prm(ss.u)+' '+prm(ss.Bj)+' '+prm(ss.g)+' '+prm(ss.Vj)+' '+prm(ss.r)+' '+prm(ss.Rc)+' '+prm(ss.i)+' '+prm(ss.Ic)+' '+prm(ss.z)
      if keyword_set(unc) then begin
      out += ' '+pru(ss.u_unc)+' '+pru(ss.Bj_unc)+' '+pru(ss.g_unc)+' '+pru(ss.Vj_unc)+' '+pru(ss.r_unc)+' '+pru(ss.Rc_unc)+' '+pru(ss.i_unc)+' '+pru(ss.Ic_unc)+' '+pru(ss.z_unc)
      endif
   endif
   if catalog eq 'landolt' or catalog eq 'sdss2' then begin
      out += prm(ss.Uj)+' '+prm(ss.us)+' '+prm(ss.Bj)+' '+prm(ss.gs)+' '+prm(ss.Vj)+' '+prm(ss.rs)+' '+prm(ss.Rc)+' '+prm(ss.is)+' '+prm(ss.Ic)+' '+prm(ss.zs)
      if keyword_set(unc) then begin
      out += ' '+pru(ss.eUj)+' '+prm(ss.eus)+' '+pru(ss.eBj)+' '+pru(ss.egs)+' '+pru(ss.eVj)+' '+pru(ss.ers)+' '+pru(ss.eRc)+' '+pru(ss.eis)+' '+pru(ss.eIc)+' '+pru(ss.ezs)
      endif
   endif
   if catalog eq 'ps1' then begin
      if n_elements(magthresh) gt 0 then if ss.r gt magthresh or ss.r le 0 then continue
      out += prm(ss.Bj)+' '+prm(ss.g)+' '+prm(ss.Vj)+' '+prm(ss.r)+' '+prm(ss.Rc)+' '+prm(ss.i)+' '+prm(ss.Ic)+' '+prm(ss.z)+' '+prm(ss.y)
      if keyword_set(unc) then begin
      out += ' '+pru(ss.Bj_unc)+' '+pru(ss.g_unc)+' '+pru(ss.Vj_unc)+' '+pru(ss.r_unc)+' '+pru(ss.Rc_unc)+' '+pru(ss.i_unc)+' '+pru(ss.Ic_unc)+' '+pru(ss.z_unc)+' '+pru(ss.y_unc)
      endif
   endif
   if catalog eq '2mass' then begin
      out += prm(ss.Y)+' '+prm(ss.J)+' '+prm(ss.H)+' '+prm(ss.Ks)+' '+prm(ss.Kp)+' '+prm(ss.K)
      if keyword_set(unc) then begin
      out += ' '+pru(ss.Y_unc)+' '+pru(ss.J_unc)+' '+pru(ss.H_unc)+' '+pru(ss.Ks_unc)+' '+pru(ss.Kp_unc)+' '+pru(ss.K_unc)
      endif
   endif
   if catalog eq 'usno' then begin
      if n_elements(magthresh) gt 0 then if ss.Rc gt magthresh then continue
      out += prm(ss.Bj)+' '+prm(ss.Rc)+' '+prm(ss.Ic)
      if keyword_set(unc) then begin
      out += ' '+pru(ss.Bj_unc)+' '+pru(ss.Rc_unc)+' '+pru(ss.Ic_unc)
      endif
   endif


   if n_Elements(outfile) eq 0 then print, out $
                               else printf, 1, out 

endfor

count = n_elements(cat)

if n_elements(outfile) gt 0 then  close, 1

end

