;+
; NAME:
;   pzap.pro
;
; PURPOSE:
;   Remove cosmic rays from a 2-D image using model subtraction / percolation.
;
; CALLING SEQUENCE:
;  pro pzap, name, outname, outmaskname, zealfactor=zealfactor, 
;            nsigma=nsigma, clnsigma=clnsigma, subsigma=subsigma, 
;            preiter=preiter, postiter=postiter, $
;            mask=mask, weight=weight, zero=zero, nan=nan, $
;            statsec=statsec, usamp=usamp, saturation=saturation,$
;            verbose=verbose, quiet=quiet, silent=silent, debug=debug, $
;            nzap=nzap
;
; INPUTS:
;   name        - 2-D image array, or name of input FITS file.
;   outname     - Output image array, or name of output FITS file.
;
; OPTIONAL INPUTS:
;   outmaskname - Output mask array, or name of output FITS file.
;   zealfactor  - Generic control parameter to make cleaning more or less aggressive.  Default 1.0, higher is more aggressive.
;   nsigma      - Rejection threshhold in sigma
;   clnsigma    - Sigma for immediate neighbors to also be considered cosmic rays
;   subsigma    - Sigma for neighbors to also be considered cosmic rays during percolation
;   preiter     - Number of iterations for initial cosmic ray-finding
;   postiter    - Number of additional iterations after replacement
;   mask        - Write out a mask identifying zapped pixels.
;   weight      - Write out weight map (boolean inverse of mask).
;   zero        - Replace zapped pixels with 0 instead of model value (for easy visual identification)
;   nan         - Replace zapped pixels with nan
;   statsec     - Specify a sub-section of the image to calculate image statistics
;   usamp       - Flag to indicate cosmic rays are undersampled (set for a traditional detector or binned data)
;   saturation  - Specify the saturation threshold of the image
;   verbose     - Print details of the percolation process
;   quiet       - Print less information
;   silent      - Print no information
;   debug       - Flag to write intermediate image products to disk to test algorithm functionality
;
; OUTPUTS:
;   nzap        - Number of pixels zapped.
;
; COMMENTS:
;   A robust cosmic ray rejection and replacement script, optimized for
;   the new LRIS-red, a deep-depletion CCD susceptible to muon-trailed
;   cosmic ray impacts which leave long, snaking patterns across the chip.
;   Replacing nnzaps, this program creates a model of the image using
;   nearby pixels and and identifies cosmic rays as large excesses above
;   this background.  A percolation algorithm is used to flag the entire
;   ray trail.
;  
;   The routine is effective but not perfect, typically removing about 99% of
;   cosmic ray flux.  Missed pixels tend to be the weakest cosmic rays with
;   wide, diagonal trails which the algorithm has difficulty distinguishing
;   from real sources in images with significant sky noise (usually, some but
;   not all of the ray trail is typically flagged in such cases.)
;
;  Last modified 12 Mar 2021
;-
;------------------------------------------------------------------------------

; Some bugs/todo's:
;   - handle edges properly by reflecting the image across the edges for a few pixels.
;   - do some tests with adaptive PSF/parameter modeling to optimize sigma parameters.
;   - Enable use in array format, plot histograms, other user-friendly stuff
;   - See if it's possible to reduce the number of iterations / model runs for faster processing
;   - Break the image into chunks and process these individually.
;   - It looks like there are some bugs with treating nzap, nbad, etc. inconsistently (# this iter vs. # total)

@modelimage

pro pzap, name, outname, outmaskname, nsigma=nsigma, clnsigma=clnsigma, subsigma=subsigma, preiter=preiter, postiter=postiter, $
 zealfactor=zealfactor, nrings=nrings, nzap=nzap, mask=mask, weight=weight, zero=zero, nan=nan, statsec=statsec, debug=debug, usamp=usamp, $
 verbose=verbose, quiet=quiet, silent=silent, $
saturation=saturation

   ; Parameter 'name' is generally a string filename - but can also be an array that has already been read in from disk, or a text file
   ; that contains a list of files to process, as in IRAF.  These alternative formats are rarely used and not tested recently.

   if (size(name, /tname) EQ 'STRING') then begin
      if n_elements(name) eq 0 then begin
         print, 'A filename must be specified.'
         return
      endif
      if n_elements(name) gt 1 then begin
         print, 'WARNING - this feature is buggy - some variables are persistent'
         for f = 0, n_elements(name)-1 do begin
            print, name[f]
            pzap, name[f],  $
               nsigma=nsigma, clnsigma=clnsigma, subsigma=subsigma, preiter=preiter, postiter=postiter, zealfactor=zealfactor, $
               nrings=nrings, nzap=nzap, mask=mask, weight=weight, zero=zero, nan=nan, debug=debug, usamp=usamp, $
               verbose=verbose, quiet=quiet, silent=silent, saturation=saturation
         endfor
         return
      endif
      if n_elements(name) eq 1 then begin
         namearr = strsplit(name,'.', /extract)
         nameext = namearr[1]
         if nameext eq 'cat' or nameext eq 'lis' or nameext eq 'list' or nameext eq 'txt' then begin
            files = grabcolumn(name,0,/str)
            pzap, files,  $
              nsigma=nsigma, clnsigma=clnsigma, subsigma=subsigma, preiter=preiter, postiter=postiter, zealfactor=zealfactor,  $
              nrings=nrings, nzap=nzap, mask=mask, weight=weight, zero=zero, nan=nan, debug=debug, usamp=usamp, $
              verbose=verbose, quiet=quiet, silent=silent, saturation=saturation
         return
         endif


         img=readfits(name, header, /silent)

         if n_elements(img) eq 1 then begin
            logprint, 'Error reading ', name
            return
         endif
         dum = where(finite(img), ctnotnan)
         if ctnotnan eq 0 then begin
            logprint, 'No valid pixels in image ', name
            return
         endif

         if n_elements(outname) eq 0 then begin
            slashpos = strpos(name,'/')
            dir = strmid(name,0,slashpos+1)
            outname = dir + 'z' + strmid(name,slashpos+1)
         endif
      endif
   endif else begin
      img=name
   endelse

   ; Set default values of parameters

   if n_elements(nsigma) eq 0 then      nsigma = 6.7
   if n_elements(clnsigma) eq 0 then    clnsigma = 5.0
   if n_elements(clsubnsigma) eq 0 then clsubnsigma = 3.9  ; this is not actually a user parameter
   if n_elements(subsigma) eq 0 then    subsigma = 2.0
   if n_elements(zealfactor) eq 0 then  zealfactor = 1.0    
   if n_elements(preiter) eq 0  then    preiter=3
   if n_elements(postiter) eq 0 then    postiter=1
   ; if n_elements(nrings) eq 0 then    nrings=1

   nsigma = nsigma / zealfactor         
   clnsigma = clnsigma / zealfactor        ; clnsigma - is a neighbor a candidate for CR check?
   clsubnsigma = clsubnsigma / zealfactor  ; clsubnismg - are the neighbor's neighbors high enough to flag neighbor as a CR?
   subsigma = subsigma / zealfactor

   ; More parameter-to-variable assignment
   if keyword_set(mask) then begin
      if strtrim(string(mask),2) eq '1' then begin
         outmaskname = repstr(outname,'.fit','.mask.fit')
         if outmaskname eq outname then outmaskname = outname+'.mask'
         ;namearr = strsplit(outname, '.', /extract)
         ;outmaskname = namearr[0] + '.mask.' + namearr[1]
      endif else begin
         outmaskname = mask
      endelse
   endif
   if keyword_set(weight) then begin
      if strtrim(string(weight),2) eq '1' then begin
         outweightname = repstr(outname,'.fit','.weight.fit')
         if outweightname eq outname then outweightname = outname+'.weight'
         ;namearr = strsplit(outname, '.', /extract)
         ;outweightname = namearr[0] + '.weight.' + namearr[1]
      endif else begin
         outweightname = weight
      endelse
   endif

   tstart = systime(/seconds)

   ; Sometimes the image already has NaN's, but NaN's are used in the algorithm to denote flagged pixels
   ; (since IDL helpfully excludes NaN's from median calculations) and replaced later.  
   ; Deal with this by recording where NaN's are now, so they can be set back to NaN at the end.

   winitnegnan = where(finite(img, /nan, sign=-1), ctinitnegnan)
   if keyword_set(nan) then winitnan = where(finite(img) eq 0, ctinitnan) else ctinitnan = 0

   dims = size(img, /dimens)
   nx = dims[0]
   ny = dims[1]
   outmask = bytarr(dims[0], dims[1])  ;create an array to hold CR position mask

   ; Record all "bad" pixels (NaN's and infinities) to model/replace these early on
   initbad = where(finite(img) eq 0, ninitbad)

   ; Calculate the overall st.dev. of the image sky
   if keyword_set(verbose) then print, '  Computing image sigma...'
   if n_elements(statsec) eq 0 then begin
     if n_elements(img) gt 5e5 then begin
       ; if the image is large, speed things up by only using a horizontal central stripe of the image to calculate statistics.
       statsec = [0, nx-1, ny/2-100 > 0, ny/2+100 < ny-1]
     endif
   endif
   if n_elements(statsec) eq 4 then begin
     imgstat = img[statsec[0]:statsec[1],statsec[2]:statsec[3]]
   endif else begin
     if n_elements(statsec) gt 1 then begin
        print, '  statsec format:  [x0, x1, y0, y1]'
        return
     endif
     ; ... otherwise just use the whole image (note that calculations are done later)
     imgstat = img
   endelse

   
   ; --- Calculate the image statistics ---

     ; Apply a median filter to remove any gradients, halos, etc. that would screw up our statistics
   skyfiltsize = 11  ; this is hard-coded, which isn't necessarily ideal
   skyimgstat = median(imgstat, skyfiltsize)
     ; Look for any bad values in the median image (maybe there were huge blocks of NaN, etc)   
   skystatinitbad = where(finite(skyimgstat) eq 0, ct)
     ; Replace them
   if ct gt 0 then skyimgstat[skystatinitbad] = median(skyimgstat)
     ; Look for any bad values in the image statisical subsection
   statinitbad = where(finite(imgstat) eq 0, ct)
     ; Replace them with the median image
   if ct gt 0 then imgstat[statinitbad] = skyimgstat[statinitbad] ; djs barfs if nans are around.
     ; Use djs_iterstat routin to calculate the median sky level and the noise level of the *sky-subtracted* image
   djs_iterstat, imgstat-skyimgstat, median=skylevel, sigma=sigval, sigrej=4
     ; Add the sky median back in so the median is a median of the image for real
   skylevel = skylevel + median(skyimgstat)  ; skylevel is a scalar indicating the average sky value.

     ; Calculate basic sky model (basically, a 3x3 median filter of the image)
   skymodel = modelsky(img)   ; skymodel is an image indicating the broad variations in the sky.
                              ; we don't use it for statistics, but will be useful later for
                              ; properly dealing with the safety factor.
    if keyword_set(debug) then mwrfits, skymodel, 'skymodel.fits', header, /silent, /create


   ; Set up some variables

   nzap = 0      ; counter for number of pixels zapped (flagged/replaced)
   iter = 0      ; iteration counter
   prevbad = 0   ; number of pixels marked bad last iteration
   maxiter = preiter + postiter


   ; Deal with saturation

   if n_elements(saturation) gt 0 then begin
      satlevel = saturation  ; a bit more parameter handling
   endif else begin
      satlevel = sxpar(header,'SATURATE')
   endelse
   if satlevel eq 0 then begin  ; zero means no saturation info specified, so figure it out ourselves
     hipix = where(img gt skylevel + 10*sigval,cthipix)
     if cthipix gt 0 then begin
        imghipix = img[hipix]
        sr = sort(imghipix)
        satlevel = imghipix[sr[n_elements(sr)*0.98]] * 0.88 > 48000.
     endif else begin
        satlevel = 48000.   ; Most CCDs saturate at 2^16 = 65536, possibly bias/sky have ben removed, so guess 48000.
     endelse 
   endif
   if keyword_set(verbose) then $
      print, '  sky = ', strtrim(string(skylevel),2), ' (sigma = ', strtrim(string(sigval),2), ', saturation = ', strtrim(long(satlevel),2),')'
   ; Identify the saturated pixels using the original image
   satpix = where(img gt satlevel, ctsat)
   if ctsat gt n_elements(img)/2 then begin
       logprint, '  The entire image is saturated.  Cannot cosmic-ray clean.'
       return
   endif
   if ctsat gt 0 then begin
     ; Also flag the neighbors of saturated pixels.
     satpix = [satpix, satpix+1, satpix-1, satpix+nx, satpix-nx, satpix+1+nx, satpix-1+nx, satpix+1-nx, satpix-1-nx] ; grow a little 
     satpix = satpix[where(satpix ge 0 and satpix lt n_elements(img))]
   endif

   ;;model2image was an alternative algorithm that evidently doesn't work so well or is too slow.
   ;;model2 = model2image(img) 
   ;;mwrfits, model2, 'model2image.fits', header, /silent, /create
   ;;mwrfits, img-model2, 'model2imageresidual.fits', header, /silent, /create
   ;;mwrfits, (img-model2)/img, 'model2imageresidualnorm.fits', header, /silent, /create


   ; --- Process the image! ----

   titer0 = systime(/seconds)

   ; master loop for each algorithm iterations.
   while (iter LT maxiter) do begin

      ; Multiple iterations are almost always necessary because cosmics have to be iteratively cleaned out to
      ; find the less sharply-affected neighboring pixels from the same hit.
      ; There are two types of iterations: 'pre' and 'post'.
      ;   pre-iterations are BEFORE replacement of bad pixels: they are simply removed from the median filter.
      ;   post-iterations are AFTER repalcement of bad pixels: their replacement values go into the median filter.

      iter = iter + 1
      nbad = 0

      titerstart = systime(/seconds)

      if (keyword_set(quiet) eq 0 and keyword_set(silent) eq 0) then begin
        if iter le preiter then logprint, '   Pre-iteration', iter, format='($,A,I3)'
        if iter gt preiter then logprint, '  Post-iteration', iter-preiter, format='($,A,I3)'
      endif
      
      ; First flag the saturated pixels (these have sharp edges that will create too many hits)
      imgnosat = img
      if ctsat gt 0 then imgnosat[satpix] = !values.f_nan

      ; Now calculate "filterimage", which is a median-filtered image (more or less).      
      ; It represents a model for what each pixel "should" look like based on the values of its neighbors.
      if keyword_set(usamp) eq 0 then begin
        ; The non-usamp version uses the immediate neighbors.
        ;   On some iterations it is very close to being a median filter, on others it uses
        ;   the median of the interpolated (average) values between all four 
        ;   symetrically-distributed neighbor pairs.
        if iter eq 1 then filterimage = model2image(imgnosat)
        if iter eq 2 then filterimage = modelimage(imgnosat)
        if iter eq 3 then filterimage = modelimage(imgnosat)
        if iter eq 4 then filterimage = model2image(imgnosat)
        if iter ge 5 then filterimage = modelimage(imgnosat)
        ; The right number for modeluncertaintyfactor is highly uncertain.  This also varies each iteration.
        if iter eq 1 then modeluncertaintyfactor = 0.05/zealfactor 
        if iter eq 2 then modeluncertaintyfactor = 0.12/zealfactor
        if iter eq 3 then modeluncertaintyfactor = 0.12/zealfactor
        if iter eq 4 then modeluncertaintyfactor = 0.05/zealfactor
        if iter ge 5 then modeluncertaintyfactor = 0.05/zealfactor
      endif else begin
        ; The usamp version uses a variant of linear extrapolation along horizontal/vertical/diagonals.
        ; It is (supposedly) more appropriate for cosmic rays on old-style detectors.
        if iter eq 1 then filterimage = usampmodelimage(imgnosat)
        if iter eq 2 then filterimage = usampmodelimage(imgnosat)
        if iter eq 3 then filterimage = usampmodelimage(imgnosat)
        if iter eq 4 then filterimage = usampmodelimage(imgnosat)
        if iter ge 5 then filterimage = usampmodelimage(imgnosat)
        if iter eq 1 then modeluncertaintyfactor = 0.1/zealfactor 
        if iter eq 2 then modeluncertaintyfactor = 0.1/zealfactor
        if iter eq 3 then modeluncertaintyfactor = 0.1/zealfactor
        if iter eq 4 then modeluncertaintyfactor = 0.1/zealfactor
        if iter ge 5 then modeluncertaintyfactor = 0.05/zealfactor
      endelse


      ; Subtract the model (filterimage) to remove genuine sources, 
      ;  leaving only things that look like cosmic rays

      residualimage = img - filterimage
      if ctsat gt 0 then residualimage[satpix] = 0
      

      ; Estimate the amount of variation we expect to see in the residuals.
      ; This is equal to the pixel-to-pixel variance across the sky, scaled up to account for additional shot noise inside bright sources,
      ; and with an additional term added to account for imperfections in the modeling.  (This additional term is very uncertain.)
      ; This produces a "sigma image", which gives what the algorithm thinks is the expected 1-sigma variance from the model.

      sigmaimage = sigval * sqrt(filterimage/median(filterimage) > 0.9) + modeluncertaintyfactor*((filterimage-skymodel)>0)
                   ; counting noise term                                   ; modeling error term

      ; residualnsigmaimage is the noise-normalized residuals. 
      ; It is the observed deviation from the model, divided by the expected RMS deviation.  
      ;  If > "a few" it's probably a cosmic ray.

      residualnsigmaimage = residualimage / sigmaimage


      ; the below is old code occasionally useful for debugging...
      ; boxsize = sqrt(16)
      ; sourcensigmaimage = (filterimage-skymodel) *  boxsize / sigmaimage  ; formerly boxsize is because this is filtered, sqrt(N) = sqrt(box^2)
      ; extrathreshimage = safetyfactor * sqrt((sourcensigmaimage-1.5) > 0)  
      ; residualnsigmaimage = residualnsigmaimage - extrathreshimage

      ; writing these intermediate images to disk is very useful for checking on problems
      if keyword_set(debug) then begin ;and iter eq 1 then begin
         mwrfits, filterimage, 'filterimage'+clip(iter)+'.fits', header, /silent, /create
         mwrfits, residualimage, 'residualimage'+clip(iter)+'.fits', header, /silent, /create
         mwrfits, sigmaimage, 'sigmaimage'+clip(iter)+'.fits', header, /silent, /create
         ;mwrfits, sourcensigmaimage, 'sourcensigmaimage.fits', header, /silent, /create
         mwrfits, residualnsigmaimage, 'residualnsigmaimage'+clip(iter)+'.fits', header, /silent, /create
         ;mwrfits, nsigma+extrathreshimage,  'threshimage.fits', header, /silent, /create
      endif



      ;  --- Next, actually find the cosmic rays ---


      if keyword_set(verbose) then print, '  Identifying cosmic rays and percolating...'

      ; zapimage records the cosmic ray pixels found in this iteration.
      ; It's an integer array:  0=clean, 1=cosmic, 2+=cosmic found by percolation.
      zapimage = intarr(dims[0], dims[1])


      ; The simplest search - look for values (of the noise-normalized residual) in excess of nsigma.      
      ; These generally are only the most severely accepted pixels of each hit.  

      crcores = where(residualnsigmaimage gt nsigma, initzaps)
      if initzaps gt 0 then zapimage[crcores] = 1


      ; The second, more sophisticated search - look for values whose significance (residualnsigma) is lower, and flag them IF
      ; some of their neighbors are also look like posible cosmic rays.

      potentialsubcrcores = where(residualnsigmaimage gt clnsigma and residualnsigmaimage lt nsigma, ncheck4zap)
      if ncheck4zap gt 0 then begin
        subcrcores = lonarr(ncheck4zap) ; list of the pixel positions of CR pixels found this way.
        b = 0
        for i = 0L, ncheck4zap-1 do begin
           s = potentialsubcrcores[i]
           coarseblockpos = [         s-nx,       $
                             s-1,           s+1, $
                                      s+nx         ]
           if max(residualnsigmaimage[coarseblockpos]) gt clsubnsigma then begin
             subcrcores[b] = s
             b = b + 1
           endif
        endfor
        if b gt 0 then subcrcores = subcrcores[0:b-1]
         ;print, ''
         ;print, 'Additional', b
        if b gt 0 then zapimage[subcrcores] = 1  ; add them to the zap image
      endif else begin
        b = 0 ; hazardous to get here, from recent experience (all NaN image)
      endelse


      if initzaps eq 0 and b eq 0 then begin        ; make this the last iteration (stop early) if no CR's found
         if iter lt preiter then iter = preiter   
         if iter gt preiter then iter = maxiter
      endif

      if n_elements(ibad) gt 0 then begin
         if ibad[0] ne -1 then zapimage[ibad] = 1   ; retain zaps from last cycle for percolation.
      endif

      if keyword_set(debug) then begin
        mwrfits, zapimage, 'zapcoreimage'+clip(iter)+'.fits', header, /silent, /create
      endif

      titera = systime(/seconds)

      if keyword_set(verbose) then print, '    Flagged ', strtrim(initzaps,2), ' initial affected pixels before percolation.'


      ; The third search is a percolation algorithm.
      ; The four nearest-neighbors of each already-zapped pixel are examined to see if their residuals are even marginally excessive.
      ; If so, they are marked as bad and the iteration is run again, looking for the neighbors of *those* pixels.
      ; Stop when the percolation runs out of pixels to find or after 32 steps.

      nperczap = 0   ; counter for number pixels zapped in this step
      subiter = 1    ; (sub)iteration counter
      nx = dims[0]
      ny = dims[1]
      while subiter lt 32 and initzaps gt 0 do begin
         nextperc = where(zapimage eq subiter, ct)
         subiter = subiter + 1
         newzaps = 0        
         if ct le 3 then break
         nrays = n_elements(nextperc)
         for c = 0L, nrays-1 do begin
             ci = nextperc[c]
             if ci lt 2*nx or ci gt nx*ny-2*nx then continue
             if ci mod nx lt 3 or ci mod nx ge nx-3 then continue
             coarseblockpos = [         ci-nx,       $
                               ci-1,           ci+1, $
                                        ci+nx         ]
             ;coarseblockpos = [ci-nx-1, ci-nx, ci-nx+1, $
             ;                  ci-1,           ci+1,    $
             ;                  ci+nx-1, ci+nx, ci+nx+1]
             newzap = where(residualnsigmaimage[coarseblockpos] gt subsigma and zapimage[coarseblockpos] eq 0, addzap)
             if addzap gt 0 then begin
                zapimage[coarseblockpos[newzap]] = subiter
                newzaps = newzaps + addzap
             endif
             nperczap = nperczap + addzap
         endfor
         ;if keyword_set(verbose) then print, '    subiteration ', strtrim(subiter,2), ':', newzaps, ' pixel zaps.'
      endwhile
      if keyword_set(verbose) then print, '    Flagged ', strtrim(nperczap,2), ' pixels during ', strtrim(subiter,2), ' percolation steps.'

      ; Finally, zap anything hiding in a cosmic ray "corner" (three neighbors are cosmic ray pixels)
      ; (not used anymore, within the iteration loop it causes wild growth)
      ;countneighbor = intarr(nx, ny)
      ;nextperc = where(zapimage gt 0, ct)
      ;if ct gt 3 then begin
      ;   nrays = n_elements(nextperc)
      ;   for c = 0L, nrays-1 do begin
      ;       ci = nextperc[c]
      ;       coarseblockpos = [ci-nx-1, ci-nx, ci-nx+1, $
      ;                         ci-1,           ci+1,    $
      ;                         ci+nx-1, ci+nx, ci+nx+1]
      ;       countneighbor[coarseblockpos] = countneighbor[coarseblockpos] + 1
      ;   endfor
      ;   newzap = where(countneighbor ge 4 and zapimage eq 0, newzaps)
      ;   if newzaps gt 0 then begin
      ;      if keyword_set(verbose) then print, '    Flagged an additional ', strtrim(newzaps,2), ' neighboring pixels.'
      ;      zapimage[newzap] = subiter + 1
      ;   endif
      ;endif

      titerb = systime(/seconds)

      if keyword_set(debug) then begin
         mwrfits, zapimage, 'zapimage'+clip(iter)+'.fits', header, /silent, /create
      endif

      ; obsolete code preserved for posterity:
      ;if (NOT keyword_set(nofluxcompare)) then begin
      ;   ;For every pixel in high-flux regions (bright stars and crays), calculate the ratio
      ;   ; of the local (small-box) median after sky subtraction and the actual value.  If
      ;   ; it is more than 1/fluxcompare, target that pixel for zapping.
      ;   i = where(zapimage NE 0,icount)
      ;   if icount gt 0 then zapimage[i] = (filterimage[i] / residualimage[i]) LT fluxratio
      ;endif
      ;
      ;Apply the growth radius by also zapping surrounding pixels, if requested.
      ; For this to function, put it at the end!!
      ;if nrings gt 0 then $
      ; zapimage = smooth(float(zapimage), 1+2*nrings, /edge) GT 1.0E-6


      ; Collect together the CR pixels and actually zap them by replacing them with NaN.
      ; (Replacement will come later.)  

      ibad = where(zapimage NE 0, nbad) ; ibad = positions of bad pixels this iteration
      if (nbad GT 0) then begin
         img[ibad] = !values.f_nan ; set to NaN 
         outmask[ibad] = 1         ; Also mark them in the bad pixel mask.
      endif
     
      nbadnew = nbad - prevbad
      if keyword_set(silent) eq 0 and keyword_set(quiet) eq 0 then $
        logprint, ': Number zapped = ', strtrim(string(nbadnew),2)
      prevbad = nbad
      nzap = nbad


        

      ; --- Replace bad pixels with an interpolated estimate ---

      ; Replacement is only done twice: at the end of the last pre-iteration and at the end of the last post-iteration.
      ; Even if the user has set 'NaN' or 'zero' as options (no replacement), replacement is done temporarily because 
      ; it can help identify CR-affected pixels that we missed before.

      if (postiter gt 0 and iter eq preiter) or (iter eq maxiter and keyword_set(nan) eq 0) then begin 

        ; Identify and replace the bad (masked) pixels.
        ibad = where(outmask gt 0, nibad)
        if nibad gt 0 then begin
          if keyword_set(usamp) eq 0 then replacementmodel = modelimage(img) $ ; note slow step here.
                                     else replacementmodel = usampmodelimage(img)
          img[ibad] = replacementmodel[ibad]
        endif

        ; The image could potentially have had large bad regions in it.  Whether dealing with this
        ; is the purview of this algorithm is questionable but below are steps to find and replace these, too.
        nans = where(finite(img) eq 0, ct) ; Anything *still* NaN after replacement
        naniter = 0
        nel = n_elements(img)
        while ct gt 0 do begin
          holdimg = img
          for c = 0L, n_elements(nans)-1 do begin
             ci = nans[c]
             if naniter lt 5 then begin
             coarseblockpos = [ci-nx-1, ci-nx, ci-nx+1, $
                               ci-1,           ci+1,    $
                               ci+nx-1, ci+nx, ci+nx+1] 
             endif else if naniter lt 7 then begin
             coarseblockpos = [ci-2*nx-2, ci-nx*2, ci-nx*2+2,$
                               ci-2,               ci+2,     $
                               ci+2*nx-2, ci+nx*2, ci+nx*2+2] 
             endif else if naniter lt 9 then begin
             coarseblockpos = [ci-20*nx-20,ci-20*nx-10,ci-20*nx,ci-20*nx+10,ci-20*nx+20,$
                               ci-10*nx-20,ci-10*nx-10,ci-10*nx,ci-10*nx+10,ci-10*nx+20,$
                               ci     - 20,ci -     10,         ci     + 10,ci     + 20,$
                               ci+10*nx-20,ci+10*nx-10,ci+10*nx,ci+10*nx+10,ci+10*nx+20,$
                               ci+20*nx-20,ci+20*nx-10,ci+20*nx,ci+20*nx+10,ci+20*nx+20]
             endif else begin
                 img[nans[c]] = 0
                 continue
             endelse
             coarseblockpos = coarseblockpos[where(coarseblockpos gt 0 and coarseblockpos lt nel)]
             img[nans[c]] = median(holdimg[coarseblockpos], /even)

          endfor
          naniter = naniter + 1
          nans = where(finite(img) eq 0, ct)
        endwhile
      endif

     ;print, 'proc time: ', titerb-titera, titera-titerstart

     ; Loop back to the start and try again.  With some pixels flagged or replaced, the filter image (model) will be more accurate, 
     ; so another run should find more bad pixels.
   endwhile

   ; End of zapping iterations.
   ; At this point we have identified all the pixels with residuals that 
   ; are in excess of the model in any of the ways we've examined. 
   



   ; Run one final CR search that is based solely on nearest-neighbors: any pixels for which
   ; 4+ out of the 8 adjacent pixels are bad, is also marked bad. 

   countneighbor = intarr(nx, ny)
   nextperc = where(outmask gt 0, ct)
   if ct gt 3 then begin
      nrays = n_elements(nextperc)
      for c = 0L, nrays-1 do begin
          ci = nextperc[c]
          coarseblockpos = [ci-nx-1, ci-nx, ci-nx+1, $
                            ci-1,           ci+1,    $
                            ci+nx-1, ci+nx, ci+nx+1]
          countneighbor[coarseblockpos] = countneighbor[coarseblockpos] + 1
      endfor
      newzap = where(countneighbor ge 4 and outmask eq 0, newzaps)
      if newzaps gt 0 then begin
         outmask[newzap] = 1
         if keyword_set(quiet) eq 0 and keyword_set(silent) eq 0 then $
         logprint, '   Zapped ', strtrim(string(newzaps),2), ' additional neighboring pixels
      endif
   endif



   if ninitbad gt 0 then outmask[initbad] = 1   ; pixels that were NaN, etc. in input image

   ; Replace, this time 'for real' now that all CR pixels have been identified.
   ; Either simply flag the bad pixels with 0 or NaN, or do one more model calculation and assign
   ;  all bad pixes to the model value.
   ibad = where(outmask gt 0, ct)
   if keyword_set(zero) or keyword_set(nan) then begin
       if keyword_set(zero) then img[ibad] = 0
       if keyword_set(nan)  then img[ibad] = !values.f_nan
   endif else begin
       if keyword_set(usamp) eq 0 then replacementmodel = modelimage(img) $ ; note slow step here.
                                   else replacementmodel = usampmodelimage(img)
       if ct gt 0 then img[ibad] = replacementmodel[ibad]
   endelse

   if keyword_set(nan) and ctinitnan gt 0 then img[winitnan] = !values.f_nan
   if ctinitnegnan gt 0 then img[winitnegnan] = -!values.f_nan  ; negative nans are kept

   if keyword_set(silent) eq 0 then $
   logprint, '   Zapped ', strtrim(string(n_elements(ibad)-ninitbad),2), ' total affected pixels (', strtrim(string((n_elements(ibad)-ninitbad)*100./(1.0*nx*ny),format='(F6.3)'),2), '% of total).'
   ; Numbers don't add up, but whatever



   ; --- Write out processed images ---

   ; FITS header management
   sxaddpar, header, 'PZAPZEAL', zealfactor
   if keyword_set(usamp) then modeltype = 'usamp' else modeltype = 'reg'
   sxaddpar, header, 'PZAPMDL', modeltype
   sxaddpar, header, 'NPZAP', n_elements(ibad)-ninitbad, 'Num. of pixels zapped by pzap'

   ;get_date, now
   hist = 'Processed by pzap '+tstamp(/ut)+' UT' ;+now
   sxaddhist, hist, header

   ; Write cleaned image to disk

   if (size(outname, /tname) EQ 'STRING') then begin
      if n_elements(outmaskname) gt 0 then sxaddpar, header, 'BPM', outmaskname, 'Bad pixel mask'
      mwrfits, img, outname, header, /create
   endif else begin
      outname = img
   endelse

   ; Write mask/weight image to disk

   if (size(outmaskname, /tname) EQ 'STRING') then begin
      mwrfits, outmask, outmaskname, header, /create
   endif ;else begin
      ;mask = outmask
   ;endelse
   if (size(outweightname, /tname) EQ 'STRING') then begin
      mwrfits, 1-outmask, outweightname, header, /create
   endif ;else begin
      ;weight = 1-outmask
   ;endelse

end

