;+
; NAME: 
;   readcatheaders
;
; PURPOSE: 
;   Read in header metadata from multiple catalog files into a structure array. 
;   Used to help filter on specific objects, filters, and similar.   
;
; CALLING SEQUENCE:
;   info = readcatheaders(files, count=count, 
;      instkey=instkey, filtkey=filtkey, filt2key=filt2key, timekey=timekey, objkey=objkey) 
;       
; INPUTS:
;   files - Array of files to read in
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;   instkey  - Specify header keyword containing instrument name
;   filtkey  - Specify header keyword containing filter name (default FILTER)
;   filt2key - Specify header keyword containing second filter name (default FILTER2)
;   timekey  - Specify header keyword containing UT/time (default UTC)
;   objkey   - Specify header keyword containing object name (default OBJECT)
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;   count - Number of headers actually read in (should always equal length of files)
;       
; OUTPUTS:
;   info  - Structure array containing information about the catalogs.
;
; Last modified 4 Sep 2020.  Documentation added 3 July 2022.
;-

;;   /loud - debug option (not active now) to print out more information

function readcatheaders, files, count=count, instkey=instkey, filtkey=filtkey, filt2key=filt2key, timekey=timekey, objkey=objkey

   count = 0
   if n_elements(files) eq 0 then return, -1
   if n_elements(filtkey) eq 0 then filtkey = 'FILTER'
   if n_elements(objkey) eq 0 then filtkey = 'OBJECT'
   if n_elements(timekey) eq 0 then timekey = 'UTC'
   if clip(files[0]) eq '' then return, -1
   info = replicate({file:'', exp:0., inst:'', filt:'', filt2:'', config:'', object:'', airmass:0., zpt:0., zptunc:0., apcorr:0., nap:0, nstar:0, type:'', calfilt:'', seeing:0., rzpt:0., time:0.}, n_elements(files))
   for f = 0, n_elements(files)-1 do begin
      cat = readcat(files[f], h)
      info[f].file = files[f]
      info[f].exp = sxpar(h,'EXPTIME') ; this might not be standard.
      if n_elements(instkey) gt 0 then info[f].inst = sxpar(h,instkey)
      if n_elements(instkey) gt 0 then info[f].config += info[f].inst+'/'
      info[f].filt = clip(sxpar(h,filtkey))
      info[f].calfilt = clip(sxpar(h,'CALFILT'))
      info[f].config += info[f].filt
      if n_elements(filt2key) gt 0 then info[f].filt2 = sxpar(h,filt2key)
      if n_elements(filt2key) gt 0 then info[f].config += '/'+info[f].filt2
      timestr = sxpar(h,timekey)
      info[f].time = ten(strmid(timestr,strpos(timestr,'T')+1))
      info[f].airmass = sxpar(h,'AIRMASS')
      info[f].zpt = sxpar(h,'ZEROPT')
      info[f].zptunc = sxpar(h,'ZPTUNC')
      info[f].apcorr = sxpar(h,'APCORR')
      info[f].type = clip(sxpar(h,'PHOTCAL'))
      info[f].object = clip(sxpar(h, objkey))
      info[f].nap = clip(sxpar(h,'NSTARSAP'))
      info[f].nstar = clip(sxpar(h,'NSTARS'))
      info[f].seeing = clip(sxpar(h,'SEEING'))
      info[f].rzpt = clip(sxpar(h,'RZPT'))

   ;  if keyword_set(loud) then print, info[f].file, fpr(info[f].exp+0.001,4.2) , ' ', clip(info[f].filtdich,10), ' ', fpr(info[f].airmass,2.3), ' ', fpr(info[f].zpt,3.3), ' ', fpr(info[f].zptunc,2.3), ' ', fpr(info[f].apcorr,2.3)
   endfor

   count = n_elements(info)
   return, info
end

