;+
; NAME: 
;    readspec
;
; PURPOSE: 
;    Read in a spec file (ASCII-formatted spectrum with FITS-like header) to an IDL spec
;    structure and header string array.
;
; CALLING SEQUENCE: 
;    sp = readspec(filename, h, header=header, /timenorm, cutoff=cutoff, /unc)
;       
; INPUTS:
;    filename - Filename of spectrum to read in.
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;    /timenorm - Divide all flux-related values by the exposure time.
;    cutoff    - Minimum wavelength value.
;    /unc      - Specify that the input spectrum has an uncertainty column
;                (not normally needed; only for old files lacking column headers)
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;    h - Alternative way of obtaining the header.  
;  
; OUTPUTS: 
;    sp - Spec structure, containing wavelength, flux, pixel values, and other information
;    h  - Header as a string array of IDLAstro-compatible cards
;
; COMMENTS:
;    The routines readspec and writespec are meant to replicate the functionality of
;    readfits and writefits but for ASCII data (containing a 1D spectrum) instead of
;    FITS data.  See the documentation for writespec() for more details.  
;
; Last modified 29 Jun 2021.  Documentation added 3 July 2022.
;-

function readspec, filename, h, header=header, timenorm=timenorm, cutoff=cutoff, unc=unc

  ; reads an ASCII spectrum in Perley's standard format, analogous to a fits file.

  close, 1
  if file_test(filename) eq 0 then begin
     print, ' File '+filename+' not found!'
     return, ['']
  endif
  if strpos(filename,'.fit') ge 0 then begin
     print, ' Cannot process fits files.'
     return, ['']
  endif

  nl = countlines(filename)  
  spec = replicate({wav:0.0d, flux:0.0, sky:0.0, unc:0.0, pix:0L, ypix:0., resp:0.0, flag:0.0, dwav:0.0},nl)
  header = ['']
  fluxcol = -1 & wavcol = -1 & skycol = -1 & unccol = -1 & xcol = -1 & ycol = -1 & respcol = -1 & flagcol = -1
  
  openr, 1, filename
  inline = ''
  i = 0
  for l = 0l, nl-1 do begin
     readf, 1, inline
     inline = clip(inline)
     if strlen(inline) eq 0 then continue
     if strmid(inline,0,2) eq '##' and strpos(inline,'flux') gt 3 then begin
        ; column IDs
        cols = (strsplit(inline,/extract)) [1:*]
        wavcol = (where(strpos(cols,'wave') ge 0)) [0]
        fluxcol = (where(cols eq 'flux')) [0]
        skycol = (where(strpos(cols,'sky') ge 0)) [0]
        unccol = (where(strpos(cols,'unc') ge 0)) [0]
        xcol = (where(strpos(cols, 'xpix') ge 0)) [0]
        ycol = (where(strpos(cols, 'ypix') ge 0)) [0]
        respcol = (where(strpos(cols,'resp') ge 0)) [0]
        flagcol = (where(strpos(cols,'flag') ge 0)) [0]
        ; print, cols
        ; print, wavcol, fluxcol, skycol, unccol, xcol, ycol, respcol
     endif
     if strmid(inline,0,1) eq '#' then begin
        inline = clip(strmid(inline,1))
        ;hline = processheaderline(inline)
        inline = strtrim(inline,1)
        eqsym = '='
        eqpos = strpos(inline,'=')
        if eqpos lt 0 then begin
          eqsym = ':'
          eqpos = strpos(inline,':')
        endif
        if eqpos lt 0 then continue
        if eqpos lt 8 or (eqpos eq 8 and eqsym eq ':') then $
           inline = clip(strmid(inline,0,eqpos),8)+'='+strmid(inline,eqpos+1)
        if eqpos gt 8 then $
           inline = clip(strmid(inline,0,eqpos))+'='+strmid(inline,eqpos+1) ; sxpar doesn't work on these...
        if strlen(inline) gt 10 and strmid(inline,0,1) ne '#' then begin
          header = [header, inline]
        endif
     endif else begin
        splitline = strsplit(inline, /extract)
        ni = n_elements(splitline)
        if ni eq 0 then continue

        if fluxcol eq -1 then begin
          ; no column header: guess based on number of columns
          if ni eq 1 then begin
             fluxcol = 0
          endif else begin
            wavcol = 0
            fluxcol = 1
            if ni ge 3 and keyword_set(unc) eq 0 then skycol = 2
            if ni ge 3 and keyword_set(unc) eq 1 then unccol = 2
            if ni ge 4 and keyword_set(unc) eq 0 then unccol = 3
            if ni ge 4 and keyword_set(unc) eq 1 then skycol = 3
            if ni ge 5 then xcol = 4
            if ni ge 6 then ycol = 5
            if ni ge 7 then respcol = 6
            if ni ge 8 then flagcol = 7
          endelse
        endif

        if wavcol ge 0  and wavcol lt ni  then spec[i].wav  = float(splitline[wavcol])
        if fluxcol ge 0 and fluxcol lt ni then spec[i].flux = float(splitline[fluxcol])
        if skycol ge 0  and skycol lt ni  then spec[i].sky  = float(splitline[skycol])
        if unccol ge 0  and unccol lt ni  then spec[i].unc  = float(splitline[unccol])
        if xcol ge 0    and xcol lt ni    then spec[i].pix  = float(splitline[xcol])
        if ycol ge 0    and ycol lt ni    then spec[i].ypix = float(splitline[ycol])
        if respcol ge 0 and respcol lt ni then spec[i].resp = float(splitline[respcol])
        if flagcol ge 0 and flagcol lt ni then spec[i].flag = float(splitline[flagcol])

        if fluxcol eq 0 and xcol eq -1 then spec[i].pix = i

        i += 1
     endelse
  endfor
  if n_elements(header) gt 1 then header = header[1:*]
  spec = spec[0:i-1]

  if keyword_set(timenorm) then begin
       tn = sxpar(header,'EXPTIME', count=ct)
       if ct eq 0 then tn = sxpar(header,'ELAPTIME', count=ct)
       if ct eq 0 then tn = sxpar(header,'EXPOSURE', count=ct)
       if ct eq 0 then begin
         print, ' WARNING:  Cannot normalize exposure time!'
         tn = 1.0
       endif
       spec.flux /= tn
       spec.unc /= tn
       spec.sky /= tn
  endif

  n = n_elements(spec)
  spec.dwav = (shift(spec.wav,-1) - shift(spec.wav,+1))/2.
  spec[0].dwav = spec[1].dwav
  spec[n-1].dwav = spec[n-2].dwav
  negdwav = where(spec.dwav lt 0, ctnegdwav)
  if ctnegdwav gt 0 and ctnegdwav lt n then begin
      print, ' Warning: Wavelength solution for this spectrum is double-valued.'
      print, '     Removing points below the inflection. '
   ; This is a temporary solution; you won't always want this behavior.  Work on the wavelength solver to
   ;   recognize and stop such bad solutions earlier in the process.
      spec = spec[where(spec.dwav gt 0.)]
  endif

  if n_elements(cutoff) ge 1 then begin
     spec = spec[where(spec.wav gt cutoff[0])]
  endif

  close, 1
  h = header
  return, spec

end

