;+
; NAME: 
;   removelines
;
; PURPOSE:
;   Remove wavelength regions in a flux standard affected by spectral lines.
;   (For response-function fitting.)
;
; CALLING SEQUENCE: 
;   removelines, wav, flux, newflux, lines=lines, vel=vel, /interpolate, /nan,
;                linezone=linezone
;       
; INPUT/OUTPUT:
;   wav  - Wavelength array for the standard star
;   flux - Flux array for the standard star 
;
; INPUT KEYWORDS:
;   lines - List of lines (wavelengths) that require removal.
;   vel - List of "velocities" (relative widths) of each line, or a scalar to apply to all
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;   /interpolate - Interpolate over the affected regions, rather than remove them entirely
;   /nan - Set the affected regions to NaN, rather than remove them entirely.
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;   linezone - Boolean array on original coordinates; 0 if not line and 1 if line-affected
;       
; OUTPUTS: 
;   wav - In default mode, the input spectrum is directly modified to remove line regions
;   flux - In default or /nan mode, the input spectrum is directly modified to remove line
;          regions
;   newflux - In /interpolate mode, contains the interpolated-over spectrum
;
; Note - the behaviour above is obviously confusing as the treatment of input/output arrays
; depends on the settings; this should be revisited.
;
; Last modified 8 Apr 2018 .  Documentation added 3 July 2022.
;-

pro removelines, wav, flux, newflux, lines=lines, vel=vel, interpolate=interpolate, nan=nan, linezone=linezone

 ; remove lines in a flux star.

 ;good = intarr(n_elements(wav))+1   ; array indicating values to keep (1) versus remove or interpolate over (0)
 good = finite(flux) ;  pre-existing bad values are treated like lines now.

 for l = 0, n_elements(lines)-1 do begin
   cwav = lines[l]
   if n_elements(vel) gt 1 then v = vel[l] else v = vel
   linez = where(wav gt cwav*(1-v) and wav lt cwav*(1+v), ct)
   if ct gt 0 then good[linez] = 0
 endfor
 good[0:4] = 1
 good[n_elements(good)-11:n_elements(good)-1] = 1   ; never interpolate the edges

 linezone = abs(1 - good) ; a bool array for return to user
 wgood = where(good, ct, complement=notgood)
 if ct eq 0 then print, 'Entire spectrum is affected by lines?'

 if keyword_set(interpolate) eq 0 and keyword_set(nan) eq 0 then begin
    ; default behavior is just to trim out the line-affected parts of the spectrum
    wav = wav[wgood]
    flux = flux[wgood]
 endif else begin
    ; can also specify to mark with NaN, or interpolate over
    if keyword_set(nan) then begin
       flux[notgood] = !values.f_nan
    endif
    if keyword_set(interpolate) then begin
       newflux = flux
       if ct lt n_elements(good) then begin
       ; median-filter the array before interpolation
       ;fluxmedfilt = median(flux,5)
       ;newflux = interpol2(fluxmedfilt[wgood], wav[wgood], wav)
         mflux = flux
         mflux[notgood] = !values.f_nan
         fluxmedfilt = median(mflux,15)
         interpflux = interpol2(fluxmedfilt[wgood], wav[wgood], wav[notgood])
         newflux[notgood] = interpflux
       endif
    endif
 endelse

end

