;+
; NAME: selfsupersky.pro
;
; PURPOSE: 
;     Use a set of science images (only) to flat-field or sky-subtract each image in the 
;     group by dividing/subtracting each image by the median of the remaining images.  
;
; CALLING SEQUENCE: 
;    selfsupersky, filelist, outflatlist=outflatlist, outfilelist=outfilelist, 
;                 subtract=subtract
;       
; INPUTS:
;    filelist - List of files to process (should be at least 7 ideally)
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;    outflatlist - Filenames of flat-fields used for each image written to disk
;    outfilelist - Filenames of flat-divided images written to disk
;    subtract    - Instead of flattening the image, subtract away the sky
;                  (Use if already flat-fielded to subtract sky residuals)
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;       
; OUTPUTS: 
;
; REQUIREMENTS: 
;
; Last modified 30 May  2014.  Documentation added 2 July 2022.
;-

pro selfsupersky, filelist, outflatlist=outflatlist, outfilelist=outfilelist, subtract=subtract

; Take a block of files, make a supersky flat tailored to each file by taking the median of all the others, and divide by it.
; Useful when you need to take a superskyflat and have only a few files (e.g., z-band imaging with a large telescope)

; Requires the data to be cropped and already reasonably flatfielded.

   files = filelist

   nfiles = n_elements(files)

   if nfiles lt 5 then begin
      print, 'Very few files!  This will be a marginal flat...'
   endif

   h = headfits(files[0])
   nx = sxpar(h, 'NAXIS1')
   ny = sxpar(h, 'NAXIS2')

   data = fltarr(nx, ny, nfiles) + !values.f_nan
   maskdata = data
   meds = fltarr(nfiles)
   
   ; Read in the files
 
   print, '  Reading in files', format='($,A)'
   for f = 0, nfiles-1 do begin
      indata = readfits(files[f], h)
      s = size(indata)
      if s[1] ne nx or s[2] ne ny then begin
         print, files[f], ' does not have proper dimensions.'
         return
      endif
      data[*,*,f] = indata
      meds[f] = median(indata)
      maskdata[*,*,f] = indata/meds[f]
      print, '.', format='($,A)'
   endfor
   print



   ; Remove objects

   ; This is a really crappy object identifier.  The MOSFIRE one is better but slower.
   print, '  Identifying objects', format='($,A)'

   if n_elements(objthresh) eq 0 then objthresh = 6
   if n_elements(objbuffer) eq 0 then objbuffer = 12
   for f = 0, nfiles-1 do begin
      if keyword_set(verbose) then print, '  ', removepath(files[f])
      maskdataf = maskdata[*,*,f]
      mask = fltarr(nx, ny)
      satmask = fltarr(nx, ny)
      sigclipstats, maskdataf, sigmahi=5, sigmalo=5, median=datamed, stdev=datastdev

      sourcepixels = where(maskdataf ge datamed + objthresh * datastdev, ctsourcepix)
      if ctsourcepix gt 0 then begin
         mask[sourcepixels] = 1
         smoothmask = smooth(mask, 1+2*objbuffer, /edge_truncate)
      endif
      ctsatpix = 0
      if n_elements(satlevel) gt 0 then begin
         satpixels = where(maskdataf ge satlevel, ctsatpix)
         if ctsatpix gt 0 then begin
            satmask[satpixels] = 1
            smoothsatmask = smooth(mask, 1+2*satradius, /edge_truncate)
         endif
      endif

      if ctsourcepix gt 0 then begin
         sourcepixels = where(smoothmask gt 1.0E-6, ct)
         maskdataf[sourcepixels] = !Values.F_NAN
      endif
      if ctsatpix gt 0 then begin
         nearsatpixels = where(smoothsatmask gt 1.0E-6, ct)
         maskdataf[nearsatpixels] = !Values.F_NAN
      endif

      maskdata[*,*,f] = maskdataf  ; inefficient to copy over unaffected elements, but oh well
      print, '.', format='($,A)'
   endfor
   print


   outflats = fltarr(nx, ny, nfiles)

   print, '  Producing output files', format='($,A)'
   for f = 0, nfiles-1 do begin
      maskdatamod = maskdata
      maskdatamod[*,*,f] = !values.f_nan

      outflatf = median(maskdatamod, dimension=3, /even) 

      ; this is just too time-consuming.
      ;initnotnan = finite(data[*,*,f])
      ;for i = 0, 7 do begin  ; would be nice to find a way to speed this up.
      ;   wnewnan = where(finite(outflatf) eq 0 and initnotnan, ctnewnan)
      ;   if ctnewnan eq 0 then break
      ;   medflat = median(outflatf, 3, /even)
      ;   outflatf[wnewnan] = medflat[wnewnan]
      ;endfor


      fh = headfits(files[f])
      sxaddpar, fh, 'NSUPFLAT', nfiles-1
      fi = 0
      for ff = 0, nfiles-1 do begin
         fi += 1
         if ff ne f then sxaddpar, fh, 'SUPFLT'+clip(fi), files[ff]
      endfor

      print, '.', format='($,A)'

      if n_elements(outflatlist) gt f then writefits, outflatlist[f], outflatf, fh
      ;;writefits, repstr(outfilelist[f],'.fits','.refl.fits'), outflatf, fh

      if keyword_set(subtract) eq 0 then outflattenedf = data[*,*,f] / outflatf $
                                    else outflattenedf = data[*,*,f] - outflatf*meds[f]
      if n_elements(outfilelist) gt f then writefits, outfilelist[f], outflattenedf, fh
   endfor 
   print
   if n_elements(outflatlist) gt 0 then print, n_elements(outflatlist), ' flats written'
   if n_elements(outfilelist) gt 0 then print, n_elements(outfilelist), ' files written'


end

