;+
; NAME:
;   specextract
;
; PURPOSE:
;   Extraction of 2D spectra, including some capabilities to automatically trace and wavelength-calibrate
;   UPDATE - wavelength calibration moved to specwavcal, documentation not yet up to date
;
; CALLING SEQUENCE:
;   specextract, filename, ycent,
;      xsearch=xsearch, ysearch=ysearch, xinc=xinc, 
;      tracefile=tracefile, traceycent=traceycent, fixtrace=fixtrace,
;      arcfile=arcfile, skyfile=skyfile, crfile=crfile,
;      aprad=aprad, apsharp=apsharp, bakdist=bakdist, bakwidth=bakwidth, yoff=yoff, readdsky=readdsky,
;      outfile=outfile, keywords=keywords, success=success, setkey=setkey, setval=setval, traceonly=traceonly, 
;      savetrace=savetrace, quiet=quiet, maxsaturated=maxsaturated
;
; INPUTS:
;     I/O parameters
;   filename -   The 2D spectrum.
;   outfile  -   Output file for the extracted 1D spectrum.
;
;     Tracing parameters
;   ycent      - The y coordinate (+-few pix, in middle of image) of the object to extract
;   xsearch    - Region along wavelength axis to search for the object to trace
;   ysearch    - Region along spatial axis to search for the object to trace
;   xinc       - Step size to use in tracing.
;   tracefile  - External file to use for trace instead of current file (use yoff for offset)
;                Can be a text file (list of centroid positions)
;   traceycent - The y coordinate (+-few pix) of object to trace (if different from object to extract)
;   /fixtrace  - Set to use exact ycent (instead of searching nearest few pixels)
;   /traceonly - Stop after tracing and do not proceed to extraction.
;   /savetrace - Save trace function to a file (used only for debugging)
;
;     Extraction parameters
;   aprad    -   Radius from the central trace to extract (half width)
;   apsharp  -   Sharpness parameter for extraction kernel:  2=gaussian, 6=default, infinity=square, -1=subpixel square
;   bakdist  -   Optional background extraction separation.  default=7/4xaprad if not set.
;   bakwidth -   Background separation width (full width).  Set to 0 to suppress background subtraction.
;   yoff     -   Additional extraction center offset in pixels from best-fit trace center
;   /readdsky -  Add the sky spectrum to the science spectrum after extraction.
;                (Used if there are concerns the sky was not well modeled and contains source flux.)
;   maxsaturated - Maximum number of saturated pixels to allow to consider extraction valid.
;                If exceeded, does not write spectrum and returns an error.
;
;     Keywords indicating supplementary files to extract along same trace function
;   arcfile  -   2D spectrum of an arc (filename) to extract 
;   skyfile  -   2D spectrum (filename or extension number) of the sky to extract
;   crfile   -   2D map (filename or extension number) of cosmic ray mask to extract

;  
;     Other parameters
;   quiet    -   Suppress output 
;   keywords -   List of FITS header keywords to preserve in output file
;   setkey   -   Lits of keywords to add to header
;   setval   -   Values of setkey keywords to add to header
;   setkey   -   Set additional keywords (keyword names)
;   setval   -   Set additional keywords (values)
;
; OUTPUT KEYWORDS:
;   success  -   Returns a negative flag if procedure failed, or 1 on success
;
; OUTPUTS:
;   The extracted spectrum is written to disk as outfile.
;
; Last modified June 2022; documentation updated July 2022
;-
;------------------------------------------------------------------------------

function ekern, ny, yc, rad, warn=warn
   ; generate extraction kernel with appropriate subpixel weighting at the edges.

   y = findgen(ny) ; integer y location is the center of the pixel
   k = fltarr(ny) ; kernel weight
   yu = yc+rad ; sub-pixel position of upper limit
   yd = yc-rad ; sub-pixel position of lower limit
   yux = fix(round(yu)) ; the pixel intersecting the aperture boundary
   ydx = fix(round(yd)) ; the pixel intersecting the aperture boundary
   wm = where(y gt ydx and y lt yux, ctm) ; fully contained within the aperture
   if ctm gt 0 then k[wm] = 1

   ufrac = yu-(yux-0.5)
   dfrac = (ydx+0.5)-yd

   if yux lt ny-1 then k[yux] = ufrac
   if ydx ge 0    then k[ydx] = dfrac
   if abs(total(k)-2*rad) gt 0.05 then warn = 1

   return, k

end

function findobject, img, xsearch=xsearch, ysearch=ysearch, apmin=apmin, apmax=apmax

      s = size(img)
      nx = s[1]
      ny = s[2]

      ; blind search algorithm not rigorously tested for faint objects

      if n_elements(ysearch) ge 2 then begin
         ysearchmin = round(ysearch[0])
         ysearchmax = round(ysearch[1])
      endif
      if n_elements(ysearch) eq 0 then begin
         ysearchmin = 0  + 2
         ysearchmax = ny - 2
      endif

      if n_elements(xsearch) ge 2 then begin
         xmin = round(xsearch[0])
         xmax = round(xsearch[1])
      endif else begin
         xmin = (nx*4)/10
         xmax = (nx*6)/10
      endelse

      logprint, '  Searching for trace between y=',clip(ysearchmin),'-',clip(ysearchmax), '  (x=', clip(xmin),'-',clip(xmax),')'

      wavsum = fltarr(ysearchmax-ysearchmin)
      for y = ysearchmin, ysearchmax-1 do begin
         wavsum[y-ysearchmin] = median(img[xmin:xmax,y]) 
      endfor
      kernel = [0.2,1.0,1.5,1.0,0.2]
      if n_elements(kernel) gt n_elements(wavsum) then kernel = [0.3,1.0,0.3]
      if n_elements(kernel) gt n_elements(wavsum) then kernel = [1.0]
      kernel = kernel / total(kernel)
      wavsumconv = convol(wavsum, kernel, /edge_truncate)

      centersigma = max(wavsum)/stdev(wavsum)   ; stdev is a huge overestimate for a bright star being present.
      if centersigma lt 2 then begin
         logprint, '  WARNING:  Cannot find the source!   Best center is only ', clip(fpr(centersigma,5.2)), '-sigma above background.'
         logprint, '  Continuing anyway, but you will likely want to rerun after specifying ysearch.'
      endif

      avcenty = ysearchmin + (where(wavsumconv eq max(wavsumconv))) [0] ; max val after convolution
      if avcenty lt ysearchmin then avcenty = ysearchmin
      if avcenty gt ysearchmax then avcenty = ysearchmax

      profmax = max(wavsum)
      baseline = median(wavsum, /even)
      for y = avcenty, ysearchmax-1 do begin
         if wavsum[y-ysearchmin] lt baseline + (profmax-baseline)*0.2 then begin
            apmax = y
            break
         endif
      endfor
      for y = avcenty, ysearchmin, -1 do begin  ; formerly ysearchmin-1, which caused a rare error...
         if wavsum[y-ysearchmin] lt baseline + (profmax-baseline)*0.2 then begin
            apmin = y
            break
         endif
      endfor
     
      if n_elements(apmax) eq 0 then apmax = ysearchmax
      if n_elements(apmin) eq 0 then apmin = ysearchmin

      if (apmax-avcenty) gt 6 then apmax = avcenty + 5
      if (avcenty-apmin) gt 6 then apmin = avcenty - 5

      logprint, '  Found object at y=', strtrim(avcenty,2), ' (+', clip(apmax-avcenty), ' -', clip(avcenty-apmin), ')'

   return, avcenty

end

function traceobject, sci, avcenty=avcenty, xinc=xinc, startx=startx
      ; add avcentx to allow starting further from middle?

      ; trace the object across the chip
   
      ; march across the image, left to right, in blocks of 100, summing up horizontally for every y to get 
      ; the variable spatial profile

      debug = 0
      order = 3  ; default order (might change depending on significance)
      tracecheckplot = 0

      imsize = size(sci)
      nx = imsize[1]
      ny = imsize[2]

      xmax = nx-1

      ni = fix(xmax/xinc)
      centy = intarr(ni)
      centx = intarr(ni)
      centsig = fltarr(ni)
      isedge = fltarr(ni)
      isgood = intarr(ni)
      initscansize = 5;3
      scansize = 4;2  ; how far away from the most recently-found center to look for trace variation
      wavsum = fltarr(ny)  ; fltarr(scansize*2+1)
      lastgoodtracei = 0
      ntraceloss = 0
      ntracefound = 0
      lastgoodtracex = -1
      lastgoodtracey = avcenty
      ;kernel = [0.1,0.2,0.4,0.2,0.1]  ; this might be the better kernel for point sources, keep an eye on this
      kernel = [0.15,0.2,0.3,0.2,0.15]  ; should use the input aperture as the size?
      kernel = kernel / total(kernel)

      if debug gt 0 then begin
         window, 0, xsize=700, ysize=700
         !p.multi = [0, 4, 4]
      endif

      if n_elements(startx) eq 0 then starti = fix(i/2) else starti = fix(startx/xinc)
      direction = 1
      i = starti
      for j = 0, ni-1 do begin        ; now : starts in the middle
         if j eq 0 then begin
            lasty = avcenty
            rady = initscansize
         endif else begin
            i += direction
            rady = scansize
         endelse
         if i eq ni then begin
            direction = -direction
            i = starti-1
            lasty = firsty
         endif
         
         ymin = (lasty-rady) > 0
         ymax = (lasty+rady) < (ny-1)
         
         x1 = fix(i*xinc)
         x2 = fix(x1+xinc-1) < xmax

         ; get the maximum (block-summed) pixel and significance
         for y = 0, ny-1 do begin
            wavsum[y] = total(sci[x1:x2,y])  ; sum across xinc pix within the block
         endfor
         bad = where(finite(wavsum) eq 0, ct)
         if ct gt 0 then wavsum[bad] = 0.
         djs_iterstat, wavsum, mean=fullymean, median=fullymedian, sigma=noiserms
         wavsum -= fullymedian
         wavsumsearch = wavsum[ymin:ymax]
         wavsumconv = convol(wavsum, kernel, /edge_truncate) ; not worried about extra proc time
         wavsumsearchconv = wavsumconv[ymin:ymax]
         maxval = max(wavsumsearchconv)
         significance = maxval/noiserms

         ; get the y-coordinate of maximum pixel
         ymaxpix = ymin + (where(wavsumsearchconv eq maxval)) [0]
         if ymaxpix eq ymin or ymaxpix eq ymax then isedge[i] = 1

         ; print, wavsumsearch, ymaxpix

         ; get the refined centroid
         rymin = (ymaxpix-2) > 0
         rymax = (ymaxpix+2) < (ny-1)
         rwavsum = wavsum[rymin:rymax]
         ycentroid = total((findgen(rymax-rymin+1)+rymin) * rwavsum) / total(rwavsum)
         centxi = i*xinc + xinc/2
  
         if debug gt 0 then begin
            plot,  indgen(ymax-ymin+5*2+1)+ymin-5, wavsum[ymin-5:ymax+5], /nodata, /xsty, /ysty
            oplot, indgen(ymax-ymin+5*2+1)+ymin-5, wavsum[ymin-5:ymax+5], psym=10, color=rgb(128,128,128)
            oplot, indgen(ymax-ymin+5*2+1)+ymin-5, wavsumconv[ymin-5:ymax+5], color=rgb(128,128,0)
            oplot, indgen(ymax-ymin+1)+ymin, wavsumsearch, psym=10
            oplot, indgen(ymax-ymin+1)+ymin, wavsumsearchconv, color=rgb(255,255,0), thick=2
            oplot, [ymaxpix, ymaxpix], [0, 60000], color=rgb(0,0,255)
            oplot, [ycentroid, ycentroid], [0, 60000], color=rgb(0,255,0)

            print, '  ',clip(i,3),' ', rclip(x1,4)+'-'+clip(x2,4),rclip(ymin,4),'-',clip(ymax,4),  ymaxpix, ycentroid, significance
         endif

         if isedge[i] eq 0 then lasty = ymaxpix $
                           else lasty = fix((lasty+ymaxpix)/2) ; if on the edge, only move the aperture halfway
         if j eq 0 then firsty = lasty
         
         centx[i] = centxi
         if ycentroid gt rymin and ycentroid lt rymax then $
            centy[i] = ycentroid $
         else $
            centy[i] = ymaxpix
         centsig[i] = significance
      endfor

      sigthresh = 2.5
      if max(centsig) lt 5  then sigthresh = 2.0
      if max(centsig) gt 30 then sigthresh = 3.5
      if max(centsig) gt 50 then sigthresh = 5.0
      if max(centsig) gt 200 then sigthresh = 9.0
      if max(centsig) gt 800 then sigthresh = 30.0  ; this is mainly to avoid centroiding on ghosts or second order light.

      goodtrace = where(centsig gt sigthresh, ct)
      if ct lt 3 then begin
          logprint, '  Failed to trace - insufficient signal.'
          if tracecheckplot then psclose
          return, -1
      endif
      if ct lt n_elements(centsig)/2 then order = 1 
      if ct ge 6 and max(centsig) gt 500 then order += 1    ; fit a higher-order solution to very strong sources (standards)
      ; if ct ge 6 and median(centsig) gt 50 then order += 1     ; could consider turning these on with more testing...
      ; if ct ge 7 and median(centsig) gt 100 then order += 1  

      ; need to make this plot more useful - specific file output, write pixels, etc.

      !p.multi = 0
      if tracecheckplot then begin
         psopen, 'traceplot.ps', xsize=7, ysize=5, /inch
         device, /color
         colors = transpose([[0,0,0],$
                          [180,180,210],$ 
                          [64, 64, 192],$ 
                          [64, 136,128],$ 
                          [64, 158, 64],$  
                          [112,112, 64],$ 
                          [192, 64, 64],$
                          [128, 64, 128]$
                       ])
          tvlct, colors

          plot, centx, centy, /ynozero
          ;oplot, centx, centy2, linestyle=1   centy2 is not reliable 
       endif

      ; fit a polynomial to the profile, excluding possible bad traces



      tracepar = polyiterfit(centx[goodtrace], centy[goodtrace],order=order,niter=2)   ; this should be updated
      logprint, '  Trace function: '      , format='($,A)'
      for o = 0, n_elements(tracepar)-1 do begin
         logprint, clip(tracepar[o]), format='($,A)'
         if o eq 1 then logprint, 'x', format='($,A)'
         if o ge 2 then logprint, 'x^'+clip(o), format='($,A)'
         if o ne n_elements(tracepar)-1 then logprint, ' + ' , format='($,A)'
      endfor
      logprint
      if debug gt 0 then begin
         for x = 0, xmax, 250 do begin
            print, x, poly(x, tracepar)
         endfor
      endif

      tracey = poly(findgen(nx), tracepar)

      ; should consider doing the linear extrapolation method for regions beyond the end of the "goodtrace"

      if tracecheckplot then begin
         oplot, centx[goodtrace], centy[goodtrace], psym=7
         oplot, findgen(nx), tracey, linestyle=2

         psclose
      endif
      return, tracey
 
end



function extract, img, tracey, aprad, apsharp, bakdistin, bakwidthin, fbak=fbak, regfile=regfile, saturate=saturate, nsaturated=nsaturated, medbak=medbak, crmode=crmode

    if n_elements(medbak) eq 0 then medbak = 0 ; formerly 2
    filtbak = 11
      ; 0 - no median; uses a weighted sum like with the object aperture
      ; 1 - median calculated over both background zones (upper & lower), weighted
      ; 2 - each background zone is split in two and median is taken of each

    if n_elements(bakdistin) eq 1 then bakdist = [bakdistin, bakdistin] 
    if n_elements(bakdistin) ge 2 then bakdist = bakdistin
    if n_elements(bakwidthin) eq 1 then bakwidth = [bakwidthin, bakwidthin] 
    if n_elements(bakwidthin) ge 2 then bakwidth = bakwidthin

    imsize = size(img)
    nx = imsize[1]
    ny = imsize[2]

    ;extractionimage = fltarr(nx,ny) ; for debugging
    ;bkextractionimage = fltarr(nx,ny) ; for debugging

    baksharp = 12.d

    regstep = 40

    nsaturated = 0

    if n_elements(regfile) gt 0 then begin
       openw, 2, regfile
       printf, 2, '# Region file format: DS9 version 4.1'
       printf, 2, 'global color=green'
       printf, 2, 'image'
    endif

    nans = where(finite(img) eq 0, ct)
    if ct gt 0 then img[nans] = 0.
    for x = 0, nx-1 do if max(img[x,*]) eq 0 and min(img[x,*]) then img[x,*] = !values.f_nan ; blank column

    ; extract the object (and the background, if desired)
    fobj = fltarr(nx)
    fbak = fltarr(nx)

    offmsg = 0
    for x = 0, nx-1 do begin
       cy = tracey[x] 
       if cy lt 0 then begin
           if offmsg eq 0 then logprint, '   WARNING - trace extends off the chip?'
           offmsg = 1
           prevx = x
           prevy = cy
           continue
       endif

       if n_elements(regfile) gt 0 then begin
       if x mod regstep eq 0 then begin
           if x gt 0 then begin
              ; the +1 is to offset IDL/data frame (first cell 0) to DS9 frame (first cell 1)
              printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy)+','+clip(1+x)+','+clip(1+cy)+') # line=0 0 dash=1'
              printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy+aprad)+','+clip(1+x)+','+clip(1+cy+aprad)+') # line=0 0'
              printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy-aprad)+','+clip(1+x)+','+clip(1+cy-aprad)+') # line=0 0'
              if n_elements(bakwidth) gt 0 and n_elements(bakdist) gt 0 then if max(bakwidth) gt 0. then begin
               if bakdist[1] gt 0 and bakwidth[1] gt 0 then begin
               printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy+bakdist[1])+','+clip(1+x)+','+clip(1+cy+bakdist[1])+') # line=0 0 color=yellow'
               printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy+bakdist[1]+bakwidth[1])+','+clip(1+x)+','+clip(1+cy+bakdist[1]+bakwidth[1])+') # line=0 0 color=yellow'
               endif
               if bakdist[0] gt 0 and bakwidth[0] gt 0 then begin
               printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy-bakdist[0])+','+clip(1+x)+','+clip(1+cy-bakdist[0])+') # line=0 0 color=yellow'
               printf, 2, 'line('+clip(1+prevx)+','+clip(1+prevy-bakdist[0]-bakwidth[0])+','+clip(1+x)+','+clip(1+cy-bakdist[0]-bakwidth[0])+') # line=0 0 color=yellow'
               endif
             endif
             prevx = x
             prevy = cy
          endif else begin
             prevx = x
             prevy = cy
          endelse
       endif
       endif


       eky = indgen(ny)
       if apsharp gt 0 then begin
          extractionkernel = exp(-abs(((eky-cy)*1.0d/aprad)^apsharp)/2.)
       endif else begin 
          extractionkernel = ekern(ny, cy, aprad, warn=warnap)
       endelse

       ; cosmic ray pixel counter (specialized: does a pure sum over the cosmic ray mask)
       if n_elements(crmode) gt 0 then begin
         wap = where(extractionkernel gt 0.1, ctap)
         if ctap gt 0 then fobj[x] = total(img[x,wap])*1./(ctap*crmode)
         continue
       endif

       ;extractionimage[x,*] = extractionkernel

       if n_elements(bakwidth) gt 0 and n_elements(bakdist) gt 0 then if max(bakwidth) gt 0. and max(bakdist) gt 0. then begin
          bakkernelm = 0.
          bakkernelp = 0.
          if bakdist[0] gt 0 or bakwidth[0] gt 0 then begin
             if apsharp gt 0 then $
               bakkernelm = exp(-abs(((eky-(cy-(bakdist[0]+bakwidth[0]/2.)))*1.0d/(bakwidth[0]/2.))^baksharp)/2.) $
             else $
               bakkernelm = ekern(ny,cy-(bakdist[0]+bakwidth[0]/2.), bakwidth[0]/2.)
          endif
          if bakdist[1] gt 0 or bakwidth[1] gt 0 then begin
             if apsharp gt 0 then $
               bakkernelp = exp(-abs(((eky-(cy+(bakdist[1]+bakwidth[1]/2.)))*1.0d/(bakwidth[1]/2.))^baksharp)/2.) $
             else $
               bakkernelp = ekern(ny,cy+(bakdist[1]+bakwidth[1]/2.), bakwidth[1]/2.)
          endif
          bakkernel = bakkernelm + bakkernelp
       endif

       ; weight the two halves

       ; if x eq 1232 then begin
       ;    window, 2
       ;    plot, extractionkernel, xrange=[cy-bakdist-3*bakwidth/2,cy+bakdist+3*bakwidth/2], thick=3, /xstyle
       ;    if bakwidth gt 0. then oplot, bakkernel, linestyle=1, thick=2
       ;    oplot, sci[x,*]/(median(sci[x,cy-20:cy+20])+4*stdev(sci[x,cy-20:cy+20])), psym=10
       ; endif

       if n_elements(saturate) gt 0 then begin
          saturated = where(img[x,*] gt saturate and extractionkernel gt 0.01*max(extractionkernel), ctsaturated)
          nsaturated += ctsaturated
       endif

       fobj[x] = total(img[x,*]*extractionkernel)


       if n_elements(bakwidth) gt 0 and n_elements(bakdist) gt 0 then if max(bakwidth) gt 0. and max(bakdist) gt 0. then begin

          imgx = (img[x,*]) [*]
          if filtbak gt 1 then begin
                 ; remove the object pixels so they don't influece the median
             imgx[where(extractionkernel gt 0.1)] = !values.f_nan 
             imgx = median(imgx, filtbak, /even) ; median filter
             imgx[where(finite(imgx) eq 0)] = 0 ; leftover NaNs will cause problems if convolved, even if x0
          endif

          if medbak eq 0 then begin
             fbak[x] = total(imgx*bakkernel) * total(extractionkernel)/total(bakkernel)
             ;bkextractionimage[x,*] = imgx*bakkernel
          endif else begin
             if n_elements(bakkernelm) le 1 or n_elements(bakkernelp) le 1 then begin
               wb = where(bakkernel gt 0.5, ctm)
               if ct eq 0 then fbak[x] = !values.f_nan else $
                 fbak[x] = median(imgx[wb], /even) * total(extractionkernel)
             endif else begin
               if medbak eq 1 then begin
                 wbm = where(bakkernelm gt 0.5, ctm)
                 wbp = where(bakkernelp gt 0.5, ctp)
                 if ctm ge 2 then avbm = median(imgx[wbm], /even) else avbm = !values.f_nan
                 if ctp ge 2 then avbp = median(imgx[wbp], /even) else avbp = !values.f_nan
                 fbak[x] = (avbm*(ctm*bakdist[0])+avbp*(ctp*bakdist[1]))/(ctm*bakdist[0]+ctp*bakdist[1]) * total(extractionkernel)
               endif
               if medbak eq 2 then begin
                 wbm1 = where(bakkernelm gt 0.25 and eky lt cy-bakdist[0], ctm1)
                 wbm2 = where(bakkernelm gt 0.25 and eky ge cy-bakdist[0], ctm2)
                 wbp1 = where(bakkernelp gt 0.25 and eky le cy+bakdist[1], ctp1)
                 wbp2 = where(bakkernelp gt 0.25 and eky gt cy+bakdist[1], ctp2)
                 if ctm1 ge 1 then avbm1 = median(imgx[wbm1], /even) else avbm1 = !values.f_nan
                 if ctm2 ge 1 then avbm2 = median(imgx[wbm2], /even) else avbm2 = !values.f_nan
                 if ctp1 ge 1 then avbp1 = median(imgx[wbp1], /even) else avbp1 = !values.f_nan
                 if ctp2 ge 1 then avbp2 = median(imgx[wbp2], /even) else avbp2 = !values.f_nan
                 ctm = ctm1+ctm2
                 ctp = ctp1+ctp2
                 avbm = median([avbm1,avbm2], /even) ; avbm, 
                 avbp = median([avbp1,avbp2], /even) ; avbp,
                 fbak[x] = (avbm*(ctm*bakdist[0])+avbp*(ctp*bakdist[1]))/(ctm*bakdist[0]+ctp*bakdist[1]) * total(extractionkernel)
               endif
             endelse ; both upper and lower kernels exist
          endelse ; medbak gt 0
       endif ; background exists
    endfor

    if n_elements(regfile) then close, 2

    ;if n_elements(crmode) eq 0 then writefits, 'extractionimage.fits', extractionimage
    ;if n_elements(crmode) eq 0 and n_elements(bakwidth) gt 0 and n_elements(bakdist) gt 0 then writefits, 'bkextractionimage.fits', bkextractionimage

    return, fobj
end


pro specextract, filename, ycent, $
   xsearch=xsearch, ysearch=ysearch, xinc=xinc, tracefile=tracefile, traceycent=traceycent, fixtrace=fixtrace, $
   arcfile=arcfile, skyfile=skyfile, crfile=crfile, $
   aprad=aprad, apsharp=apsharp, bakdist=bakdist, bakwidth=bakwidth, yoff=yoff, readdsky=readdsky, $
   outfile=outfile, keywords=keywords, success=success, setkey=setkey, setval=setval, traceonly=traceonly, savetrace=savetrace, quiet=quiet, maxsaturated=maxsaturated

   ; some obsolete code below; these lists are no longer needed
   ;if n_elements(arclist) eq 0 then arclist = '~/research/redux/arc/HgArNegood.list'
   ;if n_elements(brightarclist) eq 0 then brightarclist = '~/research/redux/arc/HgNeArbright.list'
   ;if n_elements(skylist) eq 0 then skylist = '~/research/Useful/skyspec/gident_most.dat'
   if n_elements(skylist) eq 0 then skylist = ''

   if n_elements(outfile) eq 0 then outfile = repstr(filename,'.fits','.spec')
   if n_elements(maxsaturated) eq 0 then maxsaturated = 99999

   filefail = 0
   checkfiles = [filename]
   if n_elements(arcfile) then checkfiles = [checkfiles, arcfile]
   if n_elements(arclist) eq 1 then checkfiles = [checkfiles, arclist]
   if n_elements(brightarclist) eq 1 then checkfiles = [checkfiles, brightarclist]
   if n_elements(tracefile) gt 0 then checkfiles = [checkfiles, tracefile]
   if n_elements(skyfile) gt 0 then begin
      skyfile = clip(skyfile)
      if skylist ne '' then checkfiles = [checkfiles, skylist]
      if strlen(skyfile) gt 2 then begin
         checkfiles = [checkfiles, skyfile]
      endif else begin
         ; need to check that the extension exists - could use fits_info
      endelse
   endif
   if n_elements(crfile) gt 0 then begin
      crfile = clip(crfile)
      if strlen(crfile) gt 2 then begin
         checkfiles = [checkfiles, crfile]
      endif else begin
         ; need to check that the extension exists - could use fits_info
      endelse
   endif
   for c = 0, n_elements(checkfiles)-1 do begin
      if file_test(checkfiles[c]) eq 0 then begin
         logprint, '  Cannot find ', checkfiles[c]
         filefail += 1
      endif
   endfor
   if filefail gt 0 then begin 
      ; print
      success = -1 ; couldn't find input file
      return
   endif


   ; --- Read in file ---

   sci = mrdfits(filename,0,h, /silent)

   imsize = size(sci)
   nx = imsize[1]
   ny = imsize[2]
   if nx eq 0 or ny eq 0 then begin
     logprint, '  Not a valid 2D FITS spectrum.'
     ; print
     success = -2 ; bad file
     return
   endif

   if n_elements(skyfile) gt 0 then begin
       if strlen(clip(skyfile)) gt 1 then begin
          sky = mrdfits(skyfile, 0, skyh, /silent)
       endif else begin
          sky = mrdfits(filename, fix(skyfile), skyh, /silent) ; use extension
       endelse
   endif
   if n_elements(crfile) gt 0 then begin ; cosmic ray file
       if strlen(clip(crfile)) gt 1 then begin
          cr = mrdfits(crfile, 0, skyh, /silent)
       endif else begin
          cr = mrdfits(filename, fix(crfile), crh, /silent) ; use extension
       endelse
   endif

   if n_elements(xsearch) eq 0 then xsearch = [nx*0.4, nx*0.6]
   if xsearch[0] lt 0 then          xsearch = [nx*0.4, nx*0.6] 
   xcent = fix(round(median(xsearch,/even)))

   sat = sxpar(h,'SATURATE', count=ct)
   if ct gt 0 then saturate = sat
   

   ; --- Read in trace file or trace function ---

   if n_elements(xinc) eq 0 then xinc = nx/32.  ; formerly nx/16.

   if n_elements(tracefile) gt 0 then begin
      if strpos(tracefile,'.fit') gt 0 then begin
         if file_test(tracefile) eq 0 then begin
            logprint, 'ERROR: trace file ', tracefile, ' does not exist.'
         endif
         traceimg = mrdfits(tracefile,0,traceh, /silent)
      endif else begin
         ntl = countlines(tracefile)
         if ntl le 2 then begin
            openr, 3, tracefile
            inline = ''
            readf, 3, inline
            while eof(3) eq 0 and (strlen(inline) eq 0 or strmid(clip(inline),0,1) eq '#') do readf, 3, inline
            close, 3
            tracepar = float(strsplit(inline,/extract))
            tracey = poly(findgen(nx), tracepar)
         endif else begin
            rdfloat, tracefile, tracey
         endelse
      endelse
   endif else begin
      traceimg = sci  ; (science image *is* the trace image unless specified)
   endelse


   ; --- Find and trace tracing object, if necessary ---

   if n_elements(traceimg) gt 0 then begin

      if n_elements(traceycent) eq 0 and n_elements(ycent) gt 0 and n_elements(tracefile) eq 0 then traceycent = ycent
      if n_elements(traceycent) eq 0 and n_elements(ysearch) eq 1 then traceycent = ysearch

      ;;if n_elements(traceycent) gt 0 then stop
      if n_elements(traceycent) eq 0 then begin
         traceycent = findobject(traceimg,xsearch=xsearch,ysearch=ysearch, apmin=yfindmin, apmax=yfindmax)
      endif ;else begin
         ;dum = findobject(traceimg,xsearch=xsearch,ysearch=ysearch, apmin=yfindmin, apmax=yfindmax)
      ;endelse
      if n_elements(ycent) eq 0 then ycent = traceycent

      if n_elements(aprad) eq 0 then begin
          if n_elements(yfindmax) gt 0 then aprad = min([yfindmax-traceycent,traceycent-yfindmin]) $
                                       else aprad = 3.
          if aprad lt 2. then aprad = 2. ; need to work on improving the aperture determiner in findobject.
          logprint, '  Using aperture radius ', clip(aprad)
      endif else begin
          logprint, '  Radius was set to ', clip(aprad)
      endelse

      tracey = traceobject(traceimg, avcenty=traceycent, xinc=xinc, startx=median(xsearch,/even))
      if n_elements(tracey) eq 1 then begin
        if tracey eq -1 then begin
           ; print
           success = -3 ; trace failure
           return
        endif
      endif
   endif 

   if n_elements(aprad) eq 0 then aprad = 12
   if n_elements(apsharp) eq 0 then apsharp = -1
   if n_elements(bakdist) eq 0 then bakdist = aprad*7./4
   if n_elements(bakwidth) eq 0 then bakwidth = aprad*4.
     ; formerly didn't set these if a skyfile was given, but it's good to subtract background even with sky by default
     ; since there might be galaxy, twilight gradients, etc.


   ; --- Get optimum trace y-offset, if necessary ----


;   if keyword_set(fixtrace) eq 0 and (n_elements(tracefile) gt 0 or traceycent ne ycent)  then begin
;      ; get the optimum yoff by repeatedly extracting and measuring total flux
;      ; seems to not work
;      print, 'Using multi-extraction method which is currently buggy.....'
;      yoffsearch = ycent-tracey[nx/2] + [-3,3]
;
;      yoffarr = min(yoffsearch) + findgen(max(yoffsearch)-min(yoffsearch))
;      totfluxy = fltarr(n_elements(yoffarr))
;      for ei = 0, n_elements(yoffarr)-1 do begin
;         ftry = extract(sci, tracey+yoffarr[ei], aprad/2. > 1.5, apsharp) ; note no background
;         totfluxy[ei] = total(ftry)
;         ;print, yoffarr[ei], tracey[nx/2]+yoffarr[ei], totfluxy[ei]
;      endfor
;      yoffmax = yoffarr[(where(totfluxy eq max(totfluxy))) [0]] ; maybe fit a gaussian in the future?
;      yofffit = getlinecenter(yoffarr, totfluxy, yoffmax, fitrad=3)
;      ;print, 'Optimum yoff = ', yoff
;      ;print, yoffmax, yofffit
;      print, '  Optimum trace center at x=',clip(nx/2), ': y=', clip(yofffit+tracey[nx/2])
;      yoff = yofffit + yoff
;   endif

   ; if we used a separate tracing object, calculate the offset to the extracted object
   if keyword_set(fixtrace) and n_elements(ycent) gt 0 and n_elements(yoff) eq 0 then begin
      if n_elements(tracey) ne nx then logprint, 'WARNING - trace file has different dimensions from science file.'
      yoff = ycent-tracey[xcent]
   endif else begin
      if n_elements(yoff) eq 0 then yoff = 0
   endelse

   ypixcen = tracey+yoff

   if n_elements(savetrace) gt 0 then begin
       openw, 11, savetrace
         ;printf, 11, '# APRAD = '+aprad
         for i = 0, n_elements(ypixcen)-1 do printf, 11, clip(ypixcen[i])
       close, 11
   endif

   if keyword_set(traceonly) then return


   ; --- Extract 1D spectrum ---


   traceregfile = repstr(outfile,'.spec','.trace.reg')

   if keyword_set(readdsky) and n_elements(skyfile) gt 0 then esci = sci+sky else esci = sci

   fobj = extract(esci, ypixcen, aprad, apsharp, bakdist, bakwidth, fbak=fbak, regfile=traceregfile, saturate=saturate, nsaturated=nsaturated)

   if nsaturated gt maxsaturated then begin
     logprint, '  Error: '+clip(nsaturated)+' saturated pixels detected on the trace!!!'
     logprint, '     Not saving spectrum.'
     success = -4 ; saturation
     return
   endif
   if nsaturated gt 1 then begin
     logprint, '  WARNING: '+clip(nsaturated)+' saturated pixels detected on the trace!!!'
     logprint, '     Spectrum may not be usable.'
   endif

   pix = indgen(nx)     ; should look for crop and bin keywords...


   ; --- Extract arc at identical position ---

    if n_elements(arcfile) gt 0 then begin

      ; maxsep = nx*dispersion/2.

      arc = mrdfits(arcfile, 0, arch, /silent) ; use the arcfile

      if keyword_set(quiet) eq 0 then logprint, '  Extracting arc spectrum.'

      farc = extract(arc, ypixcen, aprad, apsharp)

   endif


   ; --- Extract sky at identical position ---

   if n_elements(skyfile) gt 0 then begin

      if keyword_set(quiet) eq 0 then logprint, '  Extracting sky spectrum.'

      fsky = extract(sky, ypixcen, aprad, apsharp)  

   endif


   ; --- Extract cosmic ray mask ---

   if n_elements(crfile) gt 0 then begin
      if keyword_set(quiet) eq 0 then logprint, '  Extracting bad pixel file.'
      nsum = sxpar(h,'NSUM') > 1
      fcr = extract(cr>(nsum*(~finite(esci))), ypixcen, aprad, apsharp, crmode=nsum)
           ; fraction of aperture hit by CR's across all exposures
   endif


   ; --- Subtract the sky from the object ---
   objspec = fobj-fbak
   if n_elements(fsky) gt 0 then $
      skyspec = fsky $
   else $
      skyspec = fbak   
   ; formerly fsky+fbak: this included wings of bright objects which added confusing signal and extra noise.

   ; --- Calculate uncertainties ---
   ; note crude treatment of uncertainties...
   readnoise = 4.6*sqrt(2.0*aprad) ; formerly, total of extractionkernel
   gain = 1.2
   uncspec = sqrt(readnoise^2 + ((objspec + skyspec) > 0)*gain)/gain


   ; --- Get keywords from FITS file---
   outh = ['']
   for k = 0, n_elements(keywords)-1 do begin
      if strlen(keywords[k]) le 8 then kval = sxpar(h, keywords[k], comment=comment, count=ct) $
                                  else kval = sxparh(h, keywords[k], comment=comment, count=ct)
      if ct ge 1 then begin
         if strlen(keywords[k]) le 8 then sxaddpar, outh, keywords[k], kval[0], comment $
                                     else h = [h, keywords[k] + ' = ' + kval[0] + ' / ' + comment]
      endif
   endfor
   sxaddpar, outh, '2DSPEC', removepath(filename), ' 2D spectrum file extracted'
   if n_elements(arcfile) gt 0 then sxaddpar, outh, 'ARCFILE', removepath(arcfile)
   if n_elements(tracefile) gt 0 then sxaddpar, outh, 'TRACFILE', removepath(tracefile), ' Tracing file'
   sxaddpar, outh, 'YEXTRACT', ypixcen[nx/2], ' Image y-coordinate in (x) middle of 2D spectrum'
   sxaddpar, outh, 'XCENT', xcent, ' Reference pixel location for aperture center'
   sxaddpar, outh, 'YCENT', ycent, ' Image y-coordinate specified by user (x=XCENT)'
   sxaddpar, outh, 'APRAD', aprad, ' Half-diameter of extracted trace'
   if apsharp gt 0 then sxaddpar, outh, 'APSHARP', apsharp, ' Sharpness parameter for extraction'
   if apsharp le 0 then sxaddpar, outh, 'APSHARP', apsharp, ' Simple aperture extraction used'
   sxaddpar, outh, 'NSATURAT', nsaturated, ' Number of saturated pixels in extraction'

   if n_elements(bakdist) eq 2 then if abs(bakdist[1]-bakdist[0]) lt 0.01 then bakdist = bakdist[0]
   sxaddpar, outh, 'BAKDIST', min(bakdist), ' Distance from trace to background inner edge'
   if n_elements(bakdist) ge 2 then sxaddpar, outh, 'BAKDISTM', bakdist[0], ' Distance to background (y-)'
   if n_elements(bakdist) ge 2 then sxaddpar, outh, 'BAKDISTP', bakdist[1], ' Distance to background (y+)'
   if n_elements(bakwidth) eq 2 then if abs(bakwidth[1]-bakwidth[0]) lt 0.01 then bakwidth = bakwidth[0]
   sxaddpar, outh, 'BAKWIDTH', min(bakwidth), ' Width of each background stripe'
   if n_elements(bakwidth) ge 2 then sxaddpar, outh, 'BAKWIDTM', bakwidth[0], ' Width of background (y-)'
   if n_elements(bakwidth) ge 2 then sxaddpar, outh, 'BAKWIDTP', bakwidth[1], ' Width of background (y-)'
   for k = 0, n_elements(setkey)-1 do begin
      if setkey[k] ne '' then sxaddpar, outh, setkey[k], setval[k]
   endfor
   ;get_date, now
   now = tstamp(/ut)
   sxaddpar, outh, 'EXTDATE', repstr(now,' ','T'), ' UT time 1D spectrum extracted by specextract'
   sxaddpar, outh, 'REDUCED', repstr(now,' ','T'), ' UT time 1D spectrum extracted by specextract'
   hist = sxpar(h,'HISTORY')
   for hi = 0, n_elements(hist)-1 do begin
     if clip(hist[hi]) ne '0' and clip(hist[hi]) ne '' then sxaddhist, hist[hi], outh
   endfor
   newhist = 'Processed by specextract '+now+' UT'
   sxaddhist, newhist, outh

   nl = n_elements(objspec)
   spec = replicate({wav:0.0d, flux:0.0, sky:0.0, unc:0.0, pix:0L, ypix:0., resp:0., flag:0.},nl)
   if n_elements(wav)     ge nl then spec.wav = wav
   spec.flux = objspec
   if n_elements(skyspec) ge nl then spec.sky = skyspec
   if n_elements(uncspec) ge nl then spec.unc = uncspec
   if n_elements(pix)     ge nl then spec.pix = pix 
   if n_elements(ypixcen) ge nl then spec.ypix = ypixcen
   if n_elements(fcr)     ge nl then spec.flag = fcr

   writespec, outfile, spec, outh
   success = 1

   if keyword_set(quiet) eq 0 then logprint, '  Spectrum saved to '+outfile


end

