;+
; NAME: subtractsky4.pro
;
; PURPOSE: 
;    Alternative sky subtraction algorithm based on the Kelson method.
;    (Has not proven robust enough for regular usage)
;
; CALLING SEQUENCE: 
;    subtractsky4, filename, /sourcefilter, outfile=outfile, nowrite=nowrite, 
;      xmin=xmin, xmax=xmax   
;
; INPUTS:
;      filename - Filename of 2D spectrum to skysubtract
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;      outfile - Filename to write subtracted spectrum to (default: 's'+filename)
;      /sourcefilter - Try to remove actual sources detected along the slit.
;      /nowrite - Debug option not to write to disk at all
;      xmin - Minimum x coordinate to use in modeling sky spectrum
;      xmax - Maximum x coordinate to use in modeling sky spectrum
;       
; OPTIONAL OUTPUT KEYWORD PARAMETERS:
;       
; OUTPUTS: 
;      The sky-subtracted spectrum is written to disk as outfile
;
; REQUIREMENTS: 
;
; Last modified 18 Jun 2018.  Documentation added 2 July 2022.
;-

; Yet another sky subtraction algorithm.

; Uses the Kelson method.

; This does not take into account and will not subtract:
;  * read noise or scattered light continuum
;  * slit width variations (they are hopefully flat-fielded out?)
;  * your object (a good thing)

; Should be very robust to cosmic rays

; Will perform poorly if there is a significant true gradient in the continuum
; (may add noise; requires testing).

; For 400/8500 data, seems to do a little worse than skysubtract2 in both speed and
; line analysis quality.  Possibly, could be improved with a better cross-correlation
; method or by fitting a higher order polynomial.  Speed could probably be boosted by 
; downsampling the data prior to the polynomial fit, or rasterizing and savgol smoothing,
; or attempting a 'b-spline' as in the traditional Kelson method.

; The new algorithm is generally better at not subtracting lines in extended objects.
; More testing is needed to evaluate if it oversubtracts tellurics or if skysubtract2
; does.

; I expect this will perform better than skysubtract2 on some other gratings with
; stronger tilts or curvature and we may want to replace within LPIPE in that case.

; Because the data is blocked this will NOT work if there are not lines throughout the
; spectrum.

; Would likely perform better if the cross-correlation correction also included a 
; wavelength offset/distortion term.

; 18 Jun  2018

function lopassfilt, f

  ; Low-pass filter to remove gradients prior to calculating cross-correlation.

  ; should run a low-pass filter on both first
  ;if nx mod 2 eq 0 then fy=[findgen(nx/2),nx/2-findgen(nx/2)]
  ;if nx mod 2 eq 1 then fy=[findgen(nx/2),nx/2-1-findgen(nx/2-1)]
  ;filter=1.0-1.0/(1+(fy/(nx/10))^10) 
  ;s1f = FFT(FFT(s1,1)*filter,-1)
  ;s2f = FFT(FFT(s2,1)*filter,-1)

  ; FFT would be ideal, but as usual let's use medians

  nxblocks = 5
  nx = n_elements(f)
  blocksize = nx/nxblocks
  blockcent = (findgen(nxblocks)+0.5)*blocksize

  xmeds = fltarr(nxblocks)
  for i = 0, nxblocks-1 do begin
     xmeds[i] = median(f[blocksize*i:(blocksize*(i+1)-1)<(nx-1)])
  endfor
  
  par = poly_fit(blockcent, xmeds, 3, status=status)
  lopassmodel = poly(findgen(nx), par)

  return, f-lopassmodel

end


function quickccshift, s1, s2, maxshift=maxshift, oversample=oversample, censhift=censhift

  ; Calculate cross-correlation peak between two spectra.
  ; Only looks at lines.
  ; Only considers shifts between censhift-maxshift and censhift+maxshift.
  
  ; this algorithm could probably be better.  Not a limiting factor due to interpolation.
  ; doesn't seem to take most of the processing time either.

  maxshift = fix(maxshift)
  oversample = fix(oversample)

  nx = n_elements(s1)
  nxover = nx*oversample

  x = findgen(nx)
  xover = findgen(nxover)/oversample

  if n_elements(censhift) eq 0 then censhift = 0

  ; the xy-medianing is supposed to remove this, but whatevers
  s1f = lopassfilt(s1)
  s2f = lopassfilt(s2)

  s1interp = interpol(s1f, x, xover)
  s2interp = interpol(s2f, x, xover-censhift)

 ; plot, xover, s1interp, /xsty;, xrange=[100,140]
 ; oplot, xover, s1interp, psym=1
 ; oplot, xover, s2interp, color=rgb(255,0,0)
 ; oplot, xover, s2interp, color=rgb(255,0,0), psym=1

  ; Lots of filtering to reduce impact of cosmic rays and edge effects.

  tryshifti = -maxshift*oversample+findgen(maxshift*oversample*2+1)
  tryshift = tryshifti/oversample
  cc = fltarr(n_elements(tryshift))
  f = 5.
  mednoise = median(abs(s1-shift(s1,1)))
  ;print, median(s1interp), mednoise
  ct = 0
  while ct lt 60 do begin
     wcheck = where(s1interp ge median(s1interp)+f*mednoise and xover gt 2*maxshift and xover lt nx-2*maxshift, ct)
        ; should replace 2*maxshift with censhift+maxshift!
     f -= 0.5
  endwhile

  for t = 0, n_elements(tryshift)-1 do begin
    s2interpshift = shift(s2interp,tryshifti[t])
    if tryshifti[t] gt 0 then s2interpshift[0:tryshifti[t]-1] = s2interpshift[tryshifti[t]]
    if tryshifti[t] lt 0 then s2interpshift[nxover-1+tryshifti[t]:nxover-1] = s2interpshift[nxover-2+tryshifti[t]]
    cc[t] = total(s1interp[wcheck]*s2interpshift[wcheck])
  ;  oplot, xover, s2interpshift, color=rgb(100,100,100)
    ;print, t, tryshifti[t], tryshift[t], cc[t], ct
  endfor

  ;plot, tryshift, cc, /ynoz
  ;oplot, tryshift, cc, psym=1
  
  w = where(cc eq max(cc))

  return, tryshift[w]+censhift

end

function modelblock, im, xmin=xmin, xmax=xmax, debugplot=debugplot

   ; Calculate the sky for a subset of an image.

   s = size(im)
   nx = s[1]
   ny = s[2]

   ; xmin and xmax set the region to use for determining the wavelength shift
   if n_elements(xmin) eq 0 then xmin = 0
   if n_elements(xmax) eq 0 then xmax = nx-1
 
   if n_elements(debugplot) eq 0 then debugplot = 0

   ; First need to solve for the wavelength offset as a function of y.

   ;skyspecu = median(im[*,ny/4-3:ny/4+3],dimension=2)
   skyspec  = median(im[*,ny/2-5:ny/2+5],dimension=2)
   ;skyspecl = median(im[*,4*ny/5-3:4*ny/5+3],dimension=2)

   xshift = fltarr(ny)+!values.f_nan
   xmed = fltarr(ny)

   xshift[ny/2-5:ny/2+5] = 0
   skyspecs = skyspec[xmin:xmax]

   ims = im[xmin:xmax,*]
   for y = ny/2, ny-1 do begin
      censhift = median(xshift[y-5:y-1])
      bad = where(finite(ims[*,y]) eq 0, ctbad)
      if ctbad gt nx/2 then continue
      xshift[y] = quickccshift(ims[*,y], skyspecs, censhift=censhift, max=2, oversample=5)
      xmed[y] = median(ims[*,y])
      ;print, y, censhift, xshift[y]
   endfor
   for y = ny/2-1, 0, -1 do begin
      censhift = median(xshift[y+1:y+5])
      bad = where(finite(ims[*,y]) eq 0, ctbad)
      if ctbad gt nx/2 then continue
      xshift[y] = quickccshift(ims[*,y], skyspecs, censhift=censhift, max=2, oversample=5)
      xmed[y] = median(ims[*,y])
   endfor
 
   if debugplot then window, 0
   if debugplot then plot, xshift, yrange=[-2,4], psym=1, /xsty

   ; Remove any big steps
   for y = 2, ny-3 do begin
     ;print, y, xshift[y], xshift[y]-xshift[y-2]
     if abs(xshift[y]-xshift[y-2]) gt 0.5 and abs(xshift[y]-xshift[y+2]) gt 0.5 then $
     xshift[y] = !values.f_nan
   endfor

   if debugplot then oplot, xmed/max(xmed), color=rgb(128,128,255) ; source profile
   sourcey = where(xmed gt 10*median(xmed)+30, ct) ; crude

   ; Remove strong sources
   modxshift = xshift
   modxshift[sourcey] = !values.f_nan
   i = 0
   while ct gt 0 do begin
      wx = where(finite(modxshift) eq 0, ct)
      filtmodxshift = median(modxshift,7)
      modxshift[wx] = filtmodxshift[wx]
      i = i + 1
      if i gt 5 then begin ; this seems to happen sometimes near the edges...
         wx = where(finite(modxshift) eq 0, ct)
         modxshift[wx] = 0      
      endif
   endwhile
   if debugplot then oplot, modxshift, col=rgb(255,255,0), psym=10

   ; Median filter to remove cosmic rays
   modxshift = median(modxshift,5)

   ; Smooth by fitting a low-order polynomial
   ; modxshift = convol(modxshift, savgol(30,30,0,4), /edge_trunc)
   wh = where(finite(modxshift))
   par = poly_fit(wh, modxshift[wh], 6, status=status)
   modxshift = poly(findgen(n_elements(modxshift)), par)

   if debugplot then oplot, modxshift, col=rgb(255,0,0)

   ; modxshift is the precise (modeled) wavelength offset as a function of y.


   ; Now, get the offset for every pixel.  Sum of the x position and modxshift.

   woff = fltarr(nx, ny) ; wavelength offset in pixels (from the pixel at the far-left and center)
   for y = 0, ny-1 do woff[*,y] = findgen(nx) - modxshift[y]

   wavorder = sort(woff[*])
   if debugplot then window, 2
   xrange =[210,230] ;[313,316];

   if debugplot then plot, woff[wavorder], im[wavorder], psym=3, /xsty, /ysty, xrange=xrange

   ;wmedmodel = median(woff[wavorder],11)
   ;smedmodel = median(im[wavorder],11)

   wmedmodel = woff[wavorder]
   smedmodel = im[wavorder]

   ; do a linear local regression to smooth this
   n2 = fix(max(woff)-min(woff))*4
   wmodel2 = min(woff)+findgen(n2)*0.25+0.5
   smodel2 = fltarr(n2) + !values.f_nan
   isfinite = finite(wmedmodel)
   for d = 0, n2-1 do begin
     ;dif = median(abs(wmedmodel-shift(wmedmodel,1)))
     wh = where(wmedmodel ge wmodel2[d]-0.75 and wmedmodel lt wmodel2[d]+0.75 and isfinite, ct)
     if ct lt 2 then continue
     ; sigma clipping
     for j = 0, 3 do begin
        med = median(smedmodel[wh])
        ;sd = stddev(smedmodel[wh], /nan)
        mad = median(abs(smedmodel[wh]-med))
        if j eq 0 then whx = where(smedmodel[wh] lt med+6*mad, ctx)
        if j gt 0 then whx = where(smedmodel[wh] lt med+3*mad and smedmodel[wh] gt med-3*mad, ctx)
        ;if wmodel2[d] gt 224.5 and wmodel2[d] lt 225.5 then print, d, wmodel2[d], j, med, mad, ctx, ct
        if ctx gt 10 then wh = wh[whx] else break
     endfor
     ;if  wmodel2[d] gt 314.7 and wmodel2[d] lt 315 then print, d, wmodel2[d]
     if  wmodel2[d] gt 224.5 and wmodel2[d] lt 225 then oplot, wmedmodel[wh], smedmodel[wh], color=rgb(255,255,0), psym=3
     ;print, ct
     ;if ct eq 0 then continue
     ;if ct eq 1 then begin
     ;  smodel2[d] = smedmodel[wh]
     ;  continue
     ;endif

     ; NOW should do the median filter!

     ;par = linfit(wmedmodel[wh], smedmodel[wh]) ; slow?  Could just use the regression equation
     par = poly_fit(wmedmodel[wh], smedmodel[wh], 2, status=status) ; if more need e.g. 4 < ct
     smodel2[d] = poly(wmodel2[d], par)
   endfor

   ;plot, wmedmodel, smedmodel, xrange=[340,400], psym=3, /ysty
   if debugplot then oplot, wmodel2, smodel2, color=rgb(0,255,0), thick=3

   skymodel = fltarr(nx,ny)
   ;skymodel[*] = interpol(smedmodel, wmedmodel, woff[*], /nan)  ; older method
   skymodel[*] = interpol(smodel2, wmodel2, woff[*], /nan)     ; use better interpolation algorithm??

   if debugplot then oplot, woff[*], skymodel, color=rgb(255,0,0), psym=3
  
   if debugplot then window, 1
   if debugplot then plot, woff[*], im[*]-skymodel, /xsty, xrange=xrange, psym=3

   return, skymodel
 
end


pro subtractsky4, filename, sourcefilter=sourcefilter, outfile=outfile, nowrite=nowrite, xmin=xmin, xmax=xmax


   im = readfits(filename, h, /silent)
   s = size(im)
   nx = s[1]
   ny = s[2]


   xmax = 4050
  
   ; do the x median (continuum) filter here if requested - for standards
   if keyword_set(sourcefilter) then begin
     subtractsources, im, sourceimg, subimg
     im = im - sourceimg
   endif else begin
     sourceimg = 0
   endelse


   sky = fltarr(nx,ny)+!values.f_nan

   nblocks = 10
   buf = 16
   blockx0 = nx*indgen(nblocks)/nblocks
   blockx1 = nx*(1+indgen(nblocks))/nblocks-1
      
  ; blockx0[9] = 3660
  ; blockx0[9] = 3686
   print, max(blockx0)

   nyblocks = 2
   ybuf = ny/5
   blocky0 = ny*indgen(nyblocks)/nyblocks
   blocky1 = ny*(1+indgen(nyblocks))/nyblocks-1

   ; need to make sure there are actually lines in each block

   skycube = fltarr(nx,ny,nyblocks)+!values.f_nan

   for b = 0, nblocks-1 do begin
   ;for b = 4, 4 do begin
     if b eq 0         then lbuf = 0 else lbuf = buf
     if b eq nblocks-1 then rbuf = 0 else rbuf = buf
     bx0 = blockx0[b]
     bx1 = blockx1[b]
     dx = bx1-bx0
     for yb = 0, nyblocks-1 do begin
     ;for yb = 1, 1 do begin
       if yb eq 0          then lybuf = 0 else lybuf = ybuf
       if yb eq nyblocks-1 then rybuf = 0 else rybuf = ybuf
       by0 = blocky0[yb]
       by1 = blocky1[yb]
       print, yb, ':', by0-lybuf, by1+rybuf
       dy = by1-by0

       if n_elements(xmin) gt 0 and b eq 0 then         bxmin = xmin else bxmin = 0
       if n_elements(xmax) gt 0 and b eq nblocks-1 then bxmax = xmax-(bx0-lbuf) else bxmax = bx1+rbuf-(bx0-lbuf)

       modelwithbuffer = modelblock(im[bx0-lbuf:bx1+rbuf, by0-lybuf:by1+rybuf], $
                                    xmin=bxmin, xmax=bxmax, $
                                    debugplot=0)
       ;sky[bx0:bx1,by0:by1] = modelwithbuffer[lbuf:lbuf+dx,lybuf:lybuf+dy]
       skycube[bx0-lbuf:bx1+rbuf, by0-lybuf:by1+rybuf, yb] = modelwithbuffer
     endfor
   endfor

   sky = median(skycube, dimension=3, /even)
     
  outim = im-sky+sourceimg

; check plots of a bunch of vertical cuts 
 inspectplot = 0
 if inspectplot then begin
 xps = 1000+200*indgen(10)
 for j = 0, n_elements(xps)-1 do begin
  xp = xps[j]
 plot, im[xp,*], yrange=[min(outim[xp,*]),max(im[xp,*]) < median(im[xp,*])*10+10]
 oplot, [0,1000],[0,0], color=rgb(100,100,100)
 oplot, sky[xp,*], color=rgb(255,0,0)
 oplot, outim[xp,*], color=rgb(0,255,0)
 dum = GET_KBRD()
  if dum eq 'q' then return
 endfor
endif


  if keyword_set(nowrite) eq 0 then begin
     hist = 'Processed by subtractsky4 '+tstamp(/ut)+' UT' ;+systime()
     sxaddhist, hist, h

     ;sourcemax = max(sourceim[where(finite(sourceimg))])
     ;sxaddpar, h, 'SOURCMAX', sourcemax, ' Maximum source continuum average'
     ;skymax = max(sky)
     ;sxaddpar, h, 'SKYMAX', skymax, ' Maximum source continuum average'
     ;skymed = median(sky, /even)
     ;sxaddpar, h, 'SKYMED', skymed, ' Median sky value; subtracted out'
     ;exptime = sxpar(h,'EXPTIME', count=ct)
     ;if ct gt 0. then sxaddpar, h, 'SKYRATE', skymed/exptime, ' Median sky rate'

     slashpos = strpos(filename,'/',/reverse_search)
     if n_elements(outfile) eq 0 then $
        outfile = strmid(filename,0,slashpos+1) + 's' + strmid(filename,slashpos+1)
     logprint, ' Writing subtracted/sky extensions to ', removepath(outfile)
     mwrfits, outim, outfile, h, /create, /silent
     mwrfits, sky, outfile, /silent               ; extension 1: sky
     return
  endif ;else begin
  ;   return, outimg
  ;endelse

  

end

