;+
; NAME:
;       downloadkoadata
;
; PURPOSE:
;       Automatically download a large number of files from the Keck archive
;       (KOA), one by one.
;
; CALLING SEQUENCE:
;       downloadkoadata, daterange, [/gzip, camera=, /confirmed, destination=,
;                        /generatedatabase]
;
; EXPLANATION:
;       Tool to script downloads (via UNIX wget) from the archive.
;       Looks up a local copy of the database metadata to determine the file
;       names and locations of all files within a certain date range, and
;       downloads those files one by one.  
;
; INPUTS:
;       daterange - A string indicating the range of dates.  Should be two 
;       8-digit sequences separated by a dash, e.g. 20070101-20070103.
;
; OPTIONAL INPUT KEYWORDS:
;       gzip - Compress the files after downloading to save space.
;              Compression won't occur if the program is aborted.
;       camera - Download only red or blue files (specify 'r' or 'b')
;       /confirmed - Specify to skip user confirmation check.
;       destination - Subdirectory to download to; otherwise the current
;              working directory.  Use sortkoadata to sort into nightly
;              subdirectories after the fact.
;       generatedatabase - Update the local copy of the archive database
;              file to include newly publicly-released data.  Will only
;              affect downloads into the current working directory unless
;              the koadatabase.sav file is copied to the lpipe refdata/ 
;              subdirectory.
;
; OUTPUTS:
;       The requested files are downloaded to disk.
; 
; EXAMPLES:
;       downloadkoadata, '20160101-20160104', /gzip
;
; NOTES:
;       Because the files are stored uncompressed server side the download 
;       is quite slow compared to other methods (e.g. using the KOA tarball
;       generator).  Refuses to download >100000 files; if you really want
;       that many you can always edit the code to remove this check.
;
;-
;------------------------------------------------------------------------------

pro downloadkoadata, daterange, camera=camera, confirmed=confirmed, destination=destination, generatedatabase=generatedatabase, gzip=gzip

  ; Tool for downloading LRIS data directly from the Keck archive (via unix wget).

  ; Examples:  requestkoadata, '20100208-20100209' to download two nights from February 2010.

  common lrisarchive, lrisautopath, koadata

  if keyword_set(generatedatabase) then begin
     databasefile = 'lris.csv'
     print, ' Reading in '+databasefile+' ...'
     csvdat = read_csv(databasefile, header=lh)
     koadata = replicate({date:'',cam:'',url:'',target:''}, n_elements(csvdat.field01))
     koadata.date = csvdat.field09
     koadata.cam = strlowcase(strmid(csvdat.field01,1,1))
     koadata.url = strmid(csvdat.field16,55) ; save space
     koadata.target = csvdat.field03
     save, koadata, filename='koadatabase.sav' ; needs to be moved to lris/refdata directory
     return
  endif

  urlroot = 'https://koa.ipac.caltech.edu/cgi-bin/getKOA/nph-getKOA?'

  if n_elements(lrisautopath) eq 0 then begin
    paths = strsplit(!path,':',/extract)
    for p = -1, n_elements(paths)-1 do begin   ; find the directory this file is in
      if p ge 0 then path = paths[p] $
                     else path = ''
      sfile = path+'/lpipe.pro'
      n = countfiles(sfile)
      if n gt 0 then lrisautopath = path
      if n gt 0 then break
    endfor
  endif

  if n_elements(koadata) eq 0 and n_elements(lrisautopath) gt 0 then begin
    print, ' Loading database...'
    databasepath = './'+'koadatabase.sav' ; check for a local copy first.
    checkfile = file_test(databasepath)
    if checkfile eq 0 then begin
      databasepath = lrisautopath+'/refdata/'+'koadatabase.sav'
      checkfile = file_test(databasepath)
      if checkfile eq 0 then databasepath = 'koadatabase.sav'
      checkfile = file_test(databasepath)
      if checkfile eq 0 then begin
        print, ' No KOA database found.  Generate using /generate option and local csv file.'
        return
      endif
    endif
    restore, databasepath 
  endif

  if n_elements(destination) eq 0 then destination = ''
  if destination ne '' then $
  if strmid(destination,strlen(destination)-1) ne '/' then destination = destination + '/'

  daterangearr = strsplit(repstr(clip(daterange),' ',''),'-',/extract)
  if n_elements(daterangearr) ne 2 and n_elements(daterangearr) ne 1 then begin
    print, ' daterange must be of format yyyymmdd (one day) or'
    print, '                 yyyymmdd-yyyymmdd (range of days)'
    return
  endif
  startdate = daterangearr[0]
  if n_elements(daterangearr) eq 1 then enddate = startdate else enddate = daterangearr[1]
  if strlen(startdate) ne 8 or strlen(enddate) ne 8 then begin
    print, ' daterange format must be yyyymmdd'
    return
  endif
  startdate = strmid(startdate,0,4)+'-'+strmid(startdate,4,2)+'-'+strmid(startdate,6,2)
  enddate = strmid(enddate,0,4)+'-'+strmid(enddate,4,2)+'-'+strmid(enddate,6,2)
  

  iscam = 1
  if n_elements(camera) gt 0 then iscam = koadata.cam eq camera
  w = where(koadata.date ge startdate and koadata.date le enddate and iscam, ct)
  if ct eq 0 then begin
      print, ' No LRIS files found between '+startdate+' and '+enddate
      return
  endif
  if ct gt 100000 then begin
      print, ' Returned '+clip(ct)+' files between '+startdate+' and '+enddate
      print, ' This is too many: narrow query.'
  endif

  print, ' Will download '+clip(ct)+' files:
  for i = 0, ct-1 do begin
      f = w[i]
      filename = strmid(koadata[f].url,1+strpos(koadata[f].url,'/',/reverse_search))
      print, '   '+koadata[f].date, ' ', filename, ' ', koadata[f].target
  endfor
  if ct gt 100 then   print, ' Will download '+clip(ct)+' files.'
   
  if keyword_set(confirmed) eq 0 then begin
    print, ' press (y) to confirm or (q) to quit'
    k = get_kbrd()
  endif

  outlist = ''
  if k eq 'y' or k eq 'Y' or keyword_set(confirmed) then begin
     for i = 0, ct-1 do begin
        f = w[i]
        filename = strmid(koadata[f].url,1+strpos(koadata[f].url,'/',/reverse_search))
        outpath = destination+filename
        if file_test(outpath) or file_test(outpath+'.gz') then begin
          print, ' '+outpath+' already exists.'
          continue
        endif
        spawn, 'wget "'+urlroot+koadata[f].url+'" -O '+outpath
        if file_test(outpath) then outlist = [outlist, outpath]
     endfor
  endif

  if keyword_set(gzip) then begin
    for i = 1, n_elements(outlist)-1 do begin
      if file_test(outlist[i]) then spawn, 'gzip '+outlist[i]
    endfor
  endif

end


