;+
; NAME:
;       getlrisversion
;
; PURPOSE:
;       Determine whether the pipeline is being run on data from the current 
;       LRIS camera (v4), the first replacement detector (v3), the initial
;       upgrade detector (v2) or from the pre-upgrade LRIS (v1).  Does not yet
;       recognize the old blue camera or pre-blue data (which will presumably be v0).
;
; CALLING SEQUENCE:
;       result = getlrisversion(filename, datadir=, /quiet, filenametype=)
;
; EXPLANATION:
;       The pipeline needs to know at the top level what LRIS cameras are being
;       processed (e.g. to decide whether to split the red or not).  This is
;       run at the start of pipeline processing to check some common data 
;       filename conventions.  Dates in the file headers are used (crudely) to 
;       decide what LRIS camera is in use.  Also determines what filename
;       convention is being used.
;
; OPTIONAL INPUTS:
;       filename - Give it a file or file wildcard, rather than search common
;                  locations.
;       datadir - Directory to check in.  (Not necessarily the raw data 
;                  directory.)
;       quiet - Do not print error messages if data cannot be found.
;
; OPTIONAL OUTPUT KEYWORDS:
;       filenametype - Three-character string indicating what filename 
;           convention is in use for this pipeline run.  Can be (for raw data):
;           std: [r/b]yymmdd_nnnn.fits (Current naming convention at telescope)
;           old: l[red/blue]nnnn.fits (Old naming convention at telescope)
;           koa: L[R/B].YYYYMMDD.nnnnn.fits (KOA convention)
;
; OUTPUTS:
;       Returns an integer indicating whether it is LRIS version 1, 2, or 3.
; 
; EXAMPLES:
;       print, getlrisversion(datadir='spredux1d/', filenametype=filenametype)
;
; NOTES:
;       LPipe won't run if it can't determine the version because there is no
;       data at all, which is why different error messages may be generated
;       under otherwise similar circumstances if the pipeline can/can't see
;       the raw data directory.
;
; Last updated 11 May 2021; documentation updated 3 Jul 2022
;
;-
;------------------------------------------------------------------------------

function getlrisversion, filename, datadir=datadir, quiet=quiet, filenametype=filenametype
   ; Look for an example file and find the date to infer the camera and file format

   if n_elements(filename) eq 0 then begin
      if n_elements(datadir) eq 0 then datadir = './'
      files =                        (file_search(datadir+'r??????_????'+'.fits*', count=nstd))
      if files[0] eq '' then files = (file_search(datadir+'b??????_????'+'.fits*', count=nstd))
      if files[0] eq '' then files = (file_search(datadir+'*r??????_????'+'.fits*', count=nstd)) 
      if files[0] eq '' then files = (file_search(datadir+'*b??????_????'+'.fits*', count=nstd)) 
      if files[0] eq '' then files = (file_search(datadir+'LB.????????.?????.fits*', count=nkoa))
      if files[0] eq '' then files = (file_search(datadir+'LR.????????.?????.fits*', count=nkoa))
      if files[0] eq '' then files = (file_search(datadir+'*r??????.?????*.fits*', count=nkoa))
      if files[0] eq '' then files = (file_search(datadir+'*b??????.?????*.fits*', count=nkoa))
      if files[0] eq '' then files = (file_search(datadir+'*r??????_????*'+'.fits*', count=nstd)) 
      if files[0] eq '' then files = (file_search(datadir+'*b??????_????*'+'.fits*', count=nstd)) 
      if files[0] eq '' then files = (file_search(datadir+'lred????'+'.fits*', count=nold)) 
      if files[0] eq '' then files = (file_search(datadir+'lblue????'+'.fits*', count=nold)) 
      if files[0] eq '' then files = (file_search(datadir+'*lred????'+'*.fits*', count=nold)) 
      if files[0] eq '' then files = (file_search(datadir+'*lblue????'+'*.fits*', count=nold)) 
      if files[0] eq '' then files = (file_search(datadir+'*.????????.?????.spec', count=nkoa)) 
      if files[0] eq '' then files = (file_search(datadir+'*.??????_????.spec', count=nstd)) 
   endif else begin
      files = (file_search(filename))
   endelse

   if files[0] eq '' then begin
      if keyword_set(quiet) eq 0 then print, 'Found no LRIS files.'
      return, 0   ; 0 = no files found
   endif

   filenametype = 'std'
   if n_elements(nkoa) then if nkoa ge 1 then filenametype = 'koa'
   if n_elements(nold) then if nold ge 1 then filenametype = 'old'

   vers = intarr(4)
   for i = 0, 3 do begin
     if i eq 0 then try = files[4*n_elements(files)/10]
     if i eq 1 then try = files[6*n_elements(files)/10]
     if i eq 2 then try = files[n_elements(files)-1]
     if i eq 3 then try = files[0]
     if try ne '' then begin
        if strpos(try[0],'.fits') gt 0 then h = headfits(try[0])
        if strpos(try[0],'.spec') gt 0 then h = headspec(try[0])
        mjd = float(sxpar(h,'MJD-OBS', count=ct))
        if ct eq 0 then mjd = float(sxpar(h,'MJD', count=ct))
        if mjd gt 0 and mjd lt 49000 then vers[i] = 0
        if mjd gt 49000 and mjd lt 54983 then vers[i] = 1
        if mjd ge 54983 and mjd lt 55529 then vers[i] = 2
        if mjd ge 55529 and mjd lt 59339 then vers[i] = 3
        if mjd ge 59339 then vers[i] = 4
        ; print, try, vers[i]
     endif 
   endfor

   if max(vers) le 0 then begin
     if keyword_set(quiet) eq 0 then print, 'Cannot determine LRIS version.'  ; -1 = unknown
     return, -1
   endif
   vers = vers[where(vers gt 0)]
   if max(vers)-min(vers) ne 0 then begin
     print, 'Conflicting LRIS version numbers!' 
     if min(vers) gt 0 then $
        print, 'This directory seems to include data from multiple runs with different red CCDs.'
     return, -2                               ; -2 = conflict
   endif

   return, vers[0]

end

