;+
; NAME:
;       getlrisview
;
; PURPOSE:
;       Determine the most recent scattering surface for an LRIS exposure
;       (the closed trapdoor, the flat screen, the sky, or something else)
;
; CALLING SEQUENCE:
;       result = getlrisview(header, update=)
;
; EXPLANATION:
;       Uses e.g. the dome angle and trapdoor status in the header to determine
;       what LRIS is actually seeing in a given exposure.
;
; INPUTS(/OUTPUTS):
;       header - A file header.  (*Not* a file or a LRIS header structure).
;
; OPTIONAL INPUTS:
;       update - Edit the header to include a VIEW keyword giving the result.
;
; OUTPUTS:
;       Returns a string: 'closed', 'dome', 'sky', or '???'.
; 
; NOTES:
;       This routine is designed to unify a bunch of header checks.  Not yet
;       integrated widely into the pipeline (similar checks are separately
;       coded in numerous locations still.)
;
;-
;------------------------------------------------------------------------------

function getlrisview, header, update=update

  ; What is the most recent scattering surface for the photons seen by LRIS?

  dooropen = (clip(sxpar(header,'TRAPDOOR')) eq 'open')
  azimuth = sxpar(header,'AZ')
  elevation = sxpar(header,'EL')
  domeazimuth = sxpar(header,'DOMEPOSN',count=checkdome)
  if checkdome gt 0 then begin
     domerelaz = domeazimuth - azimuth
     while domerelaz lt -180.0 do domerelaz += 360.
     while domerelaz ge  180.0 do domerelaz -= 360.
     if keyword_set(update) then $
        sxaddpar, header, 'DOMERLAZ', domerelaz, ' Relative angle between dome and telescope' ; alters the header
  endif
  if dooropen eq 0 then begin
    view = 'closed'
  endif else begin
    view = '???'
    decdeg = ten(sxpar(header,'DEC'))
    radeg = ten(sxpar(header,'RA'))*15.
    if checkdome gt 0 then begin
       if abs(elevation-45) lt 0.1 and abs(abs(domerelaz)-90) lt 3.0 then view = 'dome'
       if abs(domerelaz) lt 3.0 then view = 'sky' ; not really sure how wide the dome window is
    endif else begin
       if abs(elevation-45) lt 0.1 and abs(decdeg-45) lt 0.1 and abs(radeg-45) lt 0.1 then view = 'dome' $
                                                                     else view = 'sky' ; presumed
    endelse
  endelse

  if keyword_set(update) then sxaddpar, header, 'VIEW', view, ' Surface producing observed signal' ; closed/dome/sky

  return, view

end

