;+
; NAME:
;       lrisalignapertures
;
; PURPOSE:
;       Routine for automatically ensuring the red and blue apertures are
;       extracting the same source with the same aperture size and background.
;
; CALLING SEQUENCE:
;       lrisalignapertures, camchar=camchar, objfilechanged=objfilechanged
;
; EXPLANATION:
;       When the pipeline identifies sources on each slit there is initially
;       no guarantee that it has found the same source on the red and blue or
;       that it has chosen consistent aperture sizes.  This routine checks the
;       data stored in objectpos.txt to be sure that the extractons are 
;       consistent, based on comparison to the red-blue offsets for standard
;       stars in the same configuration plus a direct cross-correlation of the
;       profiles.
;
; INPUT KEYWORDS:
;       camchar - 'r' or 'b' to indicate which is the primary camera (vs. which
;       is the secondary camera being compared to).  The effect is basically 
;       the same either way.
;
; OPTIONAL OUTPUT KEYWORDS:
;       objfilechanged - Notes whether any changes were actually made (=1) 
;          or not (=0).  Used to determine whether to redraw the aperture
;          graphics.
;
; OUTPUTS:
;       objectpos.txt is edited.
; 
;-
;------------------------------------------------------------------------------

pro lrisalignapertures, camchar=camchar, objfilechanged=objfilechanged

   common lrisauto, lpar

   ;lpar = {spworkingdir1d:'spredux1d/'} for testing only

   brightobjfile = lpar.spworkingdir1d+'brightobjectpos.txt'
   objfile = lpar.spworkingdir1d+'objectpos.txt' ; consider adding camera...

   ; Identify any red-blue pairs within the bright-object position list and calculate the offset(s).

   if file_test(brightobjfile) eq 0 then begin
      logprint, ' Cannot find '+brightobjfile+'; no means of aligning apertures.'
      return
   endif

   readcol, brightobjfile, tfile, tcenter, format='a,f', comment='#', /silent
   if n_elements(tfile) eq 0 then begin
      logprint, ' WARNING - '+brightobjfile+' contains no sources; no means of aligning apertures.'
      return
   endif
   for i = 0, n_elements(tfile)-1 do if strpos(tfile[i],'.fits') le 0 then tfile[i] = tfile[i]+'.fits'

   logprint, ' Measuring average R-B aperture alignment using standards...'
   ctt = n_elements(tfile) ; the BRIGHT object files
   if n_elements(tfile) eq 1 and tfile[0] eq '' then ctt = 0
   if ctt gt 0 then begin
      tinfo = lrisheaderinfo(lpar.spworkingdir1d+tfile)
      tmatchfile = strarr(ctt)
      toffset = fltarr(ctt) ; not really necessary
   endif
   grgr = ['']
   grgrifile = [''] ; the blank zero value is not deleted later
   grgrjfile = ['']
   grgroffset = [!values.f_nan]
   for i = 0, ctt-1 do begin
      wothercam = where(tinfo.camera ne tinfo[i].camera, cto)
      if cto eq 0 then continue
      ; if tmatchfile[i] ne '' then continue  ; We want backwards grgrs?  Uncomment this to force A<->B pairings.
      tinfoothercam = tinfo[wothercam]
      isoverlap = toverlap(tinfo[i].tstart, tinfo[i].tstop, tinfoothercam.tstart, tinfoothercam.tstop)
      match = (where(isoverlap, ctmatch))[0]
      if ctmatch eq 0 then begin
        tdiff = abs(tinfo[i].tstart-tinfoothercam.tstart)
        nearest = (where(tdiff eq min(tdiff))) [0]
        if tdiff[nearest] lt 30. then match = nearest
        if tdiff[nearest] lt 30. then ctmatch = 1
      endif
      if ctmatch gt 0 then begin
        tmatchfile[i] = tinfoothercam[match].filename
        j = wothercam[match]
        ;tmatchfile[j] = tinfo[i].filename
        toffset[i] = tcenter[i]*tinfo[i].spatbin-tcenter[j]*tinfo[j].spatbin
        ;;toffset[j] = -toffset[i]
        ; print, clip(removepath(tinfo[i].filename),32)+'('+tinfo[i].grname+')', ' ', $
        ;        clip(removepath(tinfo[j].filename),32)+'('+tinfo[j].grname+')', ' ', $
        ;        fpr(tcenter[i],3.2), ' ', fpr(tcenter[j],3.2), ': ', fpr(toffset[i],3.2)
        grgr = [grgr, tinfo[i].grname+'+'+tinfo[j].grname]
        grgroffset = [grgroffset, toffset[i]]
        grgrifile = [grgrifile, repstr(removepath(tinfo[i].filename),'.fits','')]
        grgrjfile = [grgrjfile, repstr(removepath(tinfo[j].filename),'.fits','')]
      endif
   endfor
   ugrgr = unique(grgr)
   ugrgroffset = strarr(n_elements(ugrgr)) ; contains offsets for each unique grating combo
   ugrgrn = intarr(n_elements(ugrgr))      ; counts the number of pairs each offset is based on
   for k = 1, n_elements(ugrgr)-1 do begin
      wg = where(grgr eq ugrgr[k] and finite(grgroffset), ctx)
      if ctx eq 0 then continue ; impact of this untested
      stdoffset = stddev(grgroffset[wg])
      if stdoffset gt 2.0 then begin ; nan OK
        logprint, ' WARNING - significant offsets between red and blue positions for grating combination '+ugrgr[k]
        maxstrlen = max(strlen(grgrifile[wg]))
        for l = 0, n_elements(wg)-1 do logprint, '       '+clip(grgrifile[wg[l]],maxstrlen)+' '+$
                                           clip(grgrjfile[wg[l]],maxstrlen)+' '+fpr(grgroffset[wg[l]],3.2)
      endif
      ugrgroffset[k] = median(grgroffset[wg], /even)
      ugrgrn[k] = n_elements(wg)
      logprint, '  ', clip(ugrgr[k],21), ' : ', fpr(grgroffset[wg],4.2)
      ;outstr = clip(ugrgr[k],21)+' : '+ ugrgroffset[k]
      ;updatefile, lpar.spworkingdir1d+'stdoffsets.txt'
   endfor


   tr = loadobjectposfile(objfile)
   sh = lrisheaderinfo(lpar.spworkingdir1d+tr.file)
   if n_elements(unique(sh.camera)) gt 1 then begin
      logprint, ' Checking R-B aperture alignment...'      
      fileroot = removepath(repstr(tr.file,'.fits','') )
      updated = 0b

      errorstatus = 0
      catch, errorstatus
      if errorstatus ne 0 then begin
           logprint, '   Error opening display window.'
           logprint, '   Turning off window display.'
           lpar.displayoff = 1
      endif
      if lpar.displayoff eq 0 then begin
        window, 0, xsize=600, ysize=800, title='LPipe: red-blue trace alignment'
        if camchar eq 'r' then begin
          icoldark = rgb(200,20,20)
          jcoldark = rgb(40,40,222)
          icol = rgb(245,60,60)
          jcol = rgb(110,110,255)
        endif else begin
          jcoldark = rgb(200,20,20)
          icoldark = rgb(40,40,222)
          jcol = rgb(245,60,60)
          icol = rgb(110,110,255)
        endelse
        !p.position  = 0
        !p.multi = [0,1,6]
      endif
      catch, /cancel

      for i = 0, n_elements(sh)-1 do begin
         if strmid(sh[i].camera,0,1) ne camchar then continue
         ; it would probably be better to match to the 'best' j file (longest overlapping?) rather than the first.
         profilefile_i = lpar.spworkingdir1d+repstr(tr[i].file,'.fits','.profile')

         for j = 0, n_elements(sh)-1 do begin
           if i eq j then continue
           if sh[i].camera eq sh[j].camera then continue
           if seqfilter(sh[i]) then continue

           buf = 10. - min(sh[[i,j]].tstop-sh[[i,j]].tstart)*0.75  ; tstop/tstart are seconds past midnight
           if buf gt 10. then buf = 0. ; (this shouldn't happen unless there is a bad UT value)
           ; a pair must overlap for >75% of the length of the shorter exposure, minus 10 seconds, for alignment
           ; buf is in seconds; below is in days
           
           if toverlap(sh[i].mjd, sh[i].mjdstop, sh[j].mjd, sh[j].mjdstop, buffer=buf/24/3600.) then begin
              ; Note that it's possible to have more than one match, especially in /nosum mode.

              ; Determine offset as currently stored in (default) profile file objectpos.txt
              currentoffset = tr[i].yfix*sh[i].spatbin-tr[j].yfix*sh[j].spatbin

              ; Look up expected offset (the offset measured at standard stars in this configuration)
              grgrij = sh[i].grname+'+'+sh[j].grname
              w = where(grgrij eq ugrgr, ct)
              if ct gt 0 then expectedoffset = ugrgroffset[w[0]] else expectedoffset = !values.f_nan

             ; print, ' Current offset: ', currentoffset
             ; print, ' Expected offset: ', expectedoffset

              ; If profile files exist for both red and blue exposures, do a cross-correlation
              ; to determine the offset empirically
              profilefile_j = lpar.spworkingdir1d+repstr(tr[j].file,'.fits','.profile')
              if file_test(profilefile_i) gt 0 and file_test(profilefile_j) gt 0 then begin
                 ;if n_elements(profi) le 1 then $
                 readcol, profilefile_i, yi, profi, format='f,f', comment='#', /silent
                 readcol, profilefile_j, yj, profj, format='f,f', comment='#', /silent
                 ;print, profilefile_i, profilefile_j
                 ymini = 10
                 ymaxi = 500
                 yminj = 10
                 ymaxj = 500
                 if sh[i].camera eq 'red' and lpar.lrisversion eq 4 then begin
                    ymini = max(yi)*sh[i].spatbin/2 ; this might not work if some weird windowing, etc.
                    ymaxi = ymini+490*sh[i].spatbin
                 endif
                 if sh[j].camera eq 'red' and lpar.lrisversion eq 4 then begin
                    yminj = max(yj)*sh[j].spatbin/2
                    ymaxj = yminj+490*sh[j].spatbin
                 endif
                 wi = where(yi*sh[i].spatbin gt ymini and yi*sh[i].spatbin lt ymaxi, cti) ; region to use
                 wj = where(yj*sh[j].spatbin gt yminj and yj*sh[j].spatbin lt ymaxj, ctj) ;  ("good" part of slit)
                                                                                     ; cropping dependent...

                 ;ni = n_elements(yi)
                 ;nj = n_elements(yj)
                 ; median filter with special edge treatment
                 profi -= (median([reverse(profi),profi,reverse(profi)],75*sh[i].spatbin)) [n_elements(profi):2*n_elements(profi)-1]
                 profj -= (median([reverse(profj),profj,reverse(profj)],75*sh[j].spatbin)) [n_elements(profj):2*n_elements(profj)-1]

                 ; Occasionally if a source is on the edge of this region this will cause a
                 ; mis-centering because part of it is 'cutoff', shifting its effective centroid.
                 ; Try to avoid this by checking if the profile max is close to an edge, and if
                 ; so, trimming it out.  This could be made a lot more sophisticated.
                 wmaxi = where(profi[wi] eq max(profi[wi]))
                 wmaxj = where(profj[wj] eq max(profj[wj]))
                 if wmaxi lt 5 and cti gt 30 then wi = wi[15:cti-1]
                 if wmaxj lt 5 and ctj gt 30 then wj = wj[15:ctj-1]
                 cti = n_elements(wi)
                 ctj = n_elements(wj)
                 if wmaxi ge cti-5 and cti gt 30 then wi = wi[0:cti-16]
                 if wmaxj ge ctj-5 and ctj gt 30 then wj = wj[0:ctj-16]
                 cti = n_elements(wi)
                 ctj = n_elements(wj)
                

                 ; need a pixel scale adjustment for LRIS-R1
                 if finite(expectedoffset) then begin
                     eo = expectedoffset 
                     maxshiftx = 25.
                 endif else begin 
                     eo = currentoffset
                     maxshiftx = 70.
                 endelse
 
                ; plot, yi, profi, color=rgb(255,90,90)
                ; oplot, yi[wi], profi[wi], thick=3, color=rgb(255,90,90)
                ; plot, yj, profj, color=rgb(130,130,255)
                ; oplot, yj[wj], profj[wj], thick=3, color=rgb(130,130,255)

                 measuredoffset = eo+ccshift(yj[wj]*sh[j].spatbin+eo, profj[wj], $
                                             yi[wi]*sh[i].spatbin, profi[wi],  maxshiftx=maxshiftx)
                 ; might want to a multiple peak finder instead of a global ccshift peak finder.

                ; print, ' Measured offset: ', measuredoffset
                 
                 if abs(measuredoffset-eo) ge maxshiftx then measuredoffset = !values.f_nan
                 if lpar.displayoff eq 0 and finite(measuredoffset) then begin
                   ;if !p.multi[0] eq 0 then q = get_kbrd()
                   plot, [10,300], [-0.2,1], /nodata, /xsty, /ysty
                   offi = 0
                   offj = 0
                   if sh[i].camera eq 'red' and lpar.lrisversion eq 4 then offi = max(yi)/2
                   if sh[j].camera eq 'red' and lpar.lrisversion eq 4 then offj = max(yj)/2
                   off2 = offi > offj  ; this miraculously works for unbinned...
                   inorm = max(profi[where(yi gt offi and yi*sh[i].spatbin lt (offi+300)*sh[i].spatbin)])
                   jnorm = max(profj[where(yj gt offj and yj*sh[j].spatbin lt (offj+300)*sh[i].spatbin)])
                   oplot, (yi[wi]-offi)*sh[i].spatbin, profi[wi]/inorm, col=icoldark, linestyle=2, psym=10
                   oplot, (yj[wj]-offj)*sh[j].spatbin, profj[wj]/jnorm, col=jcoldark, linestyle=2, psym=10
                   xyouts, 200, 0.85, repstr(removepath(profilefile_i),'.profile',''), col=icol
                   xyouts, 200, 0.67, repstr(removepath(profilefile_j),'.profile',''), col=jcol
                   oplot, (yi[wi]-offi)*sh[i].spatbin-(measuredoffset-off2*sh[i].spatbin)/2, profi[wi]/inorm, col=icol, psym=10
                   oplot, (yj[wj]-offj)*sh[j].spatbin+(measuredoffset-off2*sh[j].spatbin)/2, profj[wj]/jnorm, col=jcol, psym=10
                 endif
              endif else begin
                 measuredoffset = !values.f_nan
              endelse
            
              ;print, clip(repstr(tr[i].file,'.fits',''),28), ' ', $
              ;       clip(repstr(tr[j].file,'.fits',''),28), ' ', $
              ;       fpr(currentoffset,4.2), fpr(expectedoffset,4.2), fpr(measuredoffset,4.2), $
              ;       ' (',fpr(tr[i].aprad*sh[i].spatbin,4.2), fpr(tr[j].aprad*sh[j].spatbin,4.2), ')'

               newoffset = !values.f_nan
               ; If the x-correlation measured offset looks reliable, use that.
               if finite(measuredoffset) and (abs(measuredoffset-expectedoffset) lt 24.0 $
                                          or finite(expectedoffset) eq 0) then begin
                 newoffset = measuredoffset
               endif else begin
                 ; Otherwise use the standard-star offset (unless it's small)
                 logprint, '  Having trouble determining R-B offset for '+fileroot[i]+'/'+fileroot[j]+': expected '+clip(fpr(expectedoffset,3.1))+', measured '+clip(fpr(measuredoffset,4.1))
                 if finite(expectedoffset) and abs(currentoffset-expectedoffset) gt 2.0+1.0*sh[i].airmass^2 then begin
                      logprint, '  Using standard star R-B offset; may result in imperfect placement.'
                      newoffset = expectedoffset
                 endif
               endelse

               if finite(newoffset) and abs(newoffset-currentoffset) gt 0.25 then begin
                 logprint, '  Different aperture placement for '+$
                                fileroot[i]+'('+clip(fpr(tr[i].yfix,3.2))+')'+$
                        ' vs. '+fileroot[j]+'('+clip(fpr(tr[j].yfix,3.2))+')'+$
                        ': DY='+clip(fpr(abs(currentoffset-newoffset)*0.135,2.2))+'"'
                 if abs(currentoffset-newoffset) lt 4 then begin
                   ; small difference suggests objects are the same: meet halfway
                   logprint, '    Small difference; aligning apertures.'
                   diff = (currentoffset-newoffset)
                   newyi = tr[i].yfix - 0.5*diff/sh[i].spatbin
                   newyj = tr[j].yfix + 0.5*diff/sh[j].spatbin
                   tr[i].yfix = newyi
                   tr[j].yfix = newyj
                   tr[i].aprad += 0.5*abs(diff) /sh[i].spatbin ; also make the aperture a little larger
                   tr[j].aprad += 0.5*abs(diff) /sh[j].spatbin
                   updated = 1b
                 endif else begin
                   ; large difference suggests the objects are different; we have to choose one. 
                   ; Compare S/N
                   readcol, repstr(sh[i].filename, '.fits', '.profile'), pry, prf, /silent
                   readcol, repstr(sh[j].filename, '.fits', '.profile'), pryrb, prfrb, /silent
                   noise = median(abs(prf-shift(prf,1)))
                   noiserb = median(abs(prfrb-shift(prfrb,1)))
                   sn   =   prf[(where(pry   eq fix(tr[i].yfix))) [0]] / noise
                   snrb = prfrb[(where(pryrb eq fix(tr[j].yfix))) [0]] / noiserb
                   ; align the lower S/N object with the higher S/N object
                   if snrb gt sn then begin
                     logprint, '    Large difference; aligning '+fileroot[i]+' with '+fileroot[j]
                     tr[i].yfix = tr[i].yfix - (currentoffset-newoffset)/sh[i].spatbin
                     tr[i].aprad = (tr[j].aprad*sh[j].spatbin)/sh[i].spatbin
                     tr[i].bakmin = (tr[j].bakmin*sh[j].spatbin)/sh[i].spatbin
                     tr[i].bakmax = (tr[j].bakmax*sh[j].spatbin)/sh[i].spatbin
                   endif else begin
                     logprint, '    Large difference; aligning '+fileroot[j]+' with '+fileroot[i]
                     tr[j].yfix = tr[j].yfix + (currentoffset-newoffset)/sh[j].spatbin
                     tr[j].aprad = (tr[i].aprad*sh[i].spatbin)/sh[j].spatbin
                     tr[j].bakmin = (tr[i].bakmin*sh[i].spatbin)/sh[j].spatbin
                     tr[j].bakmax = (tr[i].bakmax*sh[i].spatbin)/sh[j].spatbin
                   endelse
                   updated = 1b
                 endelse
              endif

              if abs(tr[i].aprad*sh[i].spatbin-tr[j].aprad*sh[j].spatbin) gt 0.01 then begin
                ; always average apertures unless difference is basically zero
                if abs(tr[i].aprad*sh[i].spatbin-tr[j].aprad*sh[j].spatbin) gt 6 then begin ; 0.61"
                  ; only print if difference is large
                  logprint, '  Different aperture sizes for '+$
                                  fileroot[i]+'('+clip(fpr(tr[i].aprad,3.2))+')'+$
                          ' and '+fileroot[i]+'('+clip(fpr(tr[j].aprad,3.2))+')'+$
                          ': DR='+clip(fpr(abs(tr[i].aprad*sh[i].spatbin-tr[j].aprad*sh[j].spatbin)*0.135,2.2))+'"'
                  logprint, '    Averaging...'
                endif
                newradi = ((tr[i].aprad*sh[i].spatbin)+(tr[j].aprad*sh[j].spatbin))/2. / sh[i].spatbin
                newradj = ((tr[i].aprad*sh[i].spatbin)+(tr[j].aprad*sh[j].spatbin))/2. / sh[j].spatbin
                tr[i].aprad = newradi
                tr[j].aprad = newradj
                updated = 1b
              endif
              for abi = 0, 1 do begin ; lower and upper backgrounds (silent change)
              if abs(tr[i].bakmin[abi]*sh[i].spatbin-tr[j].bakmin[abi]*sh[j].spatbin) gt 0.15 then begin
                newbak =  ((tr[i].bakmin[abi]*sh[i].spatbin)+(tr[j].bakmin[abi]*sh[j].spatbin))/2.
                newbaki = newbak / sh[i].spatbin
                newbakj = newbak / sh[j].spatbin
                tr[i].bakmin[abi] = newbaki
                tr[j].bakmin[abi] = newbakj
                updated = 1b
              endif
              if abs(tr[i].bakmax[abi]*sh[i].spatbin-tr[j].bakmax[abi]*sh[j].spatbin) gt 0.05 then begin
                newbak  = ((tr[i].bakmax[abi]*sh[i].spatbin)+(tr[j].bakmax[abi]*sh[j].spatbin))/2.
                newbaki = newbak / sh[i].spatbin
                newbakj = newbak / sh[j].spatbin
                tr[i].bakmax[abi] = newbaki
                tr[j].bakmax[abi] = newbakj
                updated = 1b
              endif
              endfor
              continue ; only match to the first file if there are multiple matches
              ;inarrpos = strsplit(oinfo[i].str)
              ;newoutstr = strmid(oinfo[i].str,0, inarrpos[2]) + fpr(objectrad,3.2) + strmid(oinfo[i].str,inarrpos[3])
              ;updatefile, objfile, newoutstr
           endif

         endfor
         ;qq = get_kbrd()
      endfor 
      if updated then begin
          logprint, ' Updating aperture file...'
          saveobjectposfile, objfile, tr ; Could also fall back on the updatefile approach...
          objfilechanged = 1
      endif
   endif

   if lpar.displayoff eq 0 then begin
      !p.position  = 0
      !p.multi = 0
   endif

end
