;+
; NAME:
;       lriscrawler
;
; PURPOSE:
;       Tool for reduction of multiple LRIS nights.
;
; CALLING SEQUENCE:
;       lriscrawler, [directorystr, flatdaydiff=, arcdaydiff=,
;                     /nofailuresummary,
;                     step=, start=, stop=, chip=, mode=, camera=, 
;                     /norebin, /redo, /forceflat, /requirepair]
;
; EXPLANATION:
;       This was designed for queuing reduction of multiple nights in series.
;       Run this one level down from the nightly reduction directories to 
;       'crawl' through each nightly directory and reduce its data before
;       moving on.  Can permit sharing of flat-fields between similar nights.
;
; INPUTS:
;       directorystr - Wildcard string indicating what directories to reduce.
;                      Defaults to all subdirectories (i.e., '*/')
;
; OPTIONAL INPUT KEYWORDS:
;       flatdaydiff - Number of days to share flats between.  Will only look at
;                     *past* flats because future flats aren't generated yet.
;                     Use lrismetacrawler to change the workflow to allow
;                     reduction using future flats.
;       arcdaydiff  - Same as above but for arcs.
;       All others - These are simply passed to lpipe: see the documentation there.
;
;
; EXAMPLES:
;       lriscrawler, '2017010[2-6]', mode='s'
;
; NOTES:
;       It is safe to reduct all the data from a multi-night run just using 
;       lpipe itself, without separating into subdirectories.  This is mainly
;       for much larger reduction efforts.
;       Includes two higher-level procedures: metalriscrawler and 
;       megalriscrawler.  lrismetacrawler repeats flat-generation multiple 
;       times such that nightly flats are tried first before subsequent-night
;       flats are tried.  lrismegacrawler is used for reductions starting TWO
;       subdirectory levels down, and is basically only for mass reductions
;       of data spanning e.g. multiple semesters.
;
;-
;------------------------------------------------------------------------------

@lpipe

pro lriscrawler, directorystr, step=step, start=start, stop=stop, chip=chip, mode=mode, camera=camera, flatdaydiff=flatdaydiff, arcdaydiff=arcdaydiff, nofailuresummary=nofailuresummary, norebin=norebin, redo=redo, forceflat=forceflat, requirepair=requirepair

  common crawldates, alldates, nad

  !quiet = 1

  if n_elements(directorystr) eq 0 then directorystr = '*/'

  if n_elements(alldates) eq 0 then begin
     ndm = [31,28,31,30,31,30,31,31,30,31,30,31]
     years = string(1994+indgen(25),'(I4)')
     months = string(1+indgen(12),'(I02)')
     alldates = ''
     for i = 0, n_elements(years)-1 do begin
      if strnum(years[i]) mod 4 eq 0 then ndm[1] = 29 else ndm[1] = 28
      for j = 0, n_elements(months)-1 do $
       for k = 1,ndm[j] do $
         alldates = [alldates,years[i]+months[j]+string(k,'(I02)')]
     endfor
     alldates = alldates[1:*]
     nad = n_elements(alldates)
  endif

  if n_elements(directorystr) gt 1 then begin 
     directories = ''
     for i = 0, n_elements(directorystr)-1 do directories = [directories, file_search(directorystr, /mark_dir, /test_dir)]
     w = where(directories ne '', ct)
     if ct gt 0 then directories = directories[w]
  endif else begin
     directories = file_search(directorystr, /mark_dir, /test_dir, count=ct)
  endelse

  if ct eq 0 then begin 
    print, ' No subdirectories to search.'
    return
  endif
  ;wdir = where(strpos(possdirectories,'/') eq strlen(possdirectories)-1, ct)
  ;if ct eq 0 then begin 
  ;  print, ' No subdirectories to search.'
  ;  return
  ;endif
  ;directories = possdirectories[wdir]
  proctimes = dblarr(n_elements(directories))

  print, ' Found '+clip(ct)+' subdirectores to reduce:'
  for d = 0, n_elements(directories)-1 do begin
     dfiles = [file_search(directories[d]+'L?.????????.?????.fits*'),$ 
               file_search(directories[d]+'raw/L?.????????.?????.fits*'),$
               file_search(directories[d]+'???????_????.fits*'),$
               file_search(directories[d]+'raw/???????_????.fits*')]
     w = where(dfiles ne '', ctfiles)
     print, '  ', directories[d], ': '+clip(ctfiles)+' LRIS files'
  endfor  

  nd = n_elements(directories)
  for d = 0, n_elements(directories)-1 do begin
    cd, directories[d]
    print, directories[d]

   ; if running directly in the data directory uncomment the below.
   ;  (we are now running on a different disk via symlinks instead).
   ; rfiles = [file_search('L?.????????.?????.fits*'),$ 
   ;           file_search('???????_????.fits*')]
   ; w = where(rfiles ne '', ctfiles)
   ; if file_test('raw') eq 0 then spawn, 'mkdir raw'
   ; for i = 0, ctfiles-1 do begin
   ;    if file_test('raw/'+rfiles[i]) eq 0 then $
   ;      file_move, rfiles[i], 'raw/' $
   ;    else $
   ;      print, rfiles[i]+' already exists in raw/'
   ; endfor

    xfiles = [file_search('raw/L?.????????.?????.fits*'),$ 
              file_search('raw/???????_????.fits*')]
    countsci = 0
    hlist = lrisheaderinfo(xfiles)
    sci = where(hlist.elevation gt 10 and hlist.trapdoor eq 'open' and $
                 hlist.exptime gt 30. and $                                    ; probably not sky flat
                 abs(hlist.domerelaz) lt 2. and abs(hlist.decdeg-45) gt 0.5 and $ ; not dome flats
                 (hlist.slit eq 'direct' or strpos(hlist.slit,'long') ge 0), ctsci)

    if ctsci eq 0 then begin
      print, ' '+directories[d]+' contains no reducable science data.'
    endif

    autolrisdefaults, /reset

    starttime = systime(/seconds)

    if n_elements(altflatdir) gt 0 then delvarx, altflatdir
    if n_elements(flatdaydiff) gt 0 then if flatdaydiff gt 0 then begin
      if n_elements(step) gt 0 then if step eq 'flatten' then begin
        dirdate = repstr(directories[d],'/','')
        wd = where(alldates eq dirdate, ctwd)
        if ctwd gt 0 then begin
          if step eq 'flatten' then begin
            altflatdir = '../'+[alldates[(wd-flatdaydiff)>0:(wd-1)<(nad-1)]+'/flats/',$
                                alldates[(wd+1)>0:(wd+flatdaydiff)<(nad-1)]+'/flats/']
            altflatdir = unique(altflatdir)
          endif
            ; harmless bug if wd=0 or wd=nad-1 which will not be the case in reality
            ; if start eq 'flatten' then  $
            ;   altflatdir = '../'+[directories[(wd-2)>0:(wd+2)<(nd-1)]+'spredux2d/',$
            ;                       directories[(wd-2)>0:(wd+2)<(nd-1)]+'imredux/']
          if n_elements(altflatdir) gt 0 then begin
            wv = where(file_test(altflatdir), ct)
            if ct gt 0 then altflatdir = altflatdir[wv] else delvarx, altflatdir
          endif
        endif
      endif
    endif

    if n_elements(altarcdir) gt 0 then delvarx, altarcdir
    if n_elements(arcdaydiff) gt 0 then if arcdaydiff gt 0 then begin
      if n_elements(step) gt 0 then if step eq 'wavapply' then begin
        dirdate = repstr(directories[d],'/','')
        wd = where(alldates eq dirdate, ctwd)
        if ctwd gt 0 then begin
          if step eq 'wavapply' then begin
            altarcdir = '../'+[alldates[(wd-arcdaydiff)>0:(wd-1)<(nad-1)],$
                               alldates[(wd+1)>0:(wd+arcdaydiff)<(nad-1)]]+'/arcs/'
            altarcdir = unique(altarcdir)
          endif
          if n_elements(altarcdir) gt 0 then begin
            wv = where(file_test(altarcdir), ct)
            if ct gt 0 then altarcdir = altarcdir[wv] else delvarx, altarcdir
          endif
        endif
      endif
    endif


    lpipe, /timer, reducer='Perley', data='raw/', step=step, start=start, stop=stop, chip=chip, mode=mode, camera=camera, altflatdir=altflatdir, altarcdir=altarcdir, nofailuresummary=nofailuresummary, norebin=norebin, redo=redo, forceflat=forceflat, requirepair=requirepair

    endtime = systime(/seconds)

    proctimes[d] = endtime - starttime

    ; disused
   ; flatdir = '../'+directories[d]+'flats/'
   ; if n_elements(lastflatdir) eq 0 then lastflatdir = flatdir else lastflatdir = [flatdir, lastflatdir]
   ; while n_elements(lastflatdir) gt 4 do lastflatdir = lastflatdir[0:n_elements(lastflatdir)-2]

   ; spflatdir = '../'+directories[d]+'spredux2d/' ; may not exist or contain any files, but that's fine
   ; imflatdir = '../'+directories[d]+'imredux/' ; currently IGNORED
   ; if n_elements(lastflatdir) eq 0 then lastflatdir = spflatdir else lastflatdir = [spflatdir, lastflatdir]
   ; while n_elements(lastflatdir) gt 4 do lastflatdir = lastflatdir[0:n_elements(lastflatdir)-2]

    ; cleanup would go here

    cd, '../'

    

  endfor

  print, 'All done at '+tstamp(/ut)
  print, 'Processing time summary: '

  for d = 0, n_elements(directories)-1 do begin
     print, ' ', directories[d], ' ', fpr(proctimes[d],6.1)+'s'
  endfor

 
end

pro metalriscrawler, directorystr, mode=mode, start=start, stop=stop, step=step, redo=redo, sistartin=sistartin

  ; run the pipeline with repeats of the flattening step, slowly casting a wider net
  ; to be able to flat-field any observations without nightly flats.

  if n_elements(mode) eq 0 then mode = 's'

  sistart = 0 
  sistop = 13
  start7 = 'split'
  stop7 = 'extract'
  start12 = 'response'
  stop12 = 'connect'

  steps = ['prepare', 'makeflat', 'flatten', 'reflatten', 'split', 'crclean', 'skysubtract', $
         'sum', 'trace', 'extract', 'wavcal', 'wavapply', 'response', 'fluxcal', 'combine', 'connect', $
         'astrometry', 'photometry', 'stack']
  steporder = indgen(n_elements(steps))
  sosplit = steporder[where(steps eq 'split')]
  soextract = steporder[where(steps eq 'extract')]
  soflatten = steporder[where(steps eq 'flatten')]
  sowavcal = steporder[where(steps eq 'wavcal')]
  sowavapply = steporder[where(steps eq 'wavapply')]

  if n_elements(step) gt 0 then begin
    start = step
    stop  = step
  endif
  if n_elements(start) gt 0 then begin
    sostart = steporder[where(steps eq start)]
    if start eq 'prepare' then sistart = 0 
    if start eq 'makeflat' then sistart = 1
    if start eq 'flatten' then sistart = 2
    if sostart gt soflatten and sostart lt sowavcal then sistart = 7
    if start eq 'wavcal' then sistart = 8
    if start eq 'wavapply' then sistart = 9
    if sostart gt sowavapply then sistart = 12
  endif
  if n_elements(stop) gt 0 then begin
    sostop = steporder[where(steps eq stop)]
    if stop eq 'prepare' then sistop = 0
    if stop eq 'makeflat' then sistart = 3 ; not a typo
    if stop eq 'makeflat' then sistop = 3
    if stop eq 'flatten' then sistop = 6
    if sostop gt soflatten and sostop lt sowavcal then sistop = 7
    if stop eq 'wavcal' then sistart = 8
    if stop eq 'wavapply' then sistop = 11
    if sostop gt sowavapply and stop ne 'connect' then sistop = 12
    if stop eq 'connect' then sistop = 13
  endif
  if n_elements(start) then begin
     if sostart gt soflatten and sostart lt sowavcal  then start7 = start
     if sostart gt sowavapply then start12 = start
  endif
  if n_elements(stop) then begin
     if sostop gt soflatten and sostop lt sowavcal  then stop7 = stop
     if sostop gt sowavapply then stop12 = stop
  endif

  if n_elements(directorystr) eq 0 then directorystr = '*/'

  if n_elements(sistartin) gt 0 then sistart = sistartin

  ;print, sistart, sistop
  ;print, start7, ' ', stop7
  ;print, start12, ' ', stop12
  ;return

  for si = sistart, sistop do begin
    if si eq 0 then lriscrawler, directorystr, step='prepare', mode=mode, redo=redo
    if si eq 1 then lriscrawler, directorystr, step='makeflat', mode=mode, redo=redo
    if si eq 2 then lriscrawler, directorystr, step='flatten', mode=mode, redo=redo, flatdaydiff=0, /norebin
    if si eq 3 then lriscrawler, directorystr, step='makeflat', mode=mode, redo=redo, /forceflat
    if si eq 4 then lriscrawler, directorystr, step='flatten', mode=mode, redo=redo, flatdaydiff=2, /nofailuresummary
    if si eq 5 then lriscrawler, directorystr, step='flatten', mode=mode, redo=redo, flatdaydiff=9, /nofailuresummary
    if si eq 6 then lriscrawler, directorystr, step='flatten', mode=mode, redo=redo, flatdaydiff=180 ; the whole semester
    if si eq 7 then lriscrawler, directorystr, start=start7, stop=stop7, mode=mode, redo=redo
    if si eq 8 then lriscrawler, directorystr, step='wavcal', mode=mode, redo=redo
    if si eq 9 then lriscrawler, directorystr, step='wavapply', mode=mode, redo=redo, arcdaydiff=0
    if si eq 10 then lriscrawler, directorystr, step='wavapply', mode=mode, redo=redo, arcdaydiff=2, /nofailuresummary
    if si eq 11 then lriscrawler, directorystr, step='wavapply', mode=mode, redo=redo, arcdaydiff=9
    if si eq 12 then lriscrawler, directorystr, start=start12, stop=stop12, mode=mode, redo=redo, /requirepair
    if si eq 13 then lriscrawler, directorystr, step='connect', mode=mode, redo=redo
  endfor

end

pro megalriscrawler, megadirectorystr, mode=mode, start=start, stop=stop, redo=redo
  
  directories = file_search(megadirectorystr, /mark_dir, /test_dir, count=ct)
  if ct eq 0 then begin 
    print, ' No subdirectories to search.'
    return
  endif
  print, transpose(directories)

  if n_elements(mode) eq 0 then mode = 's'
  if n_elements(stop) eq 0 then stop = 'connect'

  print, ' mode = ', mode
  print, ' stop = ', stop

  for d = 0, ct-1 do begin
    cd, directories[d]
    print, '-----------------------------------------------------------------------'
    print, '                             '+directories[d]
    print, '-----------------------------------------------------------------------'


    errorstatus = 0
    catch, errorstatus
    if errorstatus ne 0 then begin
       errmsg = clip(!error_state.msg)
       logprint, '     Error encountered:'
       ;print, '     '+errmsg ;!error_state.msg
       logprint, '     '+errmsg ;!error_state.msg
       logprint, ' --- LPIPE TERMINATED UNEXPECTEDLY, UT '+tstamp(/utc)+' ---', /logonly
       logprint, /close
       cd, '../'
       if !d.name eq 'PS' then psclose ; need to close the plot if an exception was thrown
       !p.multi = 0
       continue
    endif else begin
       metalriscrawler, mode=mode, start=start, stop=stop, redo=redo
    endelse

    cd, '../'
  endfor

end


