;+
; NAME: lrisfluxcal.pro
;
; PURPOSE: 
;       Utility to apply a flux calibration from a response file or files
;       to a science file or files.
;
; CALLING SEQUENCE: 
;       lrisfluxcal, sciencefiles, standardfiles, outfiles=outfiles, stnames=stdnames
;
; INPUTS:
;	sciencefiles  = science file name(s), should be .spec
;       standardfiles = standard file names(s), should be .spec and have a .response pair
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       outfiles      = output file name(s).  
;       stdnames      = names of standard stars for lookup/logging.  
;                       (Will use TARGNAME if not specified.)
;
;-
;-----------------------------

pro lrisfluxcal, sciencefiles, standardfiles, outfiles=outfiles, stdnames=stdnames, notelluric=notelluric

   ; Once the standards are chosen, actually perform flux calibration.
   ; Allows multiple science files for user convenience, but the pipeline itself always
   ; calls one sciencefile at a time.   

      elev = 4200.
      airp = 2^((4200-elev)/4570.)*[0.12947983, -0.13458371, 0.21833111, -0.48222307, $
                                    0.41800934, -0.11447838, 0.0032198187, -0.010178335, 0.0047375054]
      ;for n2 = 3, 20 do print,( 911.760 * 1./(1./n1^2 - 1./n2^2) ) / (1.0 + 2.735182E-4 + 131.4182 / vac^2 + 2.76249E8 / vac^4)

   ; path = getfilepath(sciencefiles[0])

   ; a slit loss correction would probably go here.

   stdh = lrisheaderinfo(standardfiles)
   nmatchstd = n_elements(standardfiles)

   if n_elements(stdnames) lt nmatchstd then stdnames = stdh.targname ; should be fixed/standardized

   for i = 0, n_elements(sciencefiles)-1 do begin
      sciencefile = sciencefiles[i]
      
      scih = lrisheaderinfo(sciencefile)

      sciair = scih.airmass
      logprint, ' Flux-calibrating '+removepath(sciencefile)+' ('+scih.target+', '+scih.grname+')  airmass='+fpr(sciair,1.4)

      ; actually read in the science frame
      sci = readspec(sciencefile, h, /timenorm, cutoff=3000.)
      wav = sci.wav
      nwav = sxpar(h,'NWAV',count=ctnwav)
      if ctnwav gt 0 and nwav eq 0 then begin  
         logprint, sciencefile, ' has no wavelength calibration.'
         continue
      endif

      stdair = stdh.airmass

      equal = where(stdair gt sciair-0.015 and stdair lt sciair+0.015, ctequal)
      if ctequal ge 1 then begin
        wequal = equal[(where(stdh[equal].mjd - scih.mjd) eq min(stdh[equal].mjd - scih.mjd)) [0]]
        nmatchstd = n_elements(wequal)
      endif    

      ; shut off the multiple standard matching by uncommenting this block
      ; if nmatchstd gt 1 then begin
      ;    wstd = wfluxcal[matchstandards[(where(stdair eq min(stdair))) [0]]]
      ;    nmatchstd = 1
      ; endif

      if nmatchstd eq 1 then begin
         ;if abs(sciair-stdair[0]) gt 0.1 then begin
         ;   print, '     Warning: Only one standard-star observation in this configuration.'
         ;   print, '     Airmass:  science ('+fpr(sciair,2.2)+'), standard ('+fpr(stdair[0],2.2)+')'
         ;endif
         stdh = stdh[0]
         stdair = stdh[0].airmass

         logprint, '      Calibrator  ', removepath(standardfiles[0]), ' ', stdnames[0], ' @ air=', fpr(stdair,1.3)

         if abs(sciair-stdair) gt 0.35 then begin
            adiff = abs(sciair-stdair)
            if adiff lt 0.45 then adj = 'Significant'
            if adiff ge 0.45 and adiff le 0.9 then adj = 'Large'
            ;if adiff gt 0.65 and adiff lt 1.0 then adj = 'VERY large'
            if adiff gt 0.9 then adj = 'VERY large'
            logprint, '      Warning:  '+adj+' airmass difference between target ('+clip(fpr(sciair,2.2))+ $ 
                                   ') and standard ('+clip(fpr(stdair[0],2.2))+')'
         endif

         if scih.slit ne stdh.slit then begin
            logprint, '        Warning: science slit '+scih.slit+' is different from standard slit '+stdh.slit
         endif
         if abs(scih.cwave - stdh.cwave) gt 10 then begin
            logprint, '        Warning: large flexure ('+clip(scih.cwave - stdh.cwave)+' A) vs. standard '+removepath(stdh.filename)
         endif

         responsefile = repstr(standardfiles[0],'.spec','.response')
         readcol, responsefile, rwav, rcounts, rflux, rresponse, rtelluricresponse, rflag, format='f,f,f,f,f,f', comment='#', /silent
         rresponse *= (scih.specbin*1.0/stdh.specbin) ;*(scih.ybin*1.0/stdh.ybin) ; ybin doesn't matter
         response = interpol2(rresponse, rwav, wav)
         telluricresponse = interpol2(rtelluricresponse, rwav, wav) < 1
         respflag = interpol2(rflag, rwav, wav)
         ; airmass correction (using a standard function)
         airtermarr = poly((wav-5000.)/2500.,airp)  ; magnitudes of extinction per airmass
         ; print, ' airmass difference' , sciair-stdh.airmass
         response *= 2.51^(-(sciair-stdh.airmass)*airtermarr) 
         telluricresponse = telluricresponse^(sciair/stdh.airmass)

         sxaddpar, h, 'FLUXCAL1', removepath(standardfiles[0])
         sxaddpar, h, 'FLUXNAM1', stdnames[0]
         sxaddpar, h, 'FLUXAIR1', stdair

      endif else begin

        stdair = stdh.airmass

        if nmatchstd eq 2 then begin
            wlo = (where(stdair eq min(stdair))) [0]
            whi = (where(stdair eq max(stdair))) [0]
        endif

        if nmatchstd ge 3 then begin
           ; probably this should go back in autolrisfluxcal
           below = where(stdair lt sciair, ctbelow)
           above = where(stdair gt sciair, ctabove)
           if ctbelow ge 1 and ctabove gt 1 then begin
              ; this would be much simpler if std array was just sorted by airmass
              wlo = below[(where(stdair[below] eq max(stdair[below]))) [0]]
              whi = above[(where(stdair[above] eq min(stdair[above]))) [0]]
           endif else begin
              nearest = (sort(abs(stdair-sciair))) [0:1] ; indices of two closest
              wlo = nearest[(where(stdair[nearest] eq min(stdair[nearest]))) [0]]
              whi = nearest[(where(stdair[nearest] eq max(stdair[nearest]))) [0]]
              ; formerly took the extrema, which is not so good
           endelse
        endif
 
        logprint, '      Calibrators ', removepath(standardfiles[wlo]), ' ', stdnames[wlo], ' @ air='+fpr(stdh[wlo].airmass,1.3)
        logprint, '                  ', removepath(standardfiles[whi]), ' ', stdnames[whi], ' @ air='+fpr(stdh[whi].airmass,1.3)

         adiff = min(abs(sciair-stdair[[wlo,whi]]))
         if adiff gt 0.35 then begin
            if adiff lt 0.45 then adj = 'Significant'
            if adiff ge 0.45 and adiff lt 0.9 then adj = 'Large'
            ;if adiff gt 0.65 and adiff lt 1.0 then adj = 'VERY large'
            if adiff ge 0.9 then adj = 'VERY large'
            logprint, '      Warning: '+adj+' airmass difference between target ('+clip(fpr(sciair,2.2))+ $ 
                                   ') and standards ('+clip(fpr(stdair[wlo],2.2))+','+clip(fpr(stdair[whi],2.2))+')'
         endif

        for j = 0, 1 do begin
         if j eq 0 then stdhp = stdh[wlo] else std = stdh[whi]
         if scih.slit ne stdhp.slit then begin
            logprint, '        Warning: science slit '+scih.slit+' is different from standard slit '+stdhp.slit
         endif
         if abs(scih.cwave - stdhp.cwave) gt 10 then begin
            logprint, '        Warning: large flexure ('+clip(fpr(scih.cwave - stdhp.cwave,4.1))+' A) vs. standard '+removepath(stdhp.filename)
         endif
        endfor

        responsefilelo = repstr(standardfiles[wlo],'.spec','.response')
        responsefilehi = repstr(standardfiles[whi],'.spec','.response')

        readcol, responsefilelo, rwavlo, rcountslo, rfluxlo, rresponselo, rtelluricresponselo, rflaglo, format='f,f,f,f,f,f', comment='#'
        readcol, responsefilehi, rwavhi, rcountshi, rfluxhi, rresponsehi, rtelluricresponsehi, rflaghi, format='f,f,f,f,f,f', comment='#'
        rresponselo *= (scih.specbin*1.0/stdh[wlo].specbin) ;*(scih.ybin*1.0/imh[wlo].ybin)
        rresponsehi *= (scih.specbin*1.0/stdh[whi].specbin) ;*(scih.ybin*1.0/imh[whi].ybin)

        responselo = interpol2(rresponselo, rwavlo, wav)
        responsehi = interpol2(rresponsehi, rwavhi, wav)
        telluricresponselo = interpol2(rtelluricresponselo, rwavlo, wav) < 1
        telluricresponsehi = interpol2(rtelluricresponsehi, rwavhi, wav) < 1
        flaglo = interpol2(rflaglo, rwavlo, wav)
        flaghi = interpol2(rflaghi, rwavhi, wav)

        ; airmass correction (using a standard function)
        airtermarr = poly((wav-5000.)/2500.,airp)  ; magnitudes of extinction per airmass
        responselo *= 2.51^(-(sciair-stdh[wlo].airmass)*airtermarr) 
        responsehi *= 2.51^(-(sciair-stdh[whi].airmass)*airtermarr) 
        telluricresponselo = telluricresponselo^(sciair/stdh[wlo].airmass)
        telluricresponsehi = telluricresponsehi^(sciair/stdh[whi].airmass)

        ; average the response curves together with a normalization.
         ; calculate weights based on airmass similarity
        weightlo = 1./abs(sciair-stdh[wlo].airmass) > 0.001
        weighthi = 1./abs(sciair-stdh[whi].airmass) > 0.001
        weightlo /= (weightlo+weighthi)
        weighthi = 1.-weightlo ; won't have to do any division below since this is enforced
         ; normalize and then average response
        onegood  = where((responselo gt 0.1*percentile75(responselo)) or  (responsehi gt 0.1*percentile75(responsehi)), ct1g)
        bothgood = where((responselo gt 0.1*percentile75(responselo)) and (responsehi gt 0.1*percentile75(responsehi)), ct2g)
        if ct1g lt 10 or ct2g lt 10 then begin
           logprint, ' ERROR - something looks very wrong with the flux calibration'
           continue
        endif
        ratiog = responselo[bothgood]/responsehi[bothgood] ; differential-cloudiness factor (as func. of wave)
        ratio = responselo[onegood]/responsehi[onegood]
        difgrey  = alog10(median(ratiog))
        difchrom = stdev(alog10(ratio/median(ratiog)))

        if abs(difgrey) gt 2 then logprint, ' WARNING - very large attenuation difference in these standards'
        if abs(difchrom) gt 0.5 then logprint, ' WARNING - very different profiles between these standards'
        normlo = 1.0
        normhi = 1./median(ratiog)
        ;normlo = percentile75(responselo)
        ;normhi = percentile75(responsehi)  
        norm = (normlo*weightlo + normhi*weighthi)
        responselo *= norm/normlo
        responsehi *= norm/normhi
        response = (responselo*weightlo + responsehi*weighthi)
         ; average telluric response
        telluricresponse = (telluricresponselo*weightlo + telluricresponsehi*weighthi)
        respflag = (flaglo*weightlo+flaghi*weighthi)

        ;window, 12
        ;plot, wav, response, yrange=[0,max([rresponselo,rresponsehi])]
        ;oplot, rwavlo, rresponselo, col=rgb(100,100,255), linestyle=1
        ;;oplot, rwavhi, rresponsehi, col=rgb(255,100,100), linestyle=1
        ;oplot, wav, responselo, col=rgb(100,100,255)
        ;oplot, wav, responsehi, col=rgb(255,100,100)

        if abs(difgrey) gt 3 or abs(difchrom) gt 1 then begin
           logprint, ' Refusing to process this spectrum with these standards!
           continue
        endif

        sxaddpar, h, 'FLUXCAL1', removepath(standardfiles[wlo])
        sxaddpar, h, 'FLUXCAL2', removepath(standardfiles[whi])
        sxaddpar, h, 'FLUXAIR1', stdh[wlo].airmass
        sxaddpar, h, 'FLUXAIR2', stdh[whi].airmass
        sxaddpar, h, 'FLUXNAM1', stdnames[wlo]
        sxaddpar, h, 'FLUXNAM2', stdnames[whi]
        sxaddpar, h, 'FLUXWT1',  weightlo
        sxaddpar, h, 'FLUXWT2',  weighthi
      endelse

      wsignal = where(response gt 0.01*max(response), ct)
      if scih.camera eq 'red' then if max(sci[wsignal].pix) eq max(sci.pix) then $
        wsignal = wsignal[0:n_elements(wsignal)-2] ; last pixel is usually bad on red
      if ct le 1 then begin
           logprint, ' ERROR: Something wrong with flux calibration...'
           continue
      endif

      outspec = sci[wsignal]

      outspec.resp = response[wsignal] * ( keyword_set(notelluric) ? 1.0 : telluricresponse[wsignal] )
      outspec.flux /= outspec.resp
      outspec.sky  /= outspec.resp
      outspec.unc  /= outspec.resp
      outspec.flag = outspec.flag > respflag[wsignal]
      wbad = where(outspec.flag gt 0.8, ctbad)
      if ctbad gt 0 then outspec[wbad].flux = !values.f_nan

      newhist = 'Processed by lrisfluxcal '+tstamp(/ut)+' UT'
      sxaddhist, newhist, h
      
      sxaddpar, h, 'TELLUCOR', ~keyword_set(notelluric), keyword_set(notelluric) ? 'Telluric correction off' : 'Telluric correction applied'
      
      if n_elements(outfiles) gt 0 then outfile = outfiles[i] $
      else outfile = fileappend(sciencefile,'c')
      writespec, outfile, outspec, h

   endfor

end


