;+
; NAME: lrisheaderinfo.pro
;
; PURPOSE: 
;       Get an IDL structure with information from the headers of an array of files.
;       Formerly called readlrisheaders.pro
;
; CALLING SEQUENCE: 
;       lrisheaderinfo, files, crval=crval, badarclist=badarclist, silent=silent
;
; INPUTS:
;	files = Array of filenames. Alternatively, a wildcard string, or a file header 
;               that has already been opened.
;
; OPTIONAL INPUT KEYWORD PARAMETERS:
;       crval = Specify to use the WCS information for ra and dec, rather than RA and DEC.
;       badarclist = List of malfunctioning arcs.  Their flags will be removed.
;       silent = Don't print out error messages,
;-


function lrisheaderinfo, filesin, crval=crval, badarclist=badarclist, silent=silent
 
  if n_elements(filesin) eq 0 then begin
     print, 'No files'
     return, -1
  endif

  if n_elements(filesin) eq 1 then begin
    ; wildcard string (can only provide one)
    if strpos(filesin,'*') ge 0 or strpos(filesin,'?') ge 0 then files = file_search(filesin) $
                                                            else files = filesin
    nfiles = n_elements(files)

  endif else begin
    ; a list of files or an already-read file header

    if n_elements(filesin) gt 10 and total(strpos(filesin,'=') eq 8) ge 5 then begin
        h = filesin ; it's a FITS header.
        nfiles = 1
        files = ''
    endif else begin
        files = filesin
        nfiles = n_elements(files)
    endelse

  endelse

  lris = replicate({filename:'', framenum:-1, date:'', ut:'', utstop:'', jd:0.D, mjd:0.D, $
                    mjdstop:0.D, exptime:!values.f_nan, elaptime:!values.f_nan, exp:!values.f_nan, $
                    tstart:0.D, tstop:0.D, $
                    sunelev:!values.f_nan, moonelev:!values.f_nan, moondist:!values.f_nan, $
                    object:'', target:'',  rastr:'', decstr:'', radeg:0.D, decdeg:0.D, $
                    airmass:!values.f_nan, azimuth:!values.f_nan, elevation:!values.f_nan, $
                    domeazimuth:!values.f_nan, domerelaz:!values.f_nan, slitpa:!values.f_nan, $
                    drive:!values.f_nan, $
                    instrument:'', camera:'', trapdoor:'', slit:'', $
                    dichroic:'', grname:'', filter:'', $
                    lamps:'', hg:0b, neon:0b, ar:0b, cd:0b, zn:0b, hal:0b, kr:0b, xe:0b, fe:0b, de:0b, nlamp:0, $
                    wavelength:!values.f_nan, cwave:!values.f_nan, cwav:!values.f_nan, $
                    skycounts:!values.f_nan, maxcounts:!values.f_nan, ctrate:!values.f_nan, maxsource:!values.f_nan, $
                    window:'', pane:'', win:'', chip:'', binning:'', bin:'', spatbin:0, specbin:0, $ 
                    nx:0, ny:0, x0:0, y0:0, pxscale:!values.f_nan, dispersion:!values.f_nan, $
                    ampmode:'', poname:'', $
                    satval:!values.f_nan, $
                    dichf:'', rfiltf:'', bfiltf:'', gratf:'', granglef:'',slitf:'', $
                    ncomb:0, koaid:'', koaframe:'', detector:''}, nfiles)  ; much more to come
 
  ; get important header information for a block of files.

   for f = 0, nfiles-1 do begin
     if files[f] eq '' then continue
     if file_test(files[f]) eq 0 then begin
        if keyword_set(silent) eq 0 then print, 'Cannot find file ', files[f]
        continue 
     endif

     lris[f].filename = files[f]
     if strpos(files[f],'.fits') ge 0 then h = headfits(files[f], /silent)
     if strpos(files[f],'.spec') ge 0 then h = headspec(files[f])
     if strpos(files[f],'.fits') lt 0 and strpos(files[f],'.spec') lt 0 then h = headread(files[f])

     if size(h,/tn) eq 'LONG' then begin
        logprint, ' Not a valid header'
        continue
     endif

     lris[f].framenum = sxpar(h,'FRAMENO')

       ; Date/time
     lris[f].date = clip(sxpar(h,'DATE', /silent))   ; date string, e.g. "2013-01-13T08:48:37"
     lris[f].ut   = clip(sxpar(h,'UT', /silent, count=ct))
        if ct eq 0 then lris[f].ut   = clip(sxpar(h,'UTC', /silent, count=ct))
     lris[f].jd   = double(sxpar(h,'JD')) ; strings are not read properly by SXPAR, unfortunately, so the precision is lost.
     lris[f].mjd   = double(sxpar(h,'MJD-OBS')) 
     lris[f].utstop   = clip(sxpar(h,'UTEND', /silent, count=ct))
        if ct eq 0 then lris[f].utstop = clip(sxpar(h,'UTC-END', /silent, count=ct))
        if ct eq 0 then lris[f].utstop = clip(sxpar(h,'DATE_END', /silent, count=ct))
        if ct eq 0 then lris[f].utstop = clip(sxpar(h,'UTSTOP', /silent, count=ct)) ; used in imaging stacks
     lris[f].ut     = strmid(lris[f].ut,strpos(lris[f].ut,'T')+1)        ; remove the date if present
     lris[f].utstop = strmid(lris[f].utstop,strpos(lris[f].utstop,'T')+1) ; remove the date if present
     lrisversion = 0
     if lris[f].mjd gt 49000 and lris[f].mjd lt 54983 then lrisversion = 1 ; installation to 2009-05-31
     if lris[f].mjd ge 54983 and lris[f].mjd lt 55529 then lrisversion = 2 ; 2009-06-01 to 2010-11-28
     if lris[f].mjd ge 55529 then lrisversion = 3                  ; 2010-11-29 to present
     if lrisversion eq 0 then begin
        year = float(clip(lris[f].date,4))+(float(strmid(lris[f].date,5,2))-1)/12.;+(float(strmid(lris[f].date,8,2))-1)/12./31.
        lrisversion = 2
        if year lt 2009.4 then lrisversion = 1
        if year gt 2010.9 then lrisversion = 3
     endif

       ; Exposure
     lris[f].exptime  = float(sxpar(h,'EXPTIME',count=ctx))   ; effective time on source (what you would divide by to get the electron rate) - this keyword is not written to raw files anymore?  is ZERO when this happens...
     lris[f].elaptime = float(sxpar(h,'ELAPTIME',count=ctl,/silent))   ; including pauses and exposure gaps (?)
     ttime   = float(sxpar(h,'TTIME',  count=ctt, /silent))
     teltime = float(sxpar(h,'TELAPSE',count=cte, /silent))
     xptime  = float(sxpar(h,'XPOSURE',count=ctxp,/silent))
     ; > float(sxpar(h,'TOTALEXP',/silent)) 
     ; float(sxpar(h,'TTIME',/silent))) >  - never use TTIME
     if ctx eq 1 then $
        lris[f].exp = lris[f].exptime $
     else if ctxp eq 1 then $                 ; <-- behaves like elif
        lris[f].exp = xptime $
     else if ctl eq 1 then $
        lris[f].exp = lris[f].elaptime $    ; use EXPTIME if possible, else ELAPTIME
     else $
        lris[f].exp = !values.f_nan

     mjdstop = sxpar(h,'MJDEND',count=ct,/silent)
     if ct gt 0 then lris[f].mjdstop = mjdstop else $
       lris[f].mjdstop  = lris[f].mjd + lris[f].elaptime/(24.*3600.)

     lris[f].tstart = ten(lris[f].ut)*3600. ; secs since midnight
     if lris[f].utstop ne '' then $
       lris[f].tstop  = ten(lris[f].utstop)*3600. $
     else $
       lris[f].tstop = lris[f].tstart+lris[f].exp

       ; Astronomical ephemeris
     lris[f].sunelev  = sxpar(h,'SUNELEV')
     lris[f].moonelev = sxpar(h,'MOONELEV')
     lris[f].moondist = sxpar(h,'MOONDIST')
 
       ; Target and sky position
     lris[f].object = clip(sxpar(h,'OBJECT', /silent))
     lris[f].target = repstr(clip(sxpar(h,'TARGNAME')),' ', '_')  ; spaces cause barfing in filenames
     lris[f].rastr  = clip(sxpar(h,'RA'))
     lris[f].decstr = clip(sxpar(h,'DEC'))
     lris[f].radeg  =  ten(sxpar(h,'RA'))*15.
     lris[f].decdeg =  ten(sxpar(h,'DEC'))
     if keyword_set(crval) then lris[f].radeg  = sxpar(h,'CRVAL1')
     if keyword_set(crval) then lris[f].decdeg = sxpar(h,'CRVAL2')

       ; Telescope position
     lris[f].airmass = sxpar(h,'AIRMASS')
     lris[f].azimuth = sxpar(h,'AZ')
     lris[f].elevation = sxpar(h,'EL')
     lris[f].domeazimuth = sxpar(h,'DOMEPOSN')
     lris[f].domerelaz = lris[f].domeazimuth - lris[f].azimuth
        while lris[f].domerelaz lt -180.0 do lris[f].domerelaz += 360.
        while lris[f].domerelaz ge  180.0 do lris[f].domerelaz -= 360.
        if lris[f].domeazimuth eq 0.0 then lris[f].domerelaz = 0.0  ; Sometimes DOMEPOSN is missing and causes data to be excluded.
     lris[f].slitpa = sxpar(h,'SLITPA')                             ; need a more robust solution.
     lris[f].drive = sxpar(h,'ROTPPOSN') ; is this really the drive angls?

       ; Instrument state
     lris[f].instrument = clip(sxpar(h, 'INSTRUME'))             ; instrument name, "LRIS" or "LRISBLUE"
     if lris[f].instrument eq 'LRISBLUE' then lris[f].camera = 'blue'
     if lris[f].instrument eq 'LRIS' then     lris[f].camera = 'red'
     if lris[f].instrument eq 'LRISBLUE' then lris[f].detector = 'LRISB'
     if lris[f].instrument eq 'LRIS' then     lris[f].detector = 'LRIS'+strupcase(strmid(lris[f].camera,0,1))+clip(lrisversion)
     lris[f].trapdoor   = clip(sxpar(h,'TRAPDOOR'))              ; trapdoor state, "open" or "closed"
     lris[f].slit       = clip(sxpar(h, 'SLITNAME'))             ; slit name, e.g. "long_1.0" or "direct"
     lris[f].dichroic   = clip(string(sxpar(h,'DICHNAME')))
     grkey = ''
     if lris[f].instrument eq 'LRISBLUE' then grkey = 'GRISNAME'
     if lris[f].instrument eq 'LRIS'     then grkey = 'GRANAME'
     if grkey ne '' then lris[f].grname = clip(sxpar(h,grkey)) $
                    else lris[f].grname = clip(sxpar(h,'GRISNAME'))+'+'+clip(sxpar(h,'GRANAME'))
     fkey = ''
     if lris[f].instrument eq 'LRISBLUE' then fkey = 'BLUFILT'
     if lris[f].instrument eq 'LRIS'     then fkey = 'REDFILT'
     if lris[f].instrument eq 'LRISBLUE' and clip(lris[f].grname,2) eq 'NB' then fkey = 'GRISNAME'
     if fkey ne '' then lris[f].filter = clip(sxpar(h,fkey))
     lampstr = clip(sxpar(h,'LAMPS', count=ctlampkey))
     if ctlampkey eq 0 then begin
        lris[f].hg = clip(fix(clip(sxpar(h,'MERCURY')) eq 'on'))
        lris[f].neon=clip(fix(clip(sxpar(h,'NEON')) eq 'on'))
        lris[f].ar = clip(fix(clip(sxpar(h,'ARGON')) eq 'on'))
        lris[f].cd = clip(fix(clip(sxpar(h,'CADMIUM')) eq 'on'))
        lris[f].zn = clip(fix(clip(sxpar(h,'ZINC')) eq 'on'))
        lris[f].hal= clip(fix(clip(sxpar(h,'HALOGEN')) eq 'on'))
        lris[f].kr = clip(fix(clip(sxpar(h,'KRYPTON')) eq 'on'))
        lris[f].xe = clip(fix(clip(sxpar(h,'XENON')) eq 'on'))
        lris[f].fe = clip(fix(clip(sxpar(h,'FEARGON')) eq 'on'))
        lris[f].de = clip(fix(clip(sxpar(h,'DEUTERI')) eq 'on'))
     endif else begin
       if strpos(lampstr,'1') ge 0 then begin
        if n_elements(badarclist) gt 0 then fixlampstr, lampstr, badarclist
        lamps = fix(strsplit(lampstr,',',/extract))
        lris[f].hg =   lamps[0]
        lris[f].neon = lamps[1]
        lris[f].ar =   lamps[2]
        lris[f].cd =   lamps[3]
        lris[f].zn =   lamps[4]
        lris[f].hal =  lamps[5]
        if n_elements(lamps) ge 10 then begin
           lris[f].kr = lamps[6]  ; formerly these were backwards... but this order matches lrisprepare
           lris[f].xe = lamps[7]
           lris[f].fe = lamps[8]
           lris[f].de = lamps[9]
        endif
       endif
     endelse
     lris[f].lamps = lampstr
     lris[f].nlamp = fix(lris[f].hg)+lris[f].neon+lris[f].ar+lris[f].cd+lris[f].zn+lris[f].hal+$
                     lris[f].xe+lris[f].kr+lris[f].fe+fix(lris[f].de)

     lris[f].wavelength = sxpar(h,'WAVELEN') 
     lris[f].cwave = sxpar(h,'CWAVE') ; initial guess
     if lris[f].cwave le 0. then begin ; backwards compatibility
      if lris[f].camera eq 'r' then begin
        lris[f].cwave = sxpar(h,'WAVELEN')
      endif else begin
        if lris[f].grname eq '300/5000' then lris[f].cwave = 4495.
        if lris[f].grname eq '400/3400' then lris[f].cwave = 3600. 
        if lris[f].grname eq '600/4000' then lris[f].cwave = 4305.
        if lris[f].grname eq '1200/3400' then lris[f].cwave = 3400.
      endelse   
     endif
     lris[f].cwav = sxpar(h,'CWAV') ; actual, measured, from solution, post flexure

       ; Background-related keywords
     lris[f].skycounts = sxpar(h,'SKYMED', /silent, count=ct)
       if ct eq 0 then lris[f].skycounts = sxpar(h,'COUNTS')
     lris[f].ctrate  = sxpar(h,'SKYRATE', /silent, count=ct)
       if ct eq 0 then lris[f].ctrate  = sxpar(h,'CTRATE')
     lris[f].maxcounts =  sxpar(h,'MAXCOUNT')
     lris[f].maxsource = sxpar(h,'SOURCMAX', /silent, count=ct)
        if ct eq 0 then lris[f].maxsource =  sxpar(h,'MAXCOUNT')-lris[f].skycounts

       ; CCD keywords
     lris[f].window = clip(sxpar(h,'WINDOW',count=ctww))
     lris[f].pane   = clip(sxpar(h,'PANE', count=ctp))
     lris[f].win    = clip(sxpar(h,'WINDET',count=ctw))
     if ctw eq 0 and ctp gt 0 then lris[f].win = lris[f].pane
     if ctw eq 0 and ctww gt 0 then lris[f].win = strmid(clip(lris[f].window),2)+'W'
           ; else lris[f].win = strmid(clip(lris[f].window),2)

     lris[f].chip = clip(sxpar(h, 'CHIP'))
     lris[f].binning = repstr(clip(sxpar(h,'BINNING')),' ','')
     lris[f].bin = repstr(lris[f].binning,',','x')
     lris[f].nx = sxpar(h,'NAXIS1')
     lris[f].ny = sxpar(h,'NAXIS2')
     lris[f].x0 = -fix(sxpar(h,'LTV1')) ; theoretically degenerate with pane...
     lris[f].y0 = -fix(sxpar(h,'LTV2'))
     binning = clip(sxpar(h,'BINNING'))
     binarr = fix(strsplit(binning,',', /extract))
     if n_elements(binarr) ge 2 then begin
       lris[f].spatbin = binarr[0] ;sxpar(h,'YBIN')
       lris[f].specbin = binarr[1] ;sxpar(h,'XBIN')
     endif

     lris[f].ampmode = sxpar(h,'AMPMODE')
     lris[f].poname = sxpar(h,'PONAME')

      ; First-of-config file references
     lris[f].dichf = sxpar(h,'DICHSEQF') 
     lris[f].rfiltf = sxpar(h,'RFLTSEQF')
     lris[f].bfiltf = sxpar(h,'BFLTSEQF')
     lris[f].gratf = sxpar(h,'GRATSEQF') 
     lris[f].granglef = sxpar(h,'GRANSEQF')
     lris[f].slitf = sxpar(h,'SLITSEQF')

      ; Other
     lris[f].pxscale = sxpar(h,'PXSCALE')
     lris[f].dispersion = sxpar(h,'DISPERSN')
     lris[f].satval =  sxpar(h,'SATURATE')
     lris[f].ncomb = sxpar(h,'NCOMBINE')

     ;lris[f].koaid = sxpar(h,'KOAID', count=ctk) ; full id
     koaid = sxpar(h,'KOAID', count=ctk)
     if ctk eq 0 then koaid = sxpar(h,'KOAID1', count=ctk)
     if ctk gt 0 then begin
       lris[f].koaid = koaid
       koaarr = strsplit(koaid,'.',/extract)
       if n_elements(koaarr) ge 3 then lris[f].koaframe = repstr(rclip(koaarr[2],5),' ','0')
     endif

   endfor

   return, lris

end

