;+
; NAME:
;       lrishtmltable
;
; PURPOSE:
;       Write out an HTML file summarizing processed LRIS data and making 
;       it available via hyperlinks.  For archiving / data sharing.
;
; CALLING SEQUENCE:
;       lrishtmltable, spectra, [sort=]
;
; EXPLANATION:
;       Used to generate the static online catalogues for the public LRIS 
;       reduction archive.
;
; INPUTS:
;       spectra - A list of filenames, or a wildcard string.
;
; OPTIONAL INPUT KEYWORDS:
;       /nosort - Sort by file order rather than by MJD.
;
; OUTPUTS:
;       The file lristable.html is written to disk.
;
;-
;------------------------------------------------------------------------------


pro lrishtmltable, spectra, nosort=nosort
 
  if n_elements(spectra) eq 0 then spectra = '*.spec'
  if n_elements(spectra) eq 1 then spectra = file_search(spectra)

  if keyword_set(nosort) eq 0 then begin
    mjd = dblarr(n_elements(spectra))
    for i = 0, n_elements(spectra)-1 do mjd[i] = sxpar(headspec(spectra[i]),'MJD')
    spectra = spectra[sort(mjd)]
  endif

  png = repstr(spectra,'.spec','.spec.png')

  openw, 1, 'lristable.html'

  printf, 1, '<!DOCTYPE html>'
  printf, 1, '<html>'
  printf, 1, '<head>'
  printf, 1, '<title>LRIS reduced longslit spectra</title>
  printf, 1, '<style>'
  printf, 1, 'th {'
  printf, 1, ' font-size: 75%;'
  printf, 1, '}'
  printf, 1, 'td {'
  printf, 1, ' font-size: 75%;'
  printf, 1, '}'
  printf, 1, 'a {'
  printf, 1, '  color: slategray;'
  printf, 1, '}'
  printf, 1, 'p {'
  printf, 1, '  margin: 10px;'
  printf, 1, '}'
  printf, 1, '</style>
  printf, 1, '</head>'
  printf, 1, '<body>'
  printf, 1, '<h2>LRIS reduced longslit spectra</h2>'
  printf, 1, '<hr/>'
  printf, 1, '<p>Spectra were reduced by Daniel Perley using <a href="http://www.astro.caltech.edu/~dperley/programs/lpipe.html">LPipe v2019.03</a>.  Reductions were fully automatic and unsupervised, and are provisional: they are meant to demonstrate the capabilities of the LRIS automated pipeline, rather than to produce archive-quality data for every object.  A more carefully supervised reduction run (which will include all LRIS data dating back to 2010 or earlier) is planned for the near future.  If you do use these spectra, please be sure to cite Perley 2019 (PASP, submitted) and acknowledge the observer, the program PI, and the Keck archive.</p>'
  printf, 1, '<p>Final output products were briefly validated to check for easily-apparent reduction issues, as described by the three QA indicators below:<br/>'
;  printf, 1, '<p style="margin-left: 40px"><u>QA validation codes</u>:<br/>'
  printf, 1, ' <span style="margin-left: 40px"><span style="color:green;">pass</span> : Spectrum is problem-free or shows only minor issues (e.g. small residuals/offsets or edge issues)</span><br/>'
  printf, 1, ' <span style="margin-left: 40px"><span style="color:#D3BB10;">warn</span> : Spectrum shows more significant issues, but these are easily correctable or impact only a limited region</span><br/>'
  printf, 1, ' <span style="margin-left: 40px"><span style="color:red;">fail</span> : Spectrum shows problems affecting a large fraction of the spectrum and should not be used.</span></p>'
  printf, 1, '<p>Validation is not guaranteed to catch all possible problems.  The primary reason for the warning/failure on a given observation can be checked using the mouseover text for the QA column entry.  (Note: some spectra may have more than one problem simultaneously.)</p>'
  printf, 1, '<p>Mouseover a column header for more information.  All spectrum files are ASCII-format and contain extensive header metadata that can be checked for more information about the observations and reductions.</p>'
  printf, 1, '<hr/>'
  printf, 1, '<table>'

  for i = -1, n_elements(spectra)-1 do begin
     file = spectra[i]
     if i ge 0 then begin
       h = headspec(file)
       status = clip(sxpar(h,'VALIDATE'))
       reason = repstr(clip(sxpar(h,'REASON')),'_',' ')
       if reason eq '0' then reason=''
       valcell = '<td>'
        if status eq 'pass' then $
          valcell = '<td style="color:green;">pass'
        if status eq 'flag' or status eq 'warn' then $
          valcell = '<td style="color:#D3BB10;" title="'+reason+'">warn'
        if status eq 'fail' then $
          valcell = '<td style="color:red;" title="'+reason+'">fail'
       ut = repstr(sxpar(h,'DATE_BEG'),'T',' ')
       targ = clip(sxpar(h,'TARGNAME'))
       progid = clip(sxpar(h,'PROGID'))
       pi = sxpar(h,'PROGPI')
       air = fpr(sxpar(h,'AIRMASS'),2.2)
       ra = sxpar(h,'RA')
       dec = sxpar(h,'DEC')
       rexp = clip(round(sxpar(h,'REDEXPTM')))
       bexp = clip(round(sxpar(h,'BLUEXPTM')))
       grat = clip(sxpar(h,'GRANAME'))
       if clip(grat) eq '0' then grat = '-'
       gris = clip(sxpar(h,'GRISNAME'))
       if clip(gris) eq '0' then gris = '-'
       if grat ne '-' then wave = '('+clip(round(sxpar(h,'WAVELEN')))+')' else wave = ''
       pa = fpr(sxpar(h,'SLITPA'),4.1)
       sun = sxpar(h,'SUNELEV')
       moon = sxpar(h,'MOONELEV')
       moonphase = sxpar(h,'MOONPHAS')
       if sun lt -16 then sunstr = '' else sunstr = clip(round(sun))
       if moon gt 0 then moonstr = clip(round(100*moonphase))+'%' else moonstr = ''
       filelink = '<a href="'+file+'">'+file+'</a>'
       if file_test('png/'+png[i]) then pnglink = '<a href="png/'+png[i]+'">preview</a>' else pnglink = ''
     endif else begin
       filelink = '<span title="Spectrum file (click to download)">filename</span>'
       pnglink = ''
       valcell = '<td><span title="Validation status">QA</span>'
       ut = '<span title="UT start time">UT start</span>'
       targ = '<span title="Target name (from observer starlist)">target</span>'
       progid = '<span title="Program ID code">program</span>'
       pi = '<span title="Name of principal investigator">PI</span>'
       air = '<span title="Airmass">airmass</span>'
       ra = '<span title="Right ascension">RA</span>'
       dec = '<span title="Declination">Dec</span>'
       rexp = '<span title="Red exposure time (s)">exp</span>'
       bexp = '<span title="Blue exposure time (s)">exp</span>'
       grat = '<span title="Red grating">grating</span>'
       wave = '<span title="Red central wavelength">(wave)</span>'
       gris = '<span title="Blue grism">grism</span>'
       pa = '<span title="Position angle (deg CCW of due north)">PA</span>'
       sunstr = '<span title="Sun elevation (blank if <-18)">twi</span>'
       moonstr = '<span title="Moon phase (blank if set)">moon</span>'
     endelse 
     begrow = '<tr>'
     endrow = '</td></tr>'

     s = '</td><td>'
     sb = '</td><td style="color:blue;">'
     sr = '</td><td style="color:red;">'
     
     out = begrow+$
                valcell+s+ $
                filelink+s+ $
                pnglink+s+ $
                ut+s + $
                targ+s + $
                ra+s + $
                dec+sb + $
                gris+sr + $
                grat+wave+sb + $
                bexp+sr + $
                rexp+s + $
                pa+s + $ 
                air+s + $ 
                sunstr+s + $ 
                moonstr+s + $
                progid+'/'+pi+ $
                endrow
     if i eq -1 then out = repstr(out,'td>','th>')
     if i eq -1 then out = repstr(out,'<td','<th')
     ;print, out
     printf, 1, out
  endfor

  printf, 1, '</table>'
  printf, 1, '</html>'
  printf, 1, '</body>'
  close, 1

end

