;+
; NAME:
;       lrisrename
;
; PURPOSE:
;       Tool to rename LRIS files following a different convention.
;
; CALLING SEQUENCE:
;       lrisrename, filestr, [/check, /koa, /orig, /std]
;
; EXPLANATION:
;       Some users prefer KOA names to match names as logged at the telescope
;       (the pipeline itself previously also required that format) and older
;       data does not include the date and thus is not unique.
;       This enables the original names to be restored.  Can also revert to 
;       KOA names (as long as they are in the headers - not available while
;       observing.)
;
; INPUTS:
;       filestr - Wildcard string indicating what files to rename.
;
; OPTIONAL INPUT KEYWORDS:
;       /check - Print out list of files to rename, but do not rename.
;       /koa - Rename to KOA names.
;       /orig - Rename to exactly-as-at-the-telescope names.
;
; EXAMPLES:
;       lrisrename, 'lblue*.fits'
;
; NOTES:
;       If neither /koa or /orig is set (default), the names are based on the
;       DATE and FRAMENO keywords.  For recent Keck data usually this is the 
;       same as at-the-telescope filenames except when there are duplicates.
;       If a duplicate file is detected lrisrename appends 'b', 'c', etc.,
;       and the sequence may not necessarily match the 'b', 'c' as at the 
;       telescope.
;       Use lrisresequence to ensure unique names without additional letters
;       (or just use the archive filenames).
;
;-
;------------------------------------------------------------------------------

pro lrisrename, filestr, check=check, koa=koa, orig=orig, std=std

  if n_elements(filestr) eq 0 then begin
     print, 'Specify files to rename (or file wildcard string).'
     print, "   For KOA files, try: 'L?.*.fits'"
     return
  endif
  ct = 1
  if n_elements(filestr) eq 1 then files = file_search(filestr, count=ct) else files = filestr
  if ct eq 0 or file_test(filestr[0]) eq 0 then begin
     print, 'File(s) not found: ', filestr
  endif

  outtype = 'std'
  if keyword_set(koa) then outtype = 'koa'
  if keyword_set(orig) then outtype = 'orig'
  if keyword_set(std) then outtype = 'std' ; default

  outfiles = ['']
  alphabet = ['','b','c','d','e','f','g','h','i','j']
 
  ; OFNAME contains the original name
  ; KOAID  contains the new name

  prevfilenum = -1
  for f = 0, n_elements(files)-1 do begin
     if files[f] eq '' then begin
        print, 'No files matching ', filestr
        continue
     endif

     path = extractpath(files[f])
     zipped = strpos(files[f],'.gz') gt 0
     zipstr = zipped ? '.gz' : ''
     h = headfits(files[f])

     ofile = sxpar(h,'OFNAME', count=ctofile) ; this is NOT necessarily a unique name.

     if outtype eq 'koa' then begin
       koaid = sxpar(h,'KOAID', ctkoa)
       if ctkoa eq 0 then begin
         print, ' File '+files[f]+' contains no KOAID!'
         print, '  Not renaming...'
         continue
       endif
       outname = koaid+zipstr
     endif
     if outtype eq 'orig' then begin
       if ctofile eq 0 then begin
         print, ' File '+files[f]+' contains no OFILE keyword!'
         print, '  Not renaming...'
         continue
       endif
       outname = ofile+zipstr
     endif
     if outtype eq 'std' then begin
       framenum = sxpar(h,'FRAMENO', count=ct)
       if ct eq 0 and ctofile gt 0 then begin
          framenum = strmid(ofile,strpos(ofile,'_')+1,strpos(ofile,'.')-strpos(ofile,'_')-1)
       endif
       if ct eq 0 and ctofile eq 0 then begin
          print, 'WARNING: File header is missing frame number! ', files[f]
          print, '   Not renaming.'
          continue
       endif

       fpos = strpos(files[f],'.fits')
       if fpos gt 0 then filestem  = strmid(files[f],0,fpos)
       if fpos gt 0 then extension = strmid(files[f],fpos) $ ; includes the .
                  else extension = '.fits'+zipstr

       ;firstdigit = 99
       ;for i = 0, 9 do firstdigit = strpos(ofile,clip(digit[i])) < firstdigit
       ;ofilepstr = strmid(ofile,0,firstdigit)
       focus = 0
       if strpos(ofile,'foc') gt 0 then focus = 1 
       filenumstr = string(framenum,'(I04)')
       instrume = clip(sxpar(h,'INSTRUME'))
       inststr = ''
       if instrume eq 'LRIS' then inststr = 'r'
       if instrume eq 'LRISBLUE' then inststr = 'b'
       if inststr eq '' then begin
         print, 'Cannot recognize instrument.'
         continue
       endif
       datestr = ''
       hdate = sxpar(h,'DATE', /silent, count=ct)
       if ct gt 0 then begin
          datestr = rundatestr(hdate, /nocentury)
       endif
       if datestr eq '' then print, 'WARNING: Cannot recognize UT date from file header'

       focstr = ''
       if focus then focstr = 'foc'
       outname = path + inststr + focstr + datestr + '_' + filenumstr + extension

       a = 0
       while total(outname eq outfiles) gt 0 do begin 
          ; we already wrote a file with this name
          a += 1        
          outname = inststr + focstr + datestr + '_' + filenumstr + alphabet[a]+extension
       endwhile
    
     endif

     if file_test(outname) then begin
        print, 'WARNING: File already exists!: ', outname
        print, '   Not overwriting.'
        continue
     endif
  
     print, files[f], ' -> ', outname
     outfiles = [outfiles, outname]
     if keyword_set(check) eq 0 then file_move, files[f], outname 

  endfor


end

