;+
; NAME:
;       lrisresequence
;
; PURPOSE:
;       Completely alter the LRIS frame numbers to ensure time-sequentiality.
;
; CALLING SEQUENCE:
;       lrisresequence, filestr, [/check]
;
; EXPLANATION:
;       On some occasions the file sequence may be reset during the night
;       (e.g. for split nights), leading to file names that are not in order
;       or which have additional letters in the filename that hide them from
;       the pipeline.  The files can be renamed(/numbered) in a logical order
;       by running this tool.  
;       (Alternatively, download data from KOA and just use the KOA names.)
;
; INPUTS:
;       filestr - Wildcard string indicating files to rename.
;
; OPTIONAL INPUT KEYWORDS:
;       /check - Print out the new names but don't actually rename the files.
;
; EXAMPLES:
;       lrisresequence, 'lr170101_*.fits'
;
; NOTES:
;       Since the sequence number depends on what files you locally have on 
;       disk, for cross-tracking purposes it's probably better to use
;       lrisrename or just stick with the KOA names.  However if running on
;       un-archived data and the file names are a mess, this may be needed.
;
;-
;------------------------------------------------------------------------------


pro lrisresequence, filestr, check=check

  if n_elements(filestr) eq 0 then begin
     print, 'Specify files to rename (or file wildcard string).'
     return
  endif
  ct = 1
  if n_elements(filestr) eq 1 then allfiles = file_search(filestr, count=ct) else allfiles = filestr
  if ct eq 0 or file_test(filestr[0]) eq 0 then begin
     print, 'File(s) not found: ', filestr
  endif

  if allfiles[0] eq '' then return
  hinfo = lrisheaderinfo(allfiles)
  ord = sort(hinfo.date+hinfo.ut)
  allfiles = allfiles[ord]
  hinfo = hinfo[ord]

  for cam = 0, 1 do begin
    if cam eq 0 then inst = 'LRIS'
    if cam eq 1 then inst = 'LRISBLUE'
    if inst eq 'LRIS' then inststr = 'r'
    if inst eq 'LRISBLUE' then inststr = 'b'
    w = where(hinfo.instrument eq inst, ct)

    for f = 0, ct-1 do begin
     h = headfits(allfiles[w[f]])
     oldfilenum = sxpar(h,'FRAMENO', /silent, count=ct)

     newfilenum = f
     filenumstr = string(newfilenum,'(I04)')

     datestr = ''
     utdate = sxpar(h,'DATE', /silent, count=ct)
     if ct gt 0 then begin
        utdate = (strsplit(utdate,'T',/extract)) [0]
        if strpos(utdate,'-') gt 0 then datestr = repstr(strmid(utdate,2),'-','') else $
        if strpos(utdate,'/') gt 0 then datestr = repstr(strmid(utdate,2),'/','')
     endif
     if datestr eq '' then print, 'WARNING: Cannot recognize UT date from file header'

     outname = inststr + datestr + '_' + filenumstr + '.fits'
     
     if file_test(outname) then begin
        print, 'WARNING: File already exists!: ', outname
        print, '   Not overwriting.'
        continue
     endif
    ;hinfo[w[f]].instrument, ' ', 
     print, utdate, ' ', hinfo[w[f]].ut, ' ', allfiles[w[f]], ' -> ', outname
     if keyword_set(check) eq 0 then spawn, 'mv '+allfiles[w[f]]+' '+outname 
    endfor
  endfor


end

