;+
; NAME:
;       lrisresponse
;
; PURPOSE:
;       Determine the response curve for a LRIS standard star observation.
;
; CALLING SEQUENCE:
;       lrisresponse, filename, [imh=imh, outfile=outfile, updated=updated]
;
; EXPLANATION:
;       Given a wavelength-calibrated but non-flux-calibrated LRIS spectrum of
;       a spectrophotometric standard, this will calculate the response curve
;       and save it to disk as a .response file.  Telluric response is treated
;       separately from instrument response and known lines are interpolated 
;       over.
;
; INPUTS:
;       filename - Spectrum to calculate response for.
;
; OPTIONAL INPUT KEYWORDS:
;       imh - Header structure for the filename. 
;             (Read from disk if not specified.)
;       outfile - Output file for the response curve.  By default, .spec is
;             replaced with .response which is what the pipeline expects.
;
; OPTIONAL OUTPUT KEYWORDS:
;       updated - Returns 1 if anything was written to disk.
;           (both of these keywords are to help with pipeline operations)
;
; OUTPUTS:
;       A .response file is written to disk.
; 
; NOTES:
;       The standard star archival spectrum is looked up in the LPipe refdata/
;       subdirectory and identified automatically given the coordinates/targname
;       from the spectrum's header.  A more general response-solution tool based
;       on this codebase and separated out is a future ambition.
;
;-
;------------------------------------------------------------------------------

pro lrisresponse, filename, imh=imh, outfile=outfile, updated=updated

   common lrisauto, lpar

   standardpath = lpar.refdatapath
    

   if n_elements(imh) eq 0 then imh = lrisheaderinfo(file)

      ; Confirm that target is in standards archive based on its name/position, and get its standardized name
   readstandard, imh.target, pos=[imh.radeg, imh.decdeg], stdname=stdname, /check, result=result, path=standardpath+'/'

      ; print, 'checking ', removepath(filename), ' (', imh.target,'@'+clip(sixtystr(imh.radeg,/ra)),',', clip(sixtystr(imh.decdeg)), '):', result

   if result eq 0 then return

   obs = readspec(filename, h, /timenorm, cutoff=3000.)

      ; cwav = sxpar(h,'CWAV')   ; not needed here
   nwav = sxpar(h,'NWAV',count=ctnwav)      
   if ctnwav gt 0 and nwav eq 0 then return

   nsaturate = sxpar(h,'NSATURAT')
   if nsaturate gt 3 then begin
       logprint, ' Calibrator ',removepath(filename), ' has '+clip(nsaturate)+' saturated pixels; skipping.'
       return
   endif

   sn80 = percentile(obs.flux/obs.unc, 80)
   if sn80 lt 20 then begin 
       logprint, ' Calibrator ',removepath(filename), ' has very low S/N; skipping.'
       return
   endif
   if sn80 lt 50 then begin 
       logprint, ' WARNING: calibrator ',removepath(filename), ' has low S/N ('+clip(fpr(sn80,3.2))+'); skipping.'
       return
   endif

   name = stdname


      ; 'obs' is the observation of the standard.
      ; 'ref' is the reference of the standard.
   logprint, ' Calibrator ', removepath(filename), ' ', stdname, ' @ air=', fpr(imh.airmass,1.3)

   wav = obs.wav        ; may want to consider oversampling this in the future?
   obsflux = obs.flux
   flag = obs.flag
    
   readstandard, stdname, refwav_in, refflux_in, unit='flambda/ang', path=standardpath+'/', stdname=stdnameconfirm
   ;print, '  Read in ', imh.target, ' as ', stdname, ' ', stdnameconfirm
   ;print, min(wav), max(Wav), min(refwav_in), max(refwav_in)
   if max(refwav_in) lt max(wav) then begin
      if max(refwav_in) lt max(wav)-100 then begin
         logprint, ' WARNING: archival standard spectrum for '+clip(stdname)+' does not cover full wavelength range!'
         logprint, '          Extrapolating to the red with a Rayleigh-Jeans law, but caution should be used...'
      endif
      addwav = max(wav) +  100 * findgen(1 + fix((max(wav)-max(refwav_in))/ 100))
      addflux = refflux_in[n_elements(refflux_in)-1]* (addwav/refwav_in[n_elements(refflux_in)-1])^(-4.)
      refwav_in = [refwav_in, addwav]
      refflux_in = [refflux_in, addflux]
   endif
   refflux = interpol2(refflux_in, refwav_in, wav)

      ; remove lines
   lookupstandardlines, stdname, lines, linev, linesnt, linevnt ; get list of absorption lines in this standard star
   removelines, wav, obsflux, obsflux_nl, lines=lines, vel=linev, /interpol      
  ;   removelines, wav, refflux, refflux_nl, lines=linesnt, vel=linevnt, /interpol  ; (tellurics already removed from ref)
   removelines, wav, refflux, refflux_nl, lines=lines, vel=linev, /interpol  ; now remove tellurics also for consistency?
                                                                                ; some standard spectra actually have tellurics in them

      ; response defined as: observed count rate (e/s) per f_lambda
   response =  obsflux_nl / refflux_nl   ; instrument+atmosphere response except tellurics

      ; airmass correction, currently not in use.  Note also in telluric section.
      ; airtermarr = poly((wav-5000.)/2500.,airp)  ; magnitudes of extinction per airmass
      ; response *= 2.51^(-(refair-stdair[0])*airtermarr)
      ; wsignal = where(response gt 0.01*max(response), ct)
      ; if ct le 1 then begin
      ;      print, 'ERROR: Something wrong with flux calibration...'
      ;      continue
      ; endif

      ; here would go the dichroic corrections, etc., to smooth out some of the complex wiggles before fitting.

   bad = where(response lt median(response)*1e-4, ct, complement=good) ; just in case any negatives snuck through
   if ct gt 0 then response[bad] = min(response[good])
   par = [0,0,0]
   n = n_elements(wav)
   c = median(wav)
   rightbuf = 3
   if clip(sxpar(h,'INSTRUME')) eq 'LRIS' then rightbuf=25
   if clip(sxpar(h,'INSTRUME')) eq 'LRISBLUE' and clip(sxpar(h,'DICHNAME')) eq '500' and clip(sxpar(h,'GRISNAME')) eq '600/4000' then rightbuf=280 ; total hack

   r = median(response[5:n_elements(wav)-rightbuf])

      ; Old polynomial method
      ; fit a polynomial function to the response to smooth out noise
      ;maxo = 36
      ;for o = 3, 36 do begin  ; previously 18, then 36
      ;  par = polyfitg((wav[5:n-rightbuf]-c)/1e3, median(alog((response[5:n-rightbuf]/r)),3), o, guess=[par, 0], chi2=chi2) 
      ;         ; in theory I shouldn't be using the interpolated regions for this however bad things happen with the fit if I don't
      ;  ;print, o, ':', par
      ;  ;print, o, chi2
      ;endfor
      ;modelresponse = r*exp(poly((wav-c)/1e3, par))
      ; Savitsky-Golay method

   ; median filter
   fresponse = median(response,3)
   fresponse[0] = fresponse[1]
   fresponse[n_elements(response)-1] = fresponse[n_elements(response)-2]
     ; median filtering (r=3) ignores the edge pixels.  
     ; edge_trunc magnifies the impact of outliers at the edge pixel, which is a bad pixel on the red.
     ; so for the edge pixel, use the median of the 3 edge pixels (same value as median for 2nd-to-edge)

   ; Savgol filter
   specbin = (fix(strsplit(sxpar(h,'BINNING'),',',/extract))) [1]
   savgolfilt = savgol(100/specbin,100/specbin,0,4)
   rtrans = median(alog((fresponse/r)),3, /even) ; log-transform
   rtrans[n-rightbuf:n-1] = rtrans[n-rightbuf-1]
   modelresponse1 = r*exp(convol(rtrans, savgolfilt, /edge_trunc))

   respratio = fresponse/modelresponse1
   modelratio = convol(respratio, savgolfilt, /edge_trunc)
   modelresponse2 = modelratio*modelresponse1
   ;print, median(modelresponse1)
   ;print, median(modelresponse2)
   ;plot, wav, response, xrange=[9000,10400], psym=10
   ;oplot, wav, fresponse, psym=10, col=rgb(0,200,0)
   ;oplot, wav, modelresponse1, col=rgb(255,0,0)
   ;oplot, wav, modelresponse2, col=rgb(255,255,0)
   ;stop
   modelresponse = (modelresponse2 > 0) ; use new method

      ; telluric correction
   removelines, wav, obs.flux, obsflux_nlt, lines=linesnt, vel=linevnt, /interpol  ; "t" version keeps tellurics
   if keyword_set(notelluric) then begin
      telluricresponse=1 
   endif else begin
      telluricresponse = obsflux_nlt / obsflux_nl  ; response due solely to tellurics
      ; telluricresponse = 1 < (telluricresponse)^(refair/stdair[0]) ; airmass correction
      ; denoisify
      savgolfilt2 = savgol(2,2,0,3)
      telluricresponse = convol(telluricresponse, savgolfilt2, /edge_trunc) < 1.
   endelse

   sxaddpar, h, 'RESPMAX', max(modelresponse), ' Maximum of response'
   sxaddpar, h, 'RESP75', percentile75(modelresponse), ' 75th percentile of response'
   sxaddpar, h, 'RESPMED', max(modelresponse), ' Median of response'

   openw, 1, outfile ; out response file

   for i = 0, n_elements(h)-1 do begin ; copy the header
     printf, 1, '# '+h[i]
     if i ge n_elements(h)-1 then lastkey = (strsplit(h[i],/extract)) [0]
   endfor
   if n_elements(h) gt 0 and lastkey ne 'END' then printf, 1, '# END'

;      printf, 1, '# AIRMASS = '+clip(imh.airmass)  ; For reference only.
;      printf, 1, '# TARGNAME= '+clip(imh.target)   ; All the actual data comes from the original .spec file.
;      printf, 1, '# OBJECT  = '+clip(imh.object)   ; (Could just copy the header...)
;      printf, 1, '# GRNAME  = '+clip(imh.grname)
;      if camchar eq 'b' then printf, 1, '# GRISNAME= '+clip(imh.grname)
;      if camchar eq 'r' then printf, 1, '# GRANAME = '+clip(imh.grname)
;      printf, 1, '# BINNING = '+repstr(clip(imh.bin),'x',',')
;      printf, 1, '# SPECBIN = '+clip(imh.specbin)
;      printf, 1, '# INSTRUME= '+clip(imh.instrument)
;      printf, 1, '# SLITNAME= '+clip(imh.slit)
;      printf, 1, '# SLITPA  = '+clip(imh.slitpa)
;      printf, 1, '# DICHNAME= '+clip(imh.dichroic)
;      printf, 1, '# FILTER  = '+clip(imh.filter)
;      printf, 1, '# EXPTIME = '+clip(imh.exptime)
;      printf, 1, '# RA      = '+clip(imh.rastr)
;      printf, 1, '# DEC     = '+clip(imh.decstr)
;      printf, 1, '# FLATFILE= '+clip(sxpar(h,'FLATFILE'))

   printf, 1, '# wav      obsflux       refflux        response       telluricresponse  flag'
   for i = 0, n_elements(wav)-1 do $
      printf, 1, clip(wav[i],10)+' '+clip(obsflux[i],13)+' '+clip(refflux[i],14)+' '+clip(modelresponse[i],14)+$
               ' '+clip(telluricresponse[i])+'    '+fpr(flag[i],2.3)
   close, 1   

   !p.font = 0
   for pl=0,1 do begin ;0=screen, 1=eps
     if lpar.displayoff and pl eq 0 then pl = 1

     xsize = 600

     if pl eq 0 then begin
        errorstatus = 0
        catch, errorstatus
        if errorstatus ne 0 then begin
           logprint, ' X-windows display not working: turning off display'
           lpar.displayoff = 1
           catch, /cancel
           continue
        endif else begin
           window, 4, xsize=xsize, ysize=800, title='LPipe: standard star response function'
        endelse
        catch, /cancel
           ;           -                  511          124890        391482905
        cols = [rgb(255,255,255), rgb(255,1,0), rgb(218,231,1), rgb(25,142,85), rgb(255,255,0), rgb(0,128,0),rgb(255,0,255)]
     endif else begin
        plotfile = repstr(outfile,'.response','.response.eps')
        psopen, plotfile, xsize=(xsize/100), ysize=8, /inches, /encaps
        device, /color
        !p.font=0
        device, /times, font_index=17  ; the stupid new default plot axis format for scientific notation
                                          ; sets it to times even if not requested!!! can't use other fonts :(
        colors = transpose([[0,0,0],$
                   [255,0,0],$ 
                   [100,100,0],$ 
                   [0,200,0],$ 
                   [160,160,0],$  
                   [10,255,10],$
                   [255,0,255]])
        tvlct, colors
        cols = indgen(7)
     endelse
     !p.multi = [2,1,2]

     xl = 0.1
     xr = 0.98
     resp = response
     telresp = telluricresponse
     name = stdname
     file = filename
 
     xrange = [min(wav), max(wav)]

     !p.position = [xl, 0.79, xr, 0.975]
     if pl eq 1 then title = '!17' else title=''
     plot, [0], [0], xrange=xrange, yrange=[0, 1.02*max([refflux_nl])], /xstyle, /ystyle, title=title
     if n_elements(origwav) eq 0 then delvarx, origwav
     oplot, refwav_in, refflux_in, color=cols[1]
     oplot, wav, refflux_nl
      xyouts, xr-(xr-xl)*0.35, !p.position[3]-0.02, /norm, title+name+' (database)'
 
     camchar = strmid(clip(sxpar(h,'CAMERA')),0,1)
     outstr = strupcase(camchar)+' '+clip(sxpar(h,'GRNAME'))+' D'+clip(sxpar(h,'DICHNAME'))
     if strlowcase(clip(sxpar(h,'FILTER'))) ne 'clear' then outstr += ' '+clip(sxpar(h,'FILTER'))
     if camchar eq 'r' then outstr += ' w='+clip(round(sxpar(h,'WAVELEN')))
     xyouts, xl, 0.98, /norm, outstr

     !p.position = [xl, 0.5, xr, 0.77]
     plot, [0], [0], yrange=[0, max([obs.flux,obsflux_nl])], xrange=xrange, /xstyle, title=title
     ; could add time-normalized right axis
     oplot, obs.wav, obs.flux, color=cols[1] ; noninterpolated
     oplot, wav, obsflux_nl                  ; interpolated (no difference...)
     xyouts, xr-(xr-xl)*0.65, !p.position[3]-0.02, /norm, title+name+' ('+removepath(file)+')'
     !p.multi = [2,1,2]

     !p.position = [xl, 0.20, xr, 0.48]
     plot, [0], [0], xrange=xrange, yrange=[0, max(response)*1.05], /xsty, /ysty, title=title;, /ylog
     ; if nmatchstd gt 1 then oplot, wav, resp*telresp, color=cols[3]
     oplot, wav, response
     oplot, wav, modelresponse*telluricresponse, color=cols[2]
     ;oplot, wav, modelresponse2, color=cols[6], linestyle=2
     xyouts, xl+(xr-xl)*0.05, !p.position[3]-0.02, /norm, title+'Solved response', color=cols[4]
      xyouts, xl+(xr-xl)*0.05, !p.position[3]-0.035, /norm, title+'Measured response', charsize=0.85

     !p.position = [xl, 0.04, xr, 0.18]
     plot, [0], [0], xrange=xrange, yrange=[min(response)*0.95, max(response)*1.05], /xsty, /ysty, title=title, /ylog
     oplot, wav, response
     oplot, wav, modelresponse*telluricresponse, color=cols[2]
     ;oplot, wav, modelresponse2, color=cols[6], linestyle=2
     xyouts, xl+(xr-xl)*0.05, !p.position[3]-0.02, /norm, title+'Log plot'

     if pl eq 1 then psclose
     if pl eq 1 then retitleplot, plotfile, removepath(filename)+' response'
     updated = 1

   endfor ; plot loop
   !p.multi = 0
   !p.position = 0

end

