;+
; NAME:
;       lrissplit
;
; PURPOSE:
;       Split the left and right chip from a merged file into individual files.
;
; CALLING SEQUENCE:
;       lrissplit, file, [/rightonly, /leftonly, renameas=, nodistort=]
;
; EXPLANATION:
;       Currently early pipeline files are saved with the right and left CCDs
;       merged together into a simple FITS file.  This breaks them apart again
;       into two files (or more typically just one corresponding to the chip of
;       interest).  Updates the astrometry and some other header information.
;
; INPUTS:
;       file - Name of the file to split apart.
;
; OPTIONAL INPUT KEYWORDS:
;       /rightonly - Only save the right chip.
;       /leftonly - Only save the left chip
;       renameas - Call the output file(s) something different from the input.
;       nodistort - Don't write distortion terms. (Currently it is not clear if
;         the distortion terms even work.)
;
; OUTPUTS:
;       One or two split-off files are written to disk.
;
;-
;------------------------------------------------------------------------------

pro lrissplit, file, rightonly=rightonly, leftonly=leftonly, renameas=renameas, nodistort=nodistort

   arr = mrdfits(file,0,h, /silent)

   ; get_date, now
   hist = 'Processed by lrissplit '+tstamp(/ut)+' UT'
   sxaddhist, hist, h

   if n_elements(renameas) eq 0 then renameas = file
   dotpos = strpos(renameas,'.',/reverse_search)
   fileroot = strmid(renameas,0,dotpos)
   rfile = fileroot+'r.fits'
   lfile = fileroot+'l.fits'

   outstr =  removepath(file) + ' -> '
   if  keyword_set(leftonly) eq 0 then outstr += removepath(rfile)
   if  keyword_set(rightonly) eq 0 and keyword_set(leftonly) eq 0 then outstr += ','
   if  keyword_set(rightonly) eq 0  then outstr += removepath(lfile)
   logprint, ' '+outstr

   slit = clip(sxpar(h,'SLITNAME'))
   if strpos(slit,'_') ge 0 then smode = 'sp' ; longslit
   if strlowcase(slit) eq 'direct' then smode = 'im'
   instrume = clip(sxpar(h,'INSTRUME'))
   if instrume eq 'LRIS'     then camera = 'r'
   if instrume eq 'LRISBLUE' then camera = 'b'
   if camera eq 'b' and clip(sxpar(h,'GRISNAME'))  eq 'clear'  then lmode = 'im'
   if camera eq 'r' and clip(sxpar(h,'GRANAME'))  eq 'mirror' then  lmode = 'im'
   if camera eq 'b' and strpos(sxpar(h,'GRISNAME'),'/') gt 0   then lmode = 'sp'
   if camera eq 'r' and strpos(sxpar(h,'GRANAME'),'/') gt 0    then lmode = 'sp'
   if n_elements(lmode) gt 0 then mode = lmode
   if n_elements(mode) eq 0 and n_elements(smode) gt 0 then mode = smode

   if n_elements(mode) eq 0 then begin
      logprint, '  Cannot identify mode.  Assuming imaging.'
      mode = 'im'
   endif

   inst = clip(sxpar(h,'INSTRUME'))
   if inst eq 'LRIS' then camchar = 'r'
   if inst eq 'LRISBLUE' then camchar = 'b'

   lrisversion = 0
   mjd = float(sxpar(h,'MJD-OBS'))
   if mjd gt 0 and mjd lt 49000 then lrisversion = 0
   if mjd gt 49000 and mjd lt 54952 then lrisversion = 1
   if mjd ge 54952 and mjd lt 55562 then lrisversion = 2
   if mjd ge 55562 then lrisversion = 3
   if lrisversion le 0 then begin
     logprint, '  Cannot identify LRIS version - assuming 3'
     lrisversion = 3
   endif

   lh = h
   rh = h
   s = size(arr)
   ;if mode eq 'im' then nx = s[1]
   ;if mode eq 'sp' then 
   ny = s[2]

   ybin = sxpar(h,'YBIN')

   ltv1 = sxpar(h,'LTV1')
   ltv2 = sxpar(h,'LTV2')
   ;if mode eq 'im' then ltvx = ltv1
   ;if mode eq 'sp' then 
   ltvy = fix(ltv2 - 2048/sxpar(h,'LTM2_2')) ; -y origin in detector coordinates (X)

   ;if mode eq 'im' then acrop = sxpar(h,'CROPX0')
   ;if mode eq 'sp' then acrop = sxpar(h,'CROPY0')
   acrop = sxpar(h,'CROPY0') 

   ampl1 = fix(strsplit(sxpar(h,'AMPL1',count=ctl1),',',/extract))-acrop
   ampl2 = fix(strsplit(sxpar(h,'AMPL2',count=ctl2),',',/extract))-acrop
   if ctl2 gt 0 then ampl = ampl2
   if ctl2 eq 0 and ctl1 gt 0 then ampl = ampl1

   ampr1 = fix(strsplit(sxpar(h,'AMPR1',count=ctr1),',',/extract))-acrop
   ampr2 = fix(strsplit(sxpar(h,'AMPR2',count=ctr2),',',/extract))-acrop
   if ctr1 gt 0 then ampr = ampr1
   if ctr1 eq 0 and ctr2 gt 0 then ampr = ampr2

   ;ampr = sxpar(h,'AMPR1',count=ctk)
   ;if ctk eq 0 then ampr = sxpar(h,'AMPR2',count=ctk)
   ;ampr -= acrop
   ;;;if n_elements(ampl) eq 1 then ampl = [ampl, 0] ; ?????
   ;;;if n_elements(ampr) lt 2 then ampr = [ampr, 0] ; ???

   cbuf = [0,0] ; number of pixels to crop around the chip gap (L, R)
   if camchar eq 'b' then imy = [360, 3790] ; (there is no gap cropping for LRIS-B)
   if camchar eq 'r' and lrisversion eq 2 then imy = [430, 3778]     ; this is NOT CONFIRMED
   if camchar eq 'r' and lrisversion eq 2 then cbuf = [44, 12]/ybin  ; this is NOT CONFIRMED
   if camchar eq 'r' and lrisversion eq 3 then imy = [430, 3778]
   if camchar eq 'r' and lrisversion eq 3 then cbuf = [12, 12]/ybin  ; buffer around gap (L chip, R chip)
      ; would be ideal not to crop, but added tape(?) noise near the gap creates problems
      ; for flats and then sky subtraction.  Some sort of special bias subtraction step might do it.
   lcrop = [imy[0]+ltvy > 0 , ampl[1]-cbuf[0]]
   rcrop = [ampr[0]+cbuf[1] > 0, imy[1] < ny-1 ]   ; the >0 shouldn't be necessary
   ;print, acrop, ampr1[0], ' ', sxpar(h,'AMPR1'), cbuf[0], rcrop[0], ' ', sxpar(h,'DETSEC')

   if mode eq 'sp' then begin

     ; In general the cropping at the 'top' is automatic via flat-fielding.
     ; Cropping a little more will avoid fake sources appearing there on occasion

     ; a little bit of extra cropping at the top for spectra
     ;if camchar eq 'r' and lrisversion ge 2 then rcrop[1] = (rcrop[1] < ampr1[0]+(564/ybin))
     ;if (camchar eq 'b' or lrisversion ge 2) then rcrop[1] = (rcrop[1] < ampr1[0]+(564/ybin))
     rcrop[1] -= 10/ybin
     if camchar eq 'b' then rcrop[1] = rcrop[1] < ((ampr[0]+584/ybin)-1)
     if camchar eq 'r' and lrisversion ge 2 then rcrop[1] = rcrop[1] < ((ampr[0]+568/ybin)-1)
     rcrop[1] = rcrop[1] < ny-1
         ; the slit trace is useful in principle, but it's often already cropped partially
         ; and dithering can put edge trace artifacts in bad places

     if keyword_set(rightonly) eq 0 then begin
        if lcrop[0] ge 0 and lcrop[1] ge 0 and lcrop[1] gt lcrop[0]  then begin
           arrl = arr[*,lcrop[0]:lcrop[1]]
        endif else begin
           logprint, '   No useable data for left chip?'
           if keyword_set(leftonly) then return
        endelse
     endif
     if keyword_set(leftonly) eq 0 then begin
        if rcrop[0] ge 0 and rcrop[1] ge 0 and rcrop[1] gt rcrop[0] then begin
           arrr = arr[*,rcrop[0]:rcrop[1]] 
        endif else begin
           logprint, '   No useable data for right chip?'
           if keyword_set(rightonly) then return
        endelse
     endif

     crpix2 = sxpar(h,'CRPIX2')
     sxaddpar, lh, 'LTV2', ltv2-lcrop[0]
     sxaddpar, rh, 'LTV2', ltv2-rcrop[0]
     sxaddpar, lh, 'CRPIX2', crpix2-lcrop[0]
     sxaddpar, rh, 'CRPIX2', crpix2-rcrop[0]

   endif

   if mode eq 'im' then begin
     if keyword_set(rightonly) eq 0 then arrl = arr[*,lcrop[0]:lcrop[1]]
     if keyword_set(leftonly) eq 0 then  arrr = arr[*,rcrop[0]:rcrop[1]]

     crpix2 = sxpar(h,'CRPIX2')
     sxaddpar, lh, 'CRPIX2', crpix2-lcrop[0]
     sxaddpar, rh, 'CRPIX2', crpix2-rcrop[0]
     sxaddpar, lh, 'LTV2', ltv2-lcrop[0]
     sxaddpar, rh, 'LTV2', ltv2-rcrop[0]

     ; re-define the center of the left chips to actually be on the chip
     if camchar eq 'r' then begin
        leftshift = (-1024/1-30./0.135)
        xyad, rh, sxpar(rh,'CRPIX1')+leftshift, sxpar(rh,'CRPIX2'), la, ld
        sxaddpar, lh, 'CRPIX2', sxpar(lh,'CRPIX2')-1024/1 ; assuming no binning
        sxaddpar, lh, 'CRVAL1', la
        sxaddpar, lh, 'CRVAL2', ld
     endif
     if camchar eq 'b' then begin
        leftshift = (-1024/1-14./0.135)
        xyad, rh, sxpar(rh,'CRPIX1')+leftshift, sxpar(rh,'CRPIX2'), la, ld
        sxaddpar, lh, 'CRPIX2', sxpar(lh,'CRPIX2')-1024/1 ; assuming no binning
        sxaddpar, lh, 'CRVAL1', la
        sxaddpar, lh, 'CRVAL2', ld
     endif

     ; add skewness and rectangularity - unfortunately, doesn't work
     ; all for un-transposed data and needs to be updated
     addskew = 0
     if addskew then begin
      thetarad = atan2(-sxpar(h,'CD2_1'), sxpar(h,'CD1_1'))
      cost = cos(thetarad)
      sint = sin(thetarad)
      if camchar eq 'r' and lrisversion eq 3 then begin
         p = 0.1349
         r = -0.9976
         s = 0.
         sxaddpar, rh, 'CD1_1',  1./3600*( p*r *  (cost + s*sint))
         sxaddpar, rh, 'CD1_2',  1./3600*((p/r)*(s*cost + sint))
         sxaddpar, rh, 'CD2_1',  1./3600*( p*r * (-sint + s*cost))
         sxaddpar, rh, 'CD2_2',  1./3600*((p/r)*(-s*sint + cost))
         p = 0.1349
         r = -0.9976 
         s = 0.
         sxaddpar, lh, 'CD1_1',  1./3600*( p*r *  (cost + s*sint))
         sxaddpar, lh, 'CD1_2',  1./3600*((p/r)*(s*cost + sint))
         sxaddpar, lh, 'CD2_1',  1./3600*( p*r * (-sint + s*cost))
         sxaddpar, lh, 'CD2_2',  1./3600*((p/r)*(-s*sint + cost))
      endif
      if camchar eq 'b' then begin
         p = 0.1355
         r = -0.998
         s = 0.
         sxaddpar, rh, 'CD1_1',  1./3600*( p*r *  (cost + s*sint))
         sxaddpar, rh, 'CD1_2',  1./3600*((p/r)*(s*cost + sint))
         sxaddpar, rh, 'CD2_1',  1./3600*( p*r * (-sint + s*cost))
         sxaddpar, rh, 'CD2_2',  1./3600*((p/r)*(-s*sint + cost))
         p = 0.1355
         r = -0.998
         s = 0.
         sxaddpar, lh, 'CD1_1',  1./3600*( p*r *  (cost + s*sint))
         sxaddpar, lh, 'CD1_2',  1./3600*((p/r)*(s*cost + sint))
         sxaddpar, lh, 'CD2_1',  1./3600*( p*r * (-sint + s*cost))
         sxaddpar, lh, 'CD2_2',  1./3600*((p/r)*(-s*sint + cost))
      endif
 
      if keyword_set(nodistort) eq 0 then begin
        ; add distortion
        sipkey = ['A_2_0','A_3_0','A_1_1','A_2_1','A_0_2','A_1_2','A_0_3',$
                   'B_2_0','B_3_0','B_1_1','B_2_1','B_0_2','B_1_2','B_0_3']
 
        if camchar eq 'r' and lrisversion eq 3 then begin
          ;sip = [4.5429166d-06, -4.3431439d-09, -1.8810083d-06, -2.0210270d-10, -1.0712478d-06, $
          ;       2.3252198d-11, 8.0474672d-10, 1.9176066d-06, -1.0084854d-09, -2.0182632d-06, $
          ;      -4.0679096d-10, -4.3712818d-07, -7.0081205d-10, 6.2149406d-10]
          sip = [4.8675553e-06, -5.4269167e-09, -7.2654544e-06, 4.1354352e-09, -4.3143095e-07, $
                 -1.8636630e-09, 3.3488706e-10, -2.3855855e-07, 1.3303440e-09, -1.0792428e-06, $
                 -1.7638335e-09, -2.4254375e-06, 4.3872773e-10, -1.0259050e-09]
          for k = 0, n_elements(sipkey)-1 do sxaddpar, rh, sipkey[k], sip[k]
          ;sip = [-3.5125614d-06, -1.8151190d-09, 3.5035019d-06, 2.3072218d-09, 3.0925328d-06, $
          ;       4.3707630e-10,  1.5245489e-10, -1.1397414d-07,  -1.3418176e-10, -7.0526093d-06, $
          ;       -4.1247492d-09, -7.1054211d-07, -2.0256577e-10, -1.6381757d-09]
          sip = [-2.2518293e-06, -2.8137610e-09, -4.2128341e-06, -2.6334235e-09, 5.0307124e-07, $
                 -1.1176543e-09, -3.2601032e-11, 1.9110552e-06, 3.2146727e-09, 3.5272010e-06, $
                  1.8675122e-11, -4.9258383e-07, -1.2088439e-09, 7.5591110e-10]
          for k = 0, n_elements(sipkey)-1 do sxaddpar, lh, sipkey[k], sip[k]
 
        endif
        if camchar eq 'b' then begin
           ;sip = [5.25475348443d-06,-3.97572591283d-09,-3.06061228064d-06,6.60619890986d-10, $
           ;       -7.35039639949d-07,-7.12587298861d-10,1.44803583697d-09,8.59834826224d-07, $
           ;       -5.62349746593d-11,2.61430913855d-07,-1.40095752059d-09,-4.69311802714d-07, $
           ;       -1.60271911589d-09,3.92593176884d-10]
           sip = [2.1951632e-06, -2.9854666e-09, -4.3662187e-06, 2.1033250e-09, -5.7961296e-07, $
                  1.5722022e-10, 4.8813969e-10, 5.2802767e-07, 1.2075610e-10, 6.9223883e-07, $
                  -2.2554050e-09, -7.1971166e-07, -8.4430150e-10, 1.8409954e-10]
           for k = 0, n_elements(sipkey)-1 do sxaddpar, rh, sipkey[k], sip[k]
           ;sip = [-3.4681918d-06, -2.1049725d-09, 3.5402198d-06,  3.4362793d-09, 1.2455002d-06, $
           ;       -3.2508281e-10, -7.8548468e-10, 6.4032455d-07, 1.6185461e-10,  -4.8094366d-06, $
           ;       -3.4307617d-09, -6.4172618d-07,  9.2343251e-10, -6.9646133e-10]
           sip = [-7.2305186e-07, -1.0972867e-09, -7.0809385e-06, -3.4131315e-09, 7.9686424e-08, $
                  -9.6289282e-10, -1.4009917e-09, 2.9964906e-06, 2.3358901e-09, 2.7125645e-06, $
                  -1.2339890e-09, -9.9749681e-07, 9.3207402e-10, 1.1163994e-09]
           for k = 0, n_elements(sipkey)-1 do sxaddpar, lh, sipkey[k], sip[k]
        endif
        sxaddpar, rh, 'CTYPE1', 'RA---TAN-SIP'
        sxaddpar, rh, 'CTYPE2', 'DEC--TAN-SIP'
        sxaddpar, rh, 'A_ORDER', 3
        sxaddpar, rh, 'B_ORDER', 3
        sxaddpar, lh, 'CTYPE1', 'RA---TAN-SIP'
        sxaddpar, lh, 'CTYPE2', 'DEC--TAN-SIP'
        sxaddpar, lh, 'A_ORDER', 3
        sxaddpar, lh, 'B_ORDER', 3
      endif    
     endif
     ;  distort doesn't work without skew
   endif

   sxdelpar, h, ['AMP1','AMP2','AMP3','AMP4','AMPL1','AMPL2','AMPR1','AMPR2']
   sxaddpar, lh, 'CHIP', 'left'
   sxaddpar, rh, 'CHIP', 'right'

   if keyword_set(rightonly) eq 0 then mwrfits, arrl, lfile, lh, /silent, /create
   if keyword_set(leftonly) eq 0 then  mwrfits, arrr, rfile, rh, /silent, /create

   if mode eq 'im' and keyword_set(nodistort) eq 0 then begin
      if keyword_set(rightonly) eq 0 then sip2pv, lfile, /modify
      if keyword_set(leftonly) eq 0  then sip2pv, rfile, /modify
   endif

end

