;+
; NAME:
;       lrissumarc
;
; PURPOSE:
;       Combine (sum) arcs together, e.g. to merge separate Ne and Ar arcs.
;
; CALLING SEQUENCE:
;       lrissumarc, file1, file2, [outfile]
;
; EXPLANATION:
;       Tool for a user to explicitly create an arc by summing existing ones,
;       either to boost S/N or (more likely) to create an arc with more lines
;       if lamps were exposed separately.
;
; INPUTS:
;       file1 - First arc to sum
;       file2 - Second arc to sum
;       outfile - Output file
;
; OUTPUTS:
;       The summed arc is written to disk.
; 
; NOTES:
;       NOT currently used by the pipeline, which does its own summation, and
;       not checked or tested recently.
;       (Unclear which is better, but note that this only allows summing two.)
;
;-
;------------------------------------------------------------------------------

pro lrissumarc, file1, file2, outfile

  ; add arcs together in case different lamps were observed individually.
  
  ; for now this only works with simple FITS files

  if file_test(file1) eq 0 or file_test(file2) eq 0 then begin
    print, 'Input file(s) not found'
    return
  endif

  im1 = readfits(file1, h1)
  im2 = readfits(file2, h2)

  if (size(im1)) [1] ne (size(im2)) [1] or $
     (size(im1)) [2] ne (size(im2)) [2] or $
     sxpar(h1,'PANE') ne sxpar(h2,'PANE') then begin
       print, 'Image binning/windowing does not match.' 
       return
  endif
  ; should also check that setup is the same since otherwise this will produce a garbage result

  im = im1 + im2
  h = h1  

  lampstr1 = clip(sxpar(h1,'LAMPS'))
  lampstr2 = clip(sxpar(h2,'LAMPS'))
  lampstr = lampstr1

  if strpos(lampstr1,'1') ge 0 and strpos(lampstr2,'1') ge 0 then begin
     for c = 0, strlen(lampstr1) do begin
        if strmid(lampstr2,c,1) eq '1' then strput, lampstr, '1', c
     endfor
     sxaddpar, h, 'LAMPS', lampstr
  endif

  lampkeys = ['MERCURY','NEON','ARGON','CADMIUM','ZINC','HALOGEN','KRYPTON','XENON','FEARGON','DEUTERI']
  for i = 0, n_elements(lampkeys)-1 do begin
     if clip(sxpar(h2, lampkeys[i])) eq 'on' then sxaddpar, h, lampkeys[i], 'on'
  endfor

  print, 'HgNeArCdZnHaKrXeFeDe'
  for j = 0, 2 do begin
     if j eq 0 then hh = h1
     if j eq 1 then hh = h2
     if j eq 2 then hh = h
     print, sxpar(hh,'LAMPS')
  endfor

  writefits, outfile, im, h

end

