;+
; NAME:
;       lrisvalidate
;
; PURPOSE:
;       LPipe spectral inspection and validation tool
;
; CALLING SEQUENCE:
;       lrisvalidate, [filestr, demo=demo, delete=delete, move=move, copy=copy,
;                     mark=mark, outdir=outdir, noview=noview, savepng=savepng,
;                     nooverwrite=nooverwrite, all=all]
;
; EXPLANATION:
;       A graphical tool for helping to ensure that a spectrum was processed
;       correctly and checking the reality of spectral features.  Will produce
;       (on screen) a plot of the spectrum (top), sky spectrum (upper middle),
;       response curves (lower middle), and 2D spectra (bottom), as well as
;       the wavelength-averaged spatial profiles for blue and red (left).
;       Cosmic ray affected areas are marked and both pre-combined and smoothed
;       post-combined spectra are shown.
;       Also provides tools for saving validated spectra and/or deleting
;       problematic spectra.
;
; OPTIONAL INPUTS:
;       filestr - A wildcard string for which spectra to validate.  If blank,
;                 all spectra in the current directory.
;
; OPTIONAL INPUT KEYWORDS:
;       /demo - Turns off file output (for testing without a working license 
;               or in a write-protected directory).
;       /delete - Deletes any failed spectra when the tool is closed.
;       /move - Moves validated spectra to a 'validated/' subdirectory when
;               the tool is closed.
;       /copy - Copies validated spectra to a 'validated/' subdirectory when
;               the tool is closed.
;       /mark - Edits spectra to add a header note about their validation
;               status when the tool is closed.
;       outdir - Call the validation subdirectory something other than
;                'validated/' (the default if left unspecified)
;       /noview - Do not view any spectra - just process the saved validation
;                 results.  (Use this if you did validation marking already
;                 but did not specify /move, /copy, or /mark.)
;       /savepng - Save .png screenshots of the validation spectra to disk
;                 as they appear (in the validation directory).
;       /nooverwrite - If using /move or /copy, do not overwrite any existing
;                 files.
;       /all - Move failed spectra to the validation subdirectory, too 
;                 (if /move or /copy is set).
;
; OUTPUTS:
;       The validation status for each file is continuously saved to 
;       validation.txt.  (This is specific to the current working directory.)
;       Spectra are also moved, copied, or deleted if specified by the 
;       appropriate option.
; 
; EXAMPLES:
;       lrisvalidate, '*.spec'  ; same as the default option
;       lrisvalidate, /move, /all, /noview ; process results
;
;
; NOTES:
;       Most users will simply use this as a viewer and do not need to worry 
;       about the input keywords.  Just type lrisvalidate after processing
;       and use the next(/back) buttons to examine all spectra.
;       For processing large numbers of files, the validation buttons allow
;       the user to record the status of each file: 'pass' to pass without
;       reservations; 'flag' to pass but mark a possible concern; 'fail' to
;       indicate a significant problem with the reductions.  If 'skip' is
;       turned on, back/next will skip over any already-evaluated spectra.
;       After the user selects 'quit', the appropriate move/copy/delete
;       action is carried out.
;
;-
;------------------------------------------------------------------------------


pro drawvbox, coords, color=color, fillcolor=fillcolor, norm=norm, data=data, label=label, charsize=charsize, textcolor=textcolor, orient=orient, savepng=savepng, textdx=textdx, textdy=textdy
  ; draws a rectangle in the active window, optionally with a text caption.

  ; coords are: x0 y0 x1 y1, like p.position
  x0 = coords[0]
  y0 = coords[1]
  x1 = coords[2]
  y1 = coords[3]
  
  if n_elements(fillcolor) gt 0 then $
  polyfill, [x0, x1, x1, x0, x0], [y0, y0, y1, y1, y0], color=fillcolor, norm=norm, data=data

  if n_elements(color) gt 0 then $
  plots, [x0, x1, x1, x0, x0], [y0, y0, y1, y1, y0], color=color, norm=norm, data=data
  
  if !p.font eq 0 then xb = 0.003
  if !p.font eq 0 then xm = 0.1
  if !p.font eq -1 then if n_elements(orient) gt 0 then xb = 0.017 else xb =0.00
  if !p.font eq -1 then xm = 0.03
  if n_elements(textdx) eq 0 then dx = (x1-x0)*xm+xb+0.001 else dx = textdx
  if n_elements(textdy) eq 0 then dy = 0.001+((y1-y0-0.01) > 0)*0.3*((n_elements(orient) eq 0)*0.7 + 0.3) else dy = textdy
  if n_elements(label) gt 0 then xyouts, x0+dx, y0+dy, label, charsize=charsize, norm=norm, data=data, color=textcolor, orient=orient
end

pro drawvbutton, coords, state, astate, oncolors=oncolors, offcolors=offcolors, darkcolors=darkcolors, norm=norm, data=data, label=label, charsize=charsize, orient=orient, textoff=textoff
  ; draws a button.  Basically a wrapper around drawbox that allows all possible colors to be passed along with two binary flags (stats and astate)
  ; indicating how to color:  "on" (state=1), "off" (state=0, astate=1), "inactive"/dark (astate=0)
  ; astate is optional.
  if n_elements(textoff) ge 1 then textdx = textoff[0]
  if n_elements(textoff) ge 2 then textdy = textoff[1]

  if state then colors = oncolors else colors = offcolors
  if n_elements(astate) gt 0 then if astate eq 0 then colors = darkcolors
  drawvbox, coords, color=colors[0], fillcolor=colors[1], textcolor=colors[2], norm=norm, data=data, label=label, charsize=charsize, orient=orient, textdx=textdx, textdy=textdy
end


pro marktelluric, yrange
   for t = 0, 3 do begin
      if t eq 0 then tel = [7590,7660+12]
      if t eq 1 then tel = [6860,6890+12]
      if t eq 2 then tel = [9280,9660]
      if t eq 3 then tel = [8950,9050]
      polyfill, tel[[0,0,1,1,0]], yrange[[0,1,1,0,0]], color=rgb(50,50,50)
   endfor
   axis, xaxis=1, xtickformat='(A1)', xticklen=1e-7
end

pro markskylines, wavrange, yrange, dim=dim
   common skydat, idn, allreflinewav, allreflinestr

   if n_elements(idn) eq 0 then begin
      paths = strsplit(!path,':',/extract)
      ; get the LPipe (and LPipe refdata) path
      for p = -1, n_elements(paths)-1 do begin
         if p ge 0 then path = paths[p] $
                        else path = ''
         sfile = path+'/lrisvalidate.pro'
         n  = countfiles(sfile)
         if n gt 0 then lrisautopath = path
         if n gt 0 then break
      endfor
      lprefdatapath = lrisautopath+'/refdata'
      skylist = lprefdatapath + '/' + 'skylines.dat'
      ; telspec = lprefdatapath + '/' + 'telluricabs.dat'
      readcol, skylist, idn, allreflinewav, allreflinestr, format='i,f,f', /silent
   endif

   if n_elements(dim) eq 0 then dim = 1
   ; mimic the line selection in specwavcal
   wlr = where(allreflinestr gt 100. and allreflinewav gt wavrange[0] and allreflinewav lt wavrange[1])
   skylinewav = allreflinewav[wlr]
   skylinestr = allreflinestr[wlr]
   nline = 100
   if wavrange[1] lt 7000. then nline = 20
   if wavrange[1] lt 6000. then nline = 5
   brightskywav = findbrightest(skylinewav, skylinestr, (nline/5)>3)   
   moreskywav = findbrightest(skylinewav, skylinestr, nline)
   for l = 0, n_elements(moreskywav)-1 do begin
       oplot, moreskywav[l]*[1,1],yrange[[0,1]], color=rgb(96*dim^2,0,0)
   endfor       
   for l = 0, n_elements(brightskywav)-1 do begin
       oplot, brightskywav[l]*[1,1],yrange[[0,1]], color=rgb(96*dim,96*dim,0)
   endfor
end



pro validateplot, spin, h, bluespin, redspin, blue2d, red2d, bluecr, redcr, filename=filename, path=path
   ; spec1dfile, spec2dfile

   sp = spin
   bluesp = bluespin
   redsp = redspin
   sp.flux *= 1e17
   sp.sky *= 1e17
 ;  sp.resp /= 1e17
   bluescale = sxpar(h,'RESCALEB', count=ct)
   if ct eq 0 then bluescale = 1.
   redscale = sxpar(h,'RESCALER', count=ct)
   if ct eq 0 then redscale = 1.
   if n_elements(bluesp) gt 1 then bluesp.flux *= 1e17 * bluescale
   if n_elements(redsp) gt 1 then  redsp.flux *= 1e17 * redscale
   if n_elements(bluesp) gt 1 then bluesp.sky *= 1e17 * bluescale
   if n_elements(redsp) gt 1 then  redsp.sky *= 1e17 * redscale
   if n_elements(bluesp) gt 1 then bluesp.resp /= bluescale ;1e17
   if n_elements(redsp) gt 1 then  redsp.resp /= redscale

   !p.multi = [0,2,6]
   !p.font = 0
   minwav = min(sp.wav)
   maxwav = max(sp.wav)
   wavrange = [minwav,maxwav]

   isb = 0
   isr = 0
   if clip(sxpar(h,'INSTRUME')) eq 'LRIS' then isr = 1
   if sxpar(h,'INSTRUME') eq 'LRISBLUE' then isb = 1
   if sxpar(h,'INSTRUME') eq 'LRIS+LRISBLUE' then isr = 1
   if sxpar(h,'INSTRUME') eq 'LRIS+LRISBLUE' then isb = 1

   atchwav = sxpar(h,'ATCHWAV', count=cta)
   ;if cta eq 0 or atchwav eq 0 then begin
     if clip(sxpar(h,'INSTRUME')) eq 'LRISBLUE' then atchwav = max(sp.wav)
     if clip(sxpar(h,'INSTRUME')) eq 'LRIS' then atchwav = min(sp.wav)
   ;endif
   blubin = sxpar(h,'BLUBIN',count=ctbb)
   redbin = sxpar(h,'REDBIN',count=ctbr)
   if ctbb+ctbr eq 0 then begin
     print, ' Not a connected file'
     return
   endif
   if isb then bybin = (fix(strsplit(blubin,',',/extract))) [0]
   if isr then rybin = (fix(strsplit(redbin,',',/extract))) [0]
   if isb then bxbin = (fix(strsplit(blubin,',',/extract))) [1]
   if isr then rxbin = (fix(strsplit(redbin,',',/extract))) [1]

   !p.position = [0.01,0.01,0.02,0.02]
   plot, [0], [0], color=rgb(0,0,0) ; fake plot
   di = rgb(128,128,128)
   br = rgb(255,255,200)
   btext = rgb(130,130,255)
   rtext = rgb(250,100,100)
   slitcol = rgb(128,128,128)
   if n_elements(filename) gt 0 then xyouts, 0.05, 0.98, /norm, repstr(filename,'.spec',''), charsize=1.1, color=rgb(200,200,200)
   xyouts, 0.05, 0.95, /norm, sxpar(h,'TARGNAME'), charsize=1.8
   xyouts, 0.05, 0.93, /norm, sxpar(h,'SLITNAME'), charsize=1.3, col=slitcol
   xyouts, 0.11, 0.93, /norm, 'D'+sxpar(h,'DICHNAME'), charsize=1.3, col=slitcol
   if isb then xyouts, 0.01, 0.92-0.01, /norm, sxpar(h,'GRISNAME'), charsize=1.4, col=btext
   if isr then xyouts, 0.09, 0.92-0.01, /norm, sxpar(h,'GRANAME')+' (w='+clip(fix(0.4+sxpar(h,'WAVELEN')))+')', charsize=1.4, col=rtext
   if isb then xyouts, 0.01, 0.92-0.03, /norm, repstr(sxpar(h,'BLUBIN'),',','x'), charsize=1.4, col=btext
   if isr then xyouts, 0.09, 0.92-0.03, /norm, repstr(sxpar(h,'REDBIN'),',','x'), charsize=1.4, col=rtext
   if isb then xyouts, 0.04, 0.92-0.03, /norm, 'resc:'+clipzero(clip(sxpar(h,'RESCALEB'),5)), charsize=1.4, col=btext
   if isr then xyouts, 0.12, 0.92-0.03, /norm, 'resc:'+clipzero(clip(sxpar(h,'RESCALER'),5)), charsize=1.4, col=rtext
   if isb then xyouts, 0.01, 0.92-0.05, /norm, clip(fix(round(sxpar(h,'BLUEXPTM'))))+'s', charsize=1.4, col=btext
   if isr then xyouts, 0.09, 0.92-0.05, /norm, clip(fix(round(sxpar(h,'REDEXPTM'))))+'s', charsize=1.4, col=rtext
   if isb then xyouts, 0.04, 0.92-0.05, /norm, 'flex:'+clip(repstr(fpr(0.01+sxpar(h,'BLUFLEX'),4.2),'-0.','-.'),4), charsize=1.4, col=btext
   if isr then xyouts, 0.12, 0.92-0.05, /norm, 'flex:'+clip(repstr(fpr(0.01+sxpar(h,'REDFLEX'),4.2),'-0.','-.'),4), charsize=1.4, col=rtext
   xyouts, -0.4+0.1+0.32,  -0.09+0.95-0.02, /norm, sxpar(h,'DATE_BEG'), charsize=1.4, col=rgb(220,220,220)
   xyouts, -0.4+0.10+0.32, -0.09+0.925-0.015, /norm, 'Sun '+fpr(sxpar(h,'SUNELEV'),3.1), charsize=1.2, color=(sxpar(h,'SUNELEV') gt -15) ? br : di
   xyouts, -0.4+0.16+0.32, -0.09+0.925-0.015, /norm, 'Moon '+fpr(sxpar(h,'MOONELEV'),3.1)+' ('+ clip(round(100*sxpar(h,'MOONPHAS')))+'%)', charsize=1.2, color=(sxpar(h,'MOONELEV') gt 1) ? br : di 
   aircol = rgb(180,180,180)
   if sxpar(h,'AIRMASS') gt 1.8 then aircol = rgb(255,255,180)
   if sxpar(h,'AIRMASS') gt 2.3 then aircol = rgb(255,50,50)
   xyouts, -0.4+0.10+0.32, -0.09+0.90-0.01, /norm, 'Airmass '+fpr(sxpar(h,'AIRMASS'),1.2), charsize=1.2, color=aircol
   ;bluair = fpr(sxpar(h,'BLUFAIR1'),1.2)+((sxpar(h,'BLUFAIR2') > 0) ? '/'+fpr(sxpar(h,'BLUFAIR1'),1.2) : '')
   ;redair = fpr(sxpar(h,'REDFAIR1'),1.2)+((sxpar(h,'REDFAIR2') > 0) ? '/'+fpr(sxpar(h,'REDFAIR1'),1.2) : '')
   if isb then $
   xyouts, -0.4+0.50, -0.09+0.9-0.01, /norm, '('+fpr(sxpar(h,'BLUFAIR1'),1.2)+',', charsize=1.2, color=rgb(130,130,255) $
   else xyouts, -0.4+0.53-0.005, -0.09+0.9-0.01, /norm, '(', color=rgb(250,100,100)
   if isr then $
   xyouts, -0.4+0.53, -0.09+0.9-0.01, /norm,     fpr(sxpar(h,'REDFAIR1'),1.2)+')', charsize=1.2, color=rgb(250,100,100)
   f2b = sxpar(h,'BLUFAIR2') gt 0
   f2r = sxpar(h,'REDFAIR2') gt 0
   if f2b then xyouts, -0.4+0.50, -0.11+0.9-0.01, /norm, '('+fpr(sxpar(h,'BLUFAIR2'),1.2)+(f2r?',':')'), charsize=1.2, color=rgb(130,130,255)
   if f2r then xyouts, -0.4+0.53, -0.11+0.9-0.01, /norm, (f2b?'':'(')+fpr(sxpar(h,'REDFAIR2'),1.2)+')', charsize=1.2, color=rgb(250,100,100)

   ; --- Spectrum ---

   wb = where(sp.wav lt atchwav-1, nb)
   wr = where(sp.wav gt atchwav+1, nr)

   !p.position = [0.2, 0.50, 0.985, 0.99]
   rms = median(abs(sp.flux-shift(sp.flux,1)))
   wgood = where(sp.wav gt 3300 and sp.wav lt 9900 and finite(sp.flux))
   yrange = [-2*rms, 1.5*percentile(sp[wgood].flux,96)]
   plot, sp.wav, sp.wav*0, yrange=yrange, /xsty, /ysty, charsize=2, /nodata, xtickformat='(A1)'
   oplot, sp.wav, sp.wav*0, color=rgb(100,100,100)
   !p.font = -1
   if max(yrange) gt 1000 then boost = 0.007 else boost = 0
   xyouts, !p.position[0]-0.025-boost, 0.64, /norm, 'Flux (1e-17 erg/cm2/Ang)', orient=90, charsize=0.9
   !p.font = 0
   marktelluric, yrange
   markskylines, wavrange, yrange, dim=0.5
   if n_elements(bluesp) gt 1 then $
        oplot, bluesp.wav, convol(bluesp.flux,[0.25,0.5,0.25], /edge_mirror), color=rgb(111,111,222)
   if n_elements(redsp) gt 1  then $
        oplot, redsp.wav,  convol(redsp.flux,[0.25,0.5,0.25], /edge_mirror),  color=rgb(222,111,111)
   if n_elements(bluesp) gt 1 then begin
        ww = where(bluesp.wav le atchwav)
        oplot, bluesp[ww].wav, convol(bluesp[ww].flux,[0.25,0.5,0.25], /edge_mirror), color=rgb(111,111,222)
   endif

   oplot, atchwav*[1,1], [-1e6,1e6], color=rgb(128,128,0), linestyle=2

   badsp = convol(sp.flux,0.1+fltarr(10), /edge_mirror)
   ;badsp[where(sp.flag lt 0.25)] = !values.f_nan
   ;oplot, sp.wav, badsp, color=rgb(255,0,235)
   badsp[where(sp.flag lt 0.4)] = !values.f_nan
   oplot, sp.wav, badsp, color=rgb(255,0,255), thick=3

   if nb gt 0 then oplot, sp[wb].wav, convol(sp[wb].flux,0.1+fltarr(10), /edge_mirror), color=rgb(255,255,255);, thick=2
   if nr gt 0 then oplot, sp[wr].wav, convol(sp[wr].flux,0.1+fltarr(10), /edge_mirror), color=rgb(255,255,255);, thick=2

   oplot, sp.wav, badsp, color=rgb(111,111,111)
   plot, sp.wav, sp.wav*0, yrange=yrange, /xsty, /ysty, /nodata, xtickformat='(A1)',ytickformat='(A1)', /noerase


   ; --- Sky ---

   !p.position = [!p.position[0], 0.35, !p.position[2], !p.position[1]]
   yrange = [0, max(sqrt(sp.sky))]
   plot, sp.wav, sqrt(0 > sp.sky), /xsty, yrange=yrange, /ysty, charsize=2, xtickformat='(A1)', /nodata
   if n_elements(bluesp) gt 1 then $
        oplot, bluesp.wav, sqrt(1+bluesp.sky > 0)-1, color=rgb(111,111,222)
   if n_elements(redsp) gt 1  then $
        oplot, redsp.wav,  sqrt(1+redsp.sky > 0)-1,  color=rgb(222,111,111)
   !p.font = -1
   xyouts, !p.position[0]-0.025, 0.4, /norm, 'sqrt(sky)', orient=90, charsize=0.9
   !p.font = 0
   markskylines, wavrange, yrange
   if nb gt 0 then oplot, sp[wb].wav, sqrt(0 > sp[wb].sky+1)-1
   if nr gt 0 then oplot, sp[wr].wav, sqrt(0 > sp[wr].sky+1)-1
   ;oplot, sp.wav, sqrt(0 > sp.sky)
   oplot, atchwav*[1,1], [-1e6,1e6], color=rgb(128,128,0), linestyle=2
   plot, sp.wav, 0*sp.sky, /xsty, yrange=yrange, /ysty, xtickformat='(A1)',ytickformat='(A1)', /nodata, /noerase

   ppos = !p.position 
   !p.position = [0.32, !p.position[3]-0.09, 0.38, !p.position[3]-0.01]
   woi = where(sp.wav gt 5561 and sp.wav lt 5599, ctoi)
   if ctoi gt 0 then begin
     polyfill, !p.position[[0,0,2,2]], !p.position[[1,3,3,1]], color=rgb(5,5,5), /norm
     yrange = [0, 1.1*max(sqrt(sp[woi].sky))]
     plot, [0], [0], xrange=[5561, 5593], yrange=yrange, /xsty, /ysty, xtickformat='(A1)', ytickformat='(A1)', color=rgb(200,200,200)
     if n_elements(bluesp) gt 1 then $
          oplot, bluesp.wav, sqrt(1+bluesp.sky > 0)-1, color=rgb(111,111,222), psym=10
     if n_elements(redsp) gt 1  then $
          oplot, redsp.wav,  sqrt(1+redsp.sky > 0)-1,  color=rgb(222,111,111), psym=10
     oplot, 5577.346680*[1,1], yrange, color=rgb(96,96,0)
     !p.font = -1
     xyouts, 0.343, !p.position[3]-0.094, /norm, '5577', color=rgb(96,96,0), charsize=0.84
     !p.font = 0
   endif

   ; --- Response ---

   !p.position = ppos
   !p.position = [!p.position[0], 0.20, !p.position[2],!p.position[1]]
   sbin = 1+fltarr(n_elements(sp)) ; spectral binning only 
   if isb then sbin[where(sp.wav le atchwav)] = bxbin
   if isr then sbin[where(sp.wav ge atchwav)] = rxbin
   dwav = sp.wav ; differential wavelength (approx.)
   if nb gt 0 then dwav[wb] = dwav[wb]-dwav[(wb-1)>0]
   if nr gt 0 then dwav[wr] = dwav[(wr+1)<(n_elements(dwav)-1)]-dwav[wr]
   if nb gt 0 then dwav[wb[0]] = dwav[wb[1]]
   if nr gt 0 then dwav[wr[n_elements(wr)-1]] = dwav[wr[n_elements(wr)-2]]
   gain = fltarr(n_elements(dwav))
   if nb gt 0 then gain[wb] = 1.63
   if nr gt 0 then gain[wr] = 1.18 ; R3, but R2 is similar.  R1 is different!
   ; resp has units (counts/pix)/(erg/cm^2/Ang)
   ; resp*(hc/lambda)
   hc = 6.6260755e-27*2.998e18 ; erg*ang
       ;  h = 6.6260755e-27 erg s
       ;  nu = 2.998e18 Ang/s
       ;  effectivearea_m2 = (rresponse * rtelluricresponse * gain * h * nu / dwav)/1e4

   outresp = gain * sp.resp*(hc/sp.wav)/dwav / (73.3e4)
   yrange = 100*[0, 1.08*max(outresp)]
   plot, sp.wav, 100*outresp, /xsty, /ysty, charsize=2, yrange=yrange, xtickformat='(A1)', /nodata
   !p.font = -1
   xyouts, !p.position[0]-0.025, !p.position[1]+0.03, /norm, 'Throughput (%)', orient=90, charsize=0.9
   !p.font = 0
   marktelluric, yrange
   
   if n_elements(bluesp) gt 1 then begin
      dwavblue = bluesp.wav
      dwavblue = dwavblue-shift(dwavblue,1)
      dwavblue[0] = dwavblue[1]
      outrespblue = 1.63*bluesp.resp*(hc/bluesp.wav)/dwavblue/ (73.3e4)
      oplot, bluesp.wav, 100*outrespblue, color=rgb(111,111,222)
   endif
   if n_elements(redsp) gt 1  then begin
      dwavred = redsp.wav
      dwavred = shift(dwavred,-1)-dwavred
      dwavred[n_elements(dwavred)-1] = dwavred[n_elements(dwavred)-2]
      outrespred = 1.18*redsp.resp*(hc/redsp.wav)/dwavred/ (73.3e4)
      oplot, redsp.wav,  100*outrespred,  color=rgb(222,111,111)
   endif
   if nb gt 0 then oplot, sp[wb].wav, 100*outresp[wb]
   if nr gt 0 then oplot, sp[wr].wav, 100*outresp[wr]
   ; sp.resp*(5500./sp.wav)/sbin
   ; redsp.resp*(5500./redsp.wav)/rxbin

   xt = !p.position[0]+0.03
   xt2 = xt+(!p.position[2]-!p.position[0])*(atchwav-min(sp.wav))/(max(sp.wav)-min(sp.wav))
   yt = !p.position[3]-0.03
   bfile1 = ' '+(strsplit(repstr(repstr(clip(sxpar(h,'BLUFCAL1')),'elsb',''),'.spec',''),'_',/extract)) [0]
   bfile2 = ' '+(strsplit(repstr(repstr(clip(sxpar(h,'BLUFCAL2')),'elsb',''),'.spec',''),'_',/extract)) [0]
   rfile1 = ' '+(strsplit(repstr(repstr(clip(sxpar(h,'REDFCAL1')),'elsr',''),'.spec',''),'_',/extract)) [0]
   rfile2 = ' '+(strsplit(repstr(repstr(clip(sxpar(h,'REDFCAL2')),'elsr',''),'.spec',''),'_',/extract)) [0]
   xyouts, xt, yt, /norm, sxpar(h,'BLUFNAM1')+bfile1, col=rgb(130,130,255), charsize=1.1
   xyouts, xt2, yt, /norm, sxpar(h,'REDFNAM1')+rfile1, col=rgb(250,100,100), charsize=1.1
   if clip(sxpar(h,'BLUFNAM2')) ne '0' then xyouts, xt, yt-0.02, /norm, sxpar(h,'BLUFNAM2')+bfile2, col=rgb(130,130,255), charsize=1.1
   if clip(sxpar(h,'REDFNAM2')) ne '0' then xyouts, xt2, yt-0.02, /norm, sxpar(h,'REDFNAM2')+rfile2, col=rgb(250,100,100), charsize=1.1
   oplot, atchwav*[1,1], 100*[-1,1e20], color=rgb(128,128,0), linestyle=2

   ; --- Traces ---

   !p.position = [!p.position[0], 0.02, !p.position[2],!p.position[1]]
;         plot, [0], [0], color=rgb(0,0,0)

   fullx = round(1280*[0.2,!p.position[2]])+[1,-1]
   if isr then minredwav = min(redsp.wav) else minredwav = max(bluesp.wav)
   if isb then maxbluewav= max(bluesp.wav) else maxbluewav=min(redsp.wav)
   bluexmax = fullx[0]+(fullx[1]-fullx[0])*((atchwav < maxbluewav)-min(sp.wav))/(max(sp.wav)-min(sp.wav))
   redxmin  = fullx[0]+(fullx[1]-fullx[0])*((atchwav > minredwav)-min(sp.wav))/(max(sp.wav)-min(sp.wav))

   ;midx    = fullx[0]+(fullx[1]-fullx[0])*((atchwav > minredwav)-min(sp.wav))/(max(sp.wav)-min(sp.wav))
   y2d = round(700*[!p.position[1],!p.position[3]])+[0,-0] ; y window pixels occupied by the trace 2D plots

   brightobjfile = path+'spredux1d/brightobjectpos.txt'

   if isb then bluycen = sxpar(h,'BLUYCENT')*blubin
   if isr then redycen = sxpar(h,'REDYCENT')*redbin
   if file_test(brightobjfile) then begin
     readcol, brightobjfile, tfile, tcenter, twid, xx, tgr, format='a,f,f,a,a', comment='#', /silent, count=nbrobj
     if nbrobj gt 0 then begin
       if isb then begin 
         bluegr = sxpar(h,'GRISNAME',ctbg)
         wgb = where(tgr eq bluegr,ctbgt)
         if ctbgt then bluycen = median(tcenter[wgb])*blubin
       endif   
       if isr then begin
         redgr = sxpar(h,'GRANAME',ctrg)
         wgr = where(tgr eq redgr,ctrgt)
         if ctrgt then redycen = median(tcenter[wgr])*redbin
       endif
     endif
   endif
   if isb then y0b = bluycen-(y2d[1]-y2d[0])/2
   if isr then y0r = redycen-(y2d[1]-y2d[0])/2

   if isr then rtracey =    sxpar(h,'REDYCENT')*redbin
   if isb then btracey =    sxpar(h,'BLUYCENT')*blubin
   if isr then if y0r+y2d[1] lt rtracey then y0r = rtracey-(y2d[1]-y2d[0])+16
   if isb then if y0b+y2d[1] lt btracey then y0b = btracey-(y2d[1]-y2d[0])+16

   ; blue trace
   ppos = !p.position
   midpos = ppos[0]+(ppos[2]-ppos[0])*(atchwav-min(sp.wav))/(max(sp.wav)-min(sp.wav))
   !p.position[2] = midpos
   ny = (y2d[1]-y2d[0])+1

   if isb then begin
   plot, [min(sp.wav),atchwav], y0b+(y2d-y2d[0]), /xsty, /ysty, /nodata, xtickformat='(A1)'
   !p.font = -1
   xyouts, !p.position[0]-0.025, 0.09, /norm, 'y pixel', orient=90, charsize=0.9
   !p.font = 0
   bluex = [fullx[0],bluexmax]
   nblue = (bluex[1]-bluex[0])+1
   resampblue2d = fltarr(nblue,ny)
   resampbluecr = bytarr(nblue,ny)
   if n_elements(blue2d) gt 2 and n_elements(bluesp) gt 2 then begin
     inbwav = bluesp.wav
     blupix = bluesp.pix
     wav1 = min([atchwav, max(inbwav)])
     outbwav = (findgen(nblue)/(nblue-1))*(wav1-min(inbwav))+inbwav[0]
     ny2 = (size(blue2d)) [2]
     for y = 0,ny-1 do if ((y0b+y)/bybin) lt (ny2-1) then resampblue2d[*,y] = interpol(blue2d[blupix,(y0b+y)/bybin], inbwav, outbwav)
     if n_elements(bluecr) gt 1 then $
     for y = 0,ny-1 do if ((y0b+y)/bybin) lt (ny2-1) then resampbluecr[*,y] = interpol(bluecr[blupix,(y0b+y)/bybin], inbwav, outbwav)
     minv = percentile(resampblue2d,25.)
     maxv = percentile(blue2d[*,min(sp[wb].ypix):max(sp[wb].ypix)],99.5)
     resampsc = (255*((resampblue2d - minv)>0)/(maxv-minv)) < 255
     ;resampblue2d = 255*(((resampblue2d > medblue) < maxblue)/maxblue)
     nstack = sxpar(h,'BLUNEXP')
     out = fltarr(nblue,ny,3)
     out[*,*,0] =(resampsc+(128*(resampbluecr)/(1.+nstack))) < 255
     out[*,*,1] = resampsc*((1+nstack-resampbluecr)/(1.+nstack))
     out[*,*,2] = resampsc*((1+nstack-resampbluecr)/(1.+nstack))
     tv, out,  bluex[0],  y2d[0], true=3
     ;tvscl, resampblue2d, bluex[0], y2d[0]
   endif
   oplot, atchwav*[1,1], [-1,500], color=rgb(128,128,0), linestyle=2
   bluap = sxpar(h,'BLUAPRAD', count=ctba)
   if ctba gt 0 and n_elements(wb) ge 3 then begin
     oplot, sp[wb].wav, (sp[wb].ypix+bluap)*bybin, color=rgb(0,255,0)
     oplot, sp[wb].wav, (sp[wb].ypix-bluap)*bybin, color=rgb(0,255,0)
   endif
   plot, [min(sp.wav),atchwav], y0b+(y2d-y2d[0]), /xsty, /ysty, /nodata, xtickformat='(A1)', /noerase
   ;plot, [min(sp.wav),max(sp.wav)], y0b+(y2d-y2d[0]), /xsty, /ysty, /nodata, xtickformat='(A1)'
   dif = 0.011*(max(sp.wav)-min(sp.wav))
   for wtick = 3000, 11000, 1000 do begin
     if wtick ge min(sp.wav) and wtick le atchwav then xyouts, wtick-dif, y0b-12, clip(wtick)
   endfor
   endif else begin
     ppos2 = !p.position
     !p.position = [0.01,0.01,0.02,0.02]
     plot, [0], [0], color=rgb(0,0,0)
     !p.position = ppos2
   endelse

   ; red trace
   if isr then begin
   redx = [redxmin,fullx[1]]
   nred = (redx[1]-redx[0])+1
   resampred2d = fltarr(nred,ny)
   resampredcr = bytarr(nred,ny)
   !p.position = ppos
   !p.position[0] = midpos
   plot, [atchwav,max(sp.wav)], y0r+(y2d-y2d[0]), /xsty, /ysty, /nodata, xtickformat='(A1)', ytickformat='(A1)'
   if isb eq 0 then begin
     !p.font = -1
     xyouts, !p.position[0]-0.025, 0.09, /norm, 'y pixel', orient=90, charsize=0.9
     !p.font = 0
     axis, yaxis=0, yrange=y0r+(y2d-y2d[0]), /ysty
   endif else begin
     axis, yaxis=1, yrange=y0r+(y2d-y2d[0]), /ysty
   endelse
   if n_elements(red2d) gt 2 and n_elements(redsp) gt 2  then begin
     inrwav = redsp.wav
     redpix = redsp.pix
     wav0 = max([atchwav, min(inrwav)])
;     if isb then wav0 = min([wav0, max(bluesp.wav)])
     outrwav = (findgen(nred)/(nred-1))*(max(inrwav)-wav0)+wav0
     ny2 = (size(red2d)) [2]
     w2r = where(outrwav gt min(inrwav))
     for y = 0,ny-1 do if ((y0r+y)/rybin) lt (ny2-1) and (y0r+y) gt 0 then $
        resampred2d[w2r,y]  = interpol(red2d[redpix,(y0r+y)/rybin], inrwav, outrwav[w2r])
     if n_elements(redcr) gt 1 then $
     for y = 0,ny-1 do if ((y0r+y)/rybin) lt (ny2-1) and (y0r+y) gt 0 then $
        resampredcr[w2r,y]  = interpol(redcr[redpix,(y0r+y)/rybin], inrwav, outrwav[w2r])
     minv = percentile(resampred2d,25.)
     inpix = red2d[*,min(sp[wr].ypix):max(sp[wr].ypix)]
     inpix = inpix[where(finite(inpix))]
     maxv = percentile(inpix,99.5)
     resampsc = (255*((resampred2d - minv)>0)/(maxv-minv)) < 255
     nstack = sxpar(h,'REDNEXP')
     out = fltarr(nred,ny,3)
     out[*,*,0] =(resampsc+(128*(resampredcr)/(1.+nstack))) < 255
     out[*,*,1] = resampsc*((1+nstack-resampredcr)/(1.+nstack))
     out[*,*,2] = resampsc*((1+nstack-resampredcr)/(1.+nstack))
     tv, out,  redx[0],  y2d[0], true=3
   endif
   redap = sxpar(h,'REDAPRAD', count=ctra)
   if ctra gt 0 and n_elements(wr) ge 3 then begin
     oplot, sp[wr].wav, (sp[wr].ypix+redap)*rybin, color=rgb(0,255,0)
     oplot, sp[wr].wav, (sp[wr].ypix-redap)*rybin, color=rgb(0,255,0)
   endif

   plot, [atchwav,max(sp.wav)], y0r+(y2d-y2d[0]), /xsty, /ysty, /nodata, xtickformat='(A1)',ytickformat='(A1)', /noerase
   ;plot, [min(sp.wav),max(sp.wav)], y0r+(y2d-y2d[0]), /xsty, /ysty, /nodata, xtickformat='(A1)'
   dif = 0.011*(max(sp.wav)-min(sp.wav))
   for wtick = 3000, 11000, 1000 do begin
     if wtick ge atchwav and wtick le max(sp.wav) then xyouts, wtick-dif, y0r-12, clip(wtick)
   endfor
   endif else begin
     !p.position = [0.01,0.01,0.02,0.02]
     plot, [0], [0], color=rgb(0,0,0), xtickformat='(A1)', ytickformat='(A1)'
   endelse
   ; end red trace

   !p.position = ppos
  
   ; --- Compass ---

   pa = sxpar(h,'SLITPA')
   parad = pa*!pi/180. ; not parallactic angle, confusingly
   paral = sxpar(h,'PARANGLE',count=ctpara)
   paralrad = paral*!pi/180
   r = 14.
   cx = 165
   cy = 1
   ysize = 40
   ynud = 0
   if ctpara gt 0 then plots, cx+25 + [-1,1]*r*sin(parad-paralrad) ,  cy+ysize/2+ynud+ [-1,1]*r*cos(parad-paralrad), thick=3, /device, color=rgb(0,128,0)    ; parallactic bar
   plots, cx+25-2, cy+[0,ysize-1], /device, color=rgb(200,200,20)
   plots, cx+25+2, cy+[0,ysize-1], /device, color=rgb(200,200,20)
   plots, cx+25 + [0,r*sin(parad)] ,     cy+ysize/2+ynud+ [0, r*cos(parad)], /device, thick=2          ; north bar
   plots, cx+25 + [0,r*sin(parad-!pi/2.)] ,  cy+ysize/2+ynud+ [0, r*cos(parad-!pi/2.)], /device    ; east bar
   xyouts, cx+25+1.28*r*sin(parad) - 3., cy+ysize/2+ynud+1.28*r*cos(parad) - 3., /device, 'N', charsize=0.7
   xyouts, cx-36, cy+20, /device, 'PA='+clip(round(float(pa))), charsize=0.75
   if ctpara gt 0 then xyouts, cx-48, cy+8, /device, 'para='+clip(round(float(paral))), charsize=0.75, color=rgb(0,128,0)


   ; --- Blue and red profiles ---

   for pp = 1, 2 do begin
      if pp eq 1 then  !p.position = [0.03, 0.45, !p.position[0]-0.04, 0.77]
      if pp eq 2 then  !p.position = [0.03, 0.09, !p.position[2], 0.40]
      if pp eq 1 then  ckp = 'BLU'
      if pp eq 2 then  ckp = 'RED'
      if pp eq 1 then  linecol = rgb(150,150,255)
      if pp eq 2 then  linecol = rgb(255,10,10)

     sp2dfile = path+extractpath(filename)+'spredux1d/'+clip(sxpar(h,ckp+'2DSPC',count=ct2d))
     if ct2d eq 0 then begin
         plot, [0], [0], color=rgb(0,0,0)
         continue
     endif
     ycen = sxpar(h,ckp+'YCENT', count=ctc)
     tracecen = ycen
     aprad = sxpar(h,ckp+'APRAD')
     bkdist = sxpar(h,ckp+'BKDST')+[0,0]
     bkwid = sxpar(h,ckp+'BKWID')+[0,0]
     if sxpar(h,ckp+'BKDM') gt 0 then  bkdist = [sxpar(h,ckp+'BKDM'),sxpar(h,ckp+'BKDP')]
     if sxpar(h,ckp+'BKDW') gt 0 then  bkwid  = [sxpar(h,ckp+'BKWM'),sxpar(h,ckp+'BKWP')]
     proffile = repstr(sp2dfile,'.fits','.profile')
     if file_test(proffile) eq 0 then begin
         plot, [0], [0], color=rgb(0,0,0)
         continue
     endif
     readcol, proffile, y, prof, comment='#', /silent
     ybin = (strsplit(sxpar(h,ckp+'BIN'),',',/extract)) [0]
     if ctc gt 0 then ycenp = ycen*ybin else ycenp = 90
     xtitle = ''
     ;if pp eq 1 then xtitle = 'Blue y pixel'
     ;if pp eq 2 then xtitle = 'Red y pixel'

     plot, y*ybin, prof, psym=10, /nodata, xrange=ycenp+[-90,90], charsize=2, /xsty, xtitle=xtitle

     if ctc eq 1 then begin
       oplot, ybin*(tracecen)+[0,0], [-1e6,1e6], col=rgb(195,115,0), linestyle=2
       oplot, ybin*(tracecen-aprad)+[0,0], [-1e6,1e6], col=rgb(0,200,0)
       oplot, ybin*(tracecen+aprad)+[0,0], [-1e6,1e6], col=rgb(0,200,0)
       oplot, ybin*(tracecen-bkdist[0])+[0,0], [-1e6,1e6], col=rgb(150,150,0), linestyle=1
       oplot, ybin*(tracecen-bkdist[0]-bkwid[0])+[0,0], [-1e6,1e6], col=rgb(150,150,0), linestyle=1
       oplot, ybin*(tracecen+bkdist[1])+[0,0], [-1e6,1e6], col=rgb(150,150,0), linestyle=1
       oplot, ybin*(tracecen+bkdist[1]+bkwid[1])+[0,0], [-1e6,1e6], col=rgb(150,150,0), linestyle=1
     endif

     oplot, y*ybin, prof, col=linecol
   endfor

end


; not used
pro vplottrace, spec1d, spec2d, x0, x1, y0, y1 ; latter four are screen window coordinates
   xsize = x1-x0+1
   ysize = y1-y0+1

   spy0 = fix( ((min(spec1d.ypix)+max(spec1d.ypix))/2) - (yrange[1]-yrange[0])/2)
   spy1 = spy0 + (yrange[1]-yrange[0])
   ygrid = spy0 + indgen(spy1-spy0+1)
   ny = n_elements(ygrid)

   nx = (xrange[1]-xrange[0]+1)
   wavgrid = wav[0]+(wav[1]-wav[0])*findgen(nx)

   outim = fltarr(nx,ny)

   for y = 0, ny-1 do begin
     outim[*,y] = interpol(spec2d[*,y], wav, wavgrid)
   endfor
   tracey = interpol(spec1d.y[*,y], wav, wavgrid)
   traceydisp = tracey - spy0

   return
end

; lrisvalidate, '*/*.spec', /copy, /noview, /savepng, /all

pro lrisvalidate, filestr, demo=demo, delete=delete, move=move, copy=copy, mark=mark, outdir=outdir, noview=noview, savepng=savepng, nooverwrite=nooverwrite, all=all, reason=reason

 if keyword_set(copy) then mark = 0
 if keyword_set(copy) then move = 0 ; 
 if keyword_set(move) then mark = 0
 reason = keyword_set(reason)

 if n_elements(outdir) eq 0 then outdir = 'validation'
 if strmid(outdir, strlen(outdir)-1) eq '/' then outdir = strmid(outdir, 0, strlen(outdir)-1)

 if file_test(outdir) eq 0 and (keyword_set(savepng) or keyword_set(copy) or keyword_set(move)) then spawn, 'mkdir '+outdir
 outpngdir = outdir+'/png'
 if file_test(outpngdir) eq 0 and keyword_set(savepng) then spawn, 'mkdir '+outpngdir

 demo = keyword_set(demo) ; for testing without a license or in a write-protected directory
 
 if keyword_set(noview) and keyword_set(savepng) then autoscroll = 1 else autoscroll = 0

 if n_elements(filestr) eq 0 then files = file_search('*.spec', count=ct)
 if n_elements(filestr) eq 1 then files = file_search(filestr, count=ct)
 if ct eq 0 then print, ' Found no LRIS spectra.'
 if ct eq 0 then return

 if keyword_set(noview) eq 0 or keyword_set(savepng) then begin

  dates = strarr(n_elements(files))
  for d = 0, n_elements(files)-1 do begin
    h = headspec(files[d])
    dates[d] = sxpar(h,'MJD') ; formely DATE
  endfor
  files = files[sort(dates)]

  result = strarr(ct)
  freason = replicate('',ct)

  black = rgb(0,0,0)
  white = rgb(255,255,255)
  red = rgb(255,0,0)
  yellow = rgb(255,255,0)
  blue = rgb(50,50,255)
  liteblue = rgb(111,111,255)
  green = rgb(0,255,0)
  gray = rgb(100,100,100)
  lightgray = rgb(200,200,200)
  yy = 0.965
  firstpos= [0.870,yy,0.89,yy+0.02]
  back5pos= [0.895,yy,0.915,yy+0.02]
  next5pos= [0.92,yy,0.94,yy+0.02]
  next20pos=[0.945,yy,0.965,yy+0.02]
  clearpos= [0.79-0.1, 0.905, 0.83-0.1, 0.96]
  passpos = [0.84-0.1, 0.905, 0.96-0.1, 0.96]
  failpos = [0.84-0.1 ,0.84,  0.895-0.1,0.895]
  flagpos = [0.905-0.1,0.84,  0.96-0.1, 0.895]
  backpos = [0.84+0.04-0.01 ,0.905, 0.895+0.03-0.01,0.96]
  nextpos = [0.905+0.025-0.01,0.905, 0.96+0.015-0.01, 0.96]
  skippos = [0.905+0.025-0.01,0.84, 0.96+0.015-0.01, 0.895]
  hidepos = [0.84+0.04-0.01 ,0.84, 0.895+0.03-0.01,0.895]
  quitpos = [0.01,0.945,0.04,0.975]
  bspecpos = [0.2, 0.02, 0.45, 0.2] ; midx boundary approximate
  rspecpos = [0.6, 0.02, 0.99, 0.2]
  passcol = green
  failcol = red
  defacol = lightgray  ; defacol
  flagcol = yellow
  quitcol = red
  clearcol = liteblue

  reasons = ['cosmic false pos.',$
             'wavecal problem',$
             'sky subtraction',$
             'not real source',$
             'trace lost','trace poor',$
             'aperture misaligned','chimeric spectrum',$               
             'background contam.','self-background',$
             'bad flux calibrator','poor response sol.',$
             'dichroic flux cal.','rescaling',$
             'poor telluric cal.',$
             'instrument problem']
  nreason = n_elements(reasons)
  reasonpos = fltarr(4,nreason)
    reasonpos[0,*] = 0.8 
    reasonpos[1,*] = 0.81-findgen(nreason)*0.025
    reasonpos[2,*] = 0.9
    reasonpos[3,*] = 0.81-findgen(nreason)*0.025+0.025

  xs = 1280
  ys = 700
  window, 12, xsize=xs, ysize=ys, title='LRIS spectrum validation'

  if file_test('validation.txt') then begin
    readcol, 'validation.txt', prevvalname, prevvalresult, format='a,a', comment='#', /silent
    prevvalreason = strarr(n_elements(prevvalresult))
    for j = 0, n_elements(prevvalresult)-1 do begin
      res = prevvalresult[j]
      cp = strpos(res,':')
      if cp gt 0 then begin
        prevvalresult[j] = strmid(res,0,cp)
        prevvalreason[j] = strmid(res,cp+1)
      endif
    endfor
    for t = 0, n_elements(prevvalname)-1 do begin
       w = where(prevvalname[t] eq files, ct) ; different w
       if ct gt 0 then result[w] = prevvalresult[t]
       if ct gt 0 then freason[w] = repstr(prevvalreason[t],'_',' ')
    endfor
  endif 

  i = 0
  if file_test('.lastspec.txt') then begin ; hidden file
    openr, 5, '.lastspec.txt'
    inline = ''
    readf, 5, inline
    close, 5
    wlast = where(inline eq files, ct)
    if ct gt 0 then i = wlast
  endif

  action = ''
  hide = 0
  skip = 0
  lastfile = ''
  iinit = i
  ifinal = iinit -1
  if ifinal eq -1 then ifinal = n_elements(files)-1
  while action ne 'quit' do begin
    if files[i] ne lastfile then begin
      sp = readspec(files[i], h)
      path = file_dirname(files[i])+'/'
      redsp = ''
      bluesp = ''
      blue2d = 0
      red2d = 0
      blue2dcr = 0
      red2dcr = 0
      bluefile = path+'spredux1d/'+clip(sxpar(h,'BLUFILE',count=ctbf)) 
      redfile = path+'spredux1d/'+clip(sxpar(h,'REDFILE',count=ctrf))  
      blue2dfile = path+'spredux1d/'+clip(sxpar(h,'BLU2DSPC',count=ctb2d))
      red2dfile = path+'spredux1d/'+clip(sxpar(h,'RED2DSPC',count=ctr2d))
      if ctbf gt 0 then if file_test(bluefile)   then bluesp = readspec(bluefile,blueh)
      if ctrf gt 0 then if file_test(redfile)    then redsp = readspec(redfile,redh)
      if ctb2d gt 0 then if file_test(blue2dfile) then begin
          fits_info, blue2dfile, /silent, n_ext=nexten
          blue2d = readfits(blue2dfile,blue2dh,/silent)
         if nexten ge 2 then blue2dcr = mrdfits(blue2dfile,2,hcr,/silent)
      endif
      if ctr2d gt 0 then if file_test(red2dfile)  then begin
         fits_info, red2dfile, /silent, n_ext=nexten
         red2d = readfits(red2dfile,red2dh,/silent)
         if nexten ge 2 then red2dcr = mrdfits(red2dfile,2,hcr,/silent)
      endif
      if demo eq 0 then begin
        openw, 5, '.lastspec.txt' ; store last spectrum read
        printf, 5, files[i]
        close, 5
      endif
      redraw = 1
    endif

    if hide eq 1 or redraw eq 1 then begin
      validateplot, sp, h, bluesp, redsp, blue2d, red2d, blue2dcr, red2dcr, filename=removepath(files[i]), path=path
      redraw = 0
      if files[i] ne lastfile and keyword_set(savepng) then begin
         outpng = repstr(files[i],'.fits','.png') ; should be .spec, not .fits...
         if outpng eq files[i] then outpng = files[i]+'.png'
         fulloutpng = outpngdir+'/'+removepath(outpng)
         write_png, fulloutpng, tvrd(/true)
      endif
      xyouts, 0.01, 0.98, /norm, clip(i+1)+'/'+clip(n_elements(files)), charsize=0.8, color=rgb(166,166,166)
    endif
    lastfile = files[i]

    if autoscroll eq 0 then begin
    if hide then begin
      wait, 0.05
      cursor, x, y, /norm
      cursor, x, y, /up
      hide = 0
    endif

    pass = result[i] eq 'pass'
    fail = result[i] eq 'fail'
    flag = result[i] eq 'flag'
    clear = result[i] eq ''

    drawvbutton, firstpos, 0,   /norm, label='1',    oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    drawvbutton, back5pos, 0,   /norm, label='-5',   oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    drawvbutton, next5pos, 0,   /norm, label='+5',   oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    drawvbutton, next20pos, 0,  /norm, label='+20',  oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    drawvbutton, backpos, 0,    /norm, label='back', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    drawvbutton, nextpos, 0,    /norm, label='next', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    drawvbutton, passpos, pass, /norm, label='pass', oncolors=[white,passcol,white], offcolors=[passcol,black,passcol], charsize=2
    drawvbutton, failpos, fail, /norm, label='fail', oncolors=[white,failcol,white], offcolors=[failcol,black,failcol], charsize=2
    drawvbutton, flagpos, flag, /norm, label='flag', oncolors=[white,flagcol,white], offcolors=[flagcol,black,flagcol], charsize=2
    drawvbutton, skippos, skip, /norm, label='skip', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    drawvbutton, quitpos, 0,    /norm, label='quit', oncolors=[white,quitcol,white], offcolors=[quitcol,black,quitcol], charsize=1.5
    if result[i] ne '' then $
    drawvbutton, clearpos, 0, /norm, label='clear', oncolors=[white,clearcol,white], offcolors=[clearcol,black,clearcol], charsize=1.9
    drawvbutton, hidepos, 0,    /norm, label='hide', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=1.5
    if reason and (fail or flag) then begin
       for r = 0, n_elements(reasons)-1 do begin
         drawvbutton, [reasonpos[0,r],reasonpos[1,r],reasonpos[2,r],reasonpos[3,r]], (freason[i] eq reasons[r]), /norm, label=reasons[r],$
                      oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], textoff=[0.005,0.005]
       endfor
    endif
    action = ''
    while action eq '' do begin
      wait, 0.025
      action = ''
      cursor, x, y, /norm;, /down
      if (x gt firstpos[0] and y gt firstpos[1] and x lt firstpos[2] and y lt firstpos[3]) then action='first'
      if (x gt back5pos[0] and y gt back5pos[1] and x lt back5pos[2] and y lt back5pos[3]) then action='back5'
      if (x gt next5pos[0] and y gt next5pos[1] and x lt next5pos[2] and y lt next5pos[3]) then action='next5'
      if (x gt next20pos[0] and y gt next20pos[1] and x lt next20pos[2] and y lt next20pos[3]) then action='next20'
      if (x gt passpos[0] and y gt passpos[1] and x lt passpos[2] and y lt passpos[3]) then action='pass'
      if (x gt failpos[0] and y gt failpos[1] and x lt failpos[2] and y lt failpos[3]) then action='fail'
      if (x gt flagpos[0] and y gt flagpos[1] and x lt flagpos[2] and y lt flagpos[3]) then action='flag'
      if (x gt backpos[0] and y gt backpos[1] and x lt backpos[2] and y lt backpos[3]) then action='back'
      if (x gt nextpos[0] and y gt nextpos[1] and x lt nextpos[2] and y lt nextpos[3]) then action='next'
      if (x gt quitpos[0] and y gt quitpos[1] and x lt quitpos[2] and y lt quitpos[3]) then action='quit'
      if result[i] ne '' then $
      if (x gt clearpos[0] and y gt clearpos[1] and x lt clearpos[2] and y lt clearpos[3]) then action='clear'
      if (x gt skippos[0] and y gt skippos[1] and x lt skippos[2] and y lt skippos[3]) then action='skip'
      if (x gt hidepos[0] and y gt hidepos[1] and x lt hidepos[2] and y lt hidepos[3]) then action='hide'
      if (x gt bspecpos[0] and y gt bspecpos[1] and x lt bspecpos[2] and y lt bspecpos[3]) then action='bspec'
      if (x gt rspecpos[0] and y gt rspecpos[1] and x lt rspecpos[2] and y lt rspecpos[3]) then action='rspec'
      if reason and (fail or flag) then begin
      for r = 0, n_elements(reasons)-1 do begin
         if x gt reasonpos[0,r] and y gt reasonpos[1,r] and x lt reasonpos[2,r] and y lt reasonpos[3,r] then begin
           action = 'reason'
           break ; current r is saved
         endif
      endfor
      endif
      ;if action ne '' then print, action
    endwhile

    if action eq 'first' then drawvbutton, firstpos, 1, /norm, label='1', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    if action eq 'back5' then drawvbutton, back5pos, 1, /norm, label='-5', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    if action eq 'next5' then drawvbutton, next5pos, 1,   /norm, label='+5',   oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    if action eq 'next20' then drawvbutton, next20pos, 1,  /norm, label='+20',  oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    if action eq 'pass' then drawvbutton, passpos, 1, /norm, label='pass', oncolors=[white,passcol,white], offcolors=[passcol,black,passcol], charsize=2
    if action eq 'fail' then drawvbutton, failpos, 1, /norm, label='fail', oncolors=[white,failcol,white], offcolors=[failcol,black,failcol], charsize=2
    if action eq 'flag' then drawvbutton, flagpos, 1, /norm, label='flag', oncolors=[white,flagcol,white], offcolors=[flagcol,black,flagcol], charsize=2
    if action eq 'back' then drawvbutton, backpos, 1, /norm, label='back', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    if action eq 'next' then drawvbutton, nextpos, 1, /norm, label='next', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    if action eq 'skip' then drawvbutton, skippos, ~skip, /norm, label='skip', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    if action eq 'quit' then drawvbutton, quitpos, 1, /norm, label='quit', oncolors=[white,quitcol,white], offcolors=[quitcol,black,quitcol], charsize=1.5
    if action eq 'clear' then drawvbutton, clearpos, 1, /norm, label='clear', oncolors=[white,clearcol,white], offcolors=[clearcol,black,clearcol], charsize=1.9
    if action eq 'hide' then drawvbutton, hidepos, 1, /norm, label='hide', oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], charsize=2
    if action eq 'reason' then begin
         drawvbutton,  [reasonpos[0,r],reasonpos[1,r],reasonpos[2,r],reasonpos[3,r]], 1, /norm, label=reasons[r],$
                      oncolors=[white,defacol,white], offcolors=[defacol,black,defacol], textoff=[0.005,0.003]
    endif

    if action eq 'bspec' or action eq 'rspec' then begin
       print, ' Opening DS9...'
       if action eq 'bspec' then spawn, 'ds9 '+blue2dfile
       if action eq 'rspec' then spawn, 'ds9 '+red2dfile
    endif

    cursor, x, y, /norm, /up
    wait, 0.1

    endif else begin ; end not-autoscroll block
      action = 'next'
      print, ' Saving '+removepath(files[i])+'.png'
      if i eq ifinal then action = 'quit'
    endelse

    if action eq 'fail' then wait, 0.1
    
    if action eq 'hide' then hide = 1

    if action eq 'skip' then skip = ~skip

    if action eq 'fail' or action eq 'flag' or action eq 'pass' or action eq 'clear' then begin
      if action eq 'pass' then result[i] = 'pass'
      if action eq 'flag' then result[i] = 'flag'
      if action eq 'fail' then result[i] = 'fail'
      if action eq 'clear' then result[i] = ''
      if action ne 'flag' and action ne 'fail' then redraw = 1
      if action ne 'flag' and action ne 'fail' then freason[i] = ''
      print, action+' '+files[i]
      if freason[i] ne '' then c = ':' else c = ''
      if demo eq 0 then updatefile, 'validation.txt', files[i]+'  '+action+c+repstr(freason[i],' ','_')+'  '+repstr(tstamp(/ut),' ','T')
    endif

    if action eq 'reason' then begin
      freason[i] = reasons[r]
      print, '     reason:', freason[i]
      if freason[i] ne '' then c = ':' else c = ''
      if demo eq 0 then updatefile, 'validation.txt', files[i]+'  '+result[i]+c+repstr(freason[i],' ','_')+'  '+repstr(tstamp(/ut),' ','T')
    endif

    if action eq 'first' or action eq 'back5' or action eq 'next5' or action eq 'next20' then begin
       if action eq 'first' then i = 0
       if action eq 'back5' then i -= 5
       if action eq 'next5' then i += 5
       if action eq 'next20' then i += 20
       if i lt 0 then i = 0
       if i ge n_elements(files) then i = n_elements(files)-1
    endif

    starti = i
    if (action eq 'pass' or action eq 'next' or $
        ((action eq 'fail' or action eq 'flag') and reason eq 0 ) ) then begin
      i += 1
      if i eq n_elements(files) then i = 0
      if skip gt 0 then begin
         while result[i] ne '' do begin
           ;print, i, result[i]
           i += 1
           if i eq n_elements(files) then i = 0
           if i eq starti then begin
              print, 'skipped back to the beginning'
              break
           endif
         endwhile
      endif
    endif
    if action eq 'back' then begin
       i -= 1
       if i lt 0 then i = n_elements(files)-1
       if skip gt 0 then begin
          while result[i] ne '' do begin
            i -= 1
            if i eq 0 then i = n_elements(files)-1
            if i eq starti then begin
              print, 'skipped back to the beginning'
              break
            endif
          endwhile           
       endif
   endif

  endwhile

  wait, 0.2
  wdelete, 12
  !p.multi = 0
  !p.position = 0

 endif

 if keyword_set(copy) or keyword_set(move) or keyword_set(mark) then begin

    if file_test('validation.txt') eq 0 then begin
       print, ' No validation file.'
       return
    endif
    readcol, 'validation.txt', vfiles, vresultstr, vtimes, format='a,a,a', /silent
    vresult = strarr(n_elements(vresultstr))
    vreason = strarr(n_elements(vresultstr))
    for j = 0, n_elements(vresultstr)-1 do begin
      res = vresultstr[j]
      cp = strpos(res,':')
      if cp gt 0 then begin
        vresult[j] = strmid(res,0,cp)
        vreason[j] = strmid(res,cp+1)
      endif else begin
        vresult[j] = res
      endelse
      ;print, vfiles[j], ' ', vresult[j], ' ', vreason[j], ' ', vtimes[j]
    endfor

    savelist = ''
    saveresult = ''
    savereason = ''
    savetime = ''
    for i = 0, n_elements(vfiles)-1 do begin
       check = total(files eq vfiles[i])
       if (vresult[i] eq 'pass' or vresult[i] eq 'flag' or $
           (vresult[i] eq 'fail' and keyword_set(all))) $
          and file_test(vfiles[i]) and check then begin
           outfile = outdir+'/'+removepath(vfiles[i])
           ;if vresult[i] eq 'fail' then print, vfiles[i], file_test(outfile)
           if (keyword_set(nooverwrite) eq 0) or (file_test(outfile) eq 0) then begin
             savelist = [savelist, vfiles[i]]
             saveresult = [saveresult, vresult[i]]
             savereason = [savereason, vreason[i]]
             savetime = [savetime, vtimes[i]]
           endif
       endif
    endfor

    if n_elements(savelist) gt 1 then begin
       if keyword_set(copy) then print, ' Will copy '+clip(n_elements(savelist)-1)+' files to '+outdir+'/:'
       if keyword_set(move) then print, ' Will move '+clip(n_elements(savelist)-1)+' files to '+outdir+'/:'
       if keyword_set(mark) then print, ' Will edit '+clip(n_elements(savelist)-1)+' files.'
       for i = 1, n_elements(savelist)-1 do begin
          existstr = ''
          if keyword_set(mark) eq 0 then $
             if file_test(outdir+'/'+removepath(savelist[i])) then $
                existstr = ' - will overwrite existing file'
          print, '   '+savelist[i]+' ('+saveresult[i]+')'+existstr
       endfor
       print, ' Press [enter] to confirm or [q] to cancel'
       key = get_kbrd()
       if key eq string(10B) then begin
         for i = 1, n_elements(savelist)-1 do begin
           if file_test(savelist[i]) then begin
             sp = readspec(savelist[i],h)
             if saveresult[i] eq 'pass' then sxaddpar, h, 'VALIDATE', 'pass', ' Passed user validation'
             if saveresult[i] eq 'flag' then sxaddpar, h, 'VALIDATE', 'warn', ' Passed user validation with warning'
             if saveresult[i] eq 'fail' then sxaddpar, h, 'VALIDATE', 'fail', ' Failed user validation'
             if saveresult[i] eq 'flag' then sxaddpar, h, 'REASON', savereason[i], ' Reason for validation warning'
             if saveresult[i] eq 'fail' then sxaddpar, h, 'REASON', savereason[i], ' Reason for validation failure'
             if saveresult[i] eq 'flag' then sxaddhist, 'This spectrum is flagged as having minor problems.', h, /comment
             if saveresult[i] eq 'fail' then sxaddhist, 'This spectrum failed validation.', h, /comment
             hist = 'Validation performed '+repstr(savetime[i],'T',' ')+' UT'
             sxaddhist, hist, h ;, /comment
             whist = where(strpos(h,'HISTORY') eq 0 or strpos(h,'END') eq 0, ct, complement=wnothist) ; put history keys at the end
             if ct gt 0 then h = [h[wnothist], h[whist]]
             wcomm = where(strpos(h,'COMMENT') eq 0 or strpos(h,'END') eq 0, ct, complement=wnotcomm) ; put history keys at the end
             if ct gt 0 then h = [h[wnotcomm], h[wcomm]]

             if keyword_set(copy) then outname = outdir+'/'+removepath(savelist[i])
             if keyword_set(move) then outname = outdir+'/'+removepath(savelist[i])
             if keyword_set(mark) then outname = savelist[i]
             writespec, outname, sp, h
             if i mod 20 eq 0 and i gt 1 then print, ' Wrote '+clip(i)+' files...'
             if keyword_set(move) and file_test(outname) then file_delete, savelist[i]
           endif else begin
             print, savelist[i]+' has vanished...?'
           endelse
         endfor
       endif else begin
         print, ' Taking no action.'
       endelse
    endif
 endif
 if keyword_set(delete) then begin
    readcol, 'validation.txt', vfiles, vresults, /silent ; can be lazy here
    deletelist = ''
    for i = 0, n_elements(vfiles)-1 do begin
       if vresults[i] eq 'fail' then $
           deletelist = [deletelist, vfiles[i]]
    endfor
    if n_elements(deletelist) gt 1 then begin
       print, ' Will delete '+clip(n_elements(deletelist)-1)+' files:'
       for i = 1, n_elements(savelist)-1 do begin
          print, '   '+deletelist[i]
       endfor
       print, ' Press [enter] to confirm or [q] to cancel'
       key = get_kbrd()
       if key eq string(10B) then begin
         for i = 1, n_elements(deletelist)-1 do begin
           file_delete, deletelist[i]
         endfor
       endif
    endif
 endif

 ; for i = 0, n_elements(result)-1 do begin
 ;    print, files[i], ' ', result[i]
 ; endfor

end
