;+
; NAME:
;       namefix
;
; PURPOSE:
;       Standardize object names using a catalog of names and positions.
;
; CALLING SEQUENCE:
;       namefix, header, tables, radius=radius, result=result
;
; EXPLANATION:
;       The OBJECT keyword is notoriously inaccurate especially for short
;       observations (standards).  This will 'fix' it by comparing against
;       a catalog such as a starlist or (in LPipe) a standard star catalog.
;       Since LPipe uses TARGNAME and not OBJECT, this is not critical, but
;       is helpful to avoid confusion.  It will also add notes about the arc
;       lines, etc.
;
; INPUTS:
;       header - FITS header (already read from disk) to check and edit.
;       tables - List of files with names and coordinates to match against.
;
; OPTIONAL INPUT KEYWORDS:
;       radius - matching radius in arcsec
;
; OPTIONAL OUTPUT KEYWORDS:
;       result - Returns 0 if unchanged, returns 1 if matched/changed
;
; OUTPUTS:
;       The header is updated.
; 
;
;-
;------------------------------------------------------------------------------

pro namefix, header, tables, radius=radius, result=result

  ; Correct incorrect object names in a FITS header using a positional catalog (specified in tables).
  ; If the RA/DEC position is within radius arcsec of anything in the catalog, the catalog name
  ; becomes the object name.

  common storeref, prevtables, ref

  if n_elements(radius) eq 0 then radius = 300.

  result = 0
  reload = 0
  if n_elements(prevtables) ne n_elements(tables) then reload = 1
  if n_elements(prevtables) eq n_elements(tables) then if total(prevtables ne tables) gt 0 then reload = 1
  if reload then begin
     for t = 0, n_elements(tables)-1 do begin
       if file_test(tables[t]) eq 0 then begin
         print, tables[t], ' not found!'
       endif
     endfor
     refname = ['']
     refra = [0.]
     refdec = [0.]
     reftype = ['']
     for t = 0, n_elements(tables)-1 do begin
       if file_test(tables[t]) eq 0 then continue
       openr, 1, tables[t]
       inline = ''
       while not eof(1) do begin
          readf, 1, inline
          inline = clip(inline)
          if strlen(inline) lt 2 then continue
          if strmid(inline,0,1) eq '#' then continue
          inarr = strsplit(inline,/extract)
          if n_elements(inarr) lt 3 then continue
          objname = (strsplit(inarr[0],'_',/extract)) [0]
          refname = [refname, objname]
          if strlen(inarr[1]) le 2 and strlen(inarr[2]) le 2 then begin  ; hh mm ss  dd mm ss format
             inra = 15*ten(inarr[1]+':'+inarr[2]+':'+inarr[3])
             indec = ten(inarr[4]+':'+inarr[5]+':'+inarr[4])
          endif else begin
             if strpos(inarr[1],':') gt 0 then inra = 15*ten(inarr[1]) else inra  = float(inarr[1])
             if strpos(inarr[2],':') gt 0 then indec =   ten(inarr[2]) else indec = float(inarr[2])
          endelse
          refra   = [refra,  inra]
          refdec  = [refdec, indec]
          if n_elements(inarr) gt 3 then  inreftype = inarr[3] else inreftype = ''
          reftype = [reftype, inreftype]
       endwhile
       close, 1
     endfor     
     nref = n_elements(refra)-1
     ref = replicate({name:'', ra:0., dec:0., type:''},nref)
     ref.name = refname[1:*]
     ref.ra = refra[1:*]
     ref.dec = refdec[1:*]
     ref.type = reftype[1:*]

     prevtables = tables
  endif else begin
     nref = n_elements(ref)
  endelse

  obj = ''
  objtype = ''

  rastr = strtrim(sxpar(header,'RA'),2)   ; should do something to allow for decimal input
  decstr = strtrim(sxpar(header,'DEC'),2)
  if strpos(rastr,':') gt 0 then ra = ten(rastr)*15  else ra = float(rastr) ; hours to degrees
  if strpos(decstr,':') gt 0 then dec = ten(decstr)  else dec = float(decstr)

  distance = fltarr(nref)

  for r = 0, nref-1 do begin
    dist = 3600*(abs(dec-ref[r].dec));
    if dist lt radius then gcirc, 2, ra, dec, ref[r].ra, ref[r].dec, dist
    distance[r] = dist
  endfor
  mindist = min(distance)
  minr = (where(distance eq mindist)) [0]
  if mindist lt radius then begin
      obj = ref[minr].name
      objtype = ref[minr].type
  endif

  trapdoor = clip(sxpar(header,'TRAPDOOR'))
  lampstr = clip(sxpar(header,'LAMPS'))
  lamps = fix(strsplit(lampstr,',',/extract))
  if n_elements(lamps) ge 6 and obj ne '' then begin
      arc = total(lamps[0:4])
      hal = lamps[5]
      if arc gt 0 then obj += ' (arc)'
      if hal gt 0 then obj += ' (hal)'
      if arc eq 0 and hal eq 0 then begin
        if trapdoor eq 'closed' then obj += ' (closed)'
      endif
  endif


  origobj = sxpar(header, 'OBJECT', /silent)
  sxaddpar, header, 'ORIGOBJ', origobj
  if obj ne origobj and obj ne '' then begin
     if obj ne clip(origobj) then print, '  [OBJECT: '+clip(origobj)+' -> '+obj+'] ' ;, format='($,A)'
     sxaddpar, header, 'OBJECT', obj, ' Object renamed from catalog'
     result = 1
  endif
  ;if objtype ne '' then sxaddpar, header, 'OBJTYPE', objtype  ; not sure what this ever did

end

