;+
; NAME:
;       readsimplelris
;
; PURPOSE:
;       Reads and bias-corrects a raw LRIS file stored in simple format.
;
; CALLING SEQUENCE:
;       readsimplelris, filename, [header=header, /nobias, /linebias]
;
; EXPLANATION:
;       Pre-upgrade LRIS blue files (and some post-upgrade red files) were 
;       written to disk by the instrument software as
;       simple FITS files with the amplifiers merged together, not as 
;       multiple-extension FITS files.  LPipe always merges the amplifiers 
;       anyway, but this tool allows simple-header files to be read in 
;       (and bias-corrected) such that the output mimics what is produced by 
;       readmhdufitslp for multiple-extension files.
;
; INPUTS:
;       filename - File to read in.
;
; OPTIONAL INPUT KEYWORDS:
;       /nobias - Do not use the overscan to bias-correct.
;       /linebias - Correct each column individually using the overscan.
;       /verbose - Does not do anything
;
; OPTIONAL OUTPUT KEYWORDS:
;       header - Contains the FITS header.
;
; OUTPUTS:
;       Returns a 2D array containing the FITS data.
; 
; EXAMPLES:
;       im = readsimplelris('r170101_0001.fits', header=h)
;
; NOTES:
;       Has not been tested recently.
;
;-
;------------------------------------------------------------------------------

function readsimplelris, filename, header=header, nobias=nobias, linebias=linebias, verbose=verbose

   ; for LRIS blue or upgraded LRIS red in old format


 fullarray = readfits(filename, header)
 nxf = (size(fullarray)) [1]
 instrume = clip(sxpar(header,'INSTRUME'))
 binning = clip(sxpar(header,'BINNING'))
 xbin = fix((strsplit(binning, ',', /extract)) [0])
 prepix = sxpar(header,'PREPIX', count=ctpre)/xbin
 postpix = sxpar(header,'POSTPIX', count=ctpost)/xbin
 window = sxpar(header,'WINDOW', count=ctwin)

 if instrume eq 'LRISBLUE' then begin
   if ctpre eq 0 then prepix = 51/xbin
   if ctpost eq 0 then postpix = 80/xbin
   if ctwin eq 0 then sxaddpar, header, 'WINDOW', '1,0,0,1024,2048', 'Missing from original file'
   x0 = (fix(strsplit(window,',',/extract))) [1]
   y0 = (fix(strsplit(window,',',/extract))) [2]

   array = 1.0*fullarray[prepix*4:nxf-1-postpix*4,*]
   if x0 gt 0 or y0 gt 0 then begin
      print, 'WARNING: The array was windowed!'
      print, '  Cannot handle this situation yet.  Not bias-correcting.'
      return, array
   endif
   bias1l = fullarray[       0:  prepix-1,*]
   bias2l = fullarray[  prepix:2*prepix-1,*]
   bias3l = fullarray[2*prepix:3*prepix-1,*]
   bias4l = fullarray[3*prepix:4*prepix-1,*]
   bias1r = fullarray[nxf-4*postpix:nxf-3*postpix-1,*]
   bias2r = fullarray[nxf-3*postpix:nxf-2*postpix-1,*]
   bias3r = fullarray[nxf-2*postpix:nxf-1*postpix-1,*]
   bias4r = fullarray[nxf-1*postpix:nxf          -1,*]
   nx = (size(array)) [1]
   ny = (size(array)) [2]
   if n_elements(nobias) eq 0 then begin
       ; this is still hard-coded and cannot deal with x-windowing.    
       ; (However blue is almost never windowed)
       
       if keyword_set(linebias) eq 0 then begin
          array[0:1024/xbin-1,*]    -= median([bias1l[*], bias1r[*]])
          array[1024/xbin:2048/xbin-1,*] -= median([bias2l[*], bias2r[*]])
          array[2048/xbin:3072/xbin-1,*] -= median([bias3l[*], bias3r[*]])
          array[3072/xbin:*,*]      -= median([bias4l[*], bias4r[*]])
       endif else begin
          for y = 0, ny-1 do begin
             array[0:1024/xbin-1,y]    -= median([bias1l[*,y], bias1r[*,y]])
             array[1024/xbin:2048/xbin-1,y] -= median([bias2l[*,y], bias2r[*,y]])
             array[2048/xbin:3072/xbin-1,y] -= median([bias3l[*,y], bias3r[*,y]])
             array[3072/xbin:*,y]      -= median([bias4l[*,y], bias4r[*,y]])
          endfor
       endelse
   endif
   array = reverse(array,2) ; seems that this is reversed
   sxaddpar, header, 'AMP1', clip(0)+','+clip(1024/xbin-1)         ;'0,1023'
   sxaddpar, header, 'AMP2', clip(1024/xbin)+','+clip(2048/xbin-1) ;'1024,2047'
   sxaddpar, header, 'AMP3', clip(2048/xbin)+','+clip(3072/xbin-1) ;'2048,3071'
   sxaddpar, header, 'AMP4', clip(3072/xbin)+','+clip(4096/xbin-1) ;'3072,4095'
   return, array

 endif else begin

   array = 1.0*fullarray[prepix*2:nxf-1-postpix*2,*]
   bias1l = fullarray[       0:  prepix-1,*]
   bias2l = fullarray[  prepix:2*prepix-1,*]
   bias1r = fullarray[nxf-2*postpix:nxf-1*postpix-1,*]
   bias2r = fullarray[nxf-1*postpix:nxf-0*postpix-1,*]
   nx = (size(array)) [1]
   ny = (size(array)) [2]
   if n_elements(nobias) eq 0 then begin
       ; this is still hard-coded and cannot deal with x-windowing.    
       if keyword_set(linebias) eq 0 then begin
          array[0:1024-1,*]    -= median([bias1l[*], bias1r[*]])
          array[1024:2048-1,*] -= median([bias2l[*], bias2r[*]])
       endif else begin
          for y = 0, ny-1 do begin
             array[0:1024-1,y]    -= median([bias1l[*,y], bias1r[*,y]])
             array[1024:2048-1,y] -= median([bias2l[*,y], bias2r[*,y]])
          endfor
       endelse
   endif
   sxaddpar, header, 'AMP1', '0,1023'
   sxaddpar, header, 'AMP2', '1024,2047'
   return, array

 endelse

end

