;+
; NAME:
;       reflatslit
;
; PURPOSE:
;       Tool to correct for flat-fielding errors associated with slit flexure
;       for slits (0.7") with strong irregularities.
;
; CALLING SEQUENCE:
;       reflatslit, filename, slitflatname, dy, outname
;
; EXPLANATION:
;       The LRIS 0.7" and (to a much lesser extent) 1.0" longslits show 
;       irregularities associated with obstructions in the slit.  These produce
;       regions of supressed sensitivity that follow the object trace 
;       horizontally across the detector and can greatly impact sky subtracction.
;       Because of slit flexure they do not divide out using a basic flat-field.
;       However, they can be removed (partially) by shifting a flat-field by
;       an appropriate amount up or down for the spatial-axis flexure that 
;       occurred.  This tool takes a flat-field, shifts it, and divides to remove
;       the irregularities.
;
; INPUTS:
;       filename - The file (generally a science observation) to correct.
;       slitflatname - The slit illumination flat filename (see notes below)
;       dy - Vertical shift to account for flexure (in pixels).
;
; OPTIONAL INPUTS:
;       outname - name of the slit-corrected output file.  By default 'c' is
;                 prepended before the 'fp' in the filename.
;
; OUTPUTS:
;       The corrected file is written to disk.
; 
; EXAMPLES:
;       reflatslit, 'fpr170101_0100r.fits', 'slitflat.fits', 1.1
;
; NOTES:
;       Requires a slit illumination flat, which is a pixel-flat-divided 
;       flat-field taken at a fixed position and with a single slit, such that
;       it contains the slit irregularities but not pixel-to-pixel variations. 
;       The user will have to produce this themselves by taking a 
;       single-position flat and dividing by a flat taken with a defect-free 
;       slit (or possibly by horizontally median-filtering a 0.7" flat.)
;       The appropriate dy correction will have to be determined by eye.
;       A future goal is to automate this correction into the pipeline. 
;       (This will be tricky, as it will require automated slit flat 
;       construction and very accurate flexure measurement.  
;       A wrongly-corrected flat makes the data worse!)
;
;
;-
;------------------------------------------------------------------------------

pro reflatslit, filename, slitflatname, dy, outname=outname

  ; did the slit move?  Use this to divide out the incorrect slit response in the flatfield and then
  ; re-add it.

  im = readfits(filename, h)
  slitflat = readfits(slitflatname, hflat)

  s = size(im)  
  nx = s[1]
  ny = s[2]
  x = findgen(nx)
  y = findgen(ny)

  dy0 = -sxpar(h,'LTV2')+sxpar(hflat,'LTV2')
  slitflat = slitflat[*,dy0:*] ; assumes flat is larger

  newx = x
  newy = y - dy

  shiftflat = interpolate(slitflat, newx, newy, /grid)

  im *= slitflat
  im /= shiftflat

  if n_elements(outname) eq 0 then begin
     outname = repstr(filename,'fp','cfp')
     if outname eq filename then outname = 'c'+removepath(filename)
     if outname eq filename then return
  endif
  writefits, outname, im, h

end

